/*  $Id: bldDbRecType.c,v 1.8 1993/07/22 14:06:22 epics Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  06-07-91        rcz     initial release
 * .02	09-27-91	rcz	Removed dimension requirement for ascii files
 * .03	mm-dd-yy	iii	Comment
 * 	...
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dbStaticLib.h>
#include <dbRecType.h>
#include <sdrHeader.h>
#include "getSdrLine.h"
#include "bldLib.h"

int             cmd_opts();

struct sdrHeader sdrHeader;
char            inFile[80];
char            outFile[80];

typedef struct enode {
    SNODE          *node;
    char            rec_name[MAXBUF];
}               ENODE;
#define SAME 0
/*
 * BLDDBRECTYPE.C
 *
 * builds record type SDR file
 *
 * INPUT:     dbRecType.ascii
 *
 * OUTPUT:     dbRecType.sdr
 */

main(argc, argv)
    char          **argv;
    int             argc;
{
    FILE           *inFile_fp;
    FILE           *outFile_fp;
    struct lineIn  *plineIn = NULL;
    struct recType *dbRecType = NULL;
    SLIST          *pList = NULL;
    SNODE          *pNew = NULL;
    ENODE          *node = NULL;
    char            rec_name[MAXBUF];
    int             status;
    int             len;
    int             i;
    long            Ssize = 0;
    long            Dsize = 0;
    long            Psize = 0;
    void            *Saddr = 0;
    long            number = 0;
    /* check cmd line usage */
    if ((cmd_opts(argc, argv)) < 0)
	return -1;

    /* allocate space for the getSdrLine input structure */
    plineIn = dbCalloc(1,sizeof(struct lineIn));
    plineIn->fname = inFile;

    if ((inFile_fp = fopen(inFile, "r")) == NULL) {
	printf("bldDbRecType: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    /* allocate a list ptr */
    pList = dbCalloc(1,sizeof(SLIST));
    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, TWO_QUOTES)) < 0) {
	    printf("bldDbRecType: illegal format in file ->>> %s <<<-\n", inFile);
	    return (-1);
	}
	if (plineIn->eof == EOF) {
	    break;
	}
	if ((status = sscanf(plineIn->line, " \"%[^\"]", rec_name)) != 1) {
	    printf("bldDbRecType: illegal field in file %s\n", inFile);
	    printf("line=%i\n", plineIn->line_no);
	    return (-1);
	}
	len = strlen(rec_name);
	Ssize += len + 1;
	/* add a definition to the list */
	pNew = dbCalloc(1,sizeof(ENODE));
	node = (ENODE *) pNew;
	memcpy(node->rec_name,rec_name,len);

	sllAdd(pList, pNew);
    }
    if ((fclose(inFile_fp)) == EOF) {
	printf("bldDbRecType: fclose error. errno=%i\nfile=%s\n",
	       errno, inFile);
	return (-1);
    }
    if (pList->count == 0) {
	printf("bldDbRecType: no entries in file %s\n", inFile);
	return (-1);
    }
    number = pList->count +1;
    /* make sure we have padding to long words */
    Ssize = ((Ssize + sizeof(long) - 1) / sizeof(long)) * sizeof(long);
    Psize = number  * sizeof(void *);
    /* total contiguous area */
    Dsize = sizeof(struct recType) + Psize + Ssize;
    dbRecType = dbCalloc(1,Dsize);
    dbRecType->papName = (void *)((char *)dbRecType + sizeof(struct recType));
    dbRecType->number = number;
    i = 1;

    /* starting address of 1st string */
    Saddr = (char *)dbRecType + sizeof(struct recType) + Psize;
    pNew = sllFirst(pList);
    /* copy strings and change pointers to offsets */
    while (pNew) {
	node = (ENODE *) pNew;
        dbRecType->papName[i] = Saddr;
	len = strlen(node->rec_name);
	memcpy((char *)Saddr,&node->rec_name[0],len + 1);
	/* make ptr an offset */
	dbRecType->papName[i] = (void *)((char *)dbRecType->papName[i] - (char *)dbRecType);
	Saddr = (char *)Saddr + len+1;
	pNew = sllNext(pNew);
	i++;
    }
    dbRecType->papName = (void *)((char *)dbRecType->papName - (char *)dbRecType);

    /* init the sdrHeader header */
    memset((char *)&sdrHeader,0, sizeof(struct sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Dsize;
    sdrHeader.type = SDR_DB_RECTYPE;

    /* write the SDR structure */
    if ((outFile_fp = fopen(outFile, "w")) == NULL) {
	printf("bldDbRecType: fopen errno=%i status=%i\nfile=%s\n",
	       errno, status, outFile);
	return (-1);
    }
    if ((status = fwrite(&sdrHeader, sizeof(struct sdrHeader), 1,
			 outFile_fp)) != 1) {
	printf("bldDbRecType: fwrite error. errno=%i status=%i\nfile=%s\n",
	       errno, status, outFile);
	return (-1);
    }
    if ((status = fwrite(dbRecType, (int) Dsize, 1, outFile_fp)) != 1) {
	printf("bldDbRecType: fwrite error. errno=%i status=%i\nfile=%s\n",
	       errno, status, outFile);
	return (-1);
    }
    if ((fclose(outFile_fp)) == EOF) {
	printf("bldDbRecType: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    return (0);
}

/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_OPTS
 */
cmd_opts(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    char           *ptrToMem;
    int             c;
    int             err = 0;
    short           aflg = 0;
    short           oflg = 0;

    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldDbRecType: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile, optarg);
	    break;
	}

    if (!aflg || !oflg || err) {
	printf("Usage: \n");
	printf("    <bldDbRecType -a dbRecType.ascii -o dbRecType.sdr>\n");
	return (-1);
    }
    return (0);
}
