/*  $Id: bldDevSup.c,v 1.13 1995/03/02 18:50:37 jba Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  06-07-91        rcz     initial release
 * .02	10-29-91	rcz	Removed dimension requirement for ascii files
 * .03	10-26-91	rcz	Bug fix - devSup ordering
 * .04  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 * 	...
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <choice.h>
#include <dbDefs.h>
#include <dbStaticLib.h>
#include <dbBase.h>
#include <devSup.h>
#include <dbRecType.h>
#include <sdrHeader.h>
#include <getSdrLine.h>
#include "bldLib.h"
#define SIX_QUOTES 6
#define SAME 0
#define BYTES50		50
#define SCAN_CHOICELINES 1

typedef struct recList {
    SLIST         **papList;
}               SREC;



char            inFile1[BYTES50];
char            inFile2[BYTES50];
char            outFile1[BYTES50];
char            outFile2[BYTES50];
int             errno;

struct sdrHeader sdrHeader;
struct lineIn   lineIn;		/* buffer for getSdrLine */


typedef struct enode {
    SNODE          *node;
    char            RecTyp_name[BYTES50];
    long            link_type;
    char            dsetName_str[BYTES50];
    char            choice_str[BYTES50];
}               ENODE;

ENODE           inLineInfo;


struct recType *dbRecType=NULL;
struct recDevSup *devSup=NULL;
struct devChoiceRec *choiceDev=NULL;
struct dbBase *pdbBase=NULL;

static int verifyType();
static int inScan();

/*
 * BLDDEVSUP
 *
 * INPUT:	devSup.ascii
 *       	dbRecType.sdr
 * OUTPUT:	devSup.sdr
 *              choiceDev.sdr
 */
main(argc, argv)
    int             argc;
    char          **argv;
{
    FILE           *inFile1_fp;
    FILE           *outFile1_fp;
    FILE           *outFile2_fp;
    ENODE          *pInfo = &inLineInfo;
    ENODE          *node = NULL;
    SREC           *pRecTypeList = NULL;
    SNODE          *pNew = NULL;
    SNODE          *pt = NULL;
    SLIST          *pList = NULL;
    char           *ptrToMem = NULL;
    struct lineIn  *plineIn = &lineIn;	/* ptr to buffer */
    int             rec_no;
    int             len;
    int             i;
    int             j;
    long            Ssize = 0;
    long            Dsize = 0;
    long            totalC = 0;
    long            totalDE = 0;
    long            totalF = 0;
    long            totalB;
    long            number;
    void           *addr1 = NULL;
    void           *addrS1 = NULL;
    int             rtnval = 0;
    FILE	   *fp;
    /* check cmd line */
    if ((cmd_DevSup(argc, argv)) < 0) {
	return (-1);
    }
    plineIn->fname = inFile1;

    if (chk_align() < 0)
	return (-1);
    /* method to get dbRead to skip the checksum */
    pdbBase = dbAllocBase();
    pdbBase->pvtSumFlag = 1;

    /* load ascii record type names */
    fp = fopen(inFile2,"r");
    if(!fp) return(-1);
    if(dbRead(pdbBase,fp)) return(-1);
    fclose(fp);
    devSup = pdbBase->precDevSup;
    dbRecType = pdbBase->precType;
    choiceDev = pdbBase->pchoiceDev;

    if ((inFile1_fp = fopen(inFile1, "r")) == NULL) {
	printf("bldDevSup: fopen error errno=%i\nfile=%s\n", errno, inFile1);
	return (-1);
    }
    pRecTypeList = dbCalloc(1,sizeof(SREC));
    pRecTypeList->papList = dbCalloc(1,dbRecType->number * sizeof(void *));

    /* count number of each record type */
    FOREVER {
	if ((getSdrLine(inFile1_fp, plineIn, SIX_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF) {
	    break;
	}
	memset((char *)&inLineInfo,0, sizeof(ENODE));
	if (inScan(SCAN_CHOICELINES, plineIn, pInfo) < 0)
	    return (-1);

	if (!verifyType(dbRecType, &rec_no, pInfo))
	    return (-1);
	/* if no list for this record type - create it */
	if (!pRecTypeList->papList[rec_no]) {
	    pRecTypeList->papList[rec_no] = dbCalloc(1,sizeof(SLIST));
	}
	pList = (SLIST *) pRecTypeList->papList[rec_no];
	/* add a definition to the list */
	pNew = dbCalloc(1,sizeof(ENODE));
	memcpy((char *)pNew,(char *)&inLineInfo,sizeof(ENODE));
	sllAdd(pList, pNew);
    }				/* end of input */
    if ((fclose(inFile1_fp)) == EOF) {
	printf("bldDevSup: fclose error. errno=%i\nfile=%s\n",
	       errno, inFile1);
	return (-1);
    }
    /* calc the total contiguous area */
    totalB = dbRecType->number * sizeof(long) + sizeof(struct recDevSup);
    for (rec_no = 0; rec_no < dbRecType->number; rec_no++) {
	if (pRecTypeList->papList[rec_no]) {
	    totalC += sizeof(struct devSup);
	    /* get the number from the list->count */
	    number = pRecTypeList->papList[rec_no]->count;
	    totalDE += number * sizeof(long);
	    totalDE += number * sizeof(long);
	    /* also count the total string size area */
	    pList = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(pList);
	    while (pt) {
		node = (ENODE *) pt;
		len = strlen(node->dsetName_str);
		Ssize += len + 1;
		pt = sllNext(pt);
	    }
	    /* make sure we have padding to long words */
	    Ssize = (Ssize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	}
    }
    Dsize = totalB + totalC + totalDE + totalF + Ssize;

    /* copy the data to the allocated area */
    /* allocate recDevSup structure */
    devSup = dbCalloc(1,Dsize);
    addr1 = (char *)devSup + sizeof(struct recDevSup);
    devSup->number = dbRecType->number;
    devSup->papDevSup = addr1;
    addr1 = (char *)addr1 + dbRecType->number * sizeof(long);
    /* insert the ptrs */
    for (rec_no = 0; rec_no < dbRecType->number; rec_no++) {
	Ssize = 0;
	if (pRecTypeList->papList[rec_no]) {
	    devSup->papDevSup[rec_no] = addr1;
	    number = pRecTypeList->papList[rec_no]->count;
	    devSup->papDevSup[rec_no]->number = number;
	    addr1 = (char *)addr1 + sizeof(struct devSup);
	    devSup->papDevSup[rec_no]->papDset = addr1;
	    addr1 = (char *)addr1 + number *sizeof(long);
	    devSup->papDevSup[rec_no]->papDsetName = addr1;
	    addr1 = (char *)addr1 + number *sizeof(long);
	    addrS1 = addr1;

	    pList = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(pList);
	    i = 0;
	    while (pt) {
		node = (ENODE *) pt;
		len = strlen(node->dsetName_str);
		devSup->papDevSup[rec_no]->papDsetName[i] = addrS1;
		memcpy(addrS1,node->dsetName_str,len + 1);
		addrS1 = (char *)addrS1 + len + 1;
		Ssize += len + 1;
		pt = sllNext(pt);
		i++;
	    }
	    /* make sure we have padding to long words */
	    Ssize = (Ssize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	    addr1 = (char *)addr1 + Ssize;
	}
    }
    /* change all pointers to offsets */
    for (i = 0; i < dbRecType->number; i++) {
	if (devSup->papDevSup[i] != NULL) {
	    for (j = 0; j < devSup->papDevSup[i]->number; j++) {
		if (devSup->papDevSup[i]->papDsetName[j] != NULL) {
		    devSup->papDevSup[i]->papDsetName[j] = (void *)(
			(char *)(devSup->papDevSup[i]->papDsetName[j]) - (char *)devSup);
		}
	    }
	    devSup->papDevSup[i]->papDsetName = (void *)(
		(char *)(devSup->papDevSup[i]->papDsetName) - (char *)devSup);
	    devSup->papDevSup[i]->papDset = (void *)(
		(char *)(devSup->papDevSup[i]->papDset) - (char *)devSup);
	    devSup->papDevSup[i] = (void *)(
		(char *)(devSup->papDevSup[i]) - (char *)devSup);
	}
    }
    devSup->papDevSup = (void *)((char *)devSup->papDevSup - (char *)devSup);

    /* write the header for SDR_DEVSUP */
    memset((char *)&sdrHeader,0, sizeof(struct sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Dsize;
    sdrHeader.type = SDR_DEVSUP;
    if ((outFile1_fp = fopen(outFile1, "w")) == NULL) {
	printf("bldDevSup: fopen errno=%i \nfile=%s\n",
	       errno, outFile1);
	return (-1);
    }
    if ((rtnval = fwrite(&sdrHeader,(int) sizeof(sdrHeader), 1,
			 outFile1_fp)) != 1) {
	printf("bldDevSup: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile1);
	return (-1);
    }
    if ((rtnval = fwrite(devSup, (int) Dsize, 1, outFile1_fp)) != 1) {
	printf("bldDevSup: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile1);
	return (-1);
    }
    if ((fclose(outFile1_fp)) == EOF) {
	printf("bldDevSup: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile1);
	return (-1);
    }
    /* calc the total contiguous area */
    Dsize = totalB = totalC = totalDE = totalF = Ssize = 0;
    totalB = dbRecType->number * sizeof(long) + sizeof(struct devChoiceRec);
    for (rec_no = 0; rec_no < dbRecType->number; rec_no++) {
	if (pRecTypeList->papList[rec_no]) {
	    totalC += sizeof(struct devChoiceSet);
	    /* get the number from the list->count */
	    number = pRecTypeList->papList[rec_no]->count;
	    totalDE += number * sizeof(long);
	    totalF += number * sizeof(struct devChoice);
            /* also count the total string size area */
	    pList = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(pList);
	    while (pt) {
		node = (ENODE *) pt;
		len = strlen(node->choice_str);
		Ssize += len + 1;
		pt = sllNext(pt);
	    }
	    /* make sure we have padding to long words */
	    Ssize = (Ssize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	}
    }
    Dsize = totalB + totalC + totalDE + totalF + Ssize;
    /* copy the data to the allocated area */
    /* allocate devChoiceRec structure */
    choiceDev = dbCalloc(1,Dsize);
    addr1 = (char *)choiceDev + sizeof(struct devChoiceRec);
    choiceDev->number = dbRecType->number;
    choiceDev->papDevChoiceSet = addr1;
    addr1 = (char *)addr1 + dbRecType->number * sizeof(long);
    /* insert the ptrs */
    for (rec_no = 0; rec_no < dbRecType->number; rec_no++) {
	Ssize = 0;
	if (pRecTypeList->papList[rec_no]) {
	    choiceDev->papDevChoiceSet[rec_no] = addr1;
	    number = pRecTypeList->papList[rec_no]->count;
	    choiceDev->papDevChoiceSet[rec_no]->number = number;
	    addr1 = (char *)addr1 + sizeof(struct devChoiceSet);
	    choiceDev->papDevChoiceSet[rec_no]->papDevChoice = addr1;
	    addr1 =(char *)addr1 + number *sizeof(long);
	    addrS1 = (char *)addr1 + number * sizeof(struct devChoice);
	    pList = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(pList);
	    i = 0;
	    while (pt) {
		node = (ENODE *) pt;
		choiceDev->papDevChoiceSet[rec_no]->papDevChoice[i] =  addr1;
		len = strlen(node->choice_str);
		choiceDev->papDevChoiceSet[rec_no]->papDevChoice[i]->pchoice = addrS1;
		choiceDev->papDevChoiceSet[rec_no]->papDevChoice[i]->link_type =
		    node->link_type;
		memcpy(addrS1,node->choice_str,len + 1);
		addr1 = (char *)addr1 + sizeof(struct devChoice);
		addrS1 = (char *)addrS1 + len + 1;
		Ssize += len + 1;
		pt = sllNext(pt);
		i++;
	    }
	    /* make sure we have padding to long words */
	    Ssize = (Ssize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
	    addr1 = (char *)addr1 + (long) Ssize;
	}
    }
    /* change all pointers to offsets */
    for (i = 0; i < choiceDev->number; i++) {
	if (choiceDev->papDevChoiceSet[i] != NULL) {
	    for (j = 0; j < choiceDev->papDevChoiceSet[i]->number; j++) {
		if (choiceDev->papDevChoiceSet[i]->papDevChoice[j] != NULL) {
		    choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice = (void *)(
			(char *)(choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice) - (char *)choiceDev);
		    choiceDev->papDevChoiceSet[i]->papDevChoice[j] = (void *)(
			(char *)(choiceDev->papDevChoiceSet[i]->papDevChoice[j])
			- (char *)choiceDev);
		}
	    }
	    choiceDev->papDevChoiceSet[i]->papDevChoice = (void *)(
		(char *)(choiceDev->papDevChoiceSet[i]->papDevChoice)
		- (char *)choiceDev);
	    choiceDev->papDevChoiceSet[i] = (void *)(
		(char *)(choiceDev->papDevChoiceSet[i])
		- (char *)choiceDev);
	}
    }
    choiceDev->papDevChoiceSet = (void *)((char *)(choiceDev->papDevChoiceSet)
	- (char *)choiceDev);
    /* write the header for SDR_CHOICEDEV */
    memset((char *)&sdrHeader,0, sizeof(sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Dsize;
    sdrHeader.type = SDR_CHOICEDEV;

    if ((outFile2_fp = fopen(outFile2, "w")) == NULL) {
	printf("bldDevSup: fopen errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile2);
	return (-1);
    }
    if ((rtnval = fwrite(&sdrHeader, sizeof(sdrHeader), 1,
			 outFile2_fp)) != 1) {
	printf("bldDevSup: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile2);
	return (-1);
    }
    if ((rtnval = fwrite(choiceDev, (int) Dsize, 1, outFile2_fp)) != 1) {
	printf("bldDevSup: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile2);
	return (-1);
    }
    if ((fclose(outFile2_fp)) == EOF) {
	printf("bldDevSup: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile2);
	return (-1);
    }
    return (0);
}

/****************************************************************************
    Verify record type
****************************************************************************/
/*
 * VERIFYTYPE
 */
static int
verifyType(dbRecType, pRec_no, pInfo)
    struct recType *dbRecType;
    int            *pRec_no;
    ENODE          *pInfo;
{
    short           i;
    /* Verify record type name */
    for (i = 0; i < dbRecType->number; i++) {
	if (dbRecType->papName[i] != NULL) {
	    if (strcmp(dbRecType->papName[i],
		       pInfo->RecTyp_name) == SAME) {
		*pRec_no = i;
		return (TRUE);
	    }
	}
    }
    printf("bldDevSup: verifyType - Record type '%s' not defined\n",
	   pInfo->RecTyp_name);
    return (FALSE);
}
/****************************************************************************
    process input line - (0,-1) success, failure
****************************************************************************/
/*
 * INSCAN
 *
 */
static int
inScan(scan_type, plineIn, pInfo)
    int             scan_type;
    struct lineIn  *plineIn;
    ENODE          *pInfo;
{
    int             err = 0;
    switch (scan_type) {
    case SCAN_CHOICELINES:
	if ((sscanf(plineIn->line, "\"%[^\"]\" %ld \"%50[^\"]\" \"%50[^\"]",
		    pInfo->RecTyp_name,
		    &pInfo->link_type,
		    pInfo->dsetName_str,
		    pInfo->choice_str)) != 4) {
	    err = 1;
	}
	break;
    default:
	printf("bldDevSup: inScan(scan_type) not defined internally\n");
	printf("line=%i\n", plineIn->line_no);
	return (-1);
	break;
    }
    if (err == 1) {
	printf("bldDevSup: Illegal field format file=%s\n", inFile1);
	printf("line=%i\n", plineIn->line_no);
	return (-1);
    }
    return (0);
}
/****************************************************************************
    Checks alignment of header structs
****************************************************************************/
/*
 * CHK_ALIGN
 */
chk_align()
{
    if ((sizeof(struct recType) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "recType");
	return (-1);
    }
    if ((sizeof(struct recDevSup) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "recDevSup");
	return (-1);
    }
    if ((sizeof(struct devSup) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devSup");
	return (-1);
    }
    if ((sizeof(struct dset) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "dset");
	return (-1);
    }
    if ((sizeof(struct devChoiceRec) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devChoiceRec");
	return (-1);
    }
    if ((sizeof(struct devChoiceSet) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devChoiceSet");
	return (-1);
    }
    if ((sizeof(struct devChoice) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devChoice");
	return (-1);
    }
    return (0);
}
/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_DEVSUP
 */
cmd_DevSup(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    char           *ptrToMem;
    struct stat    *buf;
    int             c;
    int             err = 0;
    short           aflg = 0;
    short           bflg = 0;
    short           oflg = 0;
    short           Oflg = 0;

    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:b:o:O:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile1, optarg);
	    if (stat(inFile1, buf) == -1) {
		printf("bldDevSup: Can't stat file=%s errno=%i\n",
		       inFile1, errno);
		err++;
	    }
	    break;
	case 'b':
	    bflg++;
	    strcpy(inFile2, optarg);
	    if (stat(inFile2, buf) == -1) {
		printf("bldDevSup: Can't stat file=%s errno=%i\n",
		       inFile2, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile1, optarg);
	    break;
	case 'O':
	    Oflg++;
	    strcpy(outFile2, optarg);
	    break;
	}
    if (!aflg || !bflg || !oflg || !Oflg || err) {
	printf("Usage: \n");
	printf(" <bldDevSup -a devSup.ascii -b dbRecType.sdr -o devSup.sdr -O choiceDev.sdr >\n");
	return (-1);
    }
    return (0);
}
