/*  $Id: bldDrvSup.c,v 1.8 1993/07/22 14:06:21 epics Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  06-07-91        rcz     initial release
 * .02  10-07-91        rcz     Removed dimension requirement for ascii files
 * .03  05-18-92        rcz     changed ->drvetName to ->papDrvName
 * 	...
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dbStaticLib.h>
#include <drvSup.h>
#include <sdrHeader.h>
#include "getSdrLine.h"
#include "bldLib.h"

int             cmd_opts();

struct sdrHeader sdrHeader;
struct drvSup  *drvSup;
char            inFile[80];
char            outFile[80];
#define SAME 0

typedef struct enode {
    SNODE          *node;
    char            drv_name[MAXBUF];
}               ENODE;



/*
 * BLDDRVSUP
 *
 * INPUT:	drvSup.ascii
 * OUTPUT:	drvSup.sdr
 */

main(argc, argv)
    int             argc;
    char          **argv;
{

    FILE           *inFile_fp;
    FILE           *outFile_fp;
    struct lineIn  *plineIn = NULL;
    char            drv_name[MAXBUF];
    long            Ssize = 0;
    void            *Saddr = 0;
    long            Dsize = 0;
    long            Psize = 0;
    long            number = 0;
    SLIST          *pList = NULL;
    SNODE          *pNew = NULL;
    ENODE          *node = NULL;
    int             len;
    int             i;
    int             status;

    /* check cmd line usage */
    if ((cmd_DrvSup(argc, argv)) < 0)
	return -1;
    /* allocate space for the getSdrLine input structure */
    plineIn = dbCalloc(1,sizeof(struct lineIn));
    plineIn->fname = inFile;

    if ((inFile_fp = fopen(inFile, "r")) == NULL) {
	printf("bldDrvSup: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    /* allocate a list ptr */
    pList = dbCalloc(1,sizeof(SLIST));

    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, TWO_QUOTES)) < 0) {
	    printf("bldDrvSup: illegal format in file ->>> %s <<<-\n", inFile);
	    return (-1);
	}
	if (plineIn->eof == EOF) {
	    break;
	}
	if ((sscanf(plineIn->line, " \"%[^\"]", drv_name)) != 1) {
	    printf("bldDrvSup: illegal field in file %s\n", inFile);
	    printf("line=%i\n", plineIn->line_no);
	    return (-1);
	}
	len = strlen(drv_name);
	Ssize += len + 1;
	/* add a definition to the list */
	pNew = dbCalloc(1,sizeof(ENODE));
	node = (ENODE *) pNew;
	memcpy( node->drv_name,drv_name,len);

	sllAdd(pList, pNew);
    }
    if ((fclose(inFile_fp)) == EOF) {
	printf("bldDrvSup: fclose error. errno=%i\nfile=%s\n",
	       errno, inFile);
	return (-1);
    }
    if (pList->count == 0) {
	printf("bldDrvSup: no entries in file %s\n", inFile);
	return (-1);
    }
    number = pList->count;
    /* make sure we have padding to long words */
    Ssize = ((Ssize + sizeof(long) - 1) / sizeof(long)) * sizeof(long);
    Psize = number * sizeof(void *);
    /* total contiguous area */
    Dsize = sizeof(struct drvSup) + (2 * Psize) + Ssize;
    drvSup = dbCalloc(1,Dsize);
    drvSup->number = number;
    drvSup->papDrvName = (void *)((char *)drvSup + sizeof(struct drvSup));
    drvSup->papDrvet = (void *)((char *)drvSup + sizeof(struct drvSup) + Psize);
    i = 0;
    /* starting address of 1st string */
	Saddr = (char *)drvSup + Dsize - Ssize;
    pNew = sllFirst(pList);
    /* copy strings and change pointers to offsets */
    while (pNew) {
	node = (ENODE *) pNew;
	drvSup->papDrvName[i] = Saddr;
	len = strlen(node->drv_name);
	memcpy((char *)Saddr,(char *)&node->drv_name[0],len + 1);
	/* make ptr an offset */
	drvSup->papDrvName[i] = (void *)(
		(char *)drvSup->papDrvName[i] - (char *)drvSup);
	Saddr = (char *)Saddr + len + 1;
	pNew = sllNext(pNew);
	i++;
    }
    drvSup->papDrvName = (void *)(
	(char *)drvSup->papDrvName - (char *)drvSup);
    drvSup->papDrvet = (void *)(
	(char *)drvSup->papDrvet - (char *)drvSup);

    /* write output file */
    memset((char *)&sdrHeader,0, sizeof(struct sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Dsize;
    sdrHeader.type = SDR_DRVSUP;
    if ((outFile_fp = fopen(outFile, "w")) == NULL) {
	printf("bldDrvSup: fopen errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((fwrite(&sdrHeader, sizeof(struct sdrHeader), 1,
		outFile_fp)) != 1) {
	printf("bldDrvSup: fwrite error. errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((fwrite(drvSup, (int) Dsize, 1, outFile_fp)) != 1) {
	printf("bldDrvSup: fwrite error. errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((fclose(outFile_fp)) == EOF) {
	printf("bldDrvSup: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    return (0);
}
/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_DRVSUP
 */
cmd_DrvSup(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    char           *ptrToMem;
    int             c;
    int             err = 0;
    short           aflg = 0;
    short           oflg = 0;
    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldDrvSup: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile, optarg);
	    break;
	}

    if (!aflg || !oflg || err) {
	printf("Usage: \n");
	printf(" <bldDrvSup -a drvSup.ascii -o drvSup.sdr>\n");
	return (-1);
    }
    return (0);
}
