/* base/src/bld $Id: bldGblChoice.c,v 1.13 1994/11/21 22:24:06 mrk Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  06-07-91        rcz     initial release
 * .02	10-08-91	rcz	Removed dimension requirement
 * .03	11-01-91	rcz	Fixed returns in func cmd_choiceGbl
 * .04	05-23-94	pg	HP Port
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dbStaticLib.h>
#include <choice.h>
#include <sdrHeader.h>
#include <getSdrLine.h>
#include "bldLib.h"

struct lineIn   lineIn;		/* buffer for getSdrLine */
struct sdrHeader sdrHeader;

typedef struct menuType {
    SNODE	*node;
    SLIST	menuChoiceHead;
    int		type;
} menuType;

typedef struct menuChoice {
    SNODE          *node;
    char            choice_name[MAXBUF];
} menuChoice;

char            inFile[50];
char            outFile[50];

/*
 * BLDGBLCHOICE.C
 *
 * INPUT:     choiceGbl.ascii
 *
 * OUTPUT     choiceGbl.sdr
 *
 */

main(argc, argv)
    int             argc;
    char          **argv;
{

    FILE           *inFile_fp;
    FILE           *outFile_fp;
    SLIST	   menuTypeHead;
    menuType	   *pmenuType=NULL;
    menuChoice	   *pmenuChoice=NULL;
    struct lineIn  *plineIn = NULL;
    struct arrChoiceSet *choiceGbl = NULL;
    int             inx_num;
    int             len;
    int             inx;
    int             i;
    int             j;
    int             no_byts;
    char            choice_name[MAXBUF];
    long            Dsize = 0;
    long            totalB = 0;
    long            totalC = 0;
    long            Ssize = 0;
    void            *Saddr = 0;
    void            *Baddr = 0;
    /* check cmd line */
    if ((cmd_choiceGbl(argc, argv)) < 0) {
	return (-1);
    }
    /* allocate space for the getSdrLine input structure */
    plineIn = dbCalloc(1,sizeof(struct lineIn));
    plineIn->fname = inFile;

    if ((inFile_fp = fopen(inFile, "r")) == NULL) {
	printf("bldGblChoice: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    sllInit(&menuTypeHead);
    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, TWO_QUOTES)) < 0) {
	    printf("bldGblChoice: illegal format in file ->>> %s <<<-\n", inFile);
	    return (-1);
	}
	if (plineIn->eof == EOF) {
	    break;
	}
	/* decode 2nd field seperately because of NULL string */
	if ((sscanf(plineIn->line, "%li \"%n",
		    &inx_num, &no_byts)) != 1) {
	    printf("bldGblChoice: illegal field in file %s\n", inFile);
	    printf("bldGblChoice: line=%i\n", plineIn->line_no);
	    return (-1);
	}
	nextInputFld(plineIn, &no_byts);
	if ((plineIn->pnext[0]) == '"') {
	    choice_name[0] = '\0';
	    no_byts = 1;
	} else if ((sscanf(plineIn->pnext, "%[^\"]%n",
			   choice_name, &no_byts)) != 1) {
	    printf("bldGblChoice: illegal field in file %s\n", inFile);
	    printf("bldGblChoice: line=%i\n", plineIn->line_no);
	    return (-1);
	}
	if(pmenuType==NULL || pmenuType->type < inx_num) {
	    menuType	*pnew;

	    pnew = (menuType *)dbCalloc(1,sizeof(menuType));
	    sllInit(&pnew->menuChoiceHead);
	    pnew->type = inx_num;
	    sllAdd(&menuTypeHead,pnew);
	    pmenuType = pnew;
	}
        pmenuChoice = (menuChoice *)dbCalloc(1,sizeof(menuChoice));
	strcpy(pmenuChoice->choice_name,choice_name);
	sllAdd(&pmenuType->menuChoiceHead,pmenuChoice);
	len = strlen(choice_name);
	Ssize += len + 1;
    }
    /* make sure we have padding to long words */
    Ssize = ((Ssize + sizeof(long) - 1) / sizeof(long)) * sizeof(long);
    /* calc total choice pointers */
    pmenuType = (menuType *)sllFirst((SNODE *)&menuTypeHead);
    while(pmenuType) {
	totalC += (pmenuType->menuChoiceHead.count * sizeof(void *));
	pmenuType = (menuType *)sllNext((SNODE *)pmenuType);
    }
    totalB = menuTypeHead.count * sizeof(void *);
    totalB += menuTypeHead.count * sizeof(struct choiceSet);
    Dsize = sizeof(struct arrChoiceSet) + totalB +  totalC + Ssize;
    /* allocate contiguous area */
    choiceGbl = dbCalloc(1,Dsize);
    /* xfer data to  the contiguous area */
    Baddr = (char *)choiceGbl + sizeof(struct arrChoiceSet);
    choiceGbl->papChoiceSet = Baddr;
    choiceGbl->number = menuTypeHead.count;
    Baddr = (char *)Baddr + menuTypeHead.count * sizeof(void *);
    /* starting address of 1st string */
    Saddr = (char *)choiceGbl + Dsize - Ssize;
    pmenuType = (menuType *)sllFirst((SNODE *)&menuTypeHead);
    while(pmenuType) {
	inx = pmenuType->type;
	pmenuChoice =
		(menuChoice *)sllFirst((SNODE *)&pmenuType->menuChoiceHead);
	choiceGbl->papChoiceSet[inx] = Baddr;
	choiceGbl->papChoiceSet[inx]->number = pmenuType->menuChoiceHead.count;
	Baddr = (char *)Baddr + sizeof(struct choiceSet);
	choiceGbl->papChoiceSet[inx]->papChoice = Baddr;
	i = 0;
	while(pmenuChoice) {
	    choiceGbl->papChoiceSet[inx]->papChoice[i] = Saddr;
	    len = strlen(pmenuChoice->choice_name);
	    memcpy((char *) Saddr,pmenuChoice->choice_name,len + 1);
	    Saddr = (char *)Saddr + len + 1;
	    pmenuChoice = (menuChoice *)sllNext((SNODE *)pmenuChoice);
	    i++;
	}
	Baddr = (char *)Baddr +
		choiceGbl->papChoiceSet[inx]->number * sizeof(void *);
	pmenuType = (menuType *)sllNext((SNODE *)pmenuType);
    }
    /* convert pointers to offsets */
    for (i = 0; i < choiceGbl->number; i++) {
	if (choiceGbl->papChoiceSet[i]) {
	    for (j = 0; j < choiceGbl->papChoiceSet[i]->number; j++) {
		if (choiceGbl->papChoiceSet[i]->papChoice[j]) {
		    choiceGbl->papChoiceSet[i]->papChoice[j] = (void *)
			((char *)choiceGbl->papChoiceSet[i]->papChoice[j] - (char *)choiceGbl);
		}
	    }
	    choiceGbl->papChoiceSet[i]->papChoice = (void *)
		((char *)choiceGbl->papChoiceSet[i]->papChoice - (char *)choiceGbl);
	}
	choiceGbl->papChoiceSet[i] = (void *)
	    ((char *)choiceGbl->papChoiceSet[i] - (char *)choiceGbl);
    }
    choiceGbl->papChoiceSet = (void *)((char *)choiceGbl->papChoiceSet - (char *)choiceGbl);

    /* init the sdrHeader header */
    memset((char *)&sdrHeader,0, sizeof(struct sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = Dsize;
    sdrHeader.type = SDR_CHOICEGBL;
    /* write output file */
    if ((outFile_fp = fopen(outFile, "w")) == NULL) {
	printf("bldGblChoice: fopen errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((fwrite(&sdrHeader, sizeof(struct sdrHeader), 1,
		outFile_fp)) != 1) {
	printf("bldGblChoice: fwrite error. errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((fwrite(choiceGbl, (int) Dsize, 1, outFile_fp)) != 1) {
	printf("bldGblChoice: fwrite error. errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((fclose(outFile_fp)) == EOF) {
	printf("bldGblChoice: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    return (0);
}

/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_CHOICEGBL
 */
cmd_choiceGbl(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c,
                    err;
    short           aflg,
                    oflg;
    aflg = oflg = err = 0;
    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldGblChoice: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile, optarg);
	    break;
	}

    if (!aflg || !oflg || err) {
	printf("Usage: \n");
	printf("    <bldGblChoice -a choiceGbl.ascii -o choiceGbl.sdr>\n");
	return (-1);
    }
    return (0);
}

