/*  $Id: bldRecChoice.c,v 1.12 1995/03/02 18:50:41 jba Exp $
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  06-07-91        rcz     initial release
 * .02	11-09-91	rcz	Removed dimension requirement
 * .03  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 * 	...
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <dbStaticLib.h>
#include <choice.h>
#include <dbRecType.h>
#include <dbBase.h>
#include <dbStaticLib.h>
#include <sdrHeader.h>
#include <getSdrLine.h>
#include "bldLib.h"


#define SAME 0
#define BYTES50		50
#define SCAN_CHOICELINES	50
#ifndef FALSE
#define FALSE 0
#define TRUE  1
#endif

char            inFile[BYTES50];
char            inFile2[BYTES50];
char            outFile[BYTES50];

struct sdrHeader sdrHeader;
char           *allocateMem();

typedef struct sublist {
    SNODE          *node;
    int             index_num;
    SLIST           slist;
}               SUBLIST;

typedef struct info {
    SNODE          *node;
    char            RecTyp_name[BYTES50];
    int             index_num;
    char            choice_str[BYTES50];
}               LINEINFO;
LINEINFO        inLineInfo;

typedef struct enode {
    SNODE          *node;
    char            choice_str[BYTES50];
}               ENODE;

typedef struct recList {
    long            number;
    SLIST         **papList;
}               SREC;


struct recType *dbRecType=NULL;
struct choiceRec *choiceRec=NULL;
struct dbBase *pdbBase=NULL;

static int verifyType();

main(argc, argv)
    int             argc;
    char          **argv;
{
    FILE           *inFile_fp;
    FILE           *outFile_fp;
    SREC           *pRecTypeList = NULL;
    LINEINFO       *pInfo = &inLineInfo;
    SLIST          *plist1 = NULL;
    SLIST          *pList2 = NULL;
    struct lineIn  *plineIn = NULL;
    SNODE          *pt = NULL;
    SNODE          *pt2 = NULL;
    SUBLIST        *pSubList = NULL;
    ENODE          *node = NULL;
    long            size1 = 0;
    long            Ssize = 0;
    void            *addr1;
    int             rec_no = 0;
    int             max_inx = 0;
    int             index_num = 0;
    int             i = 0;
    int             j = 0;
    int             k = 0;
    int             rtnval;
    int             len = 0;
    FILE	    *fp;
    /* check cmd line */
    if ((cmd_choiceRec(argc, argv)) < 0) {
	return (-1);
    }
    /* method to get dbRead to skip the checksum */
    pdbBase = dbAllocBase();
    pdbBase->pvtSumFlag = 1;

    /* load ascii record type names */
    fp = fopen(inFile2,"r");
    if(!fp) return(-1);
    if(dbRead(pdbBase,fp)) return(-1);
    fclose(fp);
    dbRecType = pdbBase->precType;
    choiceRec = pdbBase->pchoiceRec;
    /* allocate space for the getSdrLine input structure */
    plineIn = dbCalloc(1,sizeof(struct lineIn));
    plineIn->fname = inFile;
    if ((inFile_fp = fopen(inFile, "r")) == NULL) {
	printf("bldRecChoice: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    pRecTypeList = dbCalloc(1,sizeof(SREC));
    pRecTypeList->papList = dbCalloc(1,dbRecType->number * sizeof(void *));
    pRecTypeList->number = dbRecType->number;
    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, FOUR_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF) {
	    break;
	}
	memset((char *)&inLineInfo,0, sizeof(LINEINFO));
	if (inScan(SCAN_CHOICELINES, plineIn, pInfo) < 0)
	    return (-1);
	if (!verifyType(dbRecType, &rec_no, pInfo))
	    return (-1);
	/* if no list for this record type - create it */
	if (!pRecTypeList->papList[rec_no]) {
	    pRecTypeList->papList[rec_no] = dbCalloc(1,sizeof(SLIST));
	}
	plist1 = (SLIST *) pRecTypeList->papList[rec_no];
	if (appendToSubList(plist1, pInfo->index_num + 1, pInfo) < 0)
	    return (-1);
    }
    if ((fclose(inFile_fp)) == EOF) {
	printf("bldRecChoice: fclose error. errno=%i\nfile=%s\n",
	       errno, inFile);
	return (-1);
    }
    /* calc total size */
    size1 = sizeof(struct choiceRec) + dbRecType->number * sizeof(void *);
    for (rec_no = 0; rec_no < pRecTypeList->number; rec_no++) {
	if (pRecTypeList->papList[rec_no]) {
	    plist1 = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(plist1);
	    max_inx = 0;
	    while (pt) {
		pSubList = (SUBLIST *) pt;
		max_inx = (max_inx < pSubList->index_num) ? pSubList->index_num : max_inx;
		pt = sllNext(pt);
	    }
	    size1 += sizeof(struct arrChoiceSet);
	    size1 += max_inx * sizeof(void *);
	    pt = sllFirst(plist1);
	    while (pt) {
		pSubList = (SUBLIST *) pt;
		pList2 = &pSubList->slist;
		pt2 = sllFirst(pList2);
		Ssize = 0;
		while (pt2) {
		    node = (ENODE *) pt2;
		    len = strlen(node->choice_str);
		    Ssize += len + 1;
		    /* make sure we have padding to long words */
		    Ssize = (Ssize + sizeof(long) - 1) / sizeof(long) * sizeof(long);
		    pt2 = sllNext(pt2);
		}
		size1 += Ssize;
		size1 += sizeof(struct choiceSet);
		size1 += pList2->count * sizeof(void *);
		pt = sllNext(pt);
	    }
	}
    }
    /* allocate contiguous area */
    choiceRec = dbCalloc(1,size1);
    addr1 = (char *)choiceRec + sizeof(struct choiceRec);
    choiceRec->number = dbRecType->number;
    choiceRec->papArrChoiceSet = addr1;
    addr1 = (char *)addr1 + choiceRec->number * sizeof(void *);
    for (rec_no = 0; rec_no < choiceRec->number; rec_no++) {
	if (pRecTypeList->papList[rec_no]) {
	    choiceRec->papArrChoiceSet[rec_no] = addr1;
	    addr1 = (char *)addr1 + sizeof(struct arrChoiceSet);
	    choiceRec->papArrChoiceSet[rec_no]->papChoiceSet = addr1;
	    plist1 = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(plist1);
	    max_inx = 0;
	    while (pt) {
		pSubList = (SUBLIST *) pt;
		max_inx = (max_inx < pSubList->index_num) ? pSubList->index_num : max_inx;
		pt = sllNext(pt);
	    }
	    choiceRec->papArrChoiceSet[rec_no]->number = max_inx;
	    addr1 = (char *)addr1 + max_inx * sizeof(void *);
	    plist1 = (SLIST *) pRecTypeList->papList[rec_no];
	    pt = sllFirst(plist1);
	    while (pt) {
		pSubList = (SUBLIST *) pt;
		index_num = pSubList->index_num - 1;
		choiceRec->papArrChoiceSet[rec_no]->papChoiceSet[index_num] = addr1;
		pList2 = &pSubList->slist;
		pt2 = sllFirst(pList2);
		addr1 = (char *)addr1 + sizeof(struct choiceSet);
		choiceRec->papArrChoiceSet[rec_no]->papChoiceSet[index_num]->number = pList2->count;
		choiceRec->papArrChoiceSet[rec_no]->papChoiceSet[index_num]->papChoice = addr1;
		addr1 = (char *)addr1 + pList2->count * sizeof(void *);
		k = 0;
		while (pt2) {
		    node = (ENODE *) pt2;
		    len = strlen(node->choice_str);
		    choiceRec->papArrChoiceSet[rec_no]->papChoiceSet[index_num]->papChoice[k] = addr1;
		    memcpy(
			  (char *)choiceRec->papArrChoiceSet[rec_no]->papChoiceSet[index_num]->papChoice[k],
			  node->choice_str,
			  len + 1);
		    addr1 = (char *)addr1 + len + 1;
		    addr1 = (void *)(((long)addr1 + sizeof(long) - 1) / sizeof(long) * sizeof(long));
		    k++;
		    pt2 = sllNext(pt2);
		}
		pt = sllNext(pt);
	    }
	}
    }
    /* change pointers offsets */
    for (i = 0; i < dbRecType->number; i++) {
	if (choiceRec->papArrChoiceSet[i] != NULL) {
	    for (j = 0; j < choiceRec->papArrChoiceSet[i]->number; j++) {
		if (choiceRec->papArrChoiceSet[i]->papChoiceSet[j] != NULL) {
		    for (k = 0; k < choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->number; k++) {
			if (choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k] != NULL) {
			    choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k] = (void *)
				((char *)choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k] - (char *)choiceRec);
			}
		    }
		    choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice = (void *)
			((char *)choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice - (char *)choiceRec);
		    choiceRec->papArrChoiceSet[i]->papChoiceSet[j] = (void *)
			((char *)choiceRec->papArrChoiceSet[i]->papChoiceSet[j] - (char *)choiceRec);
		}
	    }
	    choiceRec->papArrChoiceSet[i]->papChoiceSet = (void *)
		((char *)choiceRec->papArrChoiceSet[i]->papChoiceSet - (char *)choiceRec);
	    choiceRec->papArrChoiceSet[i] = (void *)
		((char *)choiceRec->papArrChoiceSet[i] - (char *)choiceRec);
	}
    }
    choiceRec->papArrChoiceSet = (void *)((char *)choiceRec->papArrChoiceSet - (char *)choiceRec);
    /* write output file */
    memset((char *)&sdrHeader,0, sizeof(sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = size1;
    sdrHeader.type = SDR_CHOICEREC;

    if ((outFile_fp = fopen(outFile, "w")) == NULL) {
	printf("bldRecChoice: fopen errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    if ((rtnval = fwrite(&sdrHeader, sizeof(sdrHeader), 1,
			 outFile_fp)) != 1) {
	printf("bldRecChoice: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile);
	return (-1);
    }
    if ((rtnval = fwrite(choiceRec, (int) size1, 1, outFile_fp)) != 1) {
	printf("bldRecChoice: fwrite error. errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile);
	return (-1);
    }
    if ((fclose(outFile_fp)) == EOF) {
	printf("bldRecChoice: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile);
	return (-1);
    }
    return (0);
}
/****************************************************************************
    Appends choice_str to correct index list
****************************************************************************/
appendToSubList(plist1, index_num, pInfo)
    SLIST          *plist1;
    int             index_num;
    LINEINFO       *pInfo;
{
    SUBLIST        *pSubList = NULL;
    SLIST          *pList2 = NULL;
    SNODE          *pt = NULL;
    int             found;
    int             i;
    if ((pt = sllFirst(plist1)) == NULL) {
	if (appendSubListToList(plist1, index_num) < 0)
	    return (-1);
    }
    found = FALSE;
    pt = sllFirst(plist1);
    for (i = 0; i < plist1->count; i++) {
	pSubList = (SUBLIST *) pt;
	if (pSubList->index_num == index_num) {
	    found = TRUE;
	    break;
	}
	pt = sllNext(pt);
    }
    if (!found) {
	if (appendSubListToList(plist1, index_num) < 0)
	    return (-1);
    }
    /* search for proper sublist index */
    pt = sllFirst(plist1);
    while (pt) {
	pSubList = (SUBLIST *) pt;
	if (pSubList->index_num == index_num) {
	    pList2 = &pSubList->slist;
	    appendEnodeToList(pList2, pInfo);
	    break;
	}
	pt = sllNext(pt);
    }
    return (0);
}
/****************************************************************************
    Appends choice_str node to sub list
****************************************************************************/
appendEnodeToList(pList2, pInfo)
    SLIST          *pList2;
    LINEINFO       *pInfo;
{
    int             len = 0;
    char           *ptrToMem;
    ENODE          *pNew;
    SNODE          *pt;
    pNew = dbCalloc(1,sizeof(ENODE));
    len = strlen(pInfo->choice_str);
    memcpy((char *)pNew->choice_str,(char *)pInfo->choice_str,len);

    pt = (SNODE *) pNew;
    sllAdd(pList2, pt);
    return (0);
}
/****************************************************************************
    Appends sublist and sets index_num
****************************************************************************/
/*
 * APPENDSUBLISTTOLIST
 */
appendSubListToList(plist1, index_num)
    SLIST          *plist1;
    int             index_num;
{
    SUBLIST        *pNew;
    char           *ptrToMem;
    SNODE          *pt;
    pNew = dbCalloc(1,sizeof(SUBLIST));
    pNew->index_num = index_num;

    pt = (SNODE *) pNew;
    sllAdd(plist1, pt);
    return (0);
}
/****************************************************************************
    process input line - (0,-1) success, failure
****************************************************************************/
/*
 * INSCAN
 */
inScan(scan_type, plineIn, pInfo)
    int             scan_type;
    struct lineIn  *plineIn;
    LINEINFO       *pInfo;
{
    int             err = 0;
    switch (scan_type) {
    case SCAN_CHOICELINES:
	if ((sscanf(plineIn->line, "\"%50[^\"]\" %i \"%50[^\"]\"",
		    pInfo->RecTyp_name,
		    &pInfo->index_num,
		    pInfo->choice_str)) != 3) {
	    err = 1;
	    break;
	}
	break;
    default:
	printf("bldRecChoice: inScan(scan_type) not defined internally\n");
	printf("bldRecChoice: line=%i\n", plineIn->line_no);
	return (-1);
	break;
    }
    if (err == 1) {
	printf("bldRecChoice: Illegal field format file=%s\n", plineIn->fname);
	printf("bldRecChoice: line=%i\n", plineIn->line_no);
	return (-1);
    }
    return (0);
}
/****************************************************************************
    Verify record type
****************************************************************************/
/*
 * VERIFYTYPE
 */
static int
verifyType(dbRecType, pRec_no, pInfo)
    struct recType *dbRecType;
    int            *pRec_no;
    LINEINFO       *pInfo;
{
    short           i;
    /* Verify record type name */
    for (i = 0; i < dbRecType->number; i++) {
	if (dbRecType->papName[i] != NULL) {
	    if (strcmp(dbRecType->papName[i],
		       pInfo->RecTyp_name) == SAME) {
		*pRec_no = i;
		return (TRUE);
	    }
	}
    }
    printf("bldRecChoice: verifyType - Record type '%s' not defined\n",
	   pInfo->RecTyp_name);
    return (FALSE);
}
/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_CHOICEREC
 */
cmd_choiceRec(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    char           *ptrToMem;
    struct stat    *buf;
    int             c,
                    err;
    short           aflg,
                    bflg,
                    oflg;
    aflg = bflg = oflg = err = 0;
    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:b:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldRecChoice: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'b':
	    bflg++;
	    strcpy(inFile2, optarg);
	    if (stat(inFile2, buf) == -1) {
		printf("bldRecChoice: Can't stat file=%s errno=%i\n",
		       inFile2, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile, optarg);
	    break;
	}
    if (!aflg || !bflg || !oflg || err) {
	printf("Usage: \n");
	printf(" <bldRecChoice -a choiceRec.ascii -b dbRecType.sdr -o choiceRec.sdr >\n");
	return (-1);
    }
    return (0);
}
