/*  $Id: bldSdrSum.c,v 1.3 1993/07/22 14:05:04 epics Exp $
 *	Author: Robert Zieman (ANL)
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  01-12-92        rcz     initial release
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <dbStaticLib.h>
#include <sdrHeader.h>

int             cmd_opts();

char            inFile[80];
char            inFile2[80];

struct sdrHeader sdrHeader;
struct sdrSum  *sdrSum;
/*
 * BLDSDRSUM.C
 *
 *  INPUT:	FILE 	default.sdrSum
 *  MODIFIED:	FILE 	<appl>.dctsdr
 *
 *         Creates and appends sdrSum structure to the <appl>.dctsdr
 *         file. (sdrHeader followed by structure sdrSum)
 *
 *         Note: default.sdrSum is created after all of the SDR file(s)
 *         and header files have been built by makeSdrMake.
 *         Default.sdrSum  contains the sum of all of the sdr and header files.
 *         makeSdrMake then invokes this tool to update the <appl>.dctsdr.
 *
 *         Note: getrel removes and re-links default.sdrSum
 *               iocInit will compare the contents of default.sdrSum
 *               with the sdrSum->allSdrSums string to determine
 *               if the database being loaded is out-of-date.
 */

main(argc, argv)
    char          **argv;
    int             argc;
{
    FILE           *inFile_fp;
    FILE           *inFile2_fp;
    int             len;

    /* check cmd line usage */
    if ((cmd_opts(argc, argv)) < 0)
	return -1;
    if ((inFile_fp = fopen(inFile, "r")) == NULL) {
	printf("bldSdrSum: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    sdrSum = dbCalloc(1,sizeof(struct sdrSum));
    if ((fscanf(inFile_fp, "%s", sdrSum->allSdrSums)) != 1) {
	printf("bldSdrSum: error reading file=%s  errno=%i", inFile, errno);
	return (-1);
    }
    len = strlen(sdrSum->allSdrSums);
    if ((fclose(inFile_fp)) == EOF) {
	printf("bldSdrSum: fclose error. errno=%i\nfile=%s\n",
	       errno, inFile);
	return (-1);
    }
    if ((inFile2_fp = fopen((char *) inFile2, "a")) == NULL) {
	printf("bldSdrSum: fopen error errno=%i\nfile=%s\n", errno, inFile2);
	return (-1);
    }
    /* init the sdrHeader header */
    memset((char *)&sdrHeader,0, sizeof(struct sdrHeader));
    sdrHeader.magic = DBMAGIC;
    sdrHeader.nbytes = len + 1;
    sdrHeader.type = SDR_ALLSUMS;
    /* write output file */
    if ((fwrite(&sdrHeader, sizeof(sdrHeader), 1,
			 inFile2_fp)) != 1) {
	printf("bldSdrSum: fwrite error. errno=%i \nfile=%s\n",
	       errno, inFile2);
	return (-1);
    }
    if ((fwrite(sdrSum->allSdrSums, len + 1, 1,
			 inFile2_fp)) != 1) {
	printf("bldSdrSum: fwrite error. errno=%i \nfile=%s\n",
	       errno, inFile2);
	return (-1);
    }
    if ((fclose(inFile2_fp)) == EOF) {
	printf("bldSdrSum: fclose error. errno=%i\nfile=%s\n", errno, inFile2);
	return (-1);
    }
    return (0);
}

/****************************************************************************
    Processes command line args
****************************************************************************/
/*
 * CMD_OPTS
 */
cmd_opts(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c;
    int             err = 0;
    short           aflg = 0;
    short           bflg = 0;
    buf = dbCalloc(1,sizeof(struct stat));
    while ((c = getopt(argc, argv, "a:b:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldSdrSum: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'b':
	    bflg++;
	    strcpy(inFile2, optarg);
	    if (stat(inFile2, buf) == -1) {
		printf("bldSdrSum: Can't stat file=%s errno=%i\n",
		       inFile2, errno);
		err++;
	    }
	    break;
	}
    if (!aflg || !bflg || err) {
	printf("Usage: \n");
	printf("    [ bldSdrSum -a default.sdrSum -b <appl>.dctsdr ]\n");
	return (-1);
    }
    return (0);
}
