static char SccsId[] = "$Id: comsubs.c,v 1.13 1995/04/26 16:22:16 mrk Exp $ ";
/*  $Id: comsubs.c,v 1.13 1995/04/26 16:22:16 mrk Exp $
/*  comsubs.c	1.6	5/30/91
 *	Original Author: Bob Dalesio (LANL)
 *	Date:	11-12-86
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

/*
 * COMSUBS.C
 *
 * These are library routines used by the Database Configuration Task
 *
 */
#include        <stdlib.h>
#include        <stddef.h>
#include        <stdio.h>
#include	<ctype.h>	/* SUN support for is* routines */
#include	<curses.h>	/* SUN support for cursor routines */
#include	<string.h>
#include	"keys.h"
#include	"dctDef.h"

/* extern variables */
extern short	suppress_errors;


/*
 * Subroutines
 *
 * bufcmp	compare two strings of specified length
 *	args
 *		pstr1	pointer to string 1
 *		pstr2	pointer to string 2
 *		length	number of bytes to compare
 *	returns
 *		1	string 1 greater than string 2
 *		-1	string 1 less than string 2
 *		0	equal
 * bufcpy	copy specified number of bytes from one buffer to another
 *	args
 *		pstr1	pointer to destination
 *		pstr2	pointer to source
 *		length	number of bytes to copy
 *	returns
 *		i	number of bytes copied
 * clear_error	clears the error message area at the bottom of the screen
 * error_msg		prints an error message on the bottom of the screen
 *	args
 *		index	error number
 * fill		fill an n character buffer with the specified character
 *	args
 *		pbuffer	pointer to the buffer to fill
 *		size	number of characters to fill
 *		char	character with which to fill buffer
 * getlinen	read a line from the keyboard given some chars are read
 *	args
 *		pstr	pointer to the input string
 *		lim	limit of characters to accept
 *		row	y position of cursor
 *	returns
 *		i	the number of characters read
 * getline	read a line from the keyboard
 *	args
 *		pstr	pointer to the input string
 *		lim	limit of characters to accept
 *		row	y position of cursor
 *		col	x position of cursor
 *	returns
 *		i	the number of characters read
 * get_field	parse a field from a string
 *	args
 *		pstr	pointer to the string
 *		pfield	pointer to the field
 * getinpch	get an input character
 *	returns
 *		cursor direction or a character
 * cinstr	find a character in a string
 *	args
 *		pstr		pointer to a string
 *		check_char	character to find
 *	returns
 *		-1	did not find the character
 *		i	position of the character in the string
 * isprintable		verify a character is printable
 *	args
 *		character	character to verify
 *	returns
 *		TRUE	is printable
 *		FALSE	is not printable
 * legal_flt_numbers	verifies a string contains legal floating point number
 *	args
 *		pstring	pointer to the ascii floating point number
 *	returns
 *		TRUE	a legal floating point number
 *		FALSE	illegal floating point ascii
 * legal_hex_numbers	verifies a string contains a legal hexadecimal value
 *	args
 *		pstring	pointer to the ascii hexadecimal number
 *	returns
 *		TRUE	a legal hexadecimal number
 *		FALSE	illegal hexadecimal number
 * legal_name		verifies a string  contains legal PV name characters
 *	args
 *		pname	pointer to a PV name
 *	returns
 *		TRUE	a legal PV name
 *		FALSE	not a legal PV name
 * legal_numbers verifies a string contains only legal ascii integer chars
 *	args
 *		pstring	pointer to the ascii integer
 *	returns
 *		TRUE	all legal ascii integer
 *		FALSE	invalid ascii integer specification
 * menu_select	selects an item from a menu on the CRT
 *	args
 *		y		last column of a menu entry
 *		no_items	number of items in the menu
 *		row		current item selected
 *		usepage		flags whether the caller wants to know about
 *				page up & down
 *	returns
 *		UP	page up
 *		DOWN	page down
 *		DONE	no selection made (abort)
 *		>=0	selected item
 * name_char	verifies that a character is legal in a PV name
 *	args
 *		check_char	character to check
 *	returns
 *		TRUE	is a legal PV name character
 *		FALSE	is not a legal PV name character
 * readline	read a line from a disk file
 *	args
 *		fd	file descriptor of file
 *		pinx	pointer to the file index
 *		pline	line read from disk file
 *	returns
 *		FALSE	no lines left
 *		TRUE	read a line
 * specified_by	uses an ascii string with (*) wildcard(s)
 *	args
 *		pspec	pointer to the string with wildcards
 *		ptest	pointer to the string to test
 *	returns
 *		TRUE	ptest is specified by pspec
 *		FALSE	ptest not specified by pspec
 * to_double	convert ascii to a floating point number
 *	args
 *		pstring	pointer to the ascii value
 *		pdouble	pointer to the converted value
 *	returns
 *		0	successful
 *		-1	failed
 * to_hshort	convert hexadecimal ascii to double integer
 *	args
 *		pstring	pointer to the hexadecimal string
 *		pval	pointer to the double integer
 *	returns
 *		0	successful
 *		-1	failed
 * to_short	convert ascii to a short decimal
 *	args
 *		pstr	pointer to the ascii number
 *
 *	returns
 *		num	the integer
 * to_value	convert ascii to a double integer
 *	args
 *		pstring	pointer to the ascii string
 *		pval	pointer to the converted value
 *	returns
 *		0	successful
 *		-1	failed
 */

/*+*****************************************************************************
*
* BUFCMP
*
* compare two strings of specified length
*
*-******************************************************************************
* MODIFICATION HISTORY
*	DATE:		PROGRAMMER:	COMMENTS:
*	11-12-86	Bob Dalesio	Initial Release
*******************************************************************************/bufcmp(pstr1,pstr2,length)
char	*pstr1;
char	*pstr2;
short	length;
{
	short	i;

	for(i=0; i<length; i++){
		if(*(pstr1+i) > *(pstr2+i)) return(1);
		if(*(pstr2+i) > *(pstr1+i)) return(-1);
	}

	return(0);
}

/*+*****************************************************************************
*
* BUFCPY
*
* copy one string to another
*
*-******************************************************************************
* MODIFICATION HISTORY
*	DATE:		PROGRAMMER:	COMMENTS:
*	11-12-86	Bob Dalesio	Initial Release
*******************************************************************************/bufcpy(pstr1,pstr2,length)
char	*pstr1;
char	*pstr2;
short	length;
{
	short	i;

	for (i=0; i<length; i++)
		*(pstr1+i) = *(pstr2+i);
	return(i);
}


static char *pError[] = {
	"value out of range",			/* 0 */
	"string not found",			/* 1 */
	"illegal value",			/* 2 */
	"bad vme channel number",		/* 3 */
	"operand needed",			/* 4 */
	"operator expected",			/* 5 */
	"umatched parens",			/* 6 */
	"bad branch specification",		/* 7 */
	"illegal expression element",		/* 8 */
	"only one ? can be used",		/* 9 */
	"only one : can be used",		/* 10 */
	"need a ? (if) to use the : (else)",	/* 11 */
	"bad field name specified",		/* 12 */
	"UNR table full - execute resolve",	/* 13 */
	"link to a PV in another node illegal",	/* 14 */
	"link to a display field illegal",	/* 15 */
	"bad crate specification",		/* 16 */
	"bad card specification",		/* 17 */
	"",					/* 18 */
	"bad signal specification",		/* 19 */
	"bad bit specification",		/* 20 */
	"",					/* 21 */
	"no available channels for card type",  /* 22 */
	"could not get card type from database",/* 23 */
	"could not open this record",		/* 24 */
	"",					/* 25 */
	"bad link specification",		/* 26 */
	"bad adapter specification",		/* 27 */
	"NULL pointer for choiceSet",		/* 28 */
	"Unexpected NULL pointer for choice",	/* 29 */
	"NULL choice table",			/* 30 */
	"NULL filled choice table",		/* 31 */
	"NULL pointer for array of choice sets", /* 32 */
	"Unknown field convert type",		/* 33 */
	"Unknown record type",			/* 34 */
	"NULL pointer to Record Field Descript", /* 35 */
	"Bad Prompt value",			/* 36 */
	"Could not create .pvd file",		/* 37 */
	"Could not init .pvd file",		/* 38 */
	"malloc() failed",			/* 39 */
	"seek error on .pvd file",		/* 40 */
	"write error on .pvd file",		/* 41 */
};

/*
 * ERROR_MSG
 *
 * print an error message at the bottom of the screen
 */
error_msg(index)
short	index;
{
	char	*ptr;		/* just a dummy to get size of a pointer */
	if (!suppress_errors){
		if (index <= sizeof(pError)/sizeof(ptr)){
			clear_error();
			move(ERRLINE,0);
			printw("%s",pError[index]);
			refresh();
		}
	}
}
/*
 *  error_msg_p
 *
 *  an interface to print_error_msg_p() that uses pError[]
 */
error_msg_p(pname, error_number, pause)
char	*pname;		/* name of caller */
short	error_number;	/* index into pError[] */
int	pause;		/* pause flag */
{
	if(error_number < sizeof(pError)/sizeof(char *))
		print_error_msg_p(pname, pError[error_number], pause);
}

/*
 * FILL
 *
 * fill a buffer with the designated character
 */
 fill(pbuffer,size,fillchar)
 char	*pbuffer;
 short	size;
 char	fillchar;
 {
 	short	i;

	for (i=0; i<size; i++){
		*pbuffer = fillchar;
		pbuffer++;
	}
}

/*
 * GETLINEN
 *
 * read a line from the keyboard where a portion is already entered
 */
getline1(pstr,lim,row,col)
char	*pstr;
short	lim;
short	row;
short	col;
{
	short	i;
	char		c;

	move(row,col);
	printw("%s",pstr);
	move(row,col+strlen(pstr));
	refresh();
	c = getinpch();
	for (i=strlen(pstr); (c != EOF) && (c != 0xa);){
		if (isprintable(c) && (i < lim)){
			*(pstr+i) = c;
			printw("%c",*(pstr+i++));
		}else if ((c == K_BACKSPACE) && (i > 0)){
			i--;
			move(row,col+i);
			printw(" "); 
			move(row,col+i);
		}else if ((c == K_DELETE) && (i > 0)){
			i--;
			move(row,col+i);
			printw(" "); 
			move(row,col+i);
		}
		refresh();
		c = getinpch();
	}
	*(pstr+i) = NULL;
	return(i);
}

/*
 * GETLINE
 *
 * read a line from the keyboard
 */
getline(pstr,lim,row,col)
char	*pstr;
short		lim;
short	row,col;
{
	short	i;
	char		c;

	move(row,col);
	refresh();
	c = getinpch();
	for (i=0; (c != EOF) && (c != 0xa);){
		if (isprintable(c) && (i < lim)){
/*			if(islower(c)) *(pstr+i) = toupper(c); 
			else
*/
			*(pstr+i) = c;
			printw("%c",*(pstr+i++));
		}else if ((c == K_BACKSPACE) && (i > 0)){
			i--;
			move(row,col+i);
			printw(" "); 
			move(row,col+i);
		}else if ((c == K_DELETE) && (i > 0)){
			i--;
			move(row,col+i);
			printw(" "); 
			move(row,col+i);
		}
		refresh();
		c = getinpch();
	}	
	*(pstr+i) = NULL;
	return(i);
}

/*
 * GET_FIELD
 *
 * parse a field from a string
 */
get_field(pstr,pfield)
char	*pstr,*pfield;
{
	while (isspace(*pstr) && (*pstr)) pstr++;
	while (!isspace(*pstr) && (*pstr != '.') && (*pstr))
		*pfield++ = *pstr++;
	*pfield = NULL;
	while (isspace(*pstr) && (*pstr)) pstr++;
}

/*
 * GETINPCH
 *
 * read an entry from the keyboard
 */
getinpch()
{
	char		input;
	input = getchar();
	while (TRUE){
		if (input == K_DIRECTION){
			input = getchar();
			if(input == 0x76) return(PAGE_UP);
			else if(input == 0x3c) return(FIRST);
			else if(input == 0x3e) return(LAST);
			if (input != K_DDIRECTION){
				ungetc(input,stdin);
				continue;
			}
			input = getchar();
 			if (input == K_NEXT){
				return(NEXT);
 			}else if (input == K_PREVIOUS){
				return(PREVIOUS);
 			}else if (input == K_NEXT_ITEM){
				return(NEXT_ITEM);
 			}else if (input == K_PREV_ITEM){
				return(PREV_ITEM);
 			}else if (input == K_LONG){
				input = getchar();
				input = getchar();
 				 getchar();
				if (input == K_FIRST) return(FIRST);
				else if (input == K_LAST) return(LAST);
				else if (input == K_PAGE_UP) return(PAGE_UP);
				else if (input == K_PAGE_DOWN) return(PAGE_DOWN);
				else if (input == K_EXIT) return(EXIT);
				else if (input == K_QUIT1) return(QUIT1);
				else if (input == K_QUIT2) return(QUIT2);
				else if (input == K_QUIT3) return(QUIT3);
				input = getchar();
				continue;
			}
/* */
		} else if(input==K_LF || input==K_CR) return(EXECUTE);
		else if (input == 05) return(EXIT);
		else if (input == 0x16) return(PAGE_DOWN);
		else if (input == 0x04) return(PAGE_DOWN);
		else if (input == 0x15) return(PAGE_UP);
/* */
		return(input);
	}
}

/*
 * INSTR
 *
 * find a character in a string
 */
cinstr(pstr,check_char)
char	*pstr;
char	check_char;
{
	short	i;

	i = 0;
	while (*pstr){
		if (*pstr == check_char) return(i);
		pstr++;
		i++;
	}
	return(-1);
}


char	print_chars[40] = "!@#$%^&*()_+|\\=-][';:/.,?><~`{} \"";
/*
 * ISPRINTABLE
 *
 * checks if a character is printable
 */
isprintable(character)
char	character;
{
	char 	*pchars;

	if (isalpha(character)) return(TRUE);
	if (isdigit(character)) return(TRUE);
	pchars = &print_chars[0];
	while (*pchars){
		if(*pchars == character) return(TRUE);
		pchars++;
	}
	return(FALSE);
}

/*
 * LEGAL_FLT_NUMBERS
 *
 * verifies all the string contains only legal numbers for a floating
 * point number
 */
legal_flt_numbers(pstring)
char	*pstring;
{
	short	got_decimal;
	short   got_e;

	got_decimal = got_e = FALSE;
	if ((*pstring == '+') || (*pstring == '-')) pstring++;
	while (isdigit(*pstring)){
		pstring++;
                if (*pstring == '.'){
                        if (got_decimal) return(FALSE);
                        else  got_decimal = TRUE;
                        pstring++;
                }
                if (*pstring == 'e'){
                        if (got_e) return(FALSE);
                        else  got_e = TRUE;
                        pstring++;      /* pass e */
                        /* pass sign of exponent */
                        if ((*pstring == '+') || (*pstring == '-')) pstring++;
                        else return(FALSE);
                }
	}

	if (*pstring != NULL) return(FALSE);
	return(TRUE);
}

/*
 * LEGAL_HEX_NUMBERS
 *
 * verifies all the string contains only legal hexadecimal numbers
 */
legal_hex_numbers(pstring)
char	*pstring;
{

	while (isdigit(*pstring)
	  || ((*pstring >= 'A') && (*pstring <= 'F'))
	  || ((*pstring >= 'a') && (*pstring <= 'f')))
		pstring++;

	if (*pstring != NULL) return(FALSE);
	return(TRUE);
}

/*
 *	LEGAL_NAME
 *
 *	verifies that the Process Variable name is legal
 */
legal_name(pname)
char	*pname;
{
        /* we are now allowing names to start with numbers or letters */
        /* so we need to eliminate the possibility that it is a number */
        /* before detemining if it is a legal name - this is a little */
        /* less efficient - but this program is not intended to live */
        /* forever.                                                     */
	if (legal_flt_numbers(pname)) return(FALSE);
	if ((!isalpha(*pname)) && (!isdigit(*pname)) ) return(FALSE);
	while(*pname){
		if (!name_char(*pname)) return(FALSE);
		pname++;
	}
	return(TRUE);
}

/*
 *      LEGAL_NAME_REF
 *
 *      verifies that this a Process Variable name as opposed to a number
 */
legal_name_ref(pname)
char    *pname;
{
        /* we are now allowing names to start with numbers or letters */
        /* so we need to eliminate the possibility that it is a number */
        /* before detemining if it is a legal name - this is a little */
        /* less efficient - but this program is not intended to live */
        /* forever.                                                     */
        if (legal_flt_numbers(pname)) return(FALSE);
        if ((!isalpha(*pname)) && (!isdigit(*pname)) ) return(FALSE);
        return(TRUE);
}

/*
 * LEGAL_NUMBERS
 *
 * verifies all the string contains only legal long integers
 */
legal_numbers(pstring)
char	*pstring;
{

	if ((*pstring == '+') || (*pstring == '-')) pstring++;
	while (isdigit(*pstring)) pstring++;

	if (*pstring != NULL) return(FALSE);
	return(TRUE);
}

#define RIGHT 1
#define MIDDLE 2
#define LEFT	4
/*
 * MENU_SELECT
 *
 * get a selected item from a menu
 */
menu_select(y,no_items,row,usepage)
short		y;
short	no_items;
short	row;
short		usepage;
{
	char		input;

	/* set the cursor */
	move(row,y);
	refresh();


	input = getinpch();
		
	while (1){


		switch (input){
		case (NEXT):		/* go to next menu item */
			row++;
			if (row > no_items) row = 0;
			break;
		case (PREVIOUS):
			row--;
			if (row < 0) row = no_items;
			break;
		case (QUIT1):
		case (QUIT2):
		case (QUIT3):
		case (EXIT):
			return(DONE);
		case (PAGE_UP):
			if (usepage) return(UP);
			break;
		case (PAGE_DOWN):
			if (usepage) return(DOWN);
			break;
		case (EXECUTE):
			return(row);
		}
		move(row,y);
		refresh();
		input = getinpch();
	}

}

/*
 * NAME_CHAR
 *
 *	verifies that a character is legal in a Process Variable Name
 */
char	name_chars[14] = "!@$_';:,~`{}";

name_char(check_char){
	char	*pchars;

	if (isalpha(check_char)) return(TRUE);
	if (isdigit(check_char)) return(TRUE);
	pchars = &name_chars[0];
	while (*pchars){
		if (*pchars == check_char) return(TRUE);
		pchars++;
	}
	return(FALSE);
}

/*
 * READLINE
 *
 * read a line from the disk file specified
 */
readline(fd,pinx,pline)
int	fd;
size_t	*pinx;
char	*pline;
{
	extern size_t	lseek();
	char		buffer[120];
	short	n,i;
	char	*pline_begin;

	pline_begin = pline;

	lseek(fd,*pinx,0);
	if ((n = read(fd,buffer,120)) == 0) return(FALSE);

	i = 0;
	while ((isprintable(buffer[i]) || buffer[i] == '\t') && (i < n))
		*pline++ = buffer[i++];
	*pline = NULL;

	while (!isprintable(buffer[i++]) && (i < n));
	*pinx += (i-1);

	if (pline_begin == pline)
		return(FALSE);
	return(TRUE);
}

/*
 * SPECIFIED_BY
 *
 * check if a string is specified by a string with wildcards
 */
specified_by(pspec,ptest)
char *pspec;
char *ptest;
{
	unsigned inx;
	short	wild_card_start;

	/* check if the specification begins with a wild card */
	if (*pspec == '*') wild_card_start = TRUE;
	else wild_card_start = FALSE;

	/* check for specification */
	while (TRUE){

		/* skip any wild cards */
		while (*pspec == '*') pspec++;

		/* find the specification chars to compare */
		inx = 0;
		while ( (*(pspec+inx) != '*') && (*(pspec+inx)) )
			inx++;

		/* check for specification ending with wildcard */
		if (inx == 0) return(TRUE);

		/* find the spec chars in the test string */
		while ( (strlen(ptest) >= inx)
		  && (bufcmp(ptest,pspec,inx) != SAME) ){

			/* check variable beginning */
			if (!wild_card_start) return(FALSE);
			else ptest++;
		}

		/* check segment found */
		if (strlen(ptest) < inx) return(FALSE);

		/* adjust pointers and wild card indication */
		wild_card_start = TRUE;
		ptest += inx;
		pspec += inx;

		/* check for end of specification */
		if (*pspec == NULL){
			if (*ptest == NULL) return(TRUE);
			else return(FALSE);
		}
	}
}

/*
 * TO_DOUBLE
 *
 * convert a string to a floating point value
 */
to_double(pstring,pdouble)
char	*pstring;
double	*pdouble;
{
	double	atof();

	if (!legal_flt_numbers(pstring))
		return(-1);
	*pdouble = atof(pstring);
	return(0);
}

/*
 * TO_HSHORT
 *
 * convert a hexadecimal string to an 16 bit value
 */
to_hshort(pstring,pval)
char	*pstring;
double	*pval;
{

	if (!legal_hex_numbers(pstring))
		return(-1);

	*pval = 0;
	while (*pstring != NULL){
		if (isdigit(*pstring))
			*pval = (16 * *pval) + (*pstring - '0');
		else{
			if (islower(*pstring)) *pstring = toupper(*pstring);
			*pval = (16 * *pval) + (*pstring - 'A' + 10);
		}
		pstring++;
	}
	return(0);
}

/*
 * TO_SHORT
 *
 * convert a string to a short
 */
to_short(pstr)
char	*pstr;
{
	short	num;

	num = 0;
	while (isdigit(*pstr)){
		num = (num * 10) + (*pstr - '0');
		pstr++;
	}
	return(num);
}


/*
 * TO_VALUE
 *
 * convert a string to a double (no fractional part)
 */
to_value(pstring,pval)
char	*pstring;
double	*pval;
{
	double	atof();

	if (!legal_numbers(pstring))
		return(-1);
	*pval = atof(pstring);

	return(0);
}
