
/* share/src/db $Id: dbManipulate.c,v 1.8 1995/02/28 19:22:40 mrk Exp $ */
/* dbManipulate.c
 *
 *
 * Process variable Directory and dbManipulate Functions
 *
 *	Original Authors: Bob Dalesio, Julie Sander (Original DCT)
 *	Second Author:   Frank Lenkszus (Changes to support extensible record support)
 *	Current Author:  Robert Zieman (Access Layer for .database files)
 *      Date:	05-22-91
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	03-19-92	mrk	Replace original pvd routines
 * .02	04-17-92	rcz	installed for mrk and removed 2 printf's
 * .03  05-22-92        rcz     Added dbManipulate routines
 * .04  07-20-92        rcz     revised header doc
 */

#ifndef vxWorks
#include	<stdio.h>
#include	<malloc.h>
#include	<memory.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#else
#include	<vxWorks.h>
#include	<stdioLib.h>
#include	<types.h>
#include	<ctype.h>
#include	<stat.h>
#endif


#include	<ellLib.h>
#include	<dbRecords.h>
#include	<dbRecType.h>
#include	<dbFldTypes.h>
#include	<dbDefs.h>
#include	<link.h>
#include	<string.h>
#include	<dbBase.h>
#include	<dbRecDes.h>
#include	"dbManipulate.h"
#include	<sdrHeader.h>
#include	<choice.h>

#ifdef __STDC__
long init_record(struct dbBase *pdbBase, void *precord, short recType);
#else
long init_record();
#endif


/************************** ENUM choices ************************************/
struct	wchoice {		/* a choice */
	short	val;		/* enumerated value */
	char	*pstr;		/* string for this value */
};


struct  wchoice  Nwchoices1[] = {
	0, "NPP",
	1, "PP",	
	2, NULL				/* this terminates the list */
};
struct  wchoice  Nwchoices2[] = {
	0, "NMS",
	1, "MS",
	2, NULL				/* this terminates the list */
};

struct wchoice *Nchoicearray[] = {      /* place index of choice set in
						prompt.inx field */
	Nwchoices1,			/* choice set index 0 */
	Nwchoices2			/* choice set index 1 */
};
static char     text[120];
#define 	HASH_NO	512		/* number of hash table entries */


/*The hash algorithm is a modification of the algorithm described in	*/
/* Fast Hashing of Variable Length Text Strings, Peter K. Pearson,	*/
/* Communications of the ACM, June 1990					*/
/* The modifications were desdigned by Marty Kraimer and Ben Chin Cha	*/
/* The mods were implemented and tested by Ben Chin Cha			*/

static unsigned char T0[256] = {
 39,159,180,252, 71,  6, 13,164,232, 35,226,155, 98,120,154, 69,
157, 24,137, 29,147, 78,121, 85,112,  8,248,130, 55,117,190,160,
176,131,228, 64,211,106, 38, 27,140, 30, 88,210,227,104, 84, 77,
 75,107,169,138,195,184, 70, 90, 61,166,  7,244,165,108,219, 51,
  9,139,209, 40, 31,202, 58,179,116, 33,207,146, 76, 60,242,124,
254,197, 80,167,153,145,129,233,132, 48,246, 86,156,177, 36,187,
 45,  1, 96, 18, 19, 62,185,234, 99, 16,218, 95,128,224,123,253,
 42,109,  4,247, 72,  5,151,136,  0,152,148,127,204,133, 17, 14,
182,217, 54,199,119,174, 82, 57,215, 41,114,208,206,110,239, 23,
189, 15,  3, 22,188, 79,113,172, 28,  2,222, 21,251,225,237,105,
102, 32, 56,181,126, 83,230, 53,158, 52, 59,213,118,100, 67,142,
220,170,144,115,205, 26,125,168,249, 66,175, 97,255, 92,229, 91,
214,236,178,243, 46, 44,201,250,135,186,150,221,163,216,162, 43,
 11,101, 34, 37,194, 25, 50, 12, 87,198,173,240,193,171,143,231,
111,141,191,103, 74,245,223, 20,161,235,122, 63, 89,149, 73,238,
134, 68, 93,183,241, 81,196, 49,192, 65,212, 94,203, 10,200, 47 
};

static unsigned char T1[256] = {
  9,139,209, 40, 31,202, 58,179,116, 33,207,146, 76, 60,242,124,
254,197, 80,167,153,145,129,233,132, 48,246, 86,156,177, 36,187,
 45,  1, 96, 18, 19, 62,185,234, 99, 16,218, 95,128,224,123,253,
 42,109,  4,247, 72,  5,151,136,  0,152,148,127,204,133, 17, 14,
182,217, 54,199,119,174, 82, 57,215, 41,114,208,206,110,239, 23,
189, 15,  3, 22,188, 79,113,172, 28,  2,222, 21,251,225,237,105,
102, 32, 56,181,126, 83,230, 53,158, 52, 59,213,118,100, 67,142,
220,170,144,115,205, 26,125,168,249, 66,175, 97,255, 92,229, 91,
214,236,178,243, 46, 44,201,250,135,186,150,221,163,216,162, 43,
 11,101, 34, 37,194, 25, 50, 12, 87,198,173,240,193,171,143,231,
111,141,191,103, 74,245,223, 20,161,235,122, 63, 89,149, 73,238,
134, 68, 93,183,241, 81,196, 49,192, 65,212, 94,203, 10,200, 47,
 39,159,180,252, 71,  6, 13,164,232, 35,226,155, 98,120,154, 69,
157, 24,137, 29,147, 78,121, 85,112,  8,248,130, 55,117,190,160,
176,131,228, 64,211,106, 38, 27,140, 30, 88,210,227,104, 84, 77,
 75,107,169,138,195,184, 70, 90, 61,166,  7,244,165,108,219, 51,
};

#ifdef __STDC__
static unsigned short hash( char *pname, int length)
#else
static unsigned short hash( pname, length)
char *pname;
int length;
#endif
{
    unsigned short h=0;
    unsigned short hret;
    unsigned char  *h0=(unsigned char *)&h;
    unsigned char  *h1= h0 + 1;
    int		i;

    for(i=0; i<length; i+=2, pname+=2) {
	h = h ^ (*(unsigned short *)pname);
	*h0 = T0[*h0];
	*h1 = T1[*h1];
    }
    hret = *h0;
    return(hret + *h1);
}
/* Convert various integer types to ascii */
static char digit_to_ascii[10]={'0','1','2','3','4','5','6','7','8','9'};
static void char_to_str(source,pdest)
	char source;
	char *pdest;
{
    unsigned char val,temp;
    char	  digit[3];
    int		  i,j;

    if(source==0) {
	*pdest++ = '0';
	*pdest = 0;
	return;
    }
    if(source<0) {
	*pdest++ = '-';
	if(source == -128) {
	    strcpy(pdest,"128");
	    return;
	}
	source = -source;
    }
    val = source;
    for(i=0; val!=0; i++) {
	temp = val/10;
	digit[i] = digit_to_ascii[val - temp*10];
	val = temp;
    }
    for(j=i-1; j>=0; j--) {
	*pdest++ = digit[j];
    }
    *pdest = 0;
    return;
}

static void uchar_to_str(source,pdest)
    unsigned char source;
    char	  *pdest;
{
    unsigned char val,temp;
    char	  digit[3];
    int		  i,j;

    if(source==0) {
	*pdest++ = '0';
	*pdest = 0;
	return;
    }
    val = source;
    for(i=0; val!=0; i++) {
	temp = val/10;
	digit[i] = digit_to_ascii[val - temp*10];
	val = temp;
    }
    for(j=i-1; j>=0; j--) {
	*pdest++ = digit[j];
    }
    *pdest = 0;
    return;
}
static void short_to_str(source,pdest)
    short source;
    char  *pdest;
{
    short val,temp;
    char  digit[6];
    int	  i,j;

    if(source==0) {
	*pdest++ = '0';
	*pdest = 0;
	return;
    }
    if(source<0) {
	*pdest++ = '-';
	if(source == -32768) {
	    strcpy(pdest,"32768");
	    return;
	}
	source = -source;
    }
    val = source;
    for(i=0; val!=0; i++) {
	temp = val/10;
	digit[i] = digit_to_ascii[val - temp*10];
	val = temp;
    }
    for(j=i-1; j>=0; j--) {
	*pdest++ = digit[j];
    }
    *pdest = 0;
    return;
}
static void ushort_to_str(source,pdest)
    unsigned short source;
    char	  *pdest;
{
    unsigned short val,temp;
    char	  digit[5];
    int		  i,j;

    if(source==0) {
	*pdest++ = '0';
	*pdest = 0;
	return;
    }
    val = source;
    for(i=0; val!=0; i++) {
	temp = val/10;
	digit[i] = digit_to_ascii[val - temp*10];
	val = temp;
    }
    for(j=i-1; j>=0; j--) {
	*pdest++ = digit[j];
    }
    *pdest = 0;
    return;
}
static void long_to_str(source,pdest)
    long source;
    char  *pdest;
{
    long val,temp;
    char  digit[11];
    int	  i,j;

    if(source==0) {
	*pdest++ = '0';
	*pdest = 0;
	return;
    }
    if(source<0) {
	*pdest++ = '-';
	if(source == 0x80000000) {
	    strcpy(pdest,"2147483648");
	    return;
	}
	source = -source;
    }
    val = source;
    for(i=0; val!=0; i++) {
	temp = val/10;
	digit[i] = digit_to_ascii[val - temp*10];
	val = temp;
    }
    for(j=i-1; j>=0; j--) {
	*pdest++ = digit[j];
    }
    *pdest = 0;
    return;
}

/*
 * F_TO_STR
 *
 * converts floating point numbers to NULL terminated strings
 * 
 * This routine was written to provide fast conversion.
 * Timing figures on a 68020
 * F_TO_STR         .316 msec
 * gcvt            2.6   msec
 * sprintf[%9.3E] 33     msecs
 *
 * This routine will call gcvt for any number with an absolute value
 * greater than or equal to 10,000,000.00
 *
 */
static void f_to_str(flt_value,pstr_value,precision)
	double	flt_value;
	char	*pstr_value;
	int	precision;
{
	unsigned short	got_one;
	double		place;
	short		number;
	char	*pfirst_digit;

	pfirst_digit = pstr_value;
	if (flt_value < 0){
		*pstr_value = '-';
		pstr_value++;
		pfirst_digit++;
		flt_value = -flt_value;
	}

	if (flt_value >= 10000000 || flt_value < 1e-7 ){
		gcvt(flt_value,20,pstr_value);
		return;
	}

	/* whole numbers */
	got_one = 0;
	for (place = 1000000; place >= 1; place /= 10){
		if (flt_value >= place){
			got_one = 1;
			number = flt_value / place;
			flt_value = flt_value - (number * place);
			*pstr_value = number + '0';
			pstr_value++;
		}else if (got_one){
			*pstr_value = '0';
			pstr_value++;
		}
	}
	if (!got_one){
		*pstr_value = '0';
		pstr_value++;
	}

	/* fraction */
	if (precision > 0){
		*pstr_value = '.';
		pstr_value++;
		for (place = .1; precision > 0; place /= 10, precision--){
			number = flt_value / place;
			flt_value = flt_value - (number * place);
			*pstr_value = number + '0';
			pstr_value++;
		}
	}
	*pstr_value = 0;

	/* rounding */
	if (flt_value >= (place * 5)){
		number = pstr_value - pfirst_digit;
		pstr_value--;
		while ((pstr_value >= pfirst_digit) && 
		  ((*pstr_value == '9') || (*pstr_value == '.'))){
			if (*pstr_value == '9')
				*pstr_value = '0';
			pstr_value--;
		}
		if (pstr_value < pfirst_digit){
			while (number >= 0){
				*(pfirst_digit + number + 1) = *(pfirst_digit + number);
				number--;
			}
			*pfirst_digit = '1';
		}else{
			*pstr_value += 1;
		}
	}
	return;
}
static void ulong_to_str(source,pdest)
    unsigned long source;
    char	  *pdest;
{
    unsigned long val,temp;
    char	  digit[10];
    int		  i,j;

    if(source==0) {
	*pdest++ = '0';
	*pdest = 0;
	return;
    }
    val = source;
    for(i=0; val!=0; i++) {
	temp = val/10;
	digit[i] = digit_to_ascii[val - temp*10];
	val = temp;
    }
    for(j=i-1; j>=0; j--) {
	*pdest++ = digit[j];
    }
    *pdest = 0;
    return;
}

/****************************************************************************
 *
 *  DBFINDRECORD
 *
 *  this routine searches for a record named:
 *		"recordName" or "recordName.fieldName"
 * 
 *	if record is found sets pdbEntry->precord, and pdbEntry->record_type
 *	if the fieldName extension is included it calls dbFindField also
 *
 *
 * returns:	0,-1	success,fail
 *
 ***************************************************************************/
#ifdef __STDC__
long dbFindRecord(struct dbBase * pdbBase, char *precName, DBENTRY * pdbEntry)
#else
long            dbFindRecord(pdbBase, precName, pdbEntry)
    struct dbBase  *pdbBase;
    char           *precName;
    DBENTRY        *pdbEntry;
#endif
{
    unsigned short  hashInd;
    int             lenName;
    ELLLIST          **ppvd = (ELLLIST **) pdbBase->ppvd;
    ELLLIST           *pvdlist;
    PVDENTRY       *ppvdNode;
    int             fldnameFlag = 0;
    short           i;
    char            convName[PVNAME_SZ + 1];
    char           *pconvName = &convName[0];
    struct recTypDes *precTypDes;
    struct recDes  *precDes = pdbBase->precDes;
    if (ppvd == NULL) {
	ppvd = (ELLLIST **) calloc(HASH_NO, sizeof(ELLLIST *));
	pdbBase->ppvd = (void *) ppvd;
    }
    if ( !pdbBase || !precDes ) {
	sprintf(text, "dbFindRecord: can't access dbBase");
	errMessage(-1L, text);
	return (-1);
    }
    /* clear callers entry */
    memset((char*)pdbEntry, '\0', sizeof(DBENTRY));
    /* convert the record name */
    lenName = 0;
    while (*precName && (*precName != '.') && (lenName < PVNAME_SZ)) {
	*pconvName = *precName;
	precName++;
	pconvName++;
	lenName++;
    }
    *pconvName = 0;
    if (*precName++ == '.') 
	fldnameFlag = 1;
    pconvName = &convName[0];
    hashInd = hash(pconvName, lenName);
    if (ppvd[hashInd] == NULL)
	return (-1);
    pvdlist = ppvd[hashInd];
    for (ppvdNode = (PVDENTRY *) ellFirst(pvdlist); ppvdNode != NULL;
	    ppvdNode = (PVDENTRY *) ellNext((ELLNODE*)ppvdNode)) {
	if (strcmp(pconvName, (char *) (ppvdNode->precord)) == 0) {
	    if((precTypDes = GET_PRECTYPDES(precDes, ppvdNode->record_type)) == NULL) {
	        return(-1);
	    }
	    pdbEntry->precord = ppvdNode->precord;
	    pdbEntry->record_type = ppvdNode->record_type;
	    pdbEntry->precTypDes = precTypDes;
	    if (fldnameFlag) {
		/* Note: precName points to beg of field name here */
	        if ( (dbFindField(pdbBase, precName , pdbEntry)) != 0) {
	            /* clear callers entry */
	            memset((char*)pdbEntry, '\0', sizeof(DBENTRY));
	            return (-1);
	        }
	    }
	    return (0);
	}
    }
    return (-1);
}

/****************************************************************************
 *
 *  DBFINDFIELD
 *
 *  this routine searches for a field in a record and
 *  sets pfield and pfldDes in pdbEntry
 *
 *  depends upon dbFindRecord called to initially set up pdbEntry
 *  for this record
 *  ALGO:
 *
 *	clears pdbEntry->pfield and pdbEntry->pfldDes
 *	tests pdbEntry->precord ...
 *
 *	if pfieldName is NULL use default field name VAL
 *	if field is found, set pdbEntry->pfield, and pdbEntry->pfldDes
 * 
 *  returns 0,-1	success, fail
 *
 ***************************************************************************/
#ifdef __STDC__
long dbFindField(struct dbBase * pdbBase, char *pfieldName, DBENTRY * pdbEntry)
#else
long            dbFindField(pdbBase, pfieldName, pdbEntry)
    struct dbBase  *pdbBase;
    char           *pfieldName;
    DBENTRY        *pdbEntry;
#endif
{
    short           i;
    struct recDes  *precDes = pdbBase->precDes;
    short           type = pdbEntry->record_type;
    struct fldDes  *pfldDes;
    struct recTypDes *precTypDes;
    short           top,
                    bottom,
                    test;
    char            convName[FLDNAME_SZ];
    char           *pconvName = &convName[0];
    unsigned long  *sortFldName;
    short          *sortFldInd;

    /* assume if pdbEntry->precord set - dbFindRecord was called */
    if ( !pdbBase || !precDes ) {
	sprintf(text, "dbFindField: can't access dbBase");
	errMessage(-1L, text);
	return (-1);
    }
    /* verify that recDes and record are available */
    if (!precDes || !pdbEntry->precord) return (-1);
    /* verify that type is valid			 */
    if ((type < 0) || (type >= precDes->number)) return (-1);

    /* read the record type definitions for this record */
    if ((precTypDes = GET_PRECTYPDES(precDes, type)) == NULL) {
	return (-1);
    }
    sortFldName = precTypDes->sortFldName;
    sortFldInd = precTypDes->sortFldInd;
    /* blank pad the field name */
    i = 0;
    while (*(pfieldName + i) && (i < FLDNAME_SZ)) {
	*(pconvName + i) = *(pfieldName + i);
	i++;
    }
    while (i < FLDNAME_SZ) {
	*(pconvName + i) = ' ';
	i++;
    }
    /* check for default field name to be supplied */
    if (*pconvName == ' ') {
	memcpy(pconvName, "VAL ", FLDNAME_SZ);
    }

    /* binary search through ordered field names */
    top = precTypDes->no_fields - 1;
    bottom = 0;
    test = (top + bottom) / 2;
    while (1) {
	/* check the field name */
	if (sortFldName[test] == *(unsigned long *) pconvName) {
	    if(!(pfldDes=GET_PFLDDES(precTypDes,sortFldInd[test]))) return(-1);
	    pdbEntry->pfldDes = pfldDes;
	    pdbEntry->pfield = (char *) (pdbEntry->precord) + pfldDes->offset;
	    return (0);
	} else if (sortFldName[test] > *(unsigned long *) pconvName) {
	    top = test - 1;
	    if (top < bottom)
		return (-1);
	    test = (top + bottom) / 2;
	} else {
	    bottom = test + 1;
	    if (top < bottom)
		return (-1);
	    test = (top + bottom) / 2;
	}
    }
}

/****************************************************************************
 *
 *  DBCREATERECORD
 *
 *  this routine creates a new init'd record.
 *  Note: record must not already exist in memory db.
 *  returns 0,-1	OK,failure
 *
 ***************************************************************************/
#ifdef __STDC__
long dbCreateRecord(struct dbBase *pdbBase, int recType, char *precordName, DBENTRY *pdbEntry)
#else
long dbCreateRecord(pdbBase, recType, precordName, pdbEntry)
    struct dbBase  *pdbBase;
    int             recType;
    char           *precordName;
    DBENTRY        *pdbEntry;
#endif
{
    ELLLIST          **ppvd = (ELLLIST **) pdbBase->ppvd;
    PVDENTRY       *ppvdNode = NULL;
    ELLLIST           *ppvdlist = NULL;
    ELLLIST           *preclist = NULL;
    RECNODE        *precNode = NULL;
    RECNODE        *pNewRecNode = NULL;
    void           *pNewRecord = NULL;
    struct recDes  *precDes = pdbBase->precDes;
    struct recTypDes *precTypDes = NULL;
    struct recHeader *precHeader = (struct recHeader *) pdbBase->precHeader;
    struct recLoc  *precLoc = NULL;
    struct recType *precType = (struct recType *) pdbBase->precType;
    char            convName[PVNAME_SZ + 1];
    char           *pconvName = &convName[0];
    int             lenName;
    short           rec_size;
    unsigned short  hashInd;
    int             inserted;
    long            size;
    /* clear callers entry */
    memset((char *) pdbEntry, '\0', sizeof(DBENTRY));

    if (!pdbBase || !precDes) {
	sprintf(text, "dbCreateRecord: can't access dbBase");
	errMessage(-1L, text);
	return (-1);
    }
    /* verify that recType is valid			 */
    if ((recType < 0) || (recType >= precDes->number))
	return (-1);
    /* read the record recType definitions for this record */
    if ((precTypDes = GET_PRECTYPDES(precDes, recType)) == NULL)
	return (-1);
    if ((dbFindRecord(pdbBase, precordName, pdbEntry)) == 0) {
	sprintf(text, "dbCreateRecord: record (%s) - already exists", precordName);
	errMessage(-1L, text);
	return (-1);
    }
    /* setup the PVD */
    if (ppvd == NULL) {
	ppvd = (ELLLIST **) calloc(HASH_NO, sizeof(ELLLIST *));
	pdbBase->ppvd = (void *) ppvd;
    }
    /* convert the record name */
    lenName = 0;
    while (*precordName && (*precordName != '.') && (lenName < PVNAME_SZ)) {
	*pconvName = *precordName;
	precordName++;
	pconvName++;
	lenName++;
    }
    *pconvName = 0;
    pconvName = &convName[0];
    hashInd = hash(pconvName, lenName);
    if (ppvd[hashInd] == NULL) {
	ppvd[hashInd] = (ELLLIST *) calloc(1, sizeof(ELLLIST));
	ellInit(ppvd[hashInd]);
    }
    ppvdlist = ppvd[hashInd];
    /* make sure that it does not already exist */
    for (ppvdNode = (PVDENTRY *) ellFirst(ppvdlist); ppvdNode != NULL;
	    ppvdNode = (PVDENTRY *) ellNext((ELLNODE*)ppvdNode)) {
	if (strcmp(pconvName, (char *) (ppvdNode->precord)) == 0) {
	    return (-1);
	}
    }
    ppvdNode = (PVDENTRY *) calloc(1, (unsigned) sizeof(PVDENTRY));
    ellAdd(ppvdlist, (ELLNODE*)ppvdNode);

    /* create any missing structures and list nodes*/
    if (precHeader == NULL) {
	size = sizeof(struct recHeader) + (precType->number * sizeof(char *));
	if ((precHeader = (struct recHeader *)
	     calloc((unsigned) 1, (unsigned) size)) == NULL) {
	    sprintf(text, "dbCreateRecord: calloc error ");
	    errMessage(-1L, text);
	    return (-1);
	}
	precHeader->papRecLoc = (void *)( precHeader + 1);
	precHeader->number = precType->number;
	pdbBase->precHeader = precHeader;
    }
    if (precHeader->papRecLoc[recType] == NULL) {
	if ((precHeader->papRecLoc[recType] = (struct recLoc *)
	  calloc((unsigned) 1, (unsigned) sizeof(struct recLoc))) == NULL) {
	    sprintf(text, "dbCreateRecord: calloc error ");
	    errMessage(-1L, text);
	    return (-1);
	}
	precHeader->papRecLoc[recType]->rec_size = precTypDes->rec_size;
	precHeader->papRecLoc[recType]->record_type = recType;
    }
    precLoc = precHeader->papRecLoc[recType];
    if ((precLoc->preclist) == NULL) {
	/* allocate new list for this record type */
	if ((precLoc->preclist = (ELLLIST *)
	     calloc((unsigned) 1, (unsigned) sizeof(ELLLIST))) == NULL) {
	    printf("dbCreateRecord:  calloc error\n");
	    return (-1);
	}
	ellInit(precLoc->preclist);
    }
    rec_size = precTypDes->rec_size;
    /* setup the sorted record list */
    /* create a new record of this record type */
    if ((pNewRecord = (void *)
	 calloc((unsigned) 1, (unsigned) rec_size)) == NULL) {
	printf("dbCreateRecord:  calloc error\n");
	return (-1);
    }
    strcpy((char*)pNewRecord, (char *) pconvName);
    /* give it the default values */
    if ((init_record(pdbBase, pNewRecord, recType)) < 0) {
	sprintf(text, "dbCreateRecord: init_record (%s) failed ", pconvName);
	errMessage(-1L, text);
	return (-1);
    }
    /* make sure record doesn't already exist in this record list */
    preclist = precLoc->preclist;
    for (precNode = (RECNODE *) ellFirst(preclist);
	    precNode != NULL;
	    precNode = (RECNODE *) ellNext((ELLNODE*)precNode)) {
	if (precNode->precord == pNewRecord) {
	    return (-1);
	}
    }
    /* create a recNode */
    if ((pNewRecNode = (RECNODE *)
	 calloc((unsigned) 1, (unsigned) sizeof(RECNODE))) == NULL) {
	printf("dbCreateRecord:  calloc error\n");
	return (-1);
    }
    pNewRecNode->precord = pNewRecord;
    /* install record node in list in sorted postion */
    inserted = 0;
    for (precNode = (RECNODE *) ellFirst(preclist);
	    precNode != NULL;
	    precNode = (RECNODE *) ellNext((ELLNODE*)precNode)) {
	if ((strcmp(pconvName, (char*)precNode->precord)) < 0) {
	    precNode = (RECNODE *) ellPrevious((ELLNODE*)precNode);
	    ellInsert(preclist, (ELLNODE*)precNode, (ELLNODE*)pNewRecNode);
	    inserted = 1;
	    break;
	}
    }
    if (!inserted) {
	ellAdd(preclist, (ELLNODE*)pNewRecNode);
    }
    ppvdNode->precord = pNewRecord;
    ppvdNode->record_type = recType;
    pdbEntry->precord = ppvdNode->precord;
    pdbEntry->record_type = ppvdNode->record_type;
    pdbEntry->precTypDes = precTypDes;
    return (0);
}

/****************************************************************************
 *
 *  init_record
 *
 *   this routine initializes the field values for a record
 *
 ***************************************************************************/
#ifdef __STDC__
long init_record(struct dbBase *pdbBase, void *precord, short recType)
#else
long init_record(pdbBase, precord, recType)
    struct dbBase  *pdbBase;    /* pointer to dbBase structure */
    void *precord;
    short           recType;       /* type of process variable       */
#endif
{
    struct recDes  *precDes = pdbBase->precDes;
    struct recTypDes *precTypDes;	/* current record type definition
					 * structure */
    struct fldDes  *pfldDes;
    int             i;
    char           *iptr = (char*)precord;
    char           *ptr;

    if ( !pdbBase || !precDes ) {
	sprintf(text, "init_record: can't access dbBase");
	errMessage(-1L, text);
	return (-1);
    }
    /* verify that type is valid			 */
    if ((recType < 0) || (recType >= precDes->number)) return (-1);
    /* read the record type definitions for this record */
    if ((precTypDes = GET_PRECTYPDES(precDes, recType)) == NULL) return (-1);

    for (i = 0; i < precTypDes->no_fields; i++) {
	pfldDes = GET_PFLDDES(precTypDes, i);
	ptr = iptr + pfldDes->offset;
	switch (pfldDes->field_type) {
	case (DBF_CHAR):
	    *(char *) ptr = pfldDes->initial.char_value;
	    break;
	case (DBF_UCHAR):
	    *(unsigned char *) ptr =
		pfldDes->initial.uchar_value;
	    break;
	case (DBF_SHORT):
	    *(short *) ptr = pfldDes->initial.short_value;
	    break;
	case (DBF_USHORT):
	    *(unsigned short *) ptr =
		pfldDes->initial.ushort_value;
	    break;
	case (DBF_LONG):
	    *(long *) ptr = pfldDes->initial.long_value;
	    break;
	case (DBF_ULONG):
	    *(unsigned long *) ptr =
		pfldDes->initial.ulong_value;
	    break;
	case (DBF_FLOAT):
	    *(float *) ptr = pfldDes->initial.float_value;
	    break;
	case (DBF_DOUBLE):
	    *(double *) ptr = pfldDes->initial.double_value;
	    break;
	case (DBF_GBLCHOICE):
	case (DBF_CVTCHOICE):
	case (DBF_RECCHOICE):
	case (DBF_DEVCHOICE):
	case (DBF_ENUM):
	    *(unsigned short *) ptr =
		pfldDes->initial.enum_value;
	    break;
	case (DBF_INLINK):
	case (DBF_OUTLINK):
	case (DBF_FWDLINK):
	    ((struct link *) ptr)->value.value = 0.0;
	    break;
	}
    }
    return(0);
}

/****************************************************************************
 * DBDELETERECORD
 *
 *      removes an existing record from the memory resident database.
 *
 * returns:		0,-1	OK,ERROR
 ****************************************************************************/

#ifdef __STDC__
long dbDeleteRecord(struct dbBase *pdbBase, DBENTRY *pdbEntry)
#else
long dbDeleteRecord(pdbBase, pdbEntry)
    struct dbBase  *pdbBase;
    DBENTRY        *pdbEntry;
#endif
{
    ELLLIST       **ppvd = (ELLLIST **) pdbBase->ppvd;
    ELLLIST        *pvdlist;
    PVDENTRY       *ppvdNode;
    ELLLIST        *preclist;
    RECNODE        *precNode;
    char           *precordName;
    int             lenName;
    unsigned short  hashInd;
    int             found;

    if (!pdbBase || !ppvd ) return (-1);
    if (!(preclist = GET_PRECLIST(pdbBase, pdbEntry))) return (-1);
    if (!(ppvdNode = GET_PVDNODE(pdbEntry)))return (-1);
    precordName = (char *) pdbEntry->precord;
    found = 0;
    /* find the sorted list node */
    for (precNode = (RECNODE *) ellFirst(preclist); precNode != NULL;
            precNode = (RECNODE *) ellNext((ELLNODE *) precNode)) {
        if (strcmp(precordName, (char *) (precNode->precord)) == 0) {
            found = 1;
            break;
        }
    }
    if (!found) return (-1);
    /* get pvd entry */
    lenName = strlen(precordName);
    hashInd = hash(precordName, lenName);
    if (ppvd[hashInd] == NULL) return (-1);
    pvdlist = ppvd[hashInd];
    if (pvdlist == NULL) return (-1);

    free((char*)precNode->precord);
    ellDelete(preclist, (ELLNODE*)precNode);
    if (ellCount(preclist) == 0) {
        free((char *) preclist);
    }
    ellDelete(pvdlist, (ELLNODE*)ppvdNode);
    return (0);
}


/****************************************************************************
 * DBFREEMEM
 *
 *  removes the memory db
 *
 * returns:		0,-1	OK,ERROR
 ****************************************************************************/
#ifdef __STDC__
long dbFreeMem(struct dbBase **ppdbBase)
#else
long dbFreeMem(ppdbBase)
struct dbBase **ppdbBase;
#endif
{
    struct dbBase  *pdbBase = (struct dbBase *) * ppdbBase;
    ELLLIST          **ppvd = (ELLLIST **) pdbBase->ppvd;
    ELLLIST           *ppvdlist;
    struct recType *precType = (struct recType *) pdbBase->precType;
    struct recLoc  *precLoc;
    struct recHeader *precHeader = (struct recHeader *) pdbBase->precHeader;
    PVDENTRY       *ppvdNode;
    RECNODE        *precNode;
    ELLLIST           *preclist;
    int             i;
    int             recType;
    if (!pdbBase || !ppvd || !precType)
	return (-1);
    /* loop thru the pvd lists and remove nodes then list */
    for (i = 0; i < HASH_NO; i++) {
	ppvdlist = ppvd[i];
	if (ppvdlist == NULL)
	    continue;
	for (ppvdNode = (PVDENTRY *) ellFirst(ppvdlist); ppvdNode != NULL;
		ppvdNode = (PVDENTRY *) ellNext((ELLNODE *) ppvdNode)) {
	    /* remove each record and pvd node */
	    free((char *) ppvdNode->precord);
	    ellDelete(ppvdlist, (ELLNODE *) ppvdNode);
	}
	if (ellCount(ppvdlist) == 0) {
	    ppvd[i] = NULL;
	    free((char *) ppvdlist);
	}
    }
    /* loop thru the recLocs - removing lists then recLoc only */
    for (recType = 0; recType < precType->number; recType++) {
	if (!(precLoc = GET_PRECLOC(precHeader, recType)))
	    continue;
	preclist = precLoc->preclist;
	if (preclist == NULL)
	    continue;
	for (precNode = (RECNODE *) ellFirst(preclist); precNode != NULL;
		precNode = (RECNODE *) ellNext((ELLNODE *) precNode)) {
	    /* remove each record node */
	    ellDelete(preclist, (ELLNODE *) precNode);
	}
	if (ellCount(preclist) == 0) {
	    free((char *) preclist);
	    free((char *) precLoc);
	}
    }
    /* free the rest of the memory allocations */
    if (pdbBase->pchoiceCvt)
	free((char *) pdbBase->pchoiceCvt);
    if (pdbBase->pchoiceDev)
	free((char *) pdbBase->pchoiceDev);
    if (pdbBase->pchoiceGbl)
	free((char *) pdbBase->pchoiceGbl);
    if (pdbBase->pchoiceRec)
	free((char *) pdbBase->pchoiceRec);
    if (pdbBase->pcvtTable)
	free((char *) pdbBase->pcvtTable);
    if (pdbBase->pdbName)
	free((char *) pdbBase->pdbName);
    if (pdbBase->pdbPvd)
	free((char *) pdbBase->pdbPvd);
    if (pdbBase->pdrvSup)
	free((char *) pdbBase->pdrvSup);
    if (pdbBase->ppvd)
	free((char *) pdbBase->ppvd);
    if (pdbBase->precDes)
	free((char *) pdbBase->precDes);
    if (pdbBase->precDevSup)
	free((char *) pdbBase->precDevSup);
    if (pdbBase->precHeader)
	free((char *) pdbBase->precHeader);
    if (pdbBase->precSup)
	free((char *) pdbBase->precSup);
    if (pdbBase->precType)
	free((char *) pdbBase->precType);
    if (pdbBase->psdrSum)
	free((char *) pdbBase->psdrSum);

    /* finally */
    free((char *) pdbBase);
    *ppdbBase = 0;
    return (0);
}


/****************************************************************************
 * DBGETCHOICES
 *
 *      copies a set of choices to pdbChoices
 *
 *      depends upon complete pdbEntry setup
 *
 * returns:		0,-1	OK,ERROR
 ****************************************************************************/
#ifdef __STDC__
long dbGetChoices(struct dbBase *pdbBase,DBENTRY *pdbEntry, DBCHOICES *pdbChoices)
#else
long dbGetChoices(pdbBase, pdbEntry, pdbChoices)
struct dbBase *pdbBase;
DBENTRY *pdbEntry;
DBCHOICES *pdbChoices;
#endif
{
    struct fldDes  *pfldDes = pdbEntry->pfldDes;
    short           record_type = pdbEntry->record_type;
    short           i, num, len;
    struct choiceSet *pchoiceSet = NULL;
    struct arrChoiceSet *parrChoiceSet;
    struct devChoice *pdevChoice;
    struct devChoiceSet *pdevChoiceSet;
    if ( !pdbBase || !pdbEntry->precord || !pfldDes) return (-1);
    memset((char*)pdbChoices, '\0', sizeof(DBCHOICES));
    switch (pfldDes->field_type) {
    case DBF_GBLCHOICE:
	pchoiceSet = GET_PCHOICE_SET(pdbBase->pchoiceGbl, pfldDes->choice_set);
	goto choice_common;
    case DBF_CVTCHOICE:
	pchoiceSet = pdbBase->pchoiceCvt;
	goto choice_common;
    case DBF_RECCHOICE:
	parrChoiceSet = GET_PARR_CHOICE_SET(pdbBase->pchoiceRec, record_type);
	pchoiceSet = GET_PCHOICE_SET(parrChoiceSet, pfldDes->choice_set);
choice_common:
	if (pchoiceSet == NULL) return (-1);
	break;
    case DBF_DEVCHOICE:
	break;
    default:
	return (-1);
    }
    num = 0;
    switch (pfldDes->field_type) {
    case DBF_GBLCHOICE:
    case DBF_RECCHOICE:
    case DBF_CVTCHOICE:
	for (i = 0; i < pchoiceSet->number; i++) {
	    if ((pchoiceSet->papChoice == NULL)
	        || (pchoiceSet->papChoice[i] == NULL)) continue;
	    len = strlen(pchoiceSet->papChoice[i]);
	    if ( len >= MAX_STRING_SIZE || num >= DB_MAX_CHOICES ) continue;
	    strcpy((char*)&pdbChoices->choice[num][0],pchoiceSet->papChoice[i]);
	    pdbChoices->nchoices += 1;
	    num++;
	}
	break;
    case DBF_DEVCHOICE:
	pdevChoiceSet = GET_PDEV_CHOICE_SET(pdbBase->pchoiceDev, record_type);
	if (pdevChoiceSet == NULL) return (-1);
	for (i = 0; i < pdevChoiceSet->number; i++) {
	    pdevChoice = GET_DEV_CHOICE(pdevChoiceSet, i);
	    if (pdevChoice == NULL || pdevChoice->pchoice == NULL) continue;
	    len = strlen(pdevChoice->pchoice);
	    if ( len >= MAX_STRING_SIZE || num >= DB_MAX_CHOICES ) continue;
	    strcpy((char*)&pdbChoices->choice[num][0],pdevChoice->pchoice);
	    pdbChoices->nchoices += 1;
	    num++;
	}
	break;
    default:
	break;
    }
    return (0);
}

/****************************************************************************
 * DBGETSTRING
 *
 *      converts a db field to a string and places it into callers
 *      buffer  (pstring)
 *
 *      depends upon complete pdbEntry setup
 *
 * returns:		0,-1	OK,ERROR
 ****************************************************************************/
#ifdef __STDC__
long dbGetString(struct dbBase *pdbBase,DBENTRY *pdbEntry,char *pstring)
#else
long dbGetString(pdbBase, pdbEntry, pstring)
struct dbBase *pdbBase;
DBENTRY *pdbEntry;
char *pstring;
#endif
{
    struct fldDes  *pfldDes = pdbEntry->pfldDes;
    short           size = pfldDes->size;
    short           sizeto;
    char           *pcharsrc = (char *) pdbEntry->pfield;
    unsigned char  *pucharsrc = (unsigned char *) (pdbEntry->pfield);
    short          *pshortsrc = (short *) (pdbEntry->pfield);
    unsigned short *pushortsrc = (unsigned short *) (pdbEntry->pfield);
    long           *plongsrc = (long *) (pdbEntry->pfield);
    unsigned long  *pulongsrc = (unsigned long *) (pdbEntry->pfield);
    float          *pfloatsrc = (float *) (pdbEntry->pfield);
    double         *pdoublesrc = (double *) (pdbEntry->pfield);
    DBENTRY         entry;
    short           precision;
    short           choice_set = pfldDes->choice_set;
    short           choice_ind;
    char           *pchoice;
    struct choiceSet *pchoiceSet;
    struct arrChoiceSet *parrChoiceSet;
    struct devChoiceSet *pdevChoiceSet;
    struct devChoice *pdevChoice;
    if (!pdbBase || !pfldDes || !pdbEntry->precord || !pdbEntry->pfield)
	return (-1);
    /* always force result string to be null terminated */
    sizeto = size;
    if (sizeto >= MAX_STRING_SIZE)
	sizeto = MAX_STRING_SIZE - 1;

    *pstring = 0;
    switch (pfldDes->field_type) {
    case DBF_STRING:
	/* getStringString */
	strncpy(pstring, pcharsrc, sizeto);
	*(pstring + sizeto) = 0;
	break;
    case DBF_CHAR:
	/* getCharString */
	char_to_str(*pcharsrc, pstring);
	break;
    case DBF_UCHAR:
	/* getUcharString */
	uchar_to_str(*pucharsrc, pstring);
	break;
    case DBF_SHORT:
	/* getShortString */
	short_to_str(*pshortsrc, pstring);
	break;
    case DBF_USHORT:
	/* getUshortString */
	ushort_to_str(*pushortsrc, pstring);
	break;
    case DBF_LONG:
	/* getLongString */
	long_to_str(*plongsrc, pstring);
	break;
    case DBF_ULONG:
	/* getUlongString */
	ulong_to_str(*pulongsrc, pstring);
	break;
    case DBF_FLOAT:
	/* getFloatString */
	memcpy((char *) &entry, (char *) pdbEntry, sizeof(DBENTRY));
	entry.pfield = NULL;
	entry.pfldDes = NULL;
	if ((dbFindField(pdbBase, "PREC", &entry)) != 0)
	    return (-1);
	precision = *(short *) (((char *) pdbEntry->precord) + entry.pfldDes->offset);
	f_to_str((double) (*pfloatsrc), pstring, precision);
	break;
    case DBF_DOUBLE:
	/* getDoubleString */
	memcpy((char *) &entry, (char *) pdbEntry, sizeof(DBENTRY));
	entry.pfield = NULL;
	entry.pfldDes = NULL;
	if ((dbFindField(pdbBase, "PREC", &entry)) != 0)
	    return (-1);
	precision = *(short *) (((char *) pdbEntry->precord) + entry.pfldDes->offset);
	f_to_str(*pdoublesrc, pstring, precision);
	break;
    case DBF_GBLCHOICE:
	/* getGchoiceString */
	choice_ind = *((short *) pdbEntry->pfield);
	if ((!(pchoiceSet = GET_PCHOICE_SET(pdbBase->pchoiceGbl, choice_set)))
		|| (!(pchoice = GET_CHOICE(pchoiceSet, choice_ind)))) {
	    return (-1);
	}
	strncpy(pstring, pchoice, MAX_STRING_SIZE);
	break;
    case DBF_RECCHOICE:
	/* getRchoiceString */
	choice_ind = *((short *) pdbEntry->pfield);
	if ((!(parrChoiceSet = GET_PARR_CHOICE_SET(pdbBase->pchoiceRec, (pdbEntry->record_type))))
	     || (!(pchoiceSet = GET_PCHOICE_SET(parrChoiceSet, choice_set)))
		|| (!(pchoice = GET_CHOICE(pchoiceSet, choice_ind)))) {
	    return (-1);
	}
	strncpy(pstring, pchoice, MAX_STRING_SIZE);
	break;
    case DBF_CVTCHOICE:
	/* getCchoiceString */
	choice_ind = *((short *) pdbEntry->pfield);
	if (!(pchoice = GET_CHOICE(pdbBase->pchoiceCvt, choice_ind))) {
	    return (-1);
	}
	strncpy(pstring, pchoice, MAX_STRING_SIZE);
	break;
    case DBF_DEVCHOICE:
	/* getDchoiceString */
	choice_ind = *((short *) pdbEntry->pfield);
	if ((!(pdevChoiceSet = GET_PDEV_CHOICE_SET(pdbBase->pchoiceDev, pdbEntry->record_type)))
	   || (!(pdevChoice = GET_DEV_CHOICE(pdevChoiceSet, choice_ind)))) {
	    return (-1);
	}
	strncpy(pstring, pdevChoice->pchoice, MAX_STRING_SIZE);
	break;
    default:
	return (0);
    }
    return (0);
}

/****************************************************************************
 * DBGETSTRINGMAXACCURACY
 *
 *      same as DBGETSTRING but calculates max accuracy for field types
 *	DBF_FLOAT and DBF_DOUBLE instead of using precision field
 *
 *      depends upon complete pdbEntry setup
 *
 * returns:		0,-1	OK,ERROR
 ****************************************************************************/
#ifdef __STDC__
long dbGetStringMaxAccuracy(struct dbBase *pdbBase,DBENTRY *pdbEntry, char *pstring)
#else
long dbGetStringMaxAccuracy(pdbBase, pdbEntry, pstring)
struct dbBase *pdbBase;
DBENTRY *pdbEntry;
char *pstring; /* pointer to callers buffer (string[VALUE_STRING_SZ+8]) */
#endif
{
    char *pfield = (char*)pdbEntry->pfield;
    struct fldDes  *pfldDes = pdbEntry->pfldDes;

    if (!pdbBase || !pfldDes || !pdbEntry->precord || !pfield ) return(-1);

    switch (pfldDes->field_type) {
    case (DBF_DOUBLE):
	if ((*(double *) (pfield) >= MIN_F_FORMAT_PVAL &&
	     *(double *) (pfield) <= MAX_F_FORMAT_PVAL) ||
		(*(double *) (pfield) >= MIN_F_FORMAT_NVAL &&
		 *(double *) (pfield) <= MAX_F_FORMAT_NVAL))
	    sprintf(pstring, "%.*f", DOUBLE_E_FORMAT_PREC,
		    *(double *) (pfield));
	else
	    sprintf(pstring, "%.*e", DOUBLE_E_FORMAT_PREC,
		    *(double *) (pfield));
	break;
    case (DBF_FLOAT):
	if ((*(float *) (pfield) >= MIN_F_FORMAT_PVAL &&
	     *(float *) (pfield) <= MAX_F_FORMAT_PVAL) ||
		(*(float *) (pfield) >= MIN_F_FORMAT_NVAL &&
		 *(float *) (pfield) <= MAX_F_FORMAT_NVAL))
	    sprintf(pstring, "%.*f", FLOAT_E_FORMAT_PREC,
		    *(float *) (pfield));
	else
	    sprintf(pstring, "%.*e", FLOAT_E_FORMAT_PREC,
		    *(float *) (pfield));
	break;
    default:
	/* for other types just call dbGetString */
	if ((dbGetString(pdbBase, pdbEntry, pstring)) != 0)
	    return (-1);
	break;
    }
    return (0);
}

/****************************************************************************
 * DBPUTSTRING
 *
 *      converts a string in callers pstring buffer
 *      and places it into the record field.
 *
 *      depends upon complete pdbEntry setup
 *
 * returns:		0,-1	OK,ERROR
 ****************************************************************************/
#ifdef __STDC__
long dbPutString(struct dbBase *pdbBase,DBENTRY *pdbEntry,char *pstring)
#else
long dbPutString(pdbBase, pdbEntry, pstring)
struct dbBase *pdbBase;
DBENTRY *pdbEntry;
char *pstring;
#endif
{
    struct fldDes  *pfldDes = pdbEntry->pfldDes;
    char           *pchardest = (char *) (pdbEntry->pfield);
    unsigned char  *puchardest = (unsigned char *) (pdbEntry->pfield);
    short          *pshortdest = (short *) pdbEntry->pfield;
    unsigned short *pushortdest = (unsigned short *) (pdbEntry->pfield);
    long           *plongdest = (long *) (pdbEntry->pfield);
    unsigned long  *pulongdest = (unsigned long *) (pdbEntry->pfield);
    float          *pfloatdest = (float *) (pdbEntry->pfield);
    double         *pdoubledest = (double *) (pdbEntry->pfield);
    unsigned short *pfield = (unsigned short *) (pdbEntry->pfield);
    struct arrChoiceSet *parrChoiceSet;
    struct devChoiceSet *pdevChoiceSet;
    char           *pchoice;
    struct choiceSet *pchoiceSet;
    short           choice_set = pfldDes->choice_set;
    short           shortvalue;
    unsigned short  ushortvalue;
    long            longvalue;
    unsigned long   ulongvalue;
    float           floatvalue;
    double          doublevalue;
    unsigned short  i;
    short           len;
    short           value;
    unsigned short  uvalue;
    if (!pdbBase || !pfldDes || !pdbEntry->precord || !pdbEntry->pfield)
	return (-1);

    switch (pfldDes->field_type) {
    case DBF_STRING:
	/* putStringString */
	/* always force result string to be null terminated */
	len = strlen(pstring);
	if (len >= MAX_STRING_SIZE)
	    len = MAX_STRING_SIZE - 1;
	if (len > 0)
	    strncpy(pchardest, pstring, len);
	*(pchardest + len) = 0;
	break;
    case DBF_CHAR:
	/* putStringChar */
	if (sscanf(pstring, "%hd", &value) == 1) {
	    *pchardest = (char) value;
	} else
	    return (-1);
	break;
    case DBF_UCHAR:
	/* putStringUchar */
	if (sscanf(pstring, "%hu", &uvalue) == 1) {
	    *puchardest = (unsigned char) uvalue;
	} else
	    return (-1);
	break;
    case DBF_SHORT:
	/* putStringShort */
	if (sscanf(pstring, "%hd", &shortvalue) == 1) {
	    *pshortdest = shortvalue;
	} else
	    return (-1);
	break;
    case DBF_USHORT:
	/* putStringUshort */
	if (sscanf(pstring, "%hu", &ushortvalue) == 1) {
	    *pushortdest = ushortvalue;
	} else
	    return (-1);
	break;
    case DBF_LONG:
	/* putStringLong */
	if (sscanf(pstring, "%ld", &longvalue) == 1) {
	    *plongdest = longvalue;
	} else
	    return (-1);
	break;
    case DBF_ULONG:
	/* putStringUlong */
	if (sscanf(pstring, "%lu", &ulongvalue) == 1) {
	    *pulongdest = ulongvalue;
	} else
	    return (-1);
	break;
    case DBF_FLOAT:
	/* putStringFloat */
	if (sscanf(pstring, "%f", &floatvalue) == 1) {
	    *pfloatdest = floatvalue;
	} else
	    return (-1);
	break;
    case DBF_DOUBLE:
	/* putStringDouble */
	if (sscanf(pstring, "%lf", &doublevalue) == 1) {
	    *pdoubledest = doublevalue;
	} else
	    return (-1);
	break;
    case (DBF_GBLCHOICE):
	/* putStringGchoice */
	if (pchoiceSet = GET_PCHOICE_SET(pdbBase->pchoiceGbl, choice_set)) {
	    for (i = 0; i < pchoiceSet->number; i++) {
		if (!(pchoice = pchoiceSet->papChoice[i]))
		    continue;
		if (strcmp(pchoice, pstring) == 0) {
		    *pfield = i;
		    break;
		}
	    }
	}
	return (-1);
	break;
    case (DBF_CVTCHOICE):
	/* putStringCchoice */
	if (pchoiceSet = pdbBase->pchoiceCvt) {
	    for (i = 0; i < pchoiceSet->number; i++) {
		if (!(pchoice = pchoiceSet->papChoice[i]))
		    continue;
		if (strcmp(pchoice, pstring) == 0) {
		    *pfield = i;
		    break;
		}
	    }
	}
	return (-1);
	break;
    case (DBF_RECCHOICE):
	/* putStringRchoice */
	if ((parrChoiceSet = GET_PARR_CHOICE_SET(pdbBase->pchoiceRec, (pdbEntry->record_type)))
	     && (pchoiceSet = GET_PCHOICE_SET(parrChoiceSet, choice_set))) {
	    for (i = 0; i < pchoiceSet->number; i++) {
		if (!(pchoice = pchoiceSet->papChoice[i]))
		    continue;
		if (strcmp(pchoice, pstring) == 0) {
		    *pfield = i;
		    break;
		}
	    }
	}
	return (-1);
	break;
    case (DBF_DEVCHOICE):
	/* putStringDchoice */
	if (pdevChoiceSet = GET_PDEV_CHOICE_SET(pdbBase->pchoiceDev, pdbEntry->record_type)) {
	    for (i = 0; i < pdevChoiceSet->number; i++) {
		if (!(pchoice = pdevChoiceSet->papDevChoice[i]->pchoice))
		    continue;
		if (strcmp(pchoice, pstring) == 0) {
		    *pfield = i;
		    return (0);
		}
	    }
	}
	return (-1);
	break;
    default:
	return (-1);
    }
    return (0);
}

/******************************************************************************
DBCREATE - creates a new db by copying default.dctsdr to pdatabaseName
Returns:	 0,-1	OK,error
******************************************************************************/
#ifdef __STDC__
long dbCreate(char *pdctsdrName, char *pdatabaseName)
#else
long dbCreate(pdctsdrName, pdatabaseName)
char *pdctsdrName;
char *pdatabaseName;
#endif
{
    FILE           *fp;
    FILE           *fp2;
    struct stat     sbuf;
    struct stat     sbuf2;
    char           *buff;
    int             err = 0;
    if ((stat(pdctsdrName, &sbuf)) < 0) {
	fprintf(stderr, "dbCreate: can't stat %s\n", pdctsdrName);
	return (-1);
    }
    if ((stat(pdatabaseName, &sbuf2)) == 0) {
	fprintf(stderr, "dbCreate: DATABASE %s already exists\n", pdatabaseName);
	return (-1);
    }
    if ((fp = fopen(pdctsdrName, "r")) == NULL) {
	fprintf(stderr, "dbCreate: can't fopen %s\n", pdctsdrName);
	return (-1);
    }
    if (sbuf.st_size == 0) {
	fprintf(stderr, "dbCreate: zero length file %s\n", pdctsdrName);
	return (-1);
    }
    if ((buff = (void *) calloc(1, (unsigned) sbuf.st_size)) == NULL) {
	fprintf(stderr, "dbCreate: malloc failed\n");
	return (-1);
    }
    if ((fread(buff, (int) sbuf.st_size, 1, fp)) != 1) {
	fprintf(stderr, "dbCreate: incomplete read of %s\n", pdctsdrName);
	err++;
    }
    if (fclose(fp) != 0) {
	if (!err)
	    fprintf(stderr, "dbCreate: can't close %s\n", pdctsdrName);
	err++;
    }
    if (err)
	return (-1);
    if ((fp2 = fopen(pdatabaseName, "w")) == NULL) {
	fprintf(stderr, "dbCreate: can't fopen for writing %s\n", pdatabaseName);
	return (-1);
    }
    if ((fwrite(buff, (int) sbuf.st_size, 1, fp2)) != 1) {
	fprintf(stderr, "dbCreate: incomplete write of %s\n", pdatabaseName);
	err++;
    }
    if (fclose(fp2) != 0) {
	if (!err)
	    fprintf(stderr, "dbCreate: can't close %s\n", pdatabaseName);
	err++;
    }
    if (err)
	return (-1);
    return (0);
}
/****************************************************************************
 *
 *
 *   Nget_dct_choice_string()
 *
 *   returns a pointer to the choice string corresponding to choice
 *   have to use this routine for enum choices that aren't defined in *.sdr file
 ****************************************************************************/
char *Nget_dct_choice_string(tblinx, choice)
int	tblinx;		/* choice table index */
int	choice;		/* choice */
{
        int     i;
        int     num;
        int     found;
        struct  wchoice *pset;

        /* get the number of choices */
        num = 0;
        pset = Nchoicearray[tblinx];
        while ( (pset+num)->pstr != NULL)
                num++;
        /* find the current index for this value */
        found = i=0;
        while ( i < num)  {
                if( choice  == (pset+i)->val){
                                found++;
                                break;
                }
                i++;
        }
        if(!found)
		return("BAD_VALUE");
	else
		return((pset+i)->pstr);
}

/****************************************************************************
 *
 *
 *   Nget_dct_choice_val()
 *
 *   returns the value of a choice corresponding to a choice string 
 *	if not found returns -1
 *   have to use this routine for enum choices that aren't defined in *.sdr file
 ****************************************************************************/
short	Nget_dct_choice_val(tblinx, choicestring)
int	tblinx;		/* choice table index */
char	*choicestring;	/* choice string */
{
        int     i;
struct  wchoice *pset;

        pset = Nchoicearray[tblinx];
        i=0;
        while ( (pset+i)->pstr != NULL)  {
                if( strcmp((pset+i)->pstr, choicestring) == 0){
			return((pset+i)->val);
                }
                i++;
        }
	return(-1);
}
