/* share/epicsH $Id: dbManipulate.h,v 1.1 1994/09/02 13:31:56 mrk Exp $ */
/* dbManipulate.h
 *
 *
 *      Current Author:         Robert Zieman
 *      Date:                   05-22-92
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  05-22-92        rcz     Original
 * .02  06-08-92        rcz     Added macros GET_PRECLIST GET_PVDNODE GET_PRECNODE
 * .03  05-18-92        rcz     New database access (removed extern)
 * .04  07-16-92        jba     removed reference to dbAddRecord
 * .05  07-24-93        rcz     removed GET_PRECNODE macro
 */

#ifndef INCdbManipulateh
#define INCdbManipulateh 1

#ifndef INCdbDefsh
#include <dbDefs.h>
#endif

typedef struct recNode {
    ELLNODE            next;
    void           *precord;
} RECNODE;

typedef struct dbEntry {
	void 		*precord;
	short 		record_type;
	struct recTypDes *precTypDes;
	void	 	*pfield;
	struct fldDes 	*pfldDes;
} DBENTRY;


typedef struct pvdEntry {
    ELLNODE            pvdNode;
    void           *precord;
    short           record_type;
} PVDENTRY;

typedef struct dbChoices {
        int     nchoices;
        char    choice[DB_MAX_CHOICES][MAX_STRING_SIZE];
} DBCHOICES;




/* dbManipulate routines */
#ifdef __STDC__
long dbCreate(char *pdctsdrName,char *pdatabaseName);
long dbRead(struct dbBase **ppdbBase,char *pdatabaseName);
long dbWrite(struct dbBase *pdbBase,char *pdatabaseName);
long dbFreeMem(struct dbBase **ppdbBase);
long dbCreateRecord(struct dbBase *pdbBase,int type,
		char *precordName,DBENTRY *pdbEntry);
long dbDeleteRecord(struct dbBase *pdbBase,DBENTRY *pdbEntry);
long dbFindRecord(struct dbBase *pdbBase,char *precordName,DBENTRY *pdbEntry);
long dbFindField(struct dbBase *pdbBase,char *pfieldName,DBENTRY *pdbEntry);
long dbGetChoices(struct dbBase *pdbBase,DBENTRY *pdbEntry,
		DBCHOICES *pdbChoices);
long dbGetString(struct dbBase *pdbBase,DBENTRY *pdbEntry,char *pstring);
long dbPutString(struct dbBase *pdbBase,DBENTRY *pdbEntry,char *pstring);
long dbGetStringMaxAccuracy(struct dbBase *pdbBase,DBENTRY *pdbEntry,
		char *pstring);
long dbSave(struct dbBase *pdbBase,char *preport);
long dbRestore(struct dbBase *pdbBase,char *preport);
#else
long dbCreate();
long dbRead();
long dbWrite();
long dbFreeMem();
long dbCreateRecord();
long dbDeleteRecord();
long dbFindRecord();
long dbFindField();
long dbGetChoices();
long dbGetString();
long dbPutString();
long dbGetStringMaxAccuracy();
long dbSave();
long dbRestore();
#endif /*__STDC__*/

/************************************************************************
 * The following macro returns NULL or a ptr to record list
 * A typical usage is:
 *
 *	struct dbBase *pdbBase
 *	DBENTRY       *pdbEntry
 *	LIST          *preclist;
 *	if(!(preclist=GET_PRECLIST(pdbBase,pdbEntry))) {<not found>}
 ***********************************************************************/
#define GET_PRECLIST(PDBBASE,PDBENTRY)\
(\
  ((PDBBASE) && (PDBENTRY))\
  ?(\
    ((PDBBASE)->precHeader && (PDBENTRY)->precord)\
    ?(\
      ( ((((PDBENTRY)->record_type))<1) || (((PDBENTRY)->record_type)>=((PDBBASE)->precHeader->number)) )\
            ?(\
               NULL\
            )\
            : ((PDBBASE)->precHeader->papRecLoc[(PDBENTRY)->record_type])\
              ?(\
                 ((PDBBASE)->precHeader->papRecLoc[(PDBENTRY)->record_type]->preclist)\
              )\
              : NULL\
    )\
    : NULL\
  )\
  : NULL\
)

/************************************************************************
 * Given precord:
 * The following macro returns NULL or a ptr to pvdnode
 * A typical usage is:
 *
 *	DBENTRY       *pdbEntry
 *	PVDENTRY      *ppvdNode
 *	if(!(ppvdNode=GET_PVDNODE(pdbEntry))) {<not found>}
 ***********************************************************************/
#define GET_PVDNODE(PDBENTRY)\
(\
  (PDBENTRY)\
  ?(\
    ((PDBENTRY)->precord)\
    ?(\
      ((PVDENTRY *)((((char*)(PDBENTRY)->precord)) - sizeof(PVDENTRY))) \
    )\
    : NULL\
  )\
  : NULL\
)

/****************************************************************************
*  this macro returns the value of link_type from the devChoiceRec
*  usage:
*
*	long	link_type;
*	link_type = GET_DEV_CHOICE_LINK_VAL(record_type, card_type);
****************************************************************************/
#define GET_DEV_CHOICE_LINK_VAL(RECORD_TYPE, CARD_TYPE) \
(\
	GET_DEV_CHOICE_VAL((RECORD_TYPE), (CARD_TYPE), link_type) \
)
/**************************************************************************
* a macro to return a  value from the devChoiceRec struct 
* this macro assumes that checking for NULL pointers has been already done 
* typical usage :
*
*	long	link_type;
*	char	*pstring;
*	link_type = GET_DEV_CHOICE_VAL(record_type, card_type, link_type);
*	pstring = GET_DEV_CHOICE_VAL(record_type, carc_type, pchoice);
*
***************************************************************************/

#define GET_DEV_CHOICE_VAL(RECORD_TYPE, CARD_TYPE, VALUE) \
(\
pchoiceDev->papDevChoiceSet[(RECORD_TYPE)]->papDevChoice[(CARD_TYPE)]->VALUE \
)





/*********************************************************************************/
/* NOTE:  DO NOT USE ANY OF THE DEFINES BELOW - THEY WILL EVENTUALLY GO AWAY */
/*********************************************************************************/
#define RANGE_STRING_SZ 27	/* width of range string */
#define RANGE_X         PROMPT_SZ + 1 + FLDNAME_SZ + 2  /* place to start range string */
#define VALUE_X         RANGE_X + RANGE_STRING_SZ + 1	/* column to display field values*/
#define MAX_RECORD_SIZE	2048	/* this should be eliminated by dynamically allocating mem with malloc*/
#define VALUE_X         RANGE_X + RANGE_STRING_SZ + 1	/* column to display field values*/
#define	PDNAME_SZ	32		/* max size of a name for a database */
#define	MAX_RECNAME_SZ  12
#define  NOPROMPT 0		/* value of fldDes.promptflag for no prompt */
#define VALUE_X         RANGE_X + RANGE_STRING_SZ + 1	/* column to display field values*/

/************************FROM FRL dctDef.h **************/
/*  the following 6 defines determine the thresholds at which dct switches
	between "e" format and "f" format for displaying floats and doubles */
#define	MAX_F_FORMAT_PVAL  99999.9
#define MIN_F_FORMAT_PVAL  1e-3
#define MAX_F_FORMAT_NVAL  -1e-3
#define MIN_F_FORMAT_NVAL  -99999.9
#define MIN_F_FORMAT_PVALR .1		/* applies to range fields */
#define MAX_F_FORMAT_NVALR -.1		/* ditto */

/* the following define maximum precision for e format floats and doubles
	and precision to use for short and long reports */
#define FLOAT_E_FORMAT_PREC 7
#define DOUBLE_E_FORMAT_PREC 15
#define CONSTANT_FORMAT_PREC DOUBLE_E_FORMAT_PREC /* change this to double if
						contant link type changed 
						from float to double */

#define VALUE_STRING_SZ PVNAME_SZ+FLDNAME_SZ+9  /* size of value string */

#define SAME    0
#define MAX_STRING_LENGTH 128
#define MAX_FILENAME_SZ 80
#define  VAL_AR_SZ   80		/* array containing value size 		  */
#define  MAX_WORD_SZ 80		/* max words size of words read from line */
#define  LINE_SZ     120	/* size of array line read into		  */
#define  LOOK_PV     1		/* looking for 'PV:' status	  */
#define  LOOK_TYPE   2		/* have found 'PV:' and pv name	  */
#define  LOOK_NODE   3		/* have found 'Type:' and type name */
#define  LOOK_FIELD  4		/* accepting fields to modify	  */
#endif
