/*  $Id: dctDef.h,v 1.9 1993/07/23 11:00:22 mrk Exp $
 *	Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	3/16/92		frl	CONSTANT_FORMAT_PREC = DOUBLE_E_FORMAT_PREC
 * 	...
 */

/* dctDef.h */

#ifndef INCdctDefh
#define INCdctDefh	1

#define MAX_RECORD_SIZE	4096		/* this should be eliminated by
					dynamically allocating mem with malloc*/
#define	PDNAME_SZ	32		/* max size of a name for a database */
#define	MAXPDS		64		/* max number of database names dct
					can store in its name table.. this
					could be eliminated.  */
#define MAXFORMS	64		/* max number of report forms that 
						can be defined */
#define MAXFORMFIELDS	64		/* max # of fields in a summary form */
#define	DEFAULT_FORM	"default"	/* name of default summary form */
#define FORM_EXT	"frm"		/* file name extension for a summary
						form */
/*  the following 6 defines determine the thresholds at which dct switches
	between "e" format and "f" format for displaying floats and doubles */
#define	MAX_F_FORMAT_PVAL  99999.9
#define MIN_F_FORMAT_PVAL  1e-3
#define MAX_F_FORMAT_NVAL  -1e-3
#define MIN_F_FORMAT_NVAL  -99999.9
#define MIN_F_FORMAT_PVALR .1		/* applies to range fields */
#define MAX_F_FORMAT_NVALR -.1		/* ditto */

/* the following define maximum precision for e format floats and doubles
	and precision to use for short and long reports */
#define FLOAT_E_FORMAT_PREC 7
#define DOUBLE_E_FORMAT_PREC 15
#define CONSTANT_FORMAT_PREC DOUBLE_E_FORMAT_PREC /* change this to double if
						contant link type changed 
						from float to double */

#define	MAX_RECNAME_SZ  12
#define    NUMBER_FREE     2       /* number of free records to leave in 
					each record type db file */
#define		DONE -1

#define		MAX_FILENAME_SZ	80
#define		MAX_RECORD_TYPES 128
#define		MAX_STRING_LENGTH 128

#define	DCT_FILE_CREAT_MODE 0664	/* file create mode for database files*/

#define	FWDLINKPROMPTS   973		/* used to define prompts for 
					DBF_FWDLINK,  make sure this is not
					equal to a link type defined in link.h*/
#define DCT_INTERACTIVE	   0
#define DCT_NONINTERACTIVE 1
#define	STDOUT	1

#define YES   	1
#define NO	0
#define SAME    0

#define  NOPROMPT 0		/* value of fldDes.promptflag for no prompt */

/* positions on the window */
#define VALUE_STRING_SZ PVNAME_SZ+FLDNAME_SZ+9  /* size of value string */
#define RANGE_STRING_SZ 27	/* width of range string */
#define RANGE_X         PROMPT_SZ + 1 + FLDNAME_SZ + 2  /* place to start range
								string */
#define VALUE_X         RANGE_X + RANGE_STRING_SZ + 1	/* column to display
							 field values*/
#define INPUT_X         VALUE_X
#define PAGE_SIZE       (PAGE_LEN-3) /* useable area for info  */ 
#define HELPLINE	(PAGE_LEN-2)
#define ERRLINE		(PAGE_LEN-1)

/*   screen defines */
#define PAGE_WIDTH	110	/* min # of columns for page size */
#define PAGE_LEN	 72	/* min # of rows for a page */

#define PAUSE_F	1		/* pause flag */
#define CONTINUE_F 0		/* continue flag */
/* dct menu action defines */
#define	NOKEY		0
#define NEXT            1
#define PREVIOUS        2
#define EXIT            3
#define FIRST           4
#define LAST            5
#define PAGE_UP         6
#define PAGE_DOWN       7
#define NEXT_ITEM       8
#define PREV_ITEM       9
#define EXECUTE         10
#define QUIT1           11
#define QUIT2           12
#define QUIT3           13


/* these must be negative to distinguish them from a menu selection */
#define DONE    -1
#define UP      -2
#define DOWN    -3

#define	CLEAN_UP  0
#define CREATE_DB 1

struct	fpos	{		/* holds a screen position */
	int	x;
	int	y;
};

struct	menu_prompt {		/* defines a prompt for a menu */
	int	val;		/* value to return for this prompt */
	int	(*proc)();	/* proc to execute for this prompt */
	int	key;		/* key code for this prompt */
	char	*pmt;		/* prompt text */
};

struct	fmenu {			/* a menu set */
	struct	fpos	titlep;	/* place to put title */
	struct	fpos	pmtp;	/* starting location of prompts */
	struct	fpos	pmtc;	/* starting location of prompt cursor */
	char		**title;	/* array of title strings */
	int		nprompts; /* # of prompts */
	struct	menu_prompt *pmt;	/* array of prompts */
};

struct	fchoice {			/* a choice set */
	struct	fpos	titlep;	/* place to put title */
	struct	fpos	pmtp;	/* starting location of prompts */
	struct	fpos	pmtc;	/* starting location of prompt cursor */
	char		*title;	/* title string */
	int		nprompts; /* # of prompts */
	char		**choice;	/* array of choice strings */
};

struct	fenum_el {		/* an enum element */
	short	val;		/* the value */
	char	*string;	/* corresponding string text */
};

struct fenum {			/* a set of enums and their string vals */
	int	num;		/* # in the set */
	struct fenum_el	*item;	/* array of items */
};
	
#endif
