static char SccsId[] = "$Id: dcterr.c,v 1.5 1991/05/30 11:58:53 frl Exp $ ";

/*  $Id: dcterr.c,v 1.5 1991/05/30 11:58:53 frl Exp $
 *	Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

/*  dcterr.c                                    F. Lenkszus
*
*       error handling routines for DCT
*
*/

#include <curses.h>
#include <errno.h>
#include <string.h>
#include "dctDef.h"

extern  int     sys_nerr;       /* size of sys_errlist[] */
extern  char    *sys_errlist[]; /* vector of system error messages */
extern  short   suppress_errors;        /* if != 0 */
extern  short   suppress_errors1;        /* if != 0 */
extern	int     dct_mode;      /* mode flag  0==interactive, 1==batch */


static	char	*dcthelp[] = {
  "KPR2:QUIT KPR7:FIRST KPR9:LAST KPR11:EXIT KPR13:PGUP KPR15:PGDN UP:PREV DOWN:NEXT",			/* 0 */
"UP:Previous  DOWN:Next  ENTER:Select  KPR11:Exit",		/* 1 */
"UP:Previous Field  DOWN:Next Field   LEFT/RIGHT:Select Choice  KPR11:Exit   KPR2:Quit",		/* 2 */
"Press any key to continue or KPR11 to quit",			/* 3 */
"Press any key to continue"					/* 4 */
		};

/*
 * CLEAR_ERROR
 *
 * clear the error message area
 */
clear_error()
{
        move(ERRLINE,0);
        clrtoeol();
}

/*
 * CLEAR_HELP
 *
 * clear the help message area
 */
clear_help()
{
        move(HELPLINE,0);
        clrtoeol();
}

/* print_help_msg
*
*/

print_help_msg(help_num)
int	help_num;		/* index of help message  in dcthelp[] */

{
	if( (help_num >= 0) && (help_num < sizeof(dcthelp)/sizeof(char *))) {
                clear_help();
                move(HELPLINE,0);
                addstr(dcthelp[help_num]);
                refresh();
	}
}
/*
 * print_wait_msg
 *
 */
print_wait_msg(init)
int	init;		/* !=0 -> init sequence */
{
	static	int	next;
	static  char	wait_msg[] = {
		'|','/','-','\\'
	};

		if( dct_mode == DCT_INTERACTIVE) {
		if(init)
			next=0;
		clear_help();
		move(HELPLINE,0);
		if ( next >= sizeof(wait_msg))
			next = 0;
		printw("Please wait %c",wait_msg[next++]);
		refresh();
	}
}
/* print_error_msg
*
*/

print_error_msg(pname, msg)

char    *pname;                 /* name of proc incurring error */
char    *msg;                   /* message */

{


	if(dct_mode == DCT_INTERACTIVE) {
        	if (!suppress_errors){
               		clear_error();
                	move(ERRLINE,0);
                	printw("%s:  %s",pname, msg);
                	refresh();
        	}
	} else
		print_err(pname, msg);
}

/* print_error_msg_p
*  print an error message with pause flag
*
*/

print_error_msg_p(pname, msg, pause)

char    *pname;                 /* name of proc incurring error */
char    *msg;                   /* message */
int	pause;			/* pause flag, 1= pause, 0 = no pause */
{


	if(dct_mode == DCT_INTERACTIVE) {
        	if (!suppress_errors){
                	clear_error();
               	 	move(ERRLINE,0);
                	printw("%s:  %s",pname, msg);
                	refresh();
			if(pause) {
				getinpch();
				clear_error();
				refresh();
			}
        	}
	} else 
		print_err(pname, msg);
}

/* print_syserror_msg
*
*/

print_syserror_msg(pname, msg, errnum)

char    *pname;                 /* name of proc incurring error */
char    *msg;                   /* message */
int     errnum;                 /* for UNIX this is errno */

{


	if(dct_mode == DCT_INTERACTIVE) {
        	if (!suppress_errors){
                	if (errnum < sys_nerr) {
                        	clear_error();
                        	move(ERRLINE,0);
                        	printw("%s:  %s:  %s (%d)",pname, msg,
				 sys_errlist[errnum], errnum);
                        	refresh();
                	}
                	else {
                        	print_error_msg(pname,
					 "Bad system error number");
                	}
        	}
	} else
		print_syserr(pname, msg, errnum);
}

/* print_syserror_msg_p
*
*/

print_syserror_msg_p(pname, msg, errnum, pause)

char    *pname;                 /* name of proc incurring error */
char    *msg;                   /* message */
int     errnum;                 /* for UNIX this is errno */
int	pause;			/* pause flag, 1=pause, 0=nopause  */ 

{


	if(dct_mode == DCT_INTERACTIVE) {
        	if (!suppress_errors){
                	if (errnum < sys_nerr) {
                        	clear_error();
                        	move(ERRLINE,0);
                        	printw("%s:  %s:  %s (%d)",pname, msg,
				 sys_errlist[errnum], errnum);
                        	refresh();
                	}
                	else {
                       	 	print_error_msg(pname, 
					"Bad system error number");
                	}
			if(pause) {
				getinpch();
				clear_error();
				refresh();
			}
        	}
	} else
			print_syserr(pname, msg, errnum);
}

/*
 *	errMessage()
 *		This routine is used for sdrLoad()/sdrUnload error reporting
 */
errMessage(status,pstring)
long	status;
char	*pstring;		/* string to print */

{
	print_error_msg("", pstring);
}

/*
 *	print_syserr()
 *
 */
print_syserr(pname, msg, errnum)

char    *pname;                 /* name of proc incurring error */
char    *msg;                   /* message */
int     errnum;                 /* for UNIX this is errno */

{


        if (!suppress_errors1){
                if (errnum < sys_nerr) {
                        fprintf(stderr,"%s:  %s:  %s (%d)\n",pname, msg,
				 sys_errlist[errnum], errnum);
                }
                else {
                        fprintf(stderr,  "Bad system error number(%d)\n", errnum);
                }
        }
}
/*
 *	print_err()
 *
 */
print_err(pname, msg)

char    *pname;                 /* name of proc incurring error */
char    *msg;                   /* message */

{


        if (!suppress_errors1){
                fprintf(stderr,"%s:  %s\n",pname, msg);
        }
}
