/*  $Id: dctmacros.h,v 1.5 1992/07/11 21:41:50 rcz Exp $
 *	Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02  05-18-92        rcz     New database access - moved and renamed some old macros
 * 	...
 */

/* dctmacros.h					F. Lenkszus, 4/13/90 */
/* a collection of macros used by dct */
#ifndef	INCdctmacrosh
#define INCdctmacrosh 1

/**************************************************************************
* a macro to return a  value from the devChoiceRec struct 
* this macro assumes that checking for NULL pointers has been already done 
* typical usage :
*
*	long	link_type;
*	char	*pstring;
*	link_type = GET_DEV_CHOICE_VAL(record_type, card_type, link_type);
*	pstring = GET_DEV_CHOICE_VAL(record_type, carc_type, pchoice);
*
***************************************************************************/

#define GET_DEV_CHOICE_VAL(RECORD_TYPE, CARD_TYPE, VALUE) \
(\
choiceDev->papDevChoiceSet[(RECORD_TYPE)]->papDevChoice[(CARD_TYPE)]->VALUE \
)

/****************************************************************************
*  this macro returns the value of link_type from the devChoiceRec
*  usage:
*
*	long	link_type;
*	link_type = GET_DEV_CHOICE_LINK_VAL(record_type, card_type);
****************************************************************************/
#define GET_DEV_CHOICE_LINK_VAL(RECORD_TYPE, CARD_TYPE) \
(\
	GET_DEV_CHOICE_VAL((RECORD_TYPE), (CARD_TYPE), link_type) \
)


/****************************************************************************
*  this macro returns a pointer to a choice string  from the devChoiceRec
*  usage:
*
*	char	*pstring;
*	pstring = GET_DEV_CHOICE_STRING(record_type, card_type);
****************************************************************************/
#define GET_DEV_CHOICE_STRING(RECORD_TYPE, CARD_TYPE) \
(\
	GET_DEV_CHOICE_VAL((RECORD_TYPE), (CARD_TYPE), pchoice) \
)

/****************************************************************************
 *
 *  this macro frees previously allocated memory
 *
 ****************************************************************************/
#define FREE_MEM(MEM_PTR) \
	if(!(MEM_PTR)) \
		free(MEM_PTR)


/*****************************************************************************
 * The following macro returns NULL or a ptr to a recTypDes structure
 * A typical usage is:
 *
 *	struct recTypDes *precTypDes;
 *	if(!(precTypDes=XGET_PRECTYPDES(ind);
 ****************************************************************************/
#define XGET_PRECTYPDES(IND_REC)\
(\
    dbRecDes\
    ?(\
	( ((IND_REC)<1) || ((IND_REC)>=dbRecDes->number) )\
	? NULL\
	: (dbRecDes->papRecTypDes[(IND_REC)])\
    )\
    : NULL\
)
/**************************************************************************
 * The following macro returns NULL or a ptr to the rectype string
 * A typical usage is:
 *
 *	char *pstr;
 *	if(!(pstr=XGET_PRECTYPE(type))) {<null>}
 **************************************************************************/
#define XGET_PRECTYPE(REC_TYPE)\
(\
    dbRecType\
	?(\
		( ((REC_TYPE)<1) || ((REC_TYPE)>=dbRecType->number) )\
		? NULL\
		: (dbRecType->papName[(REC_TYPE)])\
	)\
	: NULL\
)
#endif
