static char SccsId[] = "$Id: dctsubs.c,v 1.15 1995/02/28 22:04:43 mrk Exp $ ";

/*  $Id: dctsubs.c,v 1.15 1995/02/28 22:04:43 mrk Exp $
 *	Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	3/25/92		frl	Null Terminated PV Name	
 * .02	2/23/94		mrk	For CT_HEX print leading 0x
 * 	...
 */

/*****************************************************************************
*
*	dctsubs.c				F. Lenkszus
*
*   Collection of low level routines used by DCT
*
*****************************************************************************/

#include        <stdlib.h>
#include        <stddef.h>
#include	<stdio.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<string.h>
#include	<dirent.h>
#include	<curses.h>
#include	<sdrHeader.h>	
#include	<dbDefs.h>
#include	<dbFldTypes.h>
#include	<dbRecDes.h>
#include	"dctDef.h"

#define L_SET           0       /* absolute offset */
#define L_INCR          1       /* relative to current offset */
#define L_XTND          2       /* relative to end of file */


extern size_t	lseek();
extern	short	pvdHash();
extern	short	pvdPvSearch();
extern	int	rep_fd;		/* report file descriptor */
extern int     	dis_range_check;  /* 0 = check if value within range, !=0 don't
                                        check */


/*---------------------------  SUBROUTINES ----------------------------------
 *
char    *alloc_2d_array( parray, parraysz , elemsz)
	char    *parray[];       array of pointers to strings 
	int     parraysz;        size of array[] 
	int     elemsz;          size of each element 
-----------------------------------------------------------------------------
 int build_range_string(xname, field_type, cvt_type, pstring, prange1,
        prange2)

        char    *xname;          name of caller 
        short   field_type;      the name says it all 
        short   cvt_type;        field convert type for this field type 
        char    *pstring;        place to put range string 
        union   fld_types  *prange1;   low limit 
        union   fld_types  *prange2;   hi limit 
-----------------------------------------------------------------------------
int     close_file(fd, filename, procname)
	int     fd;                      file descriptor of file to close 
	char    *filename;               name of file to close 
	char    *procname;               name of caller 
-----------------------------------------------------------------------------
check_field_range(procname, pvalue, prange1, prange2, field_type)
        char    *procname;       name of caller 
        union   fld_types       *pvalue;   field value 
        union   fld_types       *prange1;  lower limit 
        union   fld_types       *prange2;  upper limit 
        short   field_type;                field type 
-----------------------------------------------------------------------------
int	check_file_exist(path)
char	*path;		 path to check existance of 
-----------------------------------------------------------------------------
int     convert_type_to_val( procname, field_type, cvt_type, pstring, pvalue)
        char  *procname;         name of caller 
        short   field_type;      field type 
        short   cvt_type;        convert type 
        char    *pstring;        ascii string  to convert
        union   fld_types *pvalue;  pointer to place to recieve value 
-----------------------------------------------------------------------------
int     convert_type_to_ascii( procname, field_type, cvt_type, plstring,
                         pstring, ptstring, pvalue)
        char  *procname;         name of caller 
        short   field_type;      field type 
        short   cvt_type;        convert type 
        char    *plstring;       string to prepend to pstring 
        char    *pstring;        place to put resulting string 
        char    *ptstring;       string to append to pstring 
        union   fld_types *pvalue;  pointer to value to convert 
-----------------------------------------------------------------------------
int  copy_file(procname, fds, fdd, buffer, bufsize)
	char	*procname;	 name of caller 
	int	fds;		 file descriptor of source file 
	int	fdd;		 file descriptor of destination file 
	char	*buffer;	 buffer to use for read/write 
	int	bufsize;	 size of buffer 
-----------------------------------------------------------------------------
int create_file(procname,root_name, ext, filename,fnsiz)
	char    *procname;               name of calling proc 
	char    *root_name;              root name of file to create 
	char    *ext;                    file name extension 
	char    *filename;               returned complete file name 
	int     fnsiz;                   size of filename buffer 
-----------------------------------------------------------------------------
int create_file_force(procname,root_name, ext, filename,fnsiz)
	char	*procname;		 name of calling proc 
	char	*root_name;		 root name of file to create 
	char	*ext;			 file name extension 
	char	*filename;		 returned complete file name 
	int	fnsiz;			 size of filename buffer 
-----------------------------------------------------------------------------
int     delete_file(procname, pname)
	char    *procname;       name of caller 
	char    *pname;          path name of file to delete 
-----------------------------------------------------------------------------
int     dir_search( filenames, maxno, pattern)
	char    *filenames[];    place to put file names containing pattern 
	int     maxno;                   size of filenames array 
	char    *pattern;                string to search for 
-----------------------------------------------------------------------------
int	fchoice(pchoice, helpinx)
struct	fchoice	*pchoice;		 choice definition structure 
int	helpinx;	 index of help message to print for this menu	
-----------------------------------------------------------------------------
int	fdisplay(pchoice, helpinx)
struct	fchoice	*pchoice;		 choice definition structure 
int	helpinx;	 index of help message to print for this menu	
-----------------------------------------------------------------------------
char	*fenum(enum_val, pfenum)
short	enum_val;		 enum value to convert 
struct	fenum	*pfenum;	 pointer to defining structure 
-----------------------------------------------------------------------------
int	fmenu(pmenu, helpinx, choice)
struct	fmenu	*pmenu;		 menu definition structure 
int	helpinx;	 index of help message to print for this menu	
int	choice;		 start line for prompt
-----------------------------------------------------------------------------
 fill_string(pbuffer,size,fillchar)
char  *pbuffer;
short size;
char  fillchar;
-----------------------------------------------------------------------------
char	*fldname_to_string(fldname, string, size)
	char	*fldname;		 field name (non null terminated 
	char	*string;		place to put null terminated field name 
	int	size;			 size of field name 
-----------------------------------------------------------------------------
int  fieldtype_to_string(fldtype, pstring, pname)
	short	fldtype;		 field type 
	char	*pstring;		 place to return field type 
	char	*pname;			 name of caller 
-----------------------------------------------------------------------------
int  get_file_mtime( pname, path, mtime)
	char    *pname;              name of calling proc 
	char    *path;               path name of file 
	time_t  *mtime;              buffer to return file modification time 
-----------------------------------------------------------------------------
int  get_file_size( pname, path)
	char	*pname;			 name of calling proc 
	char    *path;                   path name of file 
-----------------------------------------------------------------------------
short	get_pvd_index_by_pvname(pvname, pname, report_errs)
	char	*pvname;	 name of process variable to be looked up 
	char	pname;		 caller's name 
	short	report_errs;	 0 == suppress error reporting 
-----------------------------------------------------------------------------
int     get_sdr_filetype(filename)
	char    *filename;               name of file to check 
-----------------------------------------------------------------------------
int     open_file(pname, mode, procname)
	char    *pname;          name of file to open 
	int     mode;            mode to open file 
	char    *procname;       name of caller 
-----------------------------------------------------------------------------
int sparse(pstring, pdelim, pfields, num_fields)
	char    *pstring;                string to parse 
	char    *pdelim;                 string of delimitors 
	char    *pfields[]; 		 array of pointers to buffers to receive
                                parsed fields
	int     num_fields;              max number of fields to parse 
-----------------------------------------------------------------------------
int read_file(fd, filename, procname, buf, bufsiz)
	int     fd;      file descriptor for file to read 
	char    *filename;       name of file to read, used for error reporting 
	char    *procname;       name of procedure calling this proc 
	char    *buf;            buffer to read into
	int     bufsiz;          number of bytes to read 
-----------------------------------------------------------------------------
int read_sdrHeader(fd, filename, procname, pheader )
	int     fd;      file descriptor for file to read 
	char    *filename;       name of file to read, used for error reporting 
	char    *procname;       name of procedure calling this proc 
	struct  sdrHeader *pheader;   place to put header 
-----------------------------------------------------------------------------
int     seek_file(fd, filename, procname, position)
int     fd;      file descriptor for file to write too 
char    *filename;       name of file to write, used for error reporting 
char    *procname;       name of procedure calling this proc 
size_t   position;        position to seek relative to beginning of file 
-----------------------------------------------------------------------------
int     seek_file_rel(fd, filename, procname, position)
int     fd;      file descriptor for file to write too 
char    *filename;       name of file to write, used for error reporting 
char    *procname;       name of procedure calling this proc 
size_t   position;        position to seek relative to beginning of file 
-----------------------------------------------------------------------------
int     seek_write_file(fd, filename, procname, buf, bufsiz, position)
int     fd;      file descriptor for file to write too 
char    *filename;       name of file to write, used for error reporting 
char    *procname;       name of procedure calling this proc 
char    *buf;            buffer to write 
int     bufsiz;          number of bytes to write 
size_t   position;        position to write as absolute offset within file 
-----------------------------------------------------------------------------
string_cat(buffer, bufsiz, string1, string2, string3)
char    *buffer;         buffer to receive catted strings 
int     bufsiz;          size of buffer 
char    *string1,
        *string2,
        *string3;        strings to cat 
-----------------------------------------------------------------------------
char    *string_find(s1, s2)
        char    *s1;
        char    *s2;
-----------------------------------------------------------------------------
char	*strip_white(ptr)	 strip leading white space from a string
char	*ptr;
-----------------------------------------------------------------------------
int form_filename(file_name, name_len, name, exten)
char    *file_name;      buffer to receive file name 
int     name_len;        length of file_name 
char    *name;           root name of file 
char    *exten;          file extension 
-----------------------------------------------------------------------------
int write_file(fd, filename, procname, buf, bufsiz)
int     fd;      file descriptor for file to write too 
char    *filename;       name of file to write, used for error reporting 
char    *procname;       name of procedure calling this proc 
char    *buf;            buffer to write 
int     bufsiz;          number of bytes to write 

-----------------------------------------------------------------------------
int	free_array(p, size)
	char	*p[];			 array of pointers 
	int	size;			 size of p[] 
-----------------------------------------------------------------------------
int	print_string_array(ptr, max, fd)
	char	**ptr;			 array of pointers to strings 
	int	max;			 max # to display 
	int	fd;			 file descriptor 
-----------------------------------------------------------------------------
int  add_to_string_array(pstring, parray,inx, size, len, pname)
	char	*pstring;	 string to add to the array 
	char	*parray[];	 array of string pointers 
	int	*inx;		 index of next available location 
	int	size;		 size of array 
	int	len;		 max len of a string 
	char	*pname;		 name of calling proc 
-----------------------------------------------------------------------------
int	lookup_string_array(parray, pstring, size)
	char	*parray[];	 array of string pointers 
	char	*pstring;	 string to lookup in the array 
	int	size;		 size of array 
-----------------------------------------------------------------------------
int	string_array_move(psarray, sinx, psnext, ssize, pdarray, pdnext,
		dsize, pname)
	char	*psarray[];	 source array of string pointers 
	int	sinx;		 index in source array to move  
	int	*psnext;	 pointer to next available entry in source array
	int	ssize;		 size of source array 
	char	*pdarray[];	 destination array of string pointers 
	int	*pdnext;   pointer to next available entry in destination array
	int	dsize;		 size of destination array 
	char	*pname;		 name of caller 

-----------------------------------------------------------------------------*/


/*****************************************************************************
 *
 *   alloc_2d_array( parray, parraysz, elemsz)
 *
 *	sets up a 2d array by allocating a 1d array to hold the elements
 * 	and placing pointers to each element in parray
 *
 *	Returns:
 *		pointer to element array
 *		NULL if couldn`t allocate element array
 ******************************************************************************/
char    *alloc_2d_array( parray, parraysz , elemsz)
char    *parray[];      /* array of pointers to strings */
int     parraysz;       /* size of array[] */
int     elemsz;         /* size of each element */

{
        char    *elemarray;     /* array containing elements */
	char	*ptr;
        int     elemarray_sz;
        int     i;

        elemarray_sz = parraysz*elemsz;
        if (( elemarray = malloc((unsigned)elemarray_sz)) == NULL)
                return(NULL);
        memset(elemarray,'\0', elemarray_sz);
        for( i = 0, ptr = elemarray; i < parraysz; i++) {
                parray[i] = ptr;
                ptr += elemsz;
        }
        return(elemarray);
}



/****************************************************************************
 *
 *  build_range_string
 * 
 *  forms a range string for types DBF_CHAR, DBF_UCHAR, DBF_SHORT, DBF_USHORT,
 *	 DBF_LONG, DBF_ULONG,  DBF_FLOAT, DBF_DOUBLE, DBF_ENUM
 *    string is of the form "<range1-> range2>"
 *
 * returns 
 *		0 if successful
 *		-1 if failed
 *
 ***************************************************************************/
 int build_range_string(xname, field_type, cvt_type, pstring, prange1, 
	prange2)

	char	*xname;		/* name of caller */
	short	field_type;	/* the name says it all */
	short	cvt_type;	/* field convert type for this field type */
	char	*pstring;	/* place to put range string */
	union	fld_types  *prange1;  /* low limit */
	union	fld_types  *prange2;  /* hi limit */
{
	char	tmp_str[80];
	int	status;
	char	range1_string[RANGE_STRING_SZ];
	char	range2_string[RANGE_STRING_SZ];

	status = 0;
        switch (field_type){
                case (DBF_FLOAT):
			if ( (prange1->float_value < MIN_F_FORMAT_NVAL) ||
				( (prange1->float_value > MAX_F_FORMAT_NVALR) &&
				  (prange1->float_value < MIN_F_FORMAT_PVALR))||
				(prange1->float_value > MAX_F_FORMAT_PVAL) )
				sprintf(range1_string,"%.5g",
					 prange1->float_value);
			else
				sprintf(range1_string, "%.1f",
					prange1->float_value);
			 if ((prange2->float_value < MIN_F_FORMAT_NVAL) ||
				((prange2->float_value > MAX_F_FORMAT_NVALR) &&
				  (prange2->float_value < MIN_F_FORMAT_PVALR))||
				(prange2->float_value > MAX_F_FORMAT_PVAL))
				sprintf(range2_string,"%.5g",
					 prange2->float_value);
			else
				sprintf(range2_string, "%.1f",
					prange2->float_value);
			sprintf(pstring,"<%s -> %s>", range1_string, 
					range2_string);
                        break;
                case (DBF_DOUBLE):
			if ( (prange1->double_value < MIN_F_FORMAT_NVAL) ||
				((prange1->double_value > MAX_F_FORMAT_NVALR) &&
				 (prange1->double_value < MIN_F_FORMAT_PVALR))||
				(prange1->double_value > MAX_F_FORMAT_PVAL))
				sprintf(range1_string,"%.5g",
					prange1->double_value);
			else
				sprintf(range1_string,"%.1f",
					prange1->double_value);
			if( (prange2->double_value < MIN_F_FORMAT_NVAL) ||
				((prange2->double_value > MAX_F_FORMAT_NVAL) &&
				 (prange2->double_value < MIN_F_FORMAT_PVAL)) ||
				(prange2->double_value > MAX_F_FORMAT_PVAL))
				sprintf(range2_string,"%.5g",
					prange2->double_value);
			else
				sprintf(range2_string,"%.1f",
					prange2->double_value);
			sprintf(pstring,"<%s -> %s>", range1_string, 
					range2_string);
			break;
                case (DBF_ENUM):
                        sprintf(pstring,"<%d-> %d>",
                                  prange1->enum_value, prange2->enum_value);
                        break;
		case (DBF_CHAR):
                case (DBF_UCHAR):
                case (DBF_SHORT):
		case (DBF_USHORT):
                case (DBF_LONG):
                case (DBF_ULONG):
        		if (convert_type_to_ascii(xname, field_type, cvt_type,
				 "<", tmp_str, "-> ", prange1) != 0 )
				status = -1;
        		if (convert_type_to_ascii(xname, field_type, cvt_type,
				 tmp_str, pstring, ">", prange2) != 0 )
				status = -1;
			break;
                default :
                        print_error_msg_p(xname, "Unknown range field type",
				PAUSE_F);
			status = -1;
                }
	return(status);
}

/****************************************************************************
 *
 *  close_file
 *
 *	returns:
 *		0 success
 *		-1 error
 ****************************************************************************/
int	close_file(fd, filename, procname)
int	fd;			/* file descriptor of file to close */
char	*filename;		/* name of file to close */
char	*procname;		/* name of caller */
{
	char	msg[80];
	
	if (close(fd)) {
		string_cat(msg, sizeof(msg), "File close error on ", filename,
				"");
		print_syserror_msg_p(procname, msg, errno, PAUSE_F);
		return(-1);
	}
	return(0);
}

/****************************************************************************
 *
 * check_field_range
 *
 *	check a field against it's range limits
 *
 *	returns
 *		0  if within range
 *		-1 if outside of range
 *
 ***************************************************************************/
int check_field_range(procname, pvalue, prange1, prange2, field_type, var_range)
	char	*procname;	/* name of caller */
	union	fld_types	*pvalue;  /* field value */
	union	fld_types	*prange1; /* lower limit */
	union	fld_types	*prange2; /* upper limit */
	short	field_type;		  /* field type */
        int     var_range;      /* != 0 -> current field has a variable range */
{
	int	err_stat;

	if( dis_range_check && var_range)  /* used by filedb to turn off range
						checking  for fields with var
						range depending on fields not
						yet read in */
		return(0);

	err_stat = 0;
	switch (field_type) {
		case (DBF_CHAR):
			if (( pvalue->char_value < prange1->char_value) ||
				(pvalue->char_value > prange2->char_value))
					err_stat = -1;
			break;
		case (DBF_UCHAR):
			if (( pvalue->uchar_value < prange1->uchar_value) ||
				(pvalue->uchar_value > prange2->uchar_value))
					err_stat = -1;
			break;
		case (DBF_SHORT):
			if (( pvalue->short_value < prange1->short_value) ||
				(pvalue->short_value > prange2->short_value))
					err_stat = -1;
			break;
		case (DBF_USHORT):
			if (( pvalue->ushort_value < prange1->ushort_value) ||
				(pvalue->ushort_value > prange2->ushort_value))
					err_stat = -1;
			break;
		case (DBF_LONG):
			if (( pvalue->long_value < prange1->long_value) ||
				(pvalue->long_value > prange2->long_value))
					err_stat = -1;
			break;
		case (DBF_ULONG):
			if (( pvalue->ulong_value < prange1->ulong_value) ||
				(pvalue->ulong_value > prange2->ulong_value))
					err_stat = -1;
			break;
		case (DBF_FLOAT):
			if (( pvalue->float_value < prange1->float_value) ||
				(pvalue->float_value > prange2->float_value))
					err_stat = -1;
			break;
		case (DBF_DOUBLE):
			if (( pvalue->double_value < prange1->double_value) ||
				(pvalue->double_value > prange2->double_value))
					err_stat = -1;
			break;
		case (DBF_ENUM):
			if (( pvalue->enum_value < prange1->enum_value) ||
				(pvalue->enum_value > prange2->enum_value))
					err_stat = -1;
			break;
		default :
			print_error_msg_p(procname,
				 "No range allowed for this field type", 
					PAUSE_F);
			return(-1);
	}
	if (err_stat == -1) {
		print_error_msg_p(procname, "value out of range", PAUSE_F);
		return(-1);
	}
	return(0);
}

/****************************************************************************
 *
 *	check_file_exist()
 *
 *	returns:
 *		0 if stat succeeds
 *		!=0 if can't stat file
 *
 ***************************************************************************/
int	check_file_exist(path)
char	*path;		/* path to check existance of */
{
        struct  stat    buf;

        return(stat(path, &buf));
}

/****************************************************************************
 *
 *  convert_type_to_val
 *
 *	converts a string to field value
 *
 *  returns
 *		0 if successful
 *   		-1 if error
 *
 ****************************************************************************/
int	convert_type_to_val( procname, field_type, cvt_type, pstring, pvalue)
	char  *procname;	/* name of caller */
	short	field_type;	/* field type */
	short	cvt_type;	/* convert type */
	char	*pstring;	/* ascii string  to convert*/
	union	fld_types *pvalue; /* pointer to place to recieve value */
{
	unsigned short	ushort_val;
		short	short_val;

	switch (cvt_type) {
		case (CT_DECIMAL) :
			break;
		case (CT_HEX) :
			if(pstring[0]=='0' && (pstring[1]=='x' || pstring[1]=='X')) pstring += 2;
			break;
		default :
			print_error_msg_p(procname,
			      "Unknown field convert type", PAUSE_F);
			return(-1);
	}
	switch (field_type) {
		case (DBF_CHAR) :
			switch (cvt_type) {
				case (CT_DECIMAL):
					if( sscanf(pstring, "%hd", &short_val)
						 != 1) 
						return(-1);
					break;
				case (CT_HEX):
					if( sscanf(pstring, "%hx", &short_val)
						 != 1) 
						return(-1);
 					break;
			}
			pvalue->char_value = short_val;
			break;
		case (DBF_UCHAR) :
			switch (cvt_type) {
				case (CT_DECIMAL):
					if( sscanf(pstring, "%hu", &ushort_val)
						 != 1) 
						return(-1);
					break;
				case (CT_HEX):
					if( sscanf(pstring, "%hx", &ushort_val)
						 != 1) 
						return(-1);
 					break;
			}
			pvalue->uchar_value = (unsigned char)ushort_val;
			break;
		case (DBF_SHORT) :
			switch (cvt_type) {
				case (CT_DECIMAL):
					if( sscanf(pstring, "%hd", 
						&pvalue->short_value) !=
							1) 
						return(-1);
 		
					break;
				case (CT_HEX):
					if( sscanf(pstring, "%hx", 
						&pvalue->short_value) !=
							1) 
						return(-1);
 					break;
			}
			break;
		case (DBF_USHORT) :
			switch (cvt_type) {
				case (CT_DECIMAL):
					if( sscanf(pstring, "%hu", 
						&pvalue->ushort_value) !=
							1) 
						return(-1);
 		
					break;
				case (CT_HEX):
					if( sscanf(pstring, "%hx", 
						&pvalue->ushort_value) !=
							1) 
						return(-1);
 					break;
			}
			break;
		case (DBF_LONG)  :
			switch (cvt_type) {
				case (CT_DECIMAL):
					if( sscanf(pstring, "%ld", 
						&pvalue->long_value) !=
							1) 
						return(-1);
 		
					break;
				case (CT_HEX):
					if( sscanf(pstring, "%lx", 
						&pvalue->long_value) !=
							1) 
						return(-1);
 					break;
			}
			break;
		case (DBF_ULONG) :
			switch (cvt_type) {
				case (CT_DECIMAL):
					if( sscanf(pstring, "%lu", 
						&pvalue->ulong_value) !=
							1) 
						return(-1);
 		
					break;
				case (CT_HEX):
					if( sscanf(pstring, "%lx", 
						&pvalue->ulong_value) !=
							1) 
						return(-1);
 					break;
			}
			break;
		default :
			print_error_msg_p(procname,
			      "Convert type not allowed for this field type",
					PAUSE_F);
			return(-1);
	}
	return(0);
}

/*****************************************************************************
 *
 *  convert_type_to_ascii
 *
 *  converts a field value to a string
 *
 *  returns:
 *		0  if successful
 *              -1 if error
 *
 ****************************************************************************/
int	convert_type_to_ascii( procname, field_type, cvt_type, plstring,
			 pstring, ptstring, pvalue)
	char  *procname;	/* name of caller */
	short	field_type;	/* field type */
	short	cvt_type;	/* convert type */
	char	*plstring;	/* string to prepend to pstring */
	char	*pstring;	/* place to put resulting string */
	char	*ptstring;	/* string to append to pstring */
	union	fld_types *pvalue; /* pointer to value to convert */
{
	char	tmpstring[80];
	int	err_stat;

	err_stat = 0;
	tmpstring[0] = 0;
	strcpy(pstring, plstring); 
	switch (field_type) {
		case (DBF_CHAR):
			switch (cvt_type) {
				case (CT_DECIMAL):
					sprintf(tmpstring, "%d",
						 pvalue->char_value);
					break;
				case (CT_HEX):
					sprintf(tmpstring, "0x%x",
						pvalue->char_value);
					break;
				default:
					print_error_msg_p(procname, 
						"Unknown field convert type",
							PAUSE_F);
					err_stat = -1;
			}
			break;
		case (DBF_UCHAR):
			switch (cvt_type) {
				case (CT_DECIMAL):
					sprintf(tmpstring, "%u",
						 pvalue->uchar_value);
					break;
				case (CT_HEX):
					sprintf(tmpstring, "0x%x",
						pvalue->uchar_value);
					break;
				default:
					print_error_msg_p(procname, 
						"Unknown field convert type",
							PAUSE_F);
					err_stat = -1;
			}
			break;
		case (DBF_SHORT):
			switch (cvt_type) {
				case (CT_DECIMAL):
					sprintf(tmpstring, "%d",
						 pvalue->short_value);
					break;
				case (CT_HEX):
					sprintf(tmpstring, "0x%x",
						pvalue->short_value);
					break;
				default:
					print_error_msg_p(procname, 
						"Unknown field convert type",
							PAUSE_F);
					err_stat = -1;
			}
			break;
		case (DBF_USHORT):
			switch (cvt_type) {
				case (CT_DECIMAL):
					sprintf(tmpstring, "%u",
						 pvalue->ushort_value);
					break;
				case (CT_HEX):
					sprintf(tmpstring, "0x%x",
						pvalue->ushort_value);
					break;
				default:
					print_error_msg_p(procname, 
						"Unknown field convert type",
							PAUSE_F);
					err_stat = -1;
			}
			break;
		case (DBF_LONG):
			switch (cvt_type) {
				case (CT_DECIMAL):
					sprintf(tmpstring, "%ld",
						 pvalue->long_value);
					break;
				case (CT_HEX):
					sprintf(tmpstring, "0x%lx",
						pvalue->long_value);
					break;
				default:
					print_error_msg_p(procname, 
						"Unknown field convert type",
							PAUSE_F);
					err_stat = -1;
			}
			break;
		case (DBF_ULONG):
			switch (cvt_type) {
				case (CT_DECIMAL):
					sprintf(tmpstring, "%lu",
						 pvalue->ulong_value);
					break;
				case (CT_HEX):
					sprintf(tmpstring, "0x%lx",
						pvalue->ulong_value);
					break;
				default:
					print_error_msg_p(procname, 
						"Unknown field convert type",
							PAUSE_F);
					err_stat = -1;
			}
			break;
                case (DBF_FLOAT):
                        if ( (pvalue->float_value < MIN_F_FORMAT_NVAL) ||
                                ( (pvalue->float_value > MAX_F_FORMAT_NVALR) &&
                                  (pvalue->float_value < MIN_F_FORMAT_PVALR))||
                                (pvalue->float_value > MAX_F_FORMAT_PVAL) )
                                sprintf(tmpstring,"%.5g",
                                         pvalue->float_value);
                        else
                                sprintf(tmpstring, "%.1f",
                                        pvalue->float_value);
                        break;
                case (DBF_DOUBLE):
                        if ( (pvalue->double_value < MIN_F_FORMAT_NVAL) ||
                                ( (pvalue->double_value > MAX_F_FORMAT_NVALR) &&
                                  (pvalue->double_value < MIN_F_FORMAT_PVALR))||
                                (pvalue->double_value > MAX_F_FORMAT_PVAL) )
                                sprintf(tmpstring,"%.5g",
                                         pvalue->double_value);
                        else
                                sprintf(tmpstring, "%.1f",
                                        pvalue->double_value);
                        break;
		case (DBF_ENUM) :
			sprintf(tmpstring, "%d", pvalue->enum_value);
			break;
		default:
			print_error_msg_p(procname,
		 "Convert type specifier not supported for this field type",
				PAUSE_F);
			err_stat = -1;
	}
	strcat(pstring, tmpstring);
	strcat(pstring, ptstring);
	return(err_stat);
}

/****************************************************************************
 *
 *	copy_file
 *
 *	copies one file to another,  both files must be open when this 
 *		routine is called
 *
 *	returns:
 *			0  success
 *			-1 error writing destination file
 *			-2 error reading source file
 *
 ***************************************************************************/
int  copy_file(procname, fds, fdd, buffer, bufsize)
	char	*procname;	/* name of caller */
	int	fds;		/* file descriptor of source file */
	int	fdd;		/* file descriptor of destination file */
	char	*buffer;	/* buffer to use for read/write */
	int	bufsize;	/* size of buffer */
{
	int	num_bytes;
	int	status;
	while ( (num_bytes = read(fds, buffer, bufsize)) > 0) {
		if ( ( status = write(fdd, buffer, num_bytes)) != num_bytes) {
			if(status < 0) {
				print_syserror_msg_p(procname,
					"Copy file failed on write", errno, 
						PAUSE_F);
			} else {
				print_error_msg_p(procname,
					"Copy failed, incomplete write",
						PAUSE_F);
			}	
			return(-1);
		}
	}
	if(num_bytes != 0) {
		print_syserror_msg_p(procname, "Copy failed on read", errno,
			PAUSE_F);
		return(-2);
	}
	return(0);
}

/****************************************************************************
 *
 *   create_file
 *
 *   creates a file
 *
 *	returns:
 *			file descriptor of created file if successful
 *		   -1	bad file name
 *		   -2   file already exists
 *		   -3   create error
 *
 ****************************************************************************/

int create_file(procname,root_name, ext, filename,fnsiz)
char	*procname;		/* name of calling proc */
char	*root_name;		/* root name of file to create */
char	*ext;			/* file name extension */
char	*filename;		/* returned complete file name */
int	fnsiz;			/* size of filename buffer */
{

	char	msg[80];
	int	fd;		/* file descriptor */

        if (form_filename(filename, fnsiz, root_name, ext)) {
                string_cat(msg, sizeof(msg), "Could not form  file name ",
                        root_name, ext);
                print_error_msg_p(procname, msg,PAUSE_F);
                return(-1);
        }


        if ((fd = open(filename, O_CREAT | O_EXCL | O_RDWR, DCT_FILE_CREAT_MODE))
                        < 0) {
		if ( errno == EEXIST)
			return(-2);
		string_cat(msg, sizeof(msg), "Couldn`t create file ", filename,
			"");
                print_syserror_msg_p(procname, msg, errno, PAUSE_F);
                return(-3);
        }
 
	return(fd);
}

/****************************************************************************
 *
 *   create_file_force
 *
 *   creates a file, if the file exists it is truncated
 *
 *	returns:
 *			file descriptor of created file if successful
 *		   -1	bad file name
 *		   -3   create error
 *
 ****************************************************************************/

int create_file_force(procname,root_name, ext, filename,fnsiz)
char	*procname;		/* name of calling proc */
char	*root_name;		/* root name of file to create */
char	*ext;			/* file name extension */
char	*filename;		/* returned complete file name */
int	fnsiz;			/* size of filename buffer */
{

	char	msg[80];
	int	fd;		/* file descriptor */

        if (form_filename(filename, fnsiz, root_name, ext)) {
                string_cat(msg, sizeof(msg), "Could not form  file name ",
                        root_name, ext);
                print_error_msg_p(procname, msg,PAUSE_F);
                return(-1);
        }


        if ((fd = open(filename, O_CREAT | O_TRUNC | O_RDWR, DCT_FILE_CREAT_MODE))
                        < 0) {
		string_cat(msg, sizeof(msg), "Couldn`t create file ", filename,
			"");
                print_syserror_msg_p(procname, msg, errno, PAUSE_F);
                return(-3);
        }
 
	return(fd);
}

/****************************************************************************
 *
 *  delete_file
 *
 *	deletes a file 
 *
 *  	returns:
 *		0 successful
 *		-1 error
 *****************************************************************************/
int	delete_file(procname, pname)
char	*procname;	/* name of caller */
char	*pname;		/* path name of file to delete */
{
	char	msg[80];

	if (unlink(pname)) {
		string_cat(msg, sizeof(msg), "Could not delete ", pname, "");
		print_syserror_msg_p(procname, msg, errno, PAUSE_F);
		return(-1);
	}
	return(0);
}

/***************************************************************************
 *
 *	dir_search(filenames, maxno, pattern)
 *
 * 	searches current directory for filenames containing pattern and
 *	places them in filenames[]
 *
 *	Returns:
 *		number of filenames matching pattern else`
 *    		-1 if error
 *
 *****************************************************************************/

int     dir_search( filenames, maxno, pattern)
char    *filenames[];   /* place to put file names containing pattern */
int     maxno;                  /* size of filenames array */
char    *pattern;               /* string to search for */
{
	char	*xname = "dir_search";
	char	*string_find();
        DIR     *dirp;
        struct  dirent  *dp;
        int     num_found;

        if( (dirp = opendir(".")) == NULL ) {
                print_error_msg_p(xname, "Couldnot open dir\n", PAUSE_F);
                return(-1);
        }

/* lint will complain on this because the lint library still declares
        readdir() as returning (struct direct *) instead of the new
        (struct dirent *) .  sigh!!!!! */

        for(dp=readdir(dirp), num_found=0; (num_found < maxno) && (dp != NULL);
				 dp=readdir(dirp)) {
                if ( string_find( pattern, dp->d_name) != NULL ) {
                        strcpy(filenames[num_found++], dp->d_name);
                }
        }
        closedir(dirp);
        return(num_found);
}

/****************************************************************************
 *
 *
 *  fldname_to_string(fldname, string, size)
 *
 *	converts fldname  to a null terminated string and returns string
 ****************************************************************************/
char	*fldname_to_string(fldname, string, size)
char	*fldname;		/* field name (non null terminated */
char	*string;		/*place to put null terminated field name */
int	size;			/* size of field name */
{
	memcpy(string, fldname, size);
	string[size]='\0';
	return(string);
}

/*****************************************************************************
 *
 *   fieldtype_to_string()
 *
 *	converts a field type spec to a string
 *
 ******************************************************************************/
int  fieldtype_to_string(fldtype, pstring, pname)
short	fldtype;		/* field type */
char	*pstring;		/* place to return field type */
char	*pname;			/* name of caller */
{

	switch(fldtype) {
		case (DBF_STRING) :
			strcpy(pstring, "DBF_STRING");
			break;
		case (DBF_CHAR) :
			strcpy(pstring, "DBF_CHAR");
			break;
		case (DBF_UCHAR) :
			strcpy(pstring, "DBF_UCHAR");
			break;
		case (DBF_SHORT) :
			strcpy(pstring, "DBF_SHORT");
			break;
		case (DBF_USHORT) :
			strcpy(pstring, "DBF_USHORT");
			break;
		case (DBF_LONG) :
			strcpy(pstring, "DBF_LONG");
			break;
		case (DBF_ULONG) :
			strcpy(pstring, "DBF_ULONG");
			break;
		case (DBF_FLOAT) :
			strcpy(pstring, "DBF_FLOAT");
			break;
		case (DBF_DOUBLE) :
			strcpy(pstring, "DBF_DOUBLE");
			break;
		case (DBF_ENUM) :
			strcpy(pstring, "DBF_ENUM");
			break;
		case (DBF_GBLCHOICE) :
			strcpy(pstring, "DBF_GBLCHOICE");
			break;
		case (DBF_CVTCHOICE) :
			strcpy(pstring, "DBF_CVTCHOICE");
			break;
		case (DBF_RECCHOICE) :
			strcpy(pstring, "DBF_RECCHOICE");
			break;
		case (DBF_DEVCHOICE) :
			strcpy(pstring, "DBF_DEVCHOICE");
			break;
		case (DBF_INLINK) :
			strcpy(pstring, "DBF_INLINK");
			break;
		case (DBF_OUTLINK) :
			strcpy(pstring, "DBF_OUTLINK");
			break;
		case (DBF_FWDLINK) :
			strcpy(pstring, "DBF_FWDLINK");
			break;
		case (DBF_NOACCESS) :
			strcpy(pstring, "DBF_NOACCESS");
			break;
		default :
			strcpy(pstring, "UNKNOWN_FLD");
			return(-1);
	}
	return(0);
}

/*****************************************************************************
 *
 *   get_file_mtime
 *
 *	gets a file's modification time
 *
 *	returns:
 *		0 if successful
 *		-1 if error
 *
 *****************************************************************************/
int  get_file_mtime( pname, path, mtime)
char	*pname;			/* name of calling proc */
char    *path;                  /* path name of file */
time_t  *mtime;                 /* buffer to return file modification time */
{

        struct  stat    buf;
	char	msg[80];

        if (stat(path, &buf)) {
		string_cat(msg, sizeof(msg), "Could not get mtime of ", 
			path, "");
		print_syserror_msg_p(pname, msg, errno, PAUSE_F);
                return(-1);
        }
        *mtime = buf.st_mtime;
        return(0);
}

/*****************************************************************************
 *
 *   get_file_size
 *
 *	gets a file's size
 *
 *	returns:
 *		size if successful
 *		-1 if error
 *
 *****************************************************************************/
int  get_file_size( pname, path)
char	*pname;			/* name of calling proc */
char    *path;                  /* path name of file */
{

        struct  stat    buf;
	char	msg[80];

        if (stat(path, &buf)) {
		string_cat(msg, sizeof(msg), "Could not get size of ", 
			path, "");
		print_syserror_msg_p(pname, msg, errno, PAUSE_F);
                return(-1);
        }
        return((int)buf.st_size);
}

/**************************************************************************
 *
 *  get_pvd_index_by_pvname()
 *
 *  returns index within the process variable directory of a process variable
 *	name
 ****************************************************************************/
short	get_pvd_index_by_pvname(pvname, pname, report_errs)
char	*pvname;	/* name of process variable to be looked up */
char	*pname;		/* caller's name */
short	report_errs;	/* 0 == suppress error reporting */
{
	int	i;
	char	msg[120];

	if(pvd_present(pname))
		return(-4);
        if ((!legal_name(pvname)) && report_errs){
                string_cat( msg, sizeof(msg), "Illegal pv name: ", pvname, "");
                print_error_msg_p(pname, msg, PAUSE_F);
                 return(-3);
        }
        for (i = strlen(pvname); i < PVNAME_SZ; i++)     /* NULL fill the name*/
                         *(pvname + i) = NULL;

	return(pvdPvSearch(pvname,pvdHash(pvname)));
}


/**************************************************************************
 *
 * get_sdr_filetype
 *
 *	Returns:
 *		the sdr type of a file
 *		-1 if error
 *		-2 if not an sdr file
 **************************************************************************/
int	get_sdr_filetype(filename)
char	*filename;		/* name of file to check */
{
	char	*xname = "get_sdr_filetype";
	int	fd;
	struct	sdrHeader header;

	if ( (fd= open_file(filename, O_RDONLY, xname)) < 0)
		return(-1);
	if ( read_sdrHeader(fd, filename, xname, &header) < 0)

		return(-2);
	if ( close_file(fd, filename, xname) < 0)
		return(-1);
	return(header.type);
}

	


/***************************************************************************
 *
 *   open_file
 *
 *   opens a file for read/write
 *
 *	returns:
 *		>0  file descriptor for opened file 
 *		-1	error on open
 *
 ***************************************************************************/
int	open_file(pname, mode, procname)
char	*pname;		/* name of file to open */
int	mode;		/* mode to open file */
char	*procname;	/* name of caller */
{
	char	msg[80];
	int	fd;

	if ((fd = open(pname, mode)) < 0) {
		string_cat(msg, sizeof(msg), "Open error on ", pname, "");
		print_syserror_msg_p(procname, msg, errno, PAUSE_F);
	}
	return(fd);
}

/****************************************************************************
 *
 *  sparse
 *
 *	parse an string into fields
 *	fill unused fields with a Null string
 *
 *	Returns:	The number of fields parsed
 *
 ***************************************************************************/

int sparse(pstring, pdelim, pfields, num_fields)

char    *pstring;               /* string to parse */
char    *pdelim;                /* string of delimitors */
char    *pfields[];             /* array of pointers to buffers to receive
                                parsed fields*/
int     num_fields;             /* max number of fields to parse */

{

        int     i,j;
        char    *ptr;

        for ( i=0; i < num_fields; i++, pstring=NULL) {
                ptr = strtok(pstring, pdelim);
                if(ptr)
                        strcpy(pfields[i], ptr);
                else
                        break;
        }
        for ( j=i; j < num_fields ; j++) {
                strcpy(pfields[j], "");
        }
        return(i);
}

/**************************************************************************
 *
 *  read_file
 *
 *	file must be opened before calling this routine 
 *
 * returns 
 *		number of bytes read else
 *	    -1 	read error
 *
 **************************************************************************/

int read_file(fd, filename, procname, buf, bufsiz)
int	fd;	/* file descriptor for file to read */
char	*filename;	/* name of file to read, used for error reporting */
char	*procname;	/* name of procedure calling this proc */
char	*buf;		/* buffer to read into*/
int	bufsiz;		/* number of bytes to read */
{

	int	bytes_read;	/* # of bytes actually read */
	char	msg[80];

	if ( (bytes_read = read(fd, buf, bufsiz)) != bufsiz) {
		if (bytes_read >= 0 ) {
			string_cat(msg, sizeof(msg),
			     "Warning, incomplete read of file ", filename, "");
			print_error_msg_p(procname, msg, PAUSE_F);
			return(bytes_read);
		} else {
			string_cat(msg, sizeof(msg), "Read error on ",
				filename, "");
			print_syserror_msg_p(procname, msg, errno, PAUSE_F);
			return(-1);
		}
	}

	return(bytes_read);
}

/**************************************************************************
 *
 *  read_sdrHeader
 *
 *	file must be opened before calling this routine 
 *
 * returns 
 *	     0  success	
 *	    -1 	read error
 *	    -2  not an sdr file
 *
 **************************************************************************/

int read_sdrHeader(fd, filename, procname, pheader )
int	fd;	/* file descriptor for file to read */
char	*filename;	/* name of file to read, used for error reporting */
char	*procname;	/* name of procedure calling this proc */
struct	sdrHeader *pheader;  /* place to put header */
{

	int	bytes_read;
	char	msg[80];

	if ( seek_file(fd, filename, procname, 0L)) {
		return(-1);
	}
	if (  (bytes_read = read(fd, (char *)pheader, sizeof(struct sdrHeader)
			)) != sizeof(struct sdrHeader)) {
		if (bytes_read >= 0 ) {
			string_cat(msg, sizeof(msg),
			     "Warning, incomplete read sdrHeader for  ",
					 filename, "");
			print_error_msg_p(procname, msg, PAUSE_F);
			return(-1);
		} else {
			string_cat(msg, sizeof(msg), "Read error on ",
				filename, "");
			print_syserror_msg_p(procname, msg, errno, PAUSE_F);
			return(-1);
		}
	}

	if(pheader->magic != DBMAGIC)
		return(-2);
	return(0);
}

/***************************************************************************
 *
 *  seek_file
 *
 *  seeks to specified position within an open file
 *
 * returns:
 *	 	>0 the position within the file
 *		-1 seek error
 ***************************************************************************/
int	seek_file(fd, filename, procname, position)
int     fd;     /* file descriptor for file to write too */
char    *filename;      /* name of file to write, used for error reporting */
char    *procname;      /* name of procedure calling this proc */
size_t	position;	/* position to seek relative to beginning of file */
{

        char    msg[80];

	if ( lseek(fd, position, L_SET) != position) {
		string_cat(msg, sizeof(msg), "seek error on ", filename, "");
		print_syserror_msg_p(procname, msg, errno, PAUSE_F);
		return(-1);
	}
	return(0);
}

/***************************************************************************
 *
 *  seek_file_rel
 *
 *  seeks to relative position within an open file
 *
 * returns:
 *	 	>0 the position within the file
 *		-1 seek error
 ***************************************************************************/
int	seek_file_rel(fd, filename, procname, position)
int     fd;     /* file descriptor for file to write too */
char    *filename;      /* name of file to write, used for error reporting */
char    *procname;      /* name of procedure calling this proc */
size_t	position;	/* position to seek relative to beginning of file */
{

        char    msg[80];

	if ( lseek(fd, position, L_INCR) < 0 ) {
		string_cat(msg, sizeof(msg), "seek error on ", filename, "");
		print_syserror_msg_p(procname, msg, errno, PAUSE_F);
		return(-1);
	}
	return(0);
}

/***************************************************************************
 *
 *  seek_write_file
 *
 *  seeks to specified position within an open file and writes the specified
 * number of bytes
 *
 * returns:
 *	 	>0 the number of bytes written
 *		-1 write error
 *		-2 seek error
 ***************************************************************************/
int	seek_write_file(fd, filename, procname, buf, bufsiz, position)
int     fd;     /* file descriptor for file to write too */
char    *filename;      /* name of file to write, used for error reporting */
char    *procname;      /* name of procedure calling this proc */
char    *buf;           /* buffer to write */
int     bufsiz;         /* number of bytes to write */
size_t	position;	/* position to write as absolute offset within file */
{

        char    msg[80];

	if ( lseek(fd, position, L_SET) != position) {
		string_cat(msg, sizeof(msg), "seek error on ", filename, "");
		print_syserror_msg_p(procname, msg, errno, PAUSE_F);
		return(-2);
	}
	return( write_file( fd, filename, procname, buf, bufsiz));
}


/****************************************************************************
*
*  STRING_CAT
*
*  cats strings 
*
*****************************************************************************/
string_cat(buffer, bufsiz, string1, string2, string3)
char	*buffer;	/* buffer to receive catted strings */
int	bufsiz;		/* size of buffer */
char	*string1, 
	*string2,
	*string3;	/* strings to cat */
{
	int	i;

	*buffer = 0;
	i = bufsiz -1;
	if ( i <= 0)
		return(0);
	strncat(buffer, string1, i);
	i -= strlen(string1);
	if ( i <= 0)
		return(0);
	strncat(buffer, string2, i);
	i -= strlen(string2);
	if ( i <= 0)
		return(0);
	strncat(buffer, string3, i);
	return(0);
}

/******************************************************************************
 *
 *  string_find
 *
 *     checks to see if string s1 is contained in string s2
 *
 *	returns
 *		pointer to substring in s2 where s1 was found
 *		NULL if s1 not found in s2
 *****************************************************************************/
char    *string_find(s1, s2)
        char    *s1;
        char    *s2;

{
        unsigned     s1len;

        s1len = strlen(s1);
        for (; s1len <= strlen(s2) ; s2++) {
                if ( (s2 = strchr(s2, *s1)) == NULL)
                        return(NULL);
                if ( s1len <= strlen(s2) )
                        if (strncmp( s1, s2, s1len) == 0)
                                return(s2);
        }
        return(NULL);
}

/****************************************************************************
 *
 * strip_white()
 *
 *	strips leading white space from a string in place
 *	returns a pointer to the string
 ***************************************************************************/
char	*strip_white(ptr)
char	*ptr;
{

	while ( *ptr == ' ' || *ptr == '\t' || *ptr == '\n')
		ptr++;
	return(ptr);
}

/*****************************************************************************
 * FORM_FILENAME
 *
 *
 * forms a file name from a rootname + extension
 *
 *  forms file_name = name.exten
 *  returns
 *              0  success
 *              -1 failure
 ****************************************************************************/


int form_filename(file_name, name_len, name, exten)
char    *file_name;     /* buffer to receive file name */
int     name_len;       /* length of file_name */
char    *name;          /* root name of file */
char    *exten;         /* file extension */
{
        if ( (strlen(name) + strlen(exten) + 2) > name_len)
                return(-1);
        strcpy(file_name, name);
        strcat(file_name, ".");
        strcat(file_name, exten);
        return(0);
}

/**************************************************************************
 *
 *  write_file
 *
 *	file must be opened before calling this routine 
 *
 * returns 
 *	     0 success	
 *	    -1 	write file error
 *
 **************************************************************************/

int write_file(fd, filename, procname, buf, bufsiz)
int	fd;	/* file descriptor for file to write too */
char	*filename;	/* name of file to write, used for error reporting */
char	*procname;	/* name of procedure calling this proc */
char	*buf;		/* buffer to write */
int	bufsiz;		/* number of bytes to write */
{

	int	bytes_written;	/* # of bytes actually written */
	char	msg[80];

	if ( (bytes_written = write(fd, buf, bufsiz)) != bufsiz) {
		if (bytes_written >= 0 ) {
			string_cat(msg, sizeof(msg),
			     "Warning, partial write to file ", filename, "");
			print_error_msg_p(procname, msg, PAUSE_F);
			return(-1);
		} else {
			string_cat(msg, sizeof(msg), "Write error on ",
				filename, "");
			print_syserror_msg_p(procname, msg, errno, PAUSE_F);
			return(-1);
		}
	}

	return(0);
}

/****************************************************************************
 *
 *	fchoice()
 *
 *	display and processe choices
 *****************************************************************************/
int	fchoice(pchoice, helpinx, choice)
struct	fchoice	*pchoice;		/* choice definition structure */
int	helpinx;	/* index of help message to print for this menu*/	
int	choice;		/* last choice */
{
	int	i;
	int	execute;

	clear();
	move(pchoice->titlep.y, pchoice->titlep.x);
	addstr(pchoice->title);
/*	display the prompts */

	for( i=0; i < pchoice->nprompts; i++) {
		move(pchoice->pmtp.y + i, pchoice->pmtp.x);
		addstr(pchoice->choice[i]);
	}
	print_help_msg(helpinx);
	refresh();
	if (choice >=0 && choice < pchoice->nprompts)
		i = choice;
	else
		i=0;
	do { 
		move(pchoice->pmtc.y+i, pchoice->pmtc.x);
		refresh();
		execute = 0;
		switch(getinpch()) {
			case (EXECUTE) :
				execute++;
				break;
			case (NEXT) :
				if( ++i >= pchoice->nprompts)
					i = 0;
				break;
			case (PREVIOUS) :
				if (--i < 0)
					i = pchoice->nprompts -1;
				break;
			case (NEXT_ITEM) :
			case (PREV_ITEM) :
				break;
			case (QUIT1):	
			case (QUIT2):	
			case (QUIT3):	
			case (EXIT) :
				return(DONE);
			default :
				break;
		}
	} while ( !execute);
	return(i);
}
/****************************************************************************
 *
 *	fdisplay()
 *
 *	display a page
 *****************************************************************************/
int	fdisplay(pchoice, helpinx)
struct	fchoice	*pchoice;		/* choice definition structure */
int	helpinx;	/* index of help message to print for this menu*/	
{
	int	i;

	clear();
	move(pchoice->titlep.y, pchoice->titlep.x);
	addstr(pchoice->title);
/*	display the prompts */

	for( i=0; i < pchoice->nprompts; i++) {
		move(pchoice->pmtp.y + i, pchoice->pmtp.x);
		addstr(pchoice->choice[i]);
	}
	print_help_msg(helpinx);
	move(pchoice->pmtc.y, pchoice->pmtc.x);
	refresh();
	return(0);
}
/****************************************************************************
 *
 *	fenum()
 *
 *	convert an enumerated value to a string 
 ****************************************************************************/
char	*fenum(enum_val, pfenum)
short	enum_val;		/* enum value to convert */
struct	fenum	*pfenum;	/* pointer to defining structure */
{
	int	i;
	char	string[80];

	if(!pfenum) {
		sprintf(string, "fenum: NULL pointer encountered");
		return(string);
	}else {
	    for ( i = 0 ; i < pfenum->num; i++) {
		if(pfenum->item[i].val == enum_val)
			if (pfenum->item[i].string)
				return(pfenum->item[i].string);
			else {
				sprintf(string, "fenum: NULL string for val %d",
					enum_val);
				return(string);
			}
	     }
	}
	sprintf(string, "UNKNOWN value = %d", enum_val);
	return(string);
}

/****************************************************************************
 *
 *	fmenu()
 *
 *	display and processe menus
 *****************************************************************************/
int	fmenu(pmenu, helpinx, choice)
struct	fmenu	*pmenu;		/* menu definition structure */
int	helpinx;	/* index of help message to print for this menu*/	
int	choice;		/*  choice to start cursor on */
{
	int	i,j;
	int	execute;

	clear();
	i=0;
	while(pmenu->title[i] != NULL) {
		move(pmenu->titlep.y+i, pmenu->titlep.x);
		addstr(pmenu->title[i++]);
	}

/*	display the prompts */
	for( i=0; i < pmenu->nprompts; i++) {
		move(pmenu->pmtp.y + i, pmenu->pmtp.x);
		addstr(pmenu->pmt[i].pmt);
	}
	print_help_msg(helpinx);
	refresh();
	i=choice;
	do { 
		move(pmenu->pmtc.y+i, pmenu->pmtc.x);
		refresh();
		execute = 0;
		switch(getinpch()) {
			case (EXECUTE) :
				execute++;
				break;
			case (NEXT) :
				if( ++i >= pmenu->nprompts)
					i = 0;
				break;
			case (PREVIOUS) :
				if (--i < 0)
					i = pmenu->nprompts -1;
				break;
			case (NEXT_ITEM) :
			case (PREV_ITEM) :
				break;
			case (QUIT1):	
			case (QUIT2):	
			case (QUIT3):	
				for(j=0; j < pmenu->nprompts; j++) {
					if(pmenu->pmt[j].key == QUIT1)
						return(j);
				}
				break;
			case (EXIT) :
				for(j=0; j < pmenu->nprompts; j++) {
					if(pmenu->pmt[j].key == EXIT)
						return(j);
				}
				break;
			default :
				break;
		}
	} while ( !execute);
	return(i);
}

/****************************************************************************
 * FILL_STRING
 *
 * fill a string with the designated character
 ****************************************************************************/
 fill_string(pbuffer,size,fillchar)
char  *pbuffer;
short size;
char  fillchar;
 {
        short  i;

        for (i=0; i<size; i++){
                *pbuffer = fillchar;
                pbuffer++;
        }
	*pbuffer= '\0';
}

/*****************************************************************************
 *
 *  free_array()
 * 
 *****************************************************************************/
int	free_array(p, size)
char	**p;			/* array of pointers */
int	size;			/* size of p[] */
{

	int 	i;

	if(p) {
		for(i=0; i<size; i++)
			if(*(p+i))
				free(*(p+i));
		free((char *)p);
	}
	return(0);
}

/******************************************************************************
 *  print_string_array()
 *
 *****************************************************************************/
int	print_string_array(ptr, newmax, fd)
	char	**ptr;			/* array of pointers to strings */
	int	newmax;			/* newmax # to display */
	int	fd;			/* file descriptor */
{
	int	i;
	
	i=0;
	while(ptr[i] && (i<newmax)) {
		write(fd, ptr[i], strlen(ptr[i]));
		i++;
	}
	return(0);
}

/*****************************************************************************
 *
 *  add_to_string_array()
 *
 ****************************************************************************/
int  add_to_string_array(pstring, parray,inx, size, len, pname)
char	*pstring;	/* string to add to the array */
char	*parray[];	/* array of string pointers */
int	*inx;		/* index of next available location */
int	size;		/* size of array */
int	len;		/* max len of a string */
char	*pname;		/* name of calling proc */
{
	if(strlen(pstring) > len) {
		print_error_msg_p(pname, "Bad PV name", PAUSE_F);
		return(-1);
	}
	if( *inx >= size) {
		print_error_msg_p(pname, "Index too big for array", PAUSE_F);
		return(-1);
	}
	if((parray[*inx]=calloc((unsigned)(len+1), sizeof(char))) == NULL){
		print_syserror_msg_p(pname,"malloc", errno, PAUSE_F);
		return(-2);
	}
	strcpy(parray[(*inx)++], pstring);
	return(0);
}

/****************************************************************************
 *
 *  lookup_string_array()
 *
 ****************************************************************************/
int	lookup_string_array(parray, pstring, size)
char	*parray[];	/* array of string pointers */
char	*pstring;	/* string to lookup in the array */
int	size;		/* size of array */
{
	int i;

	i = 0;
	while ( (i < size) && (parray[i] != NULL)) {
		if(!strcmp(pstring, parray[i]))
			return(i);
		i++;
	}
	return(-1);	/* didn't find the string */
}


/****************************************************************************
 *
 *  string_array_move()
 *
 ****************************************************************************/
int	string_array_move(psarray, sinx, psnext, ssize, pdarray, pdnext,
		dsize, pname)
char	*psarray[];	/* source array of string pointers */
int	sinx;		/* index in source array to move  */
int	*psnext;	/* pointer to next available entry in source array*/
int	ssize;		/* size of source array */
char	*pdarray[];	/* destination array of string pointers */
int	*pdnext;	/*pointer to next available entry in destination array*/
int	dsize;		/* size of destination array */
char	*pname;		/* name of caller */
{
	if ( *pdnext >= dsize-1) {
		print_error_msg_p(pname, "Ran out of room in string array",
			PAUSE_F);
		return(-1);
	} 
	pdarray[(*pdnext)++]=psarray[sinx];
	while ( (psarray[sinx + 1]) != NULL && (sinx < ssize)) {
		psarray[sinx]=psarray[sinx+1];
		sinx++;
	}
	if(sinx < ssize)
		psarray[sinx]=NULL;
	if(*psnext > 0)
		(*psnext)--;
	return(0);
}	

