static char SccsId[] = "$Id: dctutil.c,v 1.17 1995/04/26 13:35:12 mrk Exp $ ";

/*  $Id: dctutil.c,v 1.17 1995/04/26 13:35:12 mrk Exp $
 *	Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 *	.01	4-1-92	frl	Added SPC_HDWLINK
 * .02  05-18-92        rcz     New database access - XGET_PRECTYPE is old GET_PRECTYPE
 * .03  07-16-92        jba     Removed reference to dbr_field_type
 * 	...
 */

/****************************************************************************
 *
 *    dctutil.c
 *
 *    collection of dct utility routines 
 *
 ****************************************************************************/

#include        <stdlib.h>
#include        <stddef.h>
#include	<stdio.h>
#include	<fcntl.h>
#include	<errno.h>
#include	<curses.h>
#include	<string.h>
#include	<dbDefs.h>
#include	<dbRecords.h>
#include	<dbRecType.h>
#include	<dbRecDes.h>
#include	<sdrHeader.h>
#include	<dbFldTypes.h>
#include	<special.h>
#include	"dbPvd.h"
#include	"dctDef.h"
#include	"dctmacros.h"

extern struct recType *dbRecType;
/* the following define convert types for field description fields */
#define		DCT_FLDNAME_T	1
#define		DCT_SPECIAL_T	2
#define		DCT_DBRTYPE_T	3
#define		DCT_CVT_T	4
#define		DCT_RANGE_T	5

extern	char	*fldname_to_string();
extern	char	*alloc_2d_array();
extern	char	*convert_special_val_to_string();
extern	short	pvdPvSearch();
extern  short	pvHash();
extern	short	get_pvd_index_by_pvname();
extern	char	*fenum();

extern short   curr_rectype;           /* current record type in mem */
extern	int	dct_mode;		/* dct mode flag */
extern struct  fldDes  **pflddef;              /* current record's field
                                                definitions */
extern struct  recTypDes       *precorddes;            /* current record type
                                                definition structure */
extern struct  recLoc          pd_recLoc;              /* header for a rec db file */
extern struct  sdrHeader       pd_sdrHeader;           /* sdr header for a rec dbfile*/
extern struct  pvd		*dbPvd;
extern	short			suppress_errors;
extern	short	order[MAX_PV+4];	/* alphabetically ordered pv names */
/* process variable */
extern char    ctrl_record[];   /* PV being modified    */
extern char	rec_file[];	/* name of current record type file */	
extern char	pvd_filename[];
/* the following defines  the the utility menu */

extern	int	show_record_def();
extern	int	show_field_def();
extern	int	dump_pvd();
extern	int	check_database();
extern	int	scan_database_file();
extern	int	rebuild_pvd();

int	exit_util_menu()
{
	return(EXIT);
}

struct	menu_prompt	util_menu_prompts[] = {
	0, show_record_def,NOKEY, "Show Record Descriptions",
	1, show_field_def, NOKEY, "Show Field Descriptions",
	2, dump_pvd, NOKEY, "Dump Process Variable Directory",
	3, check_database, NOKEY, "Database Consistency Check",
	4, scan_database_file, NOKEY, "List sdr files in .database file",
	5, rebuild_pvd, NOKEY, "Regenerate Process Variable Directory (pvd)",
	6, exit_util_menu, EXIT, "Exit This Menu"
};

char	*util_title[] = {
	"DCT Utilities",
	NULL
};

struct	fmenu	util_menu_s = {
	8,1,			/* title position */
	2,3,			/* place to start prompt text */
	0,3,			/* place to start prompt cursor */
	util_title,		/* title strings */
	sizeof(util_menu_prompts)/sizeof(struct menu_prompt),  /* # of prompts*/
	util_menu_prompts		/* array of prompts */
};

struct	fchoice	f_fld_names = {
	8,1,				/* title position */
	2,3,				/* place to start prompt text */
	0,3,				/* place to start prompt cursor */
	"Select Field to Display", /* title */
	0,			/* # of choices gets set later */
	NULL	/* array of field names to be filled in later */
};

struct	fenum_el sdr_types_a[] = {
	SDR_DB_RECTYPE, 	"SDR_DB_RECTYPE",
	SDR_DB_RECORDS, 	"SDR_DB_RECORDS",
	SDR_DB_RECDES, 		"SDR_DB_RECDES",
	SDR_DB_PVD, 		"SDR_DB_PVD",
	SDR_CHOICEGBL,		"SDR_CHOICEGBL",
	SDR_CHOICECVT,		"SDR_CHOICECVT",
	SDR_CHOICEREC, 		"SDR_CHOICEREC",
	SDR_CHOICEDEV,		"SDR_CHOICEDEV",
	SDR_DEVSUP, 		"SDR_DEVSUP",
	SDR_CVTTABLE, 		"SDR_CVTTABLE",
	SDR_DRVSUP, 		"SDR_DRVSUP",
	SDR_RECSUP, 		"SDR_RECSUP",
	SDR_ALLSUMS,		"SDR_ALLSUMS"
};
struct	fenum 	sdr_types = {
	sizeof(sdr_types_a)/sizeof(struct fenum_el),
	sdr_types_a
};

/**************************************************************************
 *
 *	dct_utils
 *
 *  dct utility menu
 *
 ***************************************************************************/
int	dct_utils()
{
	char	*xname="dct_utils";
	short	selected;
	int	choice;

	choice=selected = 0;
	while( selected != EXIT) {
		selected = (*(util_menu_s.pmt[choice=fmenu(&util_menu_s, 1, 
				choice)].proc))();
	}
	return(0);
}

/**************************************************************************
 *
 *    print_pvd_entry_header
 *
 **************************************************************************/
int	print_pvd_entry_header()
{
        char *heading = "Index     PVname                  Rec Type  Rec No.   Next\n\n";

	printf(heading);
	return(0);
}

/***************************************************************************
 *
 *	print_pvd_entry()
 *
 ***************************************************************************/
int	print_pvd_entry(index)
	int	index;
{
	if ( index >= MAX_PV)
		return(-1);
	printf("%5.5d %-*.*s %5.5d    %5.5d   %8.8x\n",
		index,
		PVNAME_SZ,
		PVNAME_SZ,
		dbPvd->entry[index].name,
		dbPvd->entry[index].record_type,
		dbPvd->entry[index].record_no,
		dbPvd->entry[index].next);
	return(0);
}

/***************************************************************************
 *
 *	print_sdr_heading()
 *
 ***************************************************************************/
int	print_sdr_heading()
{
	char	*heading = "nbytes     type\n\n";

	printf("%s", heading);
}
/****************************************************************************
 *
 *	print_sdr_header()
 *
 *	prints the contents of an sdrHeader structure
 *
 *****************************************************************************/
int	print_sdr_header(p)
struct	sdrHeader	*p;
{
	char	*xname = "print_sdr_header";

	if(!p){
		print_error_msg_p(xname, "Null pointer encountered", PAUSE_F);
		return(-1);
	}
	printf("%8.8d %s\n", p->nbytes, fenum(p->type, &sdr_types));
	return(0);
}

/*************************************************************************
 *
 *  rebuild_pvd()
 *
 *************************************************************************/
int	rebuild_pvd()
{
	char	*xname = "rebuild_pvd";
	char	input[80];
	char	input1[16];
	char	file_name[80];
	char	msg[96];
	int	fd;
	char	*prectype;
	char	*ppvname;
	short	type;
	short	fldtype, fldoffset, fldsize;
	int	i;
	time_t	mtime;
	char	*msg1 = "This will reconstruct the process variable directory from the record files";
	char	*msg2 = "Continue<Y:N> ";

	if(select_db(input, "REBUILD PVD") < 0 )
		return(0);
	if(open_pd(input) < -1)
		return(0);
	clear();
	move(0,0);
	printw("%s\n%s", msg1, msg2);
	refresh();
	getline(input1,2,1,strlen(msg2));
	if(( input1[0] != 'Y') && (input1[0] != 'y'))
		return(0);
	/* lets get the mtime */
	if(form_filename(file_name, sizeof(file_name), input, "dctsdr")) {
		string_cat(msg, sizeof(msg), "Could not form filename:",
				file_name, ".database");
		print_error_msg_p(xname, msg, PAUSE_F);
		return(-1);
	}
	delete_file(xname, pvd_filename);
	if(get_file_mtime(xname, file_name, &mtime))
		return(-1);
	if(pvd_create(input, mtime) < 0)
		return(-1);
	
	for(type=1; type < dbRecType->number; type++) {
		if(!(prectype=(char*)XGET_PRECTYPE(type)))
			continue;
		printf("Rebuilding pvd for record type: %s\n", prectype);
		if(get_recfld(type, "NAME", &fldtype, &fldoffset, &fldsize)){
			printf("Couldn`t look up NAME for record type %s",
				prectype);
			continue;
		}
		if(open_record(type) < 0)
			continue;
	/* check each record */
		for(i = 0; i < pd_recLoc.no_records; i++) {
			if (read_record(i, xname) < 0) {
				close_record();
				continue;
			}
			ppvname = (char *)&ctrl_record[fldoffset];
	/* skip unused or deleted records */
			if ( !legal_name(ppvname))
				continue;
			add_pv(ppvname,type, i, 1); 
		}
		close_record();
	}
	save_pvd();
	close_pd(CREATE_DB);
}

/**************************************************************************
 *
 *	scan_database_file()
 *
 ***************************************************************************/
int	scan_database_file()
{
	char	*xname = "scan_database_file";
	char	input[80];
	char	file_name[80];
	char	msg[96];
	int	fd;
	struct sdrHeader p;
	struct recLoc	pr;

	if(select_db(input, "SCAN") < 0 )
		return(0);
	if(form_filename(file_name, sizeof(file_name), input, "database")) {
		string_cat(msg, sizeof(msg), "Could not form filename:",
				file_name, ".database");
		print_error_msg_p(xname, msg, PAUSE_F);
		return(-1);
	}
	clear();
	move(0,0);
	refresh();
	dct_mode = DCT_NONINTERACTIVE;
	if (( fd = open_file(file_name, O_RDONLY, xname)) < 0) {
		getinpch();
		dct_mode = DCT_INTERACTIVE;
		return(-1);
	}
	print_sdr_heading();
	while(1) {
	if (read_file(fd, file_name, xname, &p, sizeof(struct sdrHeader)) 
			< sizeof(struct sdrHeader)) {
		getinpch();
		dct_mode = DCT_INTERACTIVE;
		return(-1);
	}
	if(p.magic != DBMAGIC) {
		print_error_msg(xname, "Bad MAGIC in sdrHeader");
		getinpch();
		dct_mode = DCT_INTERACTIVE;
		return(-1);
	}
	print_sdr_header(&p);
	if( p.type == SDR_DB_RECORDS) {
		if (read_file(fd, file_name, xname, &pr, sizeof(struct recLoc)) 
				< sizeof(struct recLoc)) {
			getinpch();
			dct_mode = DCT_INTERACTIVE;
			return(-1);
		}
		printf("\t\t\tRecord Type = %5.5d, Record Size = %5.5d\n",
			pr.record_type, pr.rec_size);
		if(seek_file_rel(fd, file_name, xname, (size_t)(p.nbytes
				- sizeof(struct recLoc)))) 
			break;
		
	}else {
	 	if(seek_file_rel(fd, file_name, xname, (size_t)p.nbytes)) 
			break;
	}
	}	/* while */
	close_file(fd, file_name, xname);
	dct_mode = DCT_INTERACTIVE;
	printw("\nHit any key to continue\n\n");
	refresh();
	getinpch();
	return(0);		 
}

/***************************************************************************
 *
 *	check_database()
 *
 *	performs several consistency checks on the database
 *
 ***************************************************************************/
 int	check_database()
{
	char	*xname = "check_database";
	char	dbname[MAX_FILENAME_SZ];
	char	*prectype;
	int	ok;
	int	i,j;
	int	file_size;
	char	pvname[PVNAME_SZ+1];
	char	*ppvname;
	struct	pvd_entry	*pentry;
	short	type;			/* record type */
	short	rec_no;			/* current record number */
	short	pvdi;
	short	fldtype, fldoffset, fldsize;

	if (select_db(dbname, "Dump Process Variable Directory") < 0 )
		return(0);
	if(open_pd(dbname) < -1)
		return(0);

	clear();
	move(0,0);
	refresh();
	printf("Checking pvd for illegal/unknown record types\n");
	ok = 1;
	pentry = dbPvd->entry;
	for ( i = 0; i < MAX_PV; i++, pentry++) {
		if ( pentry->record_type < 0 ||
				 pentry->record_type >= dbRecType->number) {
			if(ok) {
				ok = 0;
				printf("The following pvd entries have illegal record types\n");
				print_pvd_entry_header();
			}
			print_pvd_entry(i);
		}
	}
	if(ok)
		printf("No unknown record types found in pvd\n");

	printf("Checking pvd hash values and duplicate entries\n");
	ok = 1;
	pentry = dbPvd->entry;
	for ( i = 0; i < MAX_PV; i++, pentry++) {
	/* check if hash lookup returns the present pv entry */
		if ( legal_name (&pentry->name[0] )) {
			if ((j=(int)get_pvd_index_by_pvname(pentry->name,
					 xname, 0))
				< 0) {
				ok = 0;
				printf("The following pvd entry is not in an accessible hash chain.\n");
				print_pvd_entry_header();
				print_pvd_entry(i);
			 } else if ( j != i) {
				ok = 0;
				printf("The following are duplicate entries\n");
				print_pvd_entry_header();
				print_pvd_entry(i);
				print_pvd_entry(j);
			}
		} else {
	/* check for all blanks in unused entry */
			for (j=0; j < PVNAME_SZ; j++) {
				if(pentry->name[j] != ' ') {
					ok = 0;
					printf("The following unused pvd entry has non blank characters in the name field\n");
					print_pvd_entry_header();
					print_pvd_entry(i);
					break;
				}
			}	
		}
	}
	if(ok)
		printf("All hash values resolved in pvd and no duplicate entries found\n");

	printf("Checking pvd entries against record type files\n");
	for(type=1; type < dbRecType->number; type++) {
		if(!(prectype=(char*)XGET_PRECTYPE(type)))
			continue;
		if(order_pvs(type, order, MAX_PV+4) == 0)
			continue;
		printf("Checking record type: %s\n", prectype);
		if(get_recfld(type, "NAME", &fldtype, &fldoffset, &fldsize)){
			printf("Couldn`t look up NAME for record type %s",
				prectype);
			continue;
		}
		ok = 1;
		for(i = 0; order[i] != -1; i++) {
			if(open_record(type) < 0)
				continue;
			if (read_record(dbPvd->entry[order[i]].record_no,
				xname) < 0) {
				close_record();
				continue;
			}
			close_record();
	/* check pv name in pvd against name in record */
			if(strncmp(dbPvd->entry[order[i]].name,
				 (char *)&ctrl_record[fldoffset], PVNAME_SZ)) {
				if(ok) {
					ok=0;
					printf("The following pvd entries have name mismatches:\n");
					print_pvd_entry_header();
				}
				print_pvd_entry((int)order[i]);
				printf("Control record name = %-*.*s\n",
					PVNAME_SZ, PVNAME_SZ,
					(char *)&ctrl_record[fldoffset]);
			}
		}
	}
/* check each pv record file */
	printf("Checking process variable record files\n");
	for(type=1; type < dbRecType->number; type++) {
		if(!(prectype=(char*)XGET_PRECTYPE(type)))
			continue;
		printf("Checking record type: %s\n", prectype);
		if(get_recfld(type, "NAME", &fldtype, &fldoffset, &fldsize)){
			printf("Couldn`t look up NAME for record type %s",
				prectype);
			continue;
		}
		if(open_record(type) < 0)
			continue;
	/* check the  sdrHeader */
		ok = 1;
		if( pd_sdrHeader.magic != DBMAGIC ) {
			printf("Error: magic is %x, should be %x\n",
				pd_sdrHeader.magic, DBMAGIC);
			ok = 0;
		}
		if( pd_sdrHeader.type != SDR_DB_RECORDS) {
			printf("Error: sdrHeader.type is %x, should be %x\n",
				pd_sdrHeader.type, SDR_DB_RECORDS);
			ok=0;
		}
		if ( (file_size=get_file_size(xname, rec_file)) < 0) {
			printf("Couldn't get file size\n");
			ok=0;
		} else if( file_size !=
			 (pd_sdrHeader.nbytes + sizeof(struct sdrHeader))) {
			printf("Error: filesize is %d, should be %d\n",
				file_size,
				pd_sdrHeader.nbytes + sizeof(struct sdrHeader));
			ok=0;
		}
	/* check the pd header */
		if(set_current_record_type(type))
			continue;
		if (pd_recLoc.record_type != type) {
			printf("Error: recLoc.type is %d, should be %d\n",
				pd_recLoc.record_type, type);
			ok=0;
		}
		if (pd_recLoc.rec_size != precorddes->rec_size) {
			printf("Error: recLoc.rec_size is %d, should be %d\n",
				pd_recLoc.rec_size, precorddes->rec_size);
			ok=0;
		} 
		if ( (pd_recLoc.no_records * pd_recLoc.rec_size) != 
			(pd_sdrHeader.nbytes - sizeof(struct recLoc))) {
			printf("Error: this file has incorrect record sizs\n");
			ok=0;
		}
		if(!ok)
			continue;
	/* check each record */
		for(i = 0; i < pd_recLoc.no_records; i++) {
			if (read_record(i, xname) < 0) {
				close_record();
				continue;
			}
			ppvname = (char *)&ctrl_record[fldoffset];
	/* skip unused or deleted records */
			if ( !legal_name(ppvname))
				continue;
	/* check lookup in pvd */
			if((pvdi=pvdPvSearch(ppvname, pvdHash(ppvname))) <0){
				printf(
	"Error: control record %d, pvname= %s, not found in pv directory\n",
					 i, ppvname);
				ok = 0;
				continue;
			}
	/* check consistancy of record #'s */
			if( dbPvd->entry[pvdi].record_no != i) {
				printf(
       "Error: pv = %s,  pvd entry record # = %d, this control record # = %d\n",
					ppvname,
				dbPvd->entry[pvdi].record_no, i);
				ok=0;
			}
	/* check consistancy of record types */
			if (dbPvd->entry[pvdi].record_type != type) {
				printf("Error: pv = %s, pvd entry record type = %d, this control record type = %d\n",
					ppvname,
				dbPvd->entry[pvdi].record_type, type);
				ok = 0;
			}
		}
		close_record();
	}
	if(ok)
		printf("No errors found in pv records\n");
	close_pd(CLEAN_UP);
	printw("\nHit any key to continue\n\n");
	refresh();
	getinpch();
	return(0);
}

/****************************************************************************
/***************************************************************************
 *
 *	dump_pvd()
 *
 *	prints hash table and pvd entries
 *
 ***************************************************************************/
 int	dump_pvd()
{
	char	dbname[MAX_FILENAME_SZ];

	if (select_db(dbname, "Dump Process Variable Directory") < 0 )
		return(0);
	if(open_pd(dbname) < -1)
		return(0);

	clear();
	move(0,0);
	refresh();
	pvd_dump();
	close_pd(CLEAN_UP);
	printw("\nHit any key to continue\n\n");
	refresh();
	getinpch();
	return(0);
}

/****************************************************************************
 *
 *	show_record_def()
 *
 *	displays struct recTypeDes for each record type 
 *
 *****************************************************************************/
int	show_record_def()
{
	char	dbname[MAX_FILENAME_SZ];
	short	rectype;		/* record type */
	char	*pstr;			/* pointer to record type name string*/
	int	i,j;
	short	*plinks;

	if (select_db(dbname, "SHOW Rec Defs") < 0 )
		return(0);
	if(open_pd(dbname) < -1)
		return(0);

	clear();
	move(0,0);
	refresh();
	printw("Record Definitions for %s\n\n", dbname);
	printw("Type   Name      Size  #Flds #Prompts #Links  Link Fields\n");
	for ( rectype = 1, i=0; rectype < dbRecType->number; rectype++ ) {
		/* check if this record type exists */
		if(!(pstr=(char*)XGET_PRECTYPE(rectype)))
			continue;
		if(set_current_record_type(rectype))
			continue;
		move(i+5,0);
		printw("%3.3d %-*.*s %5.5d  %3.3d    %3.3d     %3.3d  ",
			 rectype,
			MAX_RECNAME_SZ, MAX_RECNAME_SZ, pstr,
			 precorddes->rec_size, precorddes->no_fields,
			 precorddes->no_prompt, precorddes->no_links);
		plinks=precorddes->link_ind;
		for(j=0; j < precorddes->no_links; j++, plinks++)
			printw(" %3.3d", *plinks);
		printw("\n");
		i++;

	}
	close_pd(CLEAN_UP);
	printw("\nHit any key to continue\n\n");
	refresh();
	getinpch();
	return(0);
}

/***************************************************************************
 *
 *	show_field_def
 *
 *	show field definitions for a record type
 *
****************************************************************************/
int	show_field_def()
{
	char	*xname="show_field_def";
	char	dbname[MAX_FILENAME_SZ];
	short	rectype;		/* record type */
	int	i,j;
	char	*pfld_name_array;	/* space to store field names */
	char	**f_fld_names_c;	/* array of field names */
	short	*fld_map;		/* map of field description indexes */
	int	choice;
	struct	fldDes	*pfield;
	char	buffer[80];
	static	char	ptitle[80];		/* place to build title */
	static char	*fld_def_names[] = {
		"prompt string",
		"field name",
		"offset",
		"size",
		"special",
		"field_type",
		"dbr field type",
		"process_passive",
		"choice set",
		"convert type",
		"prompt flag",
		"low limit type",
		"high limit type",
		"interest level",
		"initial val",
		"low limit",
		"high limit"
	};	
	static struct fchoice	fld_def_s = {
		8,1,			/* title position */
		2,3,			/* place to start prompt text*/
		18,3,			/* place to start prompt cursor */
		ptitle,
		sizeof(fld_def_names)/sizeof(char *),
		fld_def_names		
	};

	if (select_db(dbname, "SHOW Field Defs") < 0 )
		return(0);
	if(open_pd(dbname) < -1)
		return(0);
	if(select_type(&rectype) < 0)
		return(0);
	if(set_current_record_type(rectype))
		return(-1);

	if((f_fld_names_c= (char **)calloc((unsigned)precorddes->no_fields, 
			sizeof(char *))) == NULL) {
		print_syserror_msg_p(xname, "calloc", errno, PAUSE_F);
		return(-1);
	}
	f_fld_names.choice=f_fld_names_c;
	if((pfld_name_array=alloc_2d_array(f_fld_names_c,
			 precorddes->no_fields, FLDNAME_SZ+1)) == NULL) {
		print_error_msg_p(xname, "Could not allocate field name array",
				PAUSE_F);
		return(-1);
	}
	if((fld_map=(short *)calloc((unsigned)precorddes->no_fields,
				 sizeof(short)))
			== NULL) {
		print_syserror_msg_p(xname, "calloc", errno, PAUSE_F);
		return(-1);
	}
	for(i=0, j =0; i < precorddes->no_fields; i++) {
		if (prompt_fld(rectype, pflddef[i]) == NOPROMPT)
			continue;
		fldname_to_string(pflddef[i]->fldname, f_fld_names_c[j],
				FLDNAME_SZ);
		fld_map[j++]= i;
	}	
	f_fld_names.nprompts=j;
	choice=0;
	while((choice=fchoice(&f_fld_names,1, choice)) >= 0) {
		strcpy(ptitle, "Field Description for ");
		strcat(ptitle, f_fld_names_c[choice]);
		i=0;
		fdisplay(&fld_def_s, 4);
		if(!(pfield=pflddef[fld_map[choice]])) {
			print_error_msg_p(xname,
				"NULL pointer to fldDes encountered", PAUSE_F);
			break;
		}
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(pfield->prompt);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(fldname_to_string(pfield->fldname, buffer, FLDNAME_SZ));
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		printw("%d",pfield->offset);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		printw("%d",pfield->size);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(convert_special_val_to_string(pfield->special, buffer,
				 DCT_SPECIAL_T, xname));
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(convert_special_val_to_string(pfield->field_type, buffer,
				 DCT_FLDNAME_T, xname));
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		printw("%d",pfield->process_passive);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		printw("%d",pfield->choice_set);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(convert_special_val_to_string(pfield->cvt_type, buffer,
				 DCT_CVT_T, xname));
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		printw("%d",pfield->promptflag);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(convert_special_val_to_string(pfield->lowfl, buffer,
				 DCT_RANGE_T, xname));
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(convert_special_val_to_string(pfield->highfl, buffer,
				 DCT_RANGE_T, xname));
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		printw("%d",pfield->interest);
		/* the following little dance displays the initial value
				using dct's standard calls */
		init_pv();
		fld_to_ascii(pfield, buffer, rectype);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		addstr(buffer);
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		suppress_errors=1;	/* suppress error reporting for range*/
		if(pfield->lowfl == VAR) {
			addstr(fldname_to_string(
				pflddef[pfield->range1.fldnum]->fldname,
					 buffer, FLDNAME_SZ));
		} else if(!convert_type_to_ascii(xname, pfield->field_type,
				pfield->cvt_type, "", buffer, "", 
				&pfield->range1.value)) {
			addstr(buffer);
		}
		move(fld_def_s.pmtc.y+i++, fld_def_s.pmtc.x);
		if(pfield->highfl == VAR) {
			addstr(fldname_to_string(
				pflddef[pfield->range2.fldnum]->fldname,
					 buffer, FLDNAME_SZ));
		} else if(!convert_type_to_ascii(xname, pfield->field_type,
				pfield->cvt_type, "", buffer, "", 
				&pfield->range2.value)) {
			addstr(buffer);
		}
		suppress_errors=0;
		refresh();
		getinpch();
	}
	if(pfld_name_array)
		free(pfld_name_array);
	if(f_fld_names_c)
		free((char *)f_fld_names_c);
	if(fld_map)
		free((char *)fld_map);

	close_pd(CLEAN_UP);
	return(0);
}
/****************************************************************************
 *
 *	convert_special_to_string()
 *
 *	converts values defined in special.h to strings
 *
 ****************************************************************************/
int	convert_special_to_string(spectype, pstring, pname)
short	spectype;		/* value of special field */
char	*pstring;		/* place to receive the string */
char	*pname;			/* callers name*/
{
	switch(spectype) {
		case (0) :
			strcpy(pstring, "nil");
			break;
		case (SPC_NOMOD) :
			strcpy(pstring, "SPC_NOMOD");
			break;
		case (SPC_DBADDR) :
			strcpy(pstring, "SPC_DBADDR");
			break;
		case (SPC_SCAN) :
			strcpy(pstring, "SPC_SCAN");
			break;
		case (SPC_MOD) :
			strcpy(pstring, "SPC_MOD");
			break;
		case (SPC_RESET) :
			strcpy(pstring, "SPC_RESET");
			break;
		case (SPC_LINCONV) :
			strcpy(pstring, "SPC_LINCONV");
			break;
		case (SPC_CALC) :
			strcpy(pstring, "SPC_CALC");
			break;
		case (SPC_HDWLINK) :
			strcpy(pstring, "SPC_HDWLINK");
			break;
		default :
			sprintf(pstring,"UNKNOWN_SPECIAL = %d",spectype);
			break;
	}
	return(0);
}
/****************************************************************************
 *
 *	convert_dbr_to_string()
 *
 *	converts values defined in dbFldTypes.h to strings
 *
 ****************************************************************************/
int	convert_dbr_to_string(dbrtype, pstring, pname)
short	dbrtype;		/* value of dbr_field */
char	*pstring;		/* place to receive the string */
char	*pname;			/* callers name*/
{
	switch(dbrtype) {
		case (DBR_STRING) :
			strcpy(pstring, "DBR_STRING");
			break;
		case (DBR_CHAR) :
			strcpy(pstring, "DBR_CHAR");
			break;
		case (DBR_UCHAR) :
			strcpy(pstring, "DBR_UCHAR");
			break;
		case (DBR_SHORT) :
			strcpy(pstring, "DBR_SHORT");
			break;
		case (DBR_USHORT) :
			strcpy(pstring, "DBR_USHORT");
			break;
		case (DBR_LONG) :
			strcpy(pstring, "DBR_LONG");
			break;
		case (DBR_ULONG) :
			strcpy(pstring, "DBR_ULONG");
			break;
		case (DBR_FLOAT) :
			strcpy(pstring, "DBR_FLOAT");
			break;
		case (DBR_DOUBLE) :
			strcpy(pstring, "DBR_DOUBLE");
			break;
		case (DBR_ENUM) :
			strcpy(pstring, "DBR_ENUM");
			break;
		case (DBR_NOACCESS) :
			strcpy(pstring, "DBR_NOACCESS");
			break;
		default :
			sprintf(pstring, "UNKNOWN_DBRTYPE = %d", dbrtype);
			break;
	}
	return(0);
}
/****************************************************************************
 *
 *	convert_cvt_to_string()
 *
 *	converts values dct conversion types to strings
 *
 ****************************************************************************/
int	convert_cvt_to_string(cvttype, pstring, pname)
short	cvttype;		/* value of cvt_type field */
char	*pstring;		/* place to receive the string */
char	*pname;			/* callers name*/
{
	switch(cvttype) {
		case (CT_DECIMAL) :
			strcpy(pstring, "CT_DECIMAL");
			break;
		case (CT_HEX) :
			strcpy(pstring, "CT_HEX");
			break;
		default :
			sprintf(pstring, "UNKNOWN_CVTTYPE = %d", cvttype);
			break;
	}
	return(0);
}
/****************************************************************************
 *
 *	convert_rangetype_to_string()
 *
 *	converts values range type to strings
 *
 ****************************************************************************/
int	convert_rangetype_to_string(rtype, pstring, pname)
short	rtype;			/* value of range type field */
char	*pstring;		/* place to receive the string */
char	*pname;			/* callers name*/
{
	switch(rtype) {
		case (CON) :
			strcpy(pstring, "CON");
			break;
		case (VAR) :
			strcpy(pstring, "VAR");
			break;
		default :
			sprintf(pstring, "UNKNOWN_TYPE = %d", rtype);
			break;
	}
	return(0);
}
/*****************************************************************************
 *
 *	convert_special_val_to_string()
 *
 ****************************************************************************/
char 	*convert_special_val_to_string(svalue, pstring, type, pname)
short	svalue;			/* value to convert */
char	*pstring;		/* place to receive resulting string*/
int	type;			/* the type of conversion to do */
char	*pname;			/* name of the caller */
{
	
	switch(type) {
		case (DCT_FLDNAME_T) :
			fieldtype_to_string(svalue, pstring, pname);
			break;
		case (DCT_SPECIAL_T) :
			convert_special_to_string(svalue, pstring, pname);
			break;
		case (DCT_DBRTYPE_T) :
			convert_dbr_to_string(svalue, pstring, pname);
			break;
		case (DCT_CVT_T) :
			convert_cvt_to_string(svalue, pstring, pname);
			break;
		case (DCT_RANGE_T) :
			convert_rangetype_to_string(svalue, pstring, pname);
			break;
		default :
			strcpy(pstring,"");
			break;
	}
	return(pstring);
}
