static char SccsId[] = "@(#)dctwin.c	1.8     4/17/92 ";

/*  @(#)dctwin.c	1.8	4/17/92
 *	Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	3/16/92		frl	link constant changed from float to double
 * .02	4/8/92		frl     SPC_HDWLINK support
 * .03	8/24/93		mrk	change some link fields
 *
 * 	...
 */

/*  dctwin.c		dct subwindow routines */
#include <stdio.h>
#include <curses.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#include <dbDefs.h>
#include <link.h>
#include <choice.h>
#include "keys.h"
#include "dctDef.h"

#define	BOX_H	'-'	/* Horizontal edge character for window outline */
#define BOX_V   '|'	/* Vertical edge character for window outline */
#define WIN_QUIT -1	/* quitting a window */
#define WIN_EXIT -2	/* exiting a window */

/*  conversion types  the names speak for themselves */
#define DCT_EXECUTE	1	/* not really a conversion type, but causes
					the procedure pointed to by 
					Mwindow.pmt[i].proc to be executed */
#define	DCT_WSTRING	2
#define DCT_WFLOAT	3
#define DCT_WSHORT	4
#define DCT_WENUM	5
#define DCT_WCATSTRING  6
#define DCT_WUCHAR	7
#define DCT_WDOUBLE	8

#define	DCT_WCATSTRING_SZ  INSTIO_FLD_SZ

extern short   precision;	/* display precision for floats & doubles */
extern	char	*strip_white();

int	quit_proc();
int	exit_proc();
int	vxi_d_exit_proc();
int	vxi_s_exit_proc();
int	pvlink_window();
int	constant_window();
int	pop_enum_window();
int	pop_sub_window();
int	null_proc();
int	win_pvname_in();
int	win_string_in();
int	win_string_out();
int	win_null_in();
int	win_null_out();
int	win_fldname_in();
int	win_fldname_out();
int	win_float_in();
int	win_float_out();
int	win_double_out();
int	win_double_in();
int	win_uchar_out();
int	win_uchar_in();
int	win_short_out();
int	win_short_in();
int	win_enum_in();
int	win_enum_out();
int	win_cstring_in();
int	win_cstring_out();

struct link	tlink;	/* temporary storage for link */

/* the following is used to parse concatenated strings */
static char	field0[DCT_WCATSTRING_SZ];
static char	field1[DCT_WCATSTRING_SZ];
static char	field2[DCT_WCATSTRING_SZ];
static char	field3[DCT_WCATSTRING_SZ];
static char	*pfields[] = {
		field0,
		field1,
		field2,
		field3
};
static	char	*pdelim = ".";		/* the delimiter for catted strings*/

static struct linkmaps {		/* map of link type to window index */
	long	type;			/* link type */
	int	inx;			/* window array index */
	char	*pname;			/* descriptive string */
} linkmap[] = {
	VME_IO,0,	"VME",
	CAMAC_IO,1,	"CAMAC",
	AB_IO,2,	"AB",
	GPIB_IO,3,	"GPIB",
	BITBUS_IO,4,	"BITBUS",
	CONSTANT,5,	"",
	FWDLINKPROMPTS,6,"BADVAL",
	INST_IO,7,	"INST",		/* instrument */
	BBGPIB_IO,8,	"BBGPIB",	/* bitbus to gpib */
	VXI_IO,9,	"VXI"
};

static struct lnmaps {
	int	mapsz;			/* number of linktypes */
	struct linkmaps	*pmap;			/* map array */
} lnmap = {
	sizeof(linkmap)/sizeof(struct linkmaps),
	linkmap
};

/************************** ENUM choices ************************************/
static struct	wchoice {		/* a choice */
	short	val;		/* enumerated value */
	char	*pstr;		/* string for this value */
};


struct  wchoice  wchoices1[] = {
	0, "NPP",
	1, "PP",	
	2, NULL				/* this terminates the list */
};
struct  wchoice  wchoices2[] = {
	0, "NMS",
	1, "MS",
	2, NULL				/* this terminates the list */
};

struct wchoice *choicearray[] = {      /* place index of choice set in
						prompt.inx field */
	wchoices1,			/* choice set index 0 */
	wchoices2			/* choice set index 1 */
};


static struct	wpos	{
	int	x;
	int	y;
	};
static struct  prompts {
	int	inx;		/* for DCT_EXECUTE type contains value to
					return, for DCT_WENUM type contains
					index of choice set  */
	int	(*proc)();	/* pointer to execute proc to for this menu
					 choice */
	int	(*proci)();	/* pointer to input proc to for this menu
					 choice */
	int	(*proco)();	/* pointer to output proc to for this menu
					 choice */
	int	type;		/* conversion type for this choice */
	char	*buf;		/* place to put/get value */
	int	bufsiz;		/* size of buffer */
	struct Mwindow	*pw;	/* pointer to window definition for this choic*/
	char	*pmt;		/* prompt text */
};
struct  Mwindow {		/* Window definition */
        WINDOW  *pwin;          /* pointer to curses WINDOW struct for
                                        this window */
	struct wpos	beg;		/* starting position for window */
	struct wpos	sz;		/* size of window */
	struct wpos	titlep;		/* starting position of window title */
	struct wpos	pmtp;		/* starting location of prompst */
	struct wpos	pmtc;		/* starting location of prompt cursor */
        char            *title;         /* title for window */
	int		nprompts;	/* number or prompts */
	struct	prompts *pmt;	/* array of prompts */
        };


struct	prompts	pvlinkwpvp[] = {
                0, NULL, win_pvname_in, win_string_out, DCT_WSTRING, 
			tlink.value.pv_link.pvname,
			sizeof(tlink.value.pv_link.pvname),   NULL,
		 	"    Process Variable Name :",
                0, NULL, win_fldname_in, win_fldname_out, DCT_WSTRING, 
			tlink.value.pv_link.fldname, 
			sizeof(tlink.value.pv_link.fldname),  NULL,
			"   Process Variable Field :",
                0, pop_enum_window, win_enum_in, win_enum_out, DCT_WENUM,
			(char *)&tlink.value.pv_link.process_passive, 
			sizeof(tlink.value.pv_link.process_passive),  NULL,
			" Process Passive <SELECT> :",
                1, pop_enum_window, win_enum_in, win_enum_out, DCT_WENUM,
			(char *)&tlink.value.pv_link.maximize_sevr, 
			sizeof(tlink.value.pv_link.maximize_sevr),  NULL,
			"Maximize Severity <SELECT>:",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
			NULL, 0, NULL, 
			"                     Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
			NULL, 0, NULL, 
			"                     Quit :"
};
struct  Mwindow pvlinkwpv = {		/* pvlink data entry */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        60, 16,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			29,3,		/* start prompt cursor here */
                        "Enter Link PV Name",	/* window title */
			sizeof(pvlinkwpvp)/sizeof(struct prompts),/*# of
								 prompts*/
			pvlinkwpvp		/* prompts */
        };

struct	prompts	pvlinkwpvp1[] = {	/* prompts pvlink FWDLINK */
                0, quit_proc, win_pvname_in, win_string_out,
			 DCT_WSTRING, tlink.value.pv_link.pvname,
			sizeof(tlink.value.pv_link.pvname),   NULL,
			 "Process Variable Name :",
                0, quit_proc, win_fldname_in, win_fldname_out,
			 DCT_WSTRING, tlink.value.pv_link.fldname, 
			sizeof(tlink.value.pv_link.fldname),  NULL,
			 "Process Variable Field:",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
			"                  Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
			"                  Quit :"
};
struct  Mwindow pvlinkwpv1 = {		/* pvlink data entry FWDLINK */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        56, 16,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			26,3,		/* start prompt cursor here */
                        "Enter Forward Link PV Name",	/* window title */
			sizeof(pvlinkwpvp1)/sizeof(struct prompts),/*# of
								 prompts*/
			pvlinkwpvp1		/* prompts */
        };


struct	prompts	constantwp[] = {
                0, quit_proc, win_double_in, win_double_out,
			 DCT_WDOUBLE, (char *)&tlink.value.value,
			sizeof(tlink.value.value),   NULL,
			" Constant Value :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
			"           Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
			"           Quit :"
};
struct  Mwindow constantw = {		/* constant data entry */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        48, 9,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			19,3,		/* start prompt cursor here */
                        "Enter Constant",	/* window title */
			sizeof(constantwp)/sizeof(struct prompts),/*# of
								 prompts*/
			constantwp		/* prompts */
        };
struct	prompts	vxidynamicwp[] = {	/* prompts vxi dynamic add */
                0, quit_proc, win_short_in, win_short_out,
			 DCT_WSHORT, (char *)&tlink.value.vxiio.frame,
			sizeof(tlink.value.vxiio.frame),   NULL,
			 "    Frame :",
                0, quit_proc, win_short_in, win_short_out,
			 DCT_WSHORT, (char *)&tlink.value.vxiio.slot, 
			sizeof(tlink.value.vxiio.slot),  NULL,
			 "     Slot :",
                0, quit_proc, win_short_in, win_short_out,
			 DCT_WSHORT, (char *)&tlink.value.vxiio.signal, 
			sizeof(tlink.value.vxiio.signal),  NULL,
			 "    Signal:",
                0, quit_proc, win_string_in, win_string_out,
			 DCT_WSTRING, tlink.value.vxiio.parm, 
			sizeof(tlink.value.vxiio.parm),  NULL,
			 "Parameter :",
		WIN_EXIT, vxi_d_exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, "Exit      :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, "Quit      :"
};
struct  Mwindow vxidynamicw = {		/*  */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        48, 10,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			13,3,		/* start prompt cursor here */
                        "VXI Dynamic Address",	/* window title */
			sizeof(vxidynamicwp)/sizeof(struct prompts),/*# of
								 prompts*/
			vxidynamicwp		/* prompts */
        };
struct	prompts	vxistaticwp[] = {	/* prompts vxi static add */
                0, quit_proc, win_short_in, win_short_out,
			 DCT_WSHORT, (char *)&tlink.value.vxiio.la,
			sizeof(tlink.value.vxiio.la),   NULL,
			"Static Address :",
                0, quit_proc, win_short_in, win_short_out,
			 DCT_WSHORT, (char *)&tlink.value.vxiio.signal, 
			sizeof(tlink.value.vxiio.signal),  NULL,
			"         Signal:",
                0, quit_proc, win_string_in, win_string_out,
			 DCT_WSTRING, tlink.value.vxiio.parm, 
			sizeof(tlink.value.vxiio.parm),  NULL,
			"     Parameter :",
		WIN_EXIT, vxi_s_exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
			"          Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
			"          Quit :"
};
struct  Mwindow vxistaticw = {		/*  */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        48, 9,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			18,3,		/* start prompt cursor here */
                        "VXI Static Address",	/* window title */
			sizeof(vxistaticwp)/sizeof(struct prompts),/*# of
								 prompts*/
			vxistaticwp		/* prompts */
        };

struct	prompts vmep[] = {		/* VME prompts */
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.vmeio.card,
		 sizeof(tlink.value.vmeio.card),NULL, 
		"     Card :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.vmeio.signal,
		 sizeof(tlink.value.vmeio.signal), NULL, 
		"   Signal :",
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
		tlink.value.vmeio.parm,
		 sizeof(tlink.value.vmeio.parm), NULL, 
		"Parameter :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"     Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"     Quit :"
};
struct	prompts camacp[] = {		/* CAMAC prompts */
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.camacio.b,
		 sizeof(tlink.value.camacio.b),NULL, 
		"    Branch :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.camacio.c,
		 sizeof(tlink.value.camacio.c), NULL, 
		"     Crate :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.camacio.n,
		 sizeof(tlink.value.camacio.n),NULL, 
		"   Station :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.camacio.a,
		 sizeof(tlink.value.camacio.a), NULL, 
		"Subaddress :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.camacio.f,
		 sizeof(tlink.value.camacio.f), NULL, 
		"  Function :",
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
		tlink.value.camacio.parm,
		 sizeof(tlink.value.camacio.parm), NULL, 
		" Parameter :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"      Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"      Quit :"
};

struct	prompts abp[] = {	/* ALLEN BRADLEY prompts */
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.abio.link,
		 sizeof(tlink.value.abio.link),NULL, 
		"    Link :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.abio.adapter,
		 sizeof(tlink.value.abio.adapter), NULL, 
		" Adapter :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.abio.card,
		 sizeof(tlink.value.abio.card),NULL, 
		"    Card :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.abio.signal,
		 sizeof(tlink.value.abio.signal), NULL, 
		"  Signal :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.abio.plc_flag,
		 sizeof(tlink.value.abio.plc_flag), NULL, 
		"Plc_flag :",
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
		tlink.value.abio.parm,
		 sizeof(tlink.value.abio.parm), NULL, 
		"Parameter:",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"    Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"    Quit :"
};


struct	prompts gpibp[] = {		/* GPIB prompts */
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.gpibio.link,
		 sizeof(tlink.value.gpibio.link),NULL, 
		"          Link :",
		0, quit_proc, win_short_in, win_short_out, DCT_WSHORT,
		(char *)&tlink.value.gpibio.addr,
		 sizeof(tlink.value.gpibio.addr), NULL, 
		"Device Address :",
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
		tlink.value.gpibio.parm,
		 sizeof(tlink.value.gpibio.parm), NULL, 
		"     Parameter :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"          Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"          Quit :"
};


struct	prompts bitbusp[] = {		/* BITBUS prompts */
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bitbusio.link,
		 sizeof(tlink.value.bitbusio.link),NULL, 
		"     Link :",
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bitbusio.node,
		 sizeof(tlink.value.bitbusio.node), NULL, 
		"     Node :",
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bitbusio.port,
		 sizeof(tlink.value.bitbusio.port), NULL, 
		"     Port :",
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bitbusio.signal,
		 sizeof(tlink.value.bitbusio.signal), NULL, 
		"   Signal :",
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
		tlink.value.bitbusio.parm,
		 sizeof(tlink.value.bitbusio.parm), NULL, 
		"Parameter :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"     Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"     Quit :"
};


struct	prompts	constantp[] = {		/* CONSTANT */
                0, pvlink_window, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL, 0, &pvlinkwpv, 
		"Process Variable Link:",
                0, constant_window, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL , 0, &constantw,
		"            Constant :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL, 0, NULL,
		"                Quit :"
};

struct	prompts	fwdlink1p[] = {		/* FWDLINK */
                0, pvlink_window, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL, 0, &pvlinkwpv1, 
		"Process Variable Link:",
                0, constant_window, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL , 0, &constantw,
		"            Constant :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL, 0, NULL,
		"                Quit :"
};

struct	prompts instp[] = {		/* INST_IO */
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
	 	tlink.value.instio.string,
		sizeof(tlink.value.instio.string),NULL,
		"Parameter :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"     Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"     Quit :"
};

struct	prompts bbgpibp[] = {		/* BITBUS to GPIB prompts */
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bbgpibio.link,
		 sizeof(tlink.value.bbgpibio.link),NULL, 
		"       Link :",
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bbgpibio.bbaddr,
		 sizeof(tlink.value.bbgpibio.bbaddr), NULL, 
		"Bitbus addr :",
		0, quit_proc, win_uchar_in, win_uchar_out, DCT_WUCHAR,
		(char *)&tlink.value.bbgpibio.gpibaddr,
		 sizeof(tlink.value.bbgpibio.gpibaddr), NULL, 
		"  GPIB addr :",
		0, quit_proc, win_string_in, win_string_out, DCT_WSTRING,
		tlink.value.bbgpibio.parm,
		 sizeof(tlink.value.bbgpibio.parm), NULL, 
		"  Parameter :",
		WIN_EXIT, exit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL,
		"       Exit :",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE, 
		NULL, 0, NULL, 
		"       Quit :"
};
struct	prompts	vxip[] = {		/* VXI */
                0, pop_sub_window, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL, 0, &vxidynamicw, 
		"Frame/Slot (Dynamic):",
                0, pop_sub_window, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL , 0, &vxistaticw,
		"      Static Address:",
		WIN_QUIT, quit_proc, win_null_in, win_null_out, DCT_EXECUTE,
		 NULL, 0, NULL,
		"                Quit:"
};

struct  Mwindow linkwin[] = {		/*  array of link windows */
	/* VME */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        24, 11,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			14,3,		/* start prompt cursor here */
                        "Enter VME address",	/* window title */
			sizeof(vmep)/sizeof(struct prompts),/*# of prompts*/
			vmep,		/* prompts */
	/* CAMAC */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        26, 14,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			15,3,		/* start prompt cursor here */
                        "Enter CAMAC address",	/* window title */
			sizeof(camacp)/sizeof(struct prompts),/*# of prompts*/
			camacp,		/* prompts */
	/* AB */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        22, 13,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			14,3,		/* start prompt cursor here */
                        "Enter AB address",	/* window title */
			sizeof(abp)/sizeof(struct prompts),/*# of prompts*/
			abp,		/* prompts */
	/* GPIB*/
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        54, 10,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			21,3,		/* start prompt cursor here */
                        "Enter GPIB address",	/* window title */
			sizeof(gpibp)/sizeof(struct prompts),/*# of prompts*/
			gpibp,		/* prompts */
	/* BITBUS*/
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        46, 12,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			13,3,		/* start prompt cursor here */
                        "Enter BITBUS address",	/* window title */
			sizeof(bitbusp)/sizeof(struct prompts),/*# of prompts*/
			bitbusp,		/* prompts */
	/* CONSTANT */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        28, 8,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			24,3,		/* start prompt cursor here */
                        "Select Link Type",	/* window title */
		      sizeof(constantp)/sizeof(struct prompts),/*# of prompts*/
			constantp,		/* prompts */
	/* FWDLINK */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        28, 8,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			24,3,		/* start prompt cursor here */
                        "Select Link Type",	/* window title */
		      sizeof(fwdlink1p)/sizeof(struct prompts),/*# of prompts*/
			fwdlink1p,		/* prompts */
	/* INST */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        50, 12,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			13,3,		/* start prompt cursor here */
                        "Enter INST address",	/* window title */
			sizeof(instp)/sizeof(struct prompts),/*# of prompts*/
			instp,		/* prompts */
	/* BITBUS to GPIB */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        48, 12,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			15,3,		/* start prompt cursor here */
                        "Enter BITBUS to GPIB address",	/* window title */
			sizeof(bbgpibp)/sizeof(struct prompts),/*# of prompts*/
			bbgpibp,		/* prompts */
	/* VXI */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        28, 8,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			23,3,		/* start prompt cursor here */
                        "Select Static/Dynamic",	/* window title */
			sizeof(vxip)/sizeof(struct prompts),/*# of prompts*/
			vxip,		/* prompts */
		
        };

struct Mwindow choicewin = {
	/* */
                        NULL,		/* WINDOW struct */
                        29, 16,		/* upper left corner of window */
                        24, 10,		/* window size */
			4,1,		/* start title here */
			2,3,		/* start prompt list here */
			11,3,		/* start prompt cursor here */
                        "Select Choice",	/* window title */
			0,		/*# of prompts*/
			NULL		/* prompts */
};

/**************************************************************************
 *
 *	following are the execute, input and output class routines used 
 *	to process window element types
 **************************************************************************/
int	quit_proc(pname, pwins,  plink )
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
struct	link	*plink;  /* place to put input */
{
	return(WIN_QUIT);
}

int	null_proc(pname, pwins,  plink )
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
struct	link	*plink;  /* place to put input */
{
	return(0);
}

int	exit_proc(pname, pwins,  plink )
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
struct	link	*plink;  /* place to put input */
{
	return(WIN_EXIT);
}
int	vxi_d_exit_proc(pname, pwins,  plink )
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
struct	link	*plink;  /* place to put input */
{
	plink->value.vxiio.flag=VXIDYNAMIC;
	return(WIN_EXIT);
}
int	vxi_s_exit_proc(pname, pwins,  plink )
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
struct	link	*plink;  /* place to put input */
{
	plink->value.vxiio.flag=VXISTATIC;
	return(WIN_EXIT);
}

int	win_string_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	waddstr(pwins->pwin, pwins->pmt[index].buf);
	wrefresh(pwins->pwin);
	return(0);
}


int	win_fldname_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	char	tmp[FLDNAME_SZ +1];

	memcpy(tmp,pwins->pmt[index].buf,FLDNAME_SZ);
	tmp[FLDNAME_SZ] = '\0';
	wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	waddstr(pwins->pwin, tmp);
	wrefresh(pwins->pwin);
	return(0);
}


int	win_float_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	char	buf[80];
	float	tmp;

	tmp = *(float *) pwins->pmt[index].buf;
	if ( (tmp >= MIN_F_FORMAT_PVAL && tmp <= MAX_F_FORMAT_PVAL) ||
		(tmp >= MIN_F_FORMAT_NVAL && tmp <= MAX_F_FORMAT_NVAL))
		sprintf(buf, "%.*f", precision, tmp);
	else
		sprintf(buf, "%.*e", precision, tmp);
	wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	waddstr(pwins->pwin, buf);
	wrefresh(pwins->pwin);
	return(0);
}

int	win_double_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	char	buf[80];
	double	tmp;

	tmp = *(double *) pwins->pmt[index].buf;
	if ( (tmp >= MIN_F_FORMAT_PVAL && tmp <= MAX_F_FORMAT_PVAL) ||
		(tmp >= MIN_F_FORMAT_NVAL && tmp <= MAX_F_FORMAT_NVAL))
		sprintf(buf, "%.*f", precision, tmp);
	else
		sprintf(buf, "%.*e", precision, tmp);
	wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	waddstr(pwins->pwin, buf);
	wrefresh(pwins->pwin);
	return(0);
}

int	win_uchar_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	char	buf[80];

	sprintf(buf, "%u", *(unsigned char *)pwins->pmt[index].buf);
	wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	waddstr(pwins->pwin, buf);
	wrefresh(pwins->pwin);
	return(0);
}

int	win_short_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	char	buf[80];

	sprintf(buf, "%d", *(short *)pwins->pmt[index].buf);
	wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	waddstr(pwins->pwin, buf);
	wrefresh(pwins->pwin);
	return(0);
}


int	win_enum_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	int	i;
	int	num;
	int	choicesetx;	/* index of choice set */
	int	found;
	struct	wchoice	*pset;

	choicesetx = pwins->pmt[index].inx;
	/* get the number of choices */
	num = 0;
	pset = choicearray[pwins->pmt[index].inx];
	while ( (pset+num)->pstr != NULL)
		num++;
	/* find the current index for this value */
	found = i=0;
	while ( i < num)  {
		if( *(short *)pwins->pmt[index].buf == 
			(pset+i)->val){
				found++;
				break;
		}
		i++;
	}
	if(!found) 
		return(-1);

        wmove(pwins->pwin, pwins->pmtc.y +index, pwins->pmtc.x);
	wclrtoeol(pwins->pwin);
        waddstr(pwins->pwin, (pset+i)->pstr);
        wrefresh(pwins->pwin);
	return(0);
}


int	win_cstring_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	int	i,j;
	char	*ptr, *ptr1;
	char	*string[DCT_WCATSTRING_SZ];

	i=j=0;
	memset((char *)string,'\0', sizeof(string));
	ptr=pwins->pmt[index].buf;
	while ( i < pwins->pmt[index].inx && j < DCT_WCATSTRING_SZ) {
		if(ptr=strchr(ptr, '.')) {
			i++;
			j= ptr++ - pwins->pmt[index].buf;
		} else {
			break;
		}
	}
	if (ptr || i == pwins->pmt[index].inx) {
		if(ptr1 = strchr(ptr, '.'))
			memcpy((char *)string,ptr, ptr1 - ptr);
		else
			memcpy((char *)string,ptr, strlen(ptr));
	} 
	wmove(pwins->pwin, pwins->pmtc.y+index, pwins->pmtc.x);
	waddstr(pwins->pwin, (char *)string);
	wrefresh(pwins->pwin);
	return(0);
}

int	win_null_out(pname, pwins, index)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
{
	return(0);
}

int	win_pvname_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	if(legal_name(buf)) {  /* a process variable name must start with a 
				letter, else it's a constant, not a pv */
	      strncpy(pwins->pmt[index].buf, buf, pwins->pmt[index].bufsiz -1);
	      return(0);
	} else {
		print_error_msg_p(pname, "A PV name must start with a letter",
				PAUSE_F);
		wmove(pwins->pwin, pwins->pmtc.y+index, pwins->pmtc.x);
		wclrtoeol(pwins->pwin);
		(*(pwins->pmt[index].proco))(pname, pwins,index);
		return(-1);
	}
}

int	win_string_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	strncpy(pwins->pmt[index].buf, buf, pwins->pmt[index].bufsiz -1);
	return(0);
}


int	win_fldname_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	fixup_fldname(buf,pwins->pmt[index].buf);
	return(0);
}


int	win_float_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	float	tmp;
	if (sscanf(buf, "%f", &tmp) < 1) {
		print_error_msg_p(pname, "Illegal value", PAUSE_F);
		return(-1);
	}
	*(float *)pwins->pmt[index].buf = tmp;
	return(0);
}

int	win_double_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	double	tmp;
	if (sscanf(buf, "%lf", &tmp) < 1) {
		print_error_msg_p(pname, "Illegal value", PAUSE_F);
		return(-1);
	}
	*(double *)pwins->pmt[index].buf = tmp;
	return(0);
}

int	win_uchar_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	unsigned short	tmp;
	if (sscanf(buf, "%hu", &tmp) != 1){
		print_error_msg_p(pname, "Illegal value", PAUSE_F);
		return(-1);
	}
	if(tmp < 256) {
		*(unsigned char *)pwins->pmt[index].buf = tmp;
		return(0);
	} else {
		print_error_msg_p(pname, "Illegal value", PAUSE_F);
		return(-1);
	}
}
int	win_short_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	short	tmp;
	if (sscanf(buf, "%hd", &tmp) != 1){
		print_error_msg_p(pname, "Illegal value", PAUSE_F);
		return(-1);
	}
	*(short *)pwins->pmt[index].buf = tmp;
	return(0);
}


int	win_enum_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer, this will contain the direction */
{
	int	i;
	int	num;
	int	choicesetx;	/* index of choice set */
	int	found;
        struct  wchoice *pset;


	choicesetx = pwins->pmt[index].inx;
	/* get the number of choices */
	num = 0;
        pset = choicearray[pwins->pmt[index].inx];

	while ((pset+num)->pstr != NULL)
		num++;

	/* find the current index for this value */
	found = i=0;
	while ( i < num)  {
		if( *(short *)pwins->pmt[index].buf == 
			(pset+i)->val){
				found++;
				break;
		}
		i++;
	}
	if(!found)
		return(-1);

	if( *buf == NEXT_ITEM) {
		if( ++i >= num)
			i=0;
	}else if ( *buf == PREV_ITEM) {
		if ( --i < 0 )
			i = num-1;
	}
	*(short *)pwins->pmt[index].buf=i;		
	return(0);
}

int	win_cstring_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer, */
{
	char	*ptr,*ptr1;
	char	begin[DCT_WCATSTRING_SZ];
	char	end[DCT_WCATSTRING_SZ];
	int	i,j;

	i=j=0;
	buf=strip_white(buf);
	ptr = pwins->pmt[index].buf;
	memset(begin,'\0', sizeof(begin));
	memset(end,'\0', sizeof(end));
	while( i < pwins->pmt[index].inx && j < DCT_WCATSTRING_SZ) {
		if(ptr=strchr(ptr, '.')) {
			i++;
			j = ptr++ - pwins->pmt[index].buf;
		}else {
			break;
		}
	}
	if(ptr) {
		strncpy(begin, pwins->pmt[index].buf,
			 ptr - pwins->pmt[index].buf ); 
		if(ptr1=strchr(ptr,'.'))
			strcpy(end, ptr1);
	}else {
		strcpy(begin, pwins->pmt[index].buf);
		while( i < pwins->pmt[index].inx) {
			strcat(begin,".");
			i++;
		}
	}
	string_cat(pwins->pmt[index].buf, DCT_WCATSTRING_SZ, begin, buf, end);
}

int	win_null_in(pname, pwins, index, buf)
char	*pname;		/* name of caller */
struct  Mwindow	*pwins;
int		index;	/* index into prompt structure array */
char		*buf;	/* input buffer */
{
	return(0);
}

/***************************************************************************
 *
 *	pop_link_window()
 *
 *	pops up the windows to handle DBF_INLINK and DBF_OUTLINK link types
 ***************************************************************************/
int	pop_link_window(pname, plink, linktype)
char	*pname;		/* name of caller */
struct	link	*plink;	/* place where input will go */
long	linktype;	/* the link type as defined in link.h */
{
	int	inx;	/* index into window definition array */
	int	choice; 
	char	input[80];
	
	if ( (inx= map_link(linktype)) < 0){
		print_error_msg_p(pname, "Unknown link type", PAUSE_F);
		return(-1);
	}
	memset((char *)&tlink,'\0', sizeof(struct link));
	if( linktype == CONSTANT || linktype == FWDLINKPROMPTS) {
				  /* copy so subwidow gets copy of
					original link info */
	       	memcpy((char *)&tlink,(char *)plink, sizeof(struct link));
	/* the following test always succeeds for field types INP
		or OUT because rdfsubs.mod_record forces the link
		type to agree with DTYP whenever DTYP is changed */
	}else {
		if(plink->type != linktype) {
			tlink.type = linktype;
		} else {
		       memcpy((char *)&tlink,(char *)plink, sizeof(struct link));
		}
	}
	if(open_window(pname, &linkwin[inx]))
		return(-1);
	while(1) {
		choice = fwin_entry(pname, &linkwin[inx], input, sizeof(input)); 
		if(linkwin[inx].pmt[choice].type == DCT_EXECUTE)
				break;
	}
	close_window(&linkwin[inx]);
	if((*(linkwin[inx].pmt[choice].proc))(pname, 
			linkwin[inx].pmt[choice].pw, &tlink) == WIN_EXIT)
		memcpy((char *)plink,(char *)&tlink, sizeof(struct link));
	return(0);
}	

/***************************************************************************
 *
 *	pvlink_window(pname, pwin, plink)
 *
 *	display menu window and get choice
 *
 ***************************************************************************/


int	pvlink_window(pname, pwins, plink )
char	*pname;		/* name of caller */
struct  Mwindow *pwins;	/* widow definition struct */
struct	link	*plink;  /* place to put input */
{
	int	choice;		
	char	input[80];
	short	rectype, recno;
	short	fld_type, fld_offset, fld_size;

	if(open_window(pname, pwins))
		return(-1);
	else {
		if(plink->type != PV_LINK) {
			memset((char *)plink,'\0', sizeof(struct link));
			plink->type = PV_LINK;
		}
		while (1) {
			choice = fwin_entry(pname, pwins, input, sizeof(input));
			if((pwins->pmt[choice].type == DCT_EXECUTE))  
				break;
		}
		if(plink->value.pv_link.fldname[0] == 0)
			fixup_fldname("val", plink->value.pv_link.fldname);
		if(pvdGetRec(plink->value.pv_link.pvname, &rectype, &recno))
			print_error_msg_p(pname,
		 "Warning, this process variable not defined in this database",
				 PAUSE_F);
		else if(get_recfld(rectype, plink->value.pv_link.fldname,
			&fld_type, &fld_offset, &fld_size))
			print_error_msg_p(pname,
			 "Warning, field name not defined for this pv",
				 PAUSE_F);
		close_window(pwins);
		return((*(pwins->pmt[choice].proc))(pname, pwins, plink)); 
	}
}


/***************************************************************************
 *
 *	constant_window(pname, pwin, plink)
 *
 *	display menu window and get choice
 *
 ***************************************************************************/


int	constant_window(pname, pwins, plink )
char	*pname;		/* name of caller */
struct  Mwindow *pwins;	/* widow definition struct */
struct	link	*plink;  /* place to put input */
{
	int	choice;		
	char	input[80];

	if(open_window(pname, pwins))
		return(-1);
	else {
		if(plink->type != CONSTANT) {
			memset((char *)plink,'\0', sizeof(struct link));
			plink->type = CONSTANT;
		}
		while (1) {
			choice = fwin_entry(pname, pwins, input, sizeof(input));
			if((pwins->pmt[choice].type == DCT_EXECUTE))  
				break;
		}
		close_window(pwins);
		return((*(pwins->pmt[choice].proc))(pname, pwins, &tlink)); 
	}
}


/****************************************************************************
 *
 *	open_window
 * 
 ***************************************************************************/
int	open_window(pname, pwins)
char	*pname;		/* name of caller */
struct  Mwindow *pwins;	/* widow definition struct */
{
	int	i;

	if(!(pwins->pwin= newwin(pwins->sz.y, pwins->sz.x, pwins->beg.y,
                                        pwins->beg.x))){
                print_error_msg_p(pname, "Error creating subwindow",
                                                PAUSE_F);
                return(-1);
	} else {

                box(pwins->pwin, BOX_V, BOX_H);
		wmove(pwins->pwin, pwins->titlep.y, pwins->titlep.x);
		waddstr(pwins->pwin, pwins->title);
		for( i=0; i<pwins->nprompts; i++) {
			wmove(pwins->pwin, pwins->pmtp.y +i,
				 pwins->pmtp.x);
			waddstr(pwins->pwin, pwins->pmt[i].pmt);
			(*(pwins->pmt[i].proco))(pname, pwins, i);
		}
		return(0);
	}
}


/****************************************************************************
 *
 *	close_window()
 *
 ****************************************************************************/
close_window(pwins)
struct  Mwindow *pwins;	/* widow definition struct */
{
/* most (including SYSV) curses doesn't have touchoverlap() */
#ifdef sun 
#ifdef SOLARIS 
      touchwin(stdscr);
#else
      touchoverlap(pwins->pwin, stdscr);
#endif /*SOLARIS*/
#else
      touchwin(stdscr);
#endif /*sun*/
      delwin(pwins->pwin);
      refresh();
}

/****************************************************************************
 * FMENU_SELECT
 *
 * get a selected item from a menu in a subwindow
 *
 ****************************************************************************/
int fmenu_select(pwins, pwchoice,input, inputsize, init)
struct Mwindow  *pwins;          /* pointer window definition struct */
struct	wchoice	*pwchoice;
char    *input;                 /* buffer to place input */
int     inputsize;              /* size of input buffeer */
int	init;			/* initial cursor position */
{
	int		i;
	int		execute;

        if ( init < 0 || init >= pwins->nprompts ) 
		i = 0;
	else
		i = init;

		
	do {

		wmove(pwins->pwin, pwins->pmtc.y+i, pwins->pmtc.x);
        	wrefresh(pwins->pwin);
		execute = 0;
		input[0] = getinpch();
		input[1] = 0;

		switch (input[0]){
		case (EXECUTE):
			execute++;
			break;
		case (NEXT):		/* go to next menu item */
			i++;
			if (i >= pwins->nprompts) i = 0;
			break;
		case (PREVIOUS):
			i--;
			if (i < 0) i = pwins->nprompts -1;
			break;
		case (NEXT_ITEM):
		case (PREV_ITEM):
			break;
		case (QUIT1):
		case (QUIT2):
		case (QUIT3):
			return( WIN_QUIT);
		case (EXIT):
			return( WIN_QUIT);
			/*  return( (pwchoice+i)->val); */
		default:
			break;
		}
	} while ( !execute );
	return((pwchoice+i)->val);

}

/****************************************************************************
 * fwin_entry
 *
 * do data entry on subwindow
 *
 ****************************************************************************/
fwin_entry(pname,pwins, input, inputsize)
char	*pname;			/* name of caller */
struct	Mwindow  *pwins;          /* window definition structure */
char	*input;			/* buffer to place input */
int	inputsize;		/* size of input buffeer */
{
	int		i;
	int		execute;

	i = 0;

		
	do {

		wmove(pwins->pwin, pwins->pmtc.y+i, pwins->pmtc.x);
        	wrefresh(pwins->pwin);
		execute = 0;
		input[0] = getinpch();
		input[1] = 0;

		switch (input[0]){
		case (EXECUTE):
			if( (pwins->pmt[i].type != DCT_EXECUTE) && 
				(pwins->pmt[i].proc != NULL))
			       (*(pwins->pmt[i].proc))(pname, pwins, i, input);
			execute++;
			touchwin(pwins->pwin);
			break;
		case (NEXT):		/* go to next menu item */
			i++;
			if (i >= pwins->nprompts) i = 0;
			break;
		case (PREVIOUS):
			i--;
			if (i < 0) i = pwins->nprompts -1;
			break;
		case (NEXT_ITEM):
		case (PREV_ITEM):
			if(pwins->pmt[i].type == DCT_WENUM){
			       (*(pwins->pmt[i].proci))(pname, pwins, i, input);
			       (*(pwins->pmt[i].proco))(pname, pwins, i);
			}
			break;
		case (QUIT1):
		case (QUIT2):
		case (QUIT3):
			return(find_prompt(pwins, WIN_QUIT));
			break;
		case (EXIT):
			return(find_prompt(pwins, WIN_EXIT));
			break;
		default:
			if(!isprintable(input[0]))
				continue;
			wclrtoeol(pwins->pwin);
        		wrefresh(pwins->pwin);
			wingetline(pwins->pwin,input,inputsize-1,
				pwins->pmtc.y+i,pwins->pmtc.x);
			if(!(*(pwins->pmt[i].proci))(pname, pwins, i, input))
				i++;
		}
	} while ( !execute || (pwins->pmt[i].type != DCT_EXECUTE));
	return(i);
}

/****************************************************************************
 *
 *	wingetline
 * read a line from the keyboard where a portion is already entered
 *
 */
int wingetline(pwin,pstr,lim,row,col)
WINDOW	 *pwin;          /* pointer to curses WINDOW struct for
                                        this window */
char	*pstr;			/* input buffer */
int	lim;			/* size of input buffer */
int	row;
int	col;
{
	int	i;
	char		c;

	wmove(pwin,row,col);
	wprintw(pwin,"%s",pstr);
	wmove(pwin,row,col+strlen(pstr));
	wrefresh(pwin);
	c = getinpch();
	for (i=strlen(pstr); (c != EOF) && (c != EXECUTE) && ( c != NEXT) && 
			(c != PREVIOUS) ;){
		if (isprintable(c) && (i < lim)){
			*(pstr+i) = c;
			wprintw(pwin,"%c",*(pstr+i++));
		}else if ((c == K_BACKSPACE) && (i > 0)){
			i--;
			wmove(pwin,row,col+i);
			wdelch(pwin);
		}else if ((c == K_DELETE) && (i > 0)){
			wmove(pwin,row,col+i);
			wdelch(pwin);
		}
		wrefresh(pwin);
		c = getinpch();
	}
	*(pstr+i) = NULL;
	return(i);
}

/****************************************************************************
 *
 *   find_prompt()
 *
 *   returns the prompt array index of a prompt type 
 *****************************************************************************/
int	find_prompt(pwins, type)
struct	Mwindow	*pwins;		/* window definition structure */
int	type;			/* prompt type to find */
{
	int	i;

	for ( i=0; i<pwins->nprompts; i++)
		if( pwins->pmt[i].inx == type)
			return(i);
	return(-1);
}

/******************************************************************************
 *
 *  fixup_fldname()
 *
 *  converts field name to all caps and space pads to 4 FLDNAME_SZ
 *
 ******************************************************************************/
fixup_fldname(inbuf, outbuf)
char	*inbuf;			/* input string */
char	*outbuf;		/* fixed up fieldname */
{
	int	i;
	int	fill;
	char	c;

	/* skip any leading white space */
	while( *inbuf == ' ' || *inbuf == '\t')
		inbuf++;
	if(strlen(inbuf) == 0) {
		fixup_fldname("val", outbuf);
		return(0);
	}
	for ( i=0, fill=0; i < FLDNAME_SZ; i++) {
	    if(!fill) {
		c = islower(*(inbuf+i)) ? toupper(*(inbuf+i)) : (*(inbuf+i));
		if ( (( c >= 'A') && ( c <= 'Z')) || ((c >= '0') && 
				(c <= '9'))) {
			*(outbuf+i) = c;
		}else {
			fill++;
			*(outbuf+i) = ' ';
		}
	     }else
			*(outbuf+i) = ' ';
	}
	return(0);
}

/***************************************************************************
 *
 *  map_link()
 *
 *		returns index into window defintion array for link type
 *		else -1 if unknown link type 
 ***************************************************************************/
int map_link(link_type)
long	link_type;		/* link type as defined in link.h */
{
	int i;
	for ( i=0; i < lnmap.mapsz; i++) {
		if( link_type == lnmap.pmap[i].type)
			return(lnmap.pmap[i].inx);
	}
	return(-1);
}

/****************************************************************************
 *
 *
 *   get_dct_choice_string()
 *
 *   returns a pointer to the choice string corresponding to choice
 *   have to use this routine for enum choices that aren't defined in *.sdr file
 ****************************************************************************/
char *get_dct_choice_string(tblinx, choice)
int	tblinx;		/* choice table index */
int	choice;		/* choice */
{
        int     i;
        int     num;
        int     found;
        struct  wchoice *pset;

        /* get the number of choices */
        num = 0;
        pset = choicearray[tblinx];
        while ( (pset+num)->pstr != NULL)
                num++;
        /* find the current index for this value */
        found = i=0;
        while ( i < num)  {
                if( choice  == (pset+i)->val){
                                found++;
                                break;
                }
                i++;
        }
        if(!found)
		return("BAD_VALUE");
	else
		return((pset+i)->pstr);
}

/****************************************************************************
 *
 *
 *   get_dct_choice_index()
 *
 *   returns the choice array index  corresponding to choice val
 ****************************************************************************/
int	get_dct_choice_index(tblinx, choice)
int	tblinx;		/* choice table index */
int	choice;		/* choice */
{
        int     i;
        int     num;
        int     found;
        struct  wchoice *pset;

        /* get the number of choices */
        num = 0;
        pset = choicearray[tblinx];
        while ( (pset+num)->pstr != NULL)
                num++;
        /* find the current index for this value */
        found = i=0;
        while ( i < num)  {
                if( (short)choice  == (pset+i)->val){
                                found++;
                                break;
                }
                i++;
        }
        if(!found)
		return(-1);
	else
		return(i);
}

/****************************************************************************
 *
 *
 *   get_dct_choice_val()
 *
 *   returns the value of a choice corresponding to a choice string 
 *	if not found returns -1
 *   have to use this routine for enum choices that aren't defined in *.sdr file
 ****************************************************************************/
short	get_dct_choice_val(tblinx, choicestring)
int	tblinx;		/* choice table index */
char	*choicestring;	/* choice string */
{
        int     i;
        struct  wchoice *pset;

        pset = choicearray[tblinx];
        i=0;
        while ( (pset+i)->pstr != NULL)  {
                if( strcmp((pset+i)->pstr, choicestring) == 0){
			return((pset+i)->val);
                }
                i++;
        }
	return(-1);
}

/****************************************************************************
 *
 *  pop_choice_window()
 *
 *	pops up a window of choices
 *
 ****************************************************************************/
pop_choice_window(pname, pval, pset)
char	*pname;		/* name of caller */
short	*pval;		/* place to put the choice */
struct	choiceSet	*pset;		/* the choices */
{

	int	i;
	int	tmp;
	int	num;
	int	maxlen;
	int	status;
	int	choice;
	char	input[80];
	struct	wchoice	*pwchoice;	/* place to build choice structs */

/* allocate space to build chioce struct */
	if ( (pwchoice= 
		(struct wchoice *)malloc(
			(unsigned)(pset->number + 1)*sizeof(struct wchoice)))
			== NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	status = 0;
	maxlen = strlen(choicewin.title)+ choicewin.titlep.x -choicewin.pmtp.x;
/* scan prompt strings to figure out size of window */

	for (i=0, num=0; i < pset->number; i++) {
		if( pset->papChoice[i] != NULL) {
			(pwchoice+i)->val = i;
			(pwchoice+i)->pstr = pset->papChoice[i];
			num++;
			tmp = strlen(pset->papChoice[i]);
			maxlen = tmp > maxlen ? tmp : maxlen;
		}
	}
	(pwchoice+num)->val = WIN_QUIT;
	(pwchoice+num++)->pstr = "Quit";

/*	set up the window to reflect the size and number of prompts */

	choicewin.sz.x = maxlen + 8;
	choicewin.sz.y = num +6;
	choicewin.pmtc.x = maxlen+choicewin.pmtp.x;
	choicewin.nprompts = num;

/*	open up the window */	
       if(!(choicewin.pwin= newwin(choicewin.sz.y, choicewin.sz.x, 
			choicewin.beg.y, choicewin.beg.x))){
                print_error_msg_p(pname, "Error creating subwindow",
                                                PAUSE_F);
                status= -1;
        } else {

                box(choicewin.pwin, BOX_V, BOX_H);
                wmove(choicewin.pwin, choicewin.titlep.y, choicewin.titlep.x);
                waddstr(choicewin.pwin, choicewin.title);
                for( i=0; i<num; i++) {
                        wmove(choicewin.pwin, choicewin.pmtp.y +i,
                                 choicewin.pmtp.x);
                         waddstr(choicewin.pwin, (pwchoice+i)->pstr);

                }
		if( (choice = fmenu_select(&choicewin, pwchoice, input, 
			sizeof(input), 0)) != WIN_QUIT)
				*pval = (short)choice;
        }
	close_window(&choicewin);
	free((char *)pwchoice);
	return(status);
}


/****************************************************************************
 *
 *  pop_devchoice_window()
 *
 *	pops up a window of device choices
 *
 ****************************************************************************/
pop_devchoice_window(pname, pval, pset)
char	*pname;		/* name of caller */
short	*pval;		/* place to put the choice */
struct	devChoiceSet	*pset;		/* the choices */
{

	int	i,j;
	int	tmp;
	int	num;
	int	maxlen;
	int	maxlen1;
	int	status;
	int	choice;
	char	input[80];
	struct	wchoice	*pwchoice;	/* place to build choice structs */
	struct	wchoice	*pwchoice1;	/* place to build choice structs
					to hold link types  */

/* allocate space to build chioce struct */
	if ( (pwchoice= 
		(struct wchoice *)malloc(
			(unsigned)(pset->number + 1)*sizeof(struct wchoice)))
			== NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	if ( (pwchoice1= 
		(struct wchoice *)malloc(
			(unsigned)(pset->number + 1)*sizeof(struct wchoice)))
			== NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	status = 0;
	maxlen = 0;
/*
	maxlen = strlen(choicewin.title)+ choicewin.titlep.x -choicewin.pmtp.x;
*/
/* scan prompt strings to figure out size of window */

	for (i=0, num=0; i < pset->number; i++) {
		if( pset->papDevChoice[i]->pchoice != NULL) {
			(pwchoice+num)->val = i;
			(pwchoice+num)->pstr = pset->papDevChoice[i]->pchoice;
			(pwchoice1+num)->val = i;
			(pwchoice1+num)->pstr = "BADLINK";	
			for(j=0; j < lnmap.mapsz; j++) {
				if(pset->papDevChoice[i]->link_type == 
						lnmap.pmap[j].type) {
					(pwchoice1+num)->pstr =
						 lnmap.pmap[j].pname;
					break;
				}
			}
			tmp = strlen(pset->papDevChoice[i]->pchoice);
			maxlen = tmp > maxlen ? tmp : maxlen;
			num++;
		}
	}
	(pwchoice+num)->val = WIN_QUIT;
	(pwchoice+num)->pstr = "Quit";
	(pwchoice1+num++)->pstr = "";
/*  find length of longest link type string */
	maxlen1=0;
	for(i=0, tmp=0; i < lnmap.mapsz; i++) {
		tmp = strlen(lnmap.pmap[i].pname);
		maxlen1 = tmp > maxlen1 ? tmp : maxlen1;
	}

/*	set up the window to reflect the size and number of prompts */

	choicewin.sz.x = maxlen + maxlen1 + 8;
	choicewin.sz.y = num +6;
	choicewin.pmtc.x = maxlen + maxlen1 +choicewin.pmtp.x;
	choicewin.nprompts = num;

/*	open up the window */	
       if(!(choicewin.pwin= newwin(choicewin.sz.y, choicewin.sz.x, 
			choicewin.beg.y, choicewin.beg.x))){
                print_error_msg_p(pname, "Error creating subwindow",
                                                PAUSE_F);
                status= -1;
        } else {

                box(choicewin.pwin, BOX_V, BOX_H);
                wmove(choicewin.pwin, choicewin.titlep.y, choicewin.titlep.x);
                waddstr(choicewin.pwin, choicewin.title);
                for( i=0; i<num; i++) {
                        wmove(choicewin.pwin, choicewin.pmtp.y +i,
                                 choicewin.pmtp.x);
			wprintw(choicewin.pwin, "%-*s %s", maxlen,
				 (pwchoice+i)->pstr, (pwchoice1+i)->pstr);
                }
		if( (choice = fmenu_select(&choicewin, pwchoice, input, 
			sizeof(input), 0)) != WIN_QUIT)
				*pval = (short)choice;
        }
	close_window(&choicewin);
	free((char *)pwchoice);
	free((char *)pwchoice1);
	return(status);
}
/****************************************************************************
 *
 *  pop_linktype_choice_window()
 *
 *	pops up a window of device choices
 *
 ****************************************************************************/
pop_linktype_choice_window(pname, pval, pdblink)
char	*pname;		/* name of caller */
long	*pval;		/* place to put the choice */
struct	link	*pdblink;
{

	int	i,j;
	int	tmp;
	int	num;
	int	init;
	int	maxlen;
	int	status;
	int	choice;
	char	input[80];
	struct	wchoice	*pwchoice;	/* place to build choice structs */
	char	*kluge="CONSTANT";

/* allocate space to build chioce struct */
	if ( (pwchoice= 
		(struct wchoice *)malloc(
			(unsigned)(lnmap.mapsz + 1)*sizeof(struct wchoice)))
			== NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	status = 0;
	maxlen = 0;
/* scan prompt strings to figure out size of window */

	for (i=0, num=0; i < lnmap.mapsz; i++) {
		if( lnmap.pmap[i].pname  != NULL &&
			lnmap.pmap[i].type != FWDLINKPROMPTS ) {
			(pwchoice+num)->val = lnmap.pmap[i].type;
			if ( lnmap.pmap[i].type != CONSTANT )
				(pwchoice+num)->pstr = lnmap.pmap[i].pname;
			else
				(pwchoice+num)->pstr = kluge;
			num++;
			tmp = strlen(lnmap.pmap[i].pname);
			maxlen = tmp > maxlen ? tmp : maxlen;
		}
	}
	maxlen = maxlen > strlen(choicewin.title) ? maxlen : strlen(choicewin.title);
	(pwchoice+num)->val = WIN_QUIT;
	(pwchoice+num)->pstr = "Quit";
			num++;

/*	set up the window to reflect the size and number of prompts */

	choicewin.sz.x = maxlen + 8;
	choicewin.sz.y = num +6;
	choicewin.pmtc.x = maxlen + choicewin.pmtp.x;
	choicewin.nprompts = num;

/*	open up the window */	
       if(!(choicewin.pwin= newwin(choicewin.sz.y, choicewin.sz.x, 
			choicewin.beg.y, choicewin.beg.x))){
                print_error_msg_p(pname, "Error creating subwindow",
                                                PAUSE_F);
                status= -1;
        } else {

                box(choicewin.pwin, BOX_V, BOX_H);
                wmove(choicewin.pwin, choicewin.titlep.y, choicewin.titlep.x);
                waddstr(choicewin.pwin, choicewin.title);
                for( i=0; i<num; i++) {
                        wmove(choicewin.pwin, choicewin.pmtp.y +i,
                                 choicewin.pmtp.x);
			wprintw(choicewin.pwin, "%-*s", maxlen,
				 (pwchoice+i)->pstr);
			if ( pdblink->type == (pwchoice+i)->val )
				init = i;
                }
		if( (choice = fmenu_select(&choicewin, pwchoice, input, 
			sizeof(input), init)) != WIN_QUIT)
				*pval = (long)choice;
		else
				*pval = pdblink->type;
        }
	close_window(&choicewin);
	free((char *)pwchoice);
	return(status);
}
/****************************************************************************
 *
 *  pop_enum_window()
 *
 *	pops up a window of enum choices
 *
 ****************************************************************************/
pop_enum_window(pname, pwins, index)
char	*pname;		/* name of caller */
struct   Mwindow	*pwins;
int	index;		/* index into prompt table */
{

	int	i,j;
	int	tmp;
	int	num;
	int	init;
	int	maxlen;
	int	status;
	int	choice;
	char	input[80];
	struct	wchoice	*pset;	/* pointer to choice structs */
	struct	wchoice	*pwchoice;	/* pointer to choice structs */

	pset = choicearray[pwins->pmt[index].inx];
	status = 0;
	maxlen = 0;
/* scan prompt strings to figure out size of window */

	for (num=0; (pset+num)->pstr != NULL ; num++) {
		tmp = strlen(lnmap.pmap[num].pname);
		maxlen = tmp > maxlen ? tmp : maxlen;
	}
	maxlen = maxlen > strlen(choicewin.title) ? maxlen : strlen(choicewin.title);
/* allocate space to build chioce struct */
	if ( (pwchoice= 
		(struct wchoice *)malloc(
			(unsigned)(num+1)*sizeof(struct wchoice)))
			== NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	for (i=0; i<num; i++) {
		(pwchoice+i)->val = (pset+i)->val;
		(pwchoice+i)->pstr = (pset+i)->pstr;
	}
	(pwchoice+num)->val = WIN_QUIT;
	(pwchoice+num)->pstr = "Quit";
			num++;

/*	set up the window to reflect the size and number of prompts */

	choicewin.sz.x = maxlen + 8;
	choicewin.sz.y = num +6;
	choicewin.pmtc.x = maxlen + choicewin.pmtp.x;
	choicewin.nprompts = num;

/*	open up the window */	
       if(!(choicewin.pwin= newwin(choicewin.sz.y, choicewin.sz.x, 
			choicewin.beg.y, choicewin.beg.x))){
                print_error_msg_p(pname, "Error creating subwindow",
                                                PAUSE_F);
                status= -1;
        } else {

                box(choicewin.pwin, BOX_V, BOX_H);
                wmove(choicewin.pwin, choicewin.titlep.y, choicewin.titlep.x);
                waddstr(choicewin.pwin, choicewin.title);
                for( i=0; i<num; i++) {
                        wmove(choicewin.pwin, choicewin.pmtp.y +i,
                                 choicewin.pmtp.x);
			wprintw(choicewin.pwin, "%-*s", maxlen,
				 (pwchoice+i)->pstr);
			if ( *(short *)pwins->pmt[index].buf  == (pwchoice+i)->val )
				init = i;
                }
		while(1) {
		if( (choice = fmenu_select(&choicewin, pwchoice, input, 
			sizeof(input), init)) == WIN_QUIT)
				break;
				
			*(short *)pwins->pmt[index].buf = (short)choice;
			if( pwins->pmt[index].proco != NULL)
				(*(pwins->pmt[index].proco))(pname, pwins, index);
			if((init=get_dct_choice_index(pwins->pmt[index].inx, choice)) < 0)
				init = 0;

		}
        }
	close_window(&choicewin);
	free((char *)pwchoice);
	return(status);
}
/***************************************************************************
 *
 *	pop_sub_window(pname, pwin, plink)
 *
 *	display a subwindow
 *
 ***************************************************************************/


int	pop_sub_window(pname, pwins, plink )
char	*pname;		/* name of caller */
struct  Mwindow *pwins;	/* widow definition struct */
struct	link	*plink;  /* place to put input */
{
	int	choice;		
	char	input[80];

	if(open_window(pname, pwins))
		return(-1);
	else {
		while (1) {
			choice = fwin_entry(pname, pwins, input, sizeof(input));
			if((pwins->pmt[choice].type == DCT_EXECUTE))  
				break;
		}
		close_window(pwins);
		return((*(pwins->pmt[choice].proc))(pname, pwins, plink)); 
	}
}

