static char SccsId[] = "$Id: filedb.c,v 1.15 1995/04/26 13:35:16 mrk Exp $ ";

/*  $Id: filedb.c,v 1.15 1995/04/26 13:35:16 mrk Exp $
 *	Original Author: Julie Sander (LANL)
 *	Current Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02  05-18-92        rcz   	New database access - XGET_PRECTYPE is old GET_PRECTYP
 * .03  12-06-94        jt  	Added VAL_AR_SZ_DOUBLE for fmod_pv to ensure 8-bytes boundary 
 * 	...
 */

/*
 *	filedb.c
 *
 *	Subroutines to modify the database from a file
 *
 *
 */

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <ctype.h>
#include <curses.h>
#include <string.h>
#include "dctDef.h"
#include "dbPvd.h"
#include <dbRecType.h>
#include <dbRecords.h>
#include <dbRecDes.h>
#include <sdrHeader.h>
#include "dctmacros.h"

extern struct recType *dbRecType;

#define  VAL_AR_SZ   80		/* array containing value size 		  */
#define  VAL_AR_SZ_DOUBLE   10	/* array containing value size; ensure 8-bytes boundary   */
#define  MAX_WORD_SZ 80		/* max words size of words read from line */
#define  LINE_SZ     120	/* size of array line read into		  */
#define  LOOK_PV     1		/* looking for 'PV:' status	  */
#define  LOOK_TYPE   2		/* have found 'PV:' and pv name	  */
#define  LOOK_NODE   3		/* have found 'Type:' and type name */
#define  LOOK_FIELD  4		/* accepting fields to modify	  */

extern	char	*fldname_to_string();

extern struct  recLoc    pd_recLoc;       /* header for current rec db file */
extern struct  sdrHeader pd_sdrHeader;    /* sdr header for a rec dbfile*/
extern struct  pvd      *dbPvd;         /* process variable directory */
extern struct  fldDes  **pflddef;              /* current record's field
                                                definitions */
extern struct  recTypDes       *precorddes;            /* current record type
                                                definition structure */
extern char  	*ctrl_record;	/* current database record in memory */
extern int     	pvctrl_fd;      /* file pointer for the rec db file   */
extern short	curr_rectypedef; /* record type currently in memory */
extern char     rec_file[MAX_FILENAME_SZ];      /* name of current
                                                         pd record file */
extern short	suppress_errors; /* supresses errors to the screen  */
extern short	suppress_errors1; /* supresses errors to the screen (non-
					interactive mode */
extern int      dct_mode;      /* mode flag  0==interactive, 1==batch */
extern int     	dis_range_check; /* 0 = check if value within range, !=0 don't
                                        check */
static	int	pass;		/* pass number */
short	record_no;
short   errors;		/* signals if errors occurred in process variable */
short   tot_error;	/* signals if errors occurred in modification	  */

short	mfe_fd;		/* error file identifier */

void report_error();

/* Subroutines
 *
 * filedb	top subroutine for file-to-database routines. Opens modifying
 *		file and reads it in a line at a time.
 * fadd_pv	adds a process variable
 *	args
 *		pv		process variable name
 *		type		type number
 *		node		node number
 *	returns
 *		0		variable was add
 *		-1		no available record in database
 * file_status	determines the status of a process variable
 *	args
 *		pv		process variable name
 *		type		type number
 *		node		node number
 *	returns
 *		0		pv can be modified
 *		1		pv can be added
 *		-1		pv can not be modified or added
 * fmod_pv	updates a field's value
 *	args	
 *		line		line from modification file
 *		type		type number	
 *	returns
 *		0		no error in changing field value
 *		-1		value could not be changed
 * get_type_num	gets the type number from a type name
 *	args
 *		type		type name
 *	returns
 *		>0		type names' number
 *		-1		type not found
 * report_error reports errors to the error file
 *	args
 *		index		signals which error to say
 *		descr		descriptive word
 * split_line	splits a word from the line 
 *	args 
 *		line		line from modification file
 *		pv		process variable name
 *		line_pos	position in line
 *	returns
 *		1		there is another word in the line
 *		0		end of the line, buster
 * val_new_pv	checks entered process variable for validity
 *	args
 *		pv_name		process variable name
 *		t_name		type name
 *		n_name		node name
 *		type		type number
 *	returns
 *		status		signals if pv should be skipped or modified
 */


/*
 *	FILEDB
 *
 *	main for file to database, reads file a line at a time
 *	does 2 passes
 *		pass1  range checking for VAR ranges turned off
 *		pass2  range checking for VAR ranges turned on
 *
 */

filedb(db_name,mod_name)
char	db_name[MAX_FILENAME_SZ];	/* database name for noninteractive */
char	mod_name[MAX_FILENAME_SZ];	/* report file name for noninteactive*/
{
	char	*xname = "filedb";	
	char  mod_file[MAX_FILENAME_SZ]; /* file database is modified from */
	char  err_file[MAX_FILENAME_SZ]; /* error file name */
	char  word[MAX_WORD_SZ];	/* extra words in line		  */
	char  pv_name[MAX_WORD_SZ];	/* process variable name	  */
	char  t_name[MAX_WORD_SZ];	/* type name			  */
	char  line[LINE_SZ];		/* line from modifying file	  */
	char	input[80];		
	char	buffer[80];	
	long  index;			/* index used to read a line   	  */
	int mf;			/* file desc. for modifying file  */
	short line_pos;			/* position in line		  */
	short state;			/* the state of the search algor. */
	short type;			/* type of process variable       */
	int	dct_mode_save;		/* place to save current dct_mode */

	/* initialize flags */
	errors = NO;
	tot_error = NO;
	if(dct_mode == DCT_INTERACTIVE) {		/* interactive mode */

		/* get name of file for modifying  */
		clear();
		move(0,0);	

       	 /* get the db  name and get the process database header into memory */
       	 	if (select_db(input,"MODIFY") < 0) 
			return(0);
        	if(open_pd(input) < -1)
			return(-1);

		clear();
		move(0,0);
		printw("Enter Modification File: ");
		refresh();
		getline(mod_file,MAX_WORD_SZ,0,
			strlen("Enter Modification File: "));
	} else {		/* noninteractive mode */
                if(form_filename(input, sizeof(input), db_name, "pvd")) {
                        print_err(xname, "Bad filename");
                        return(-1);
                }
                if(check_file_exist(input)) {
                       print_err(xname,"Report error, database does not exist");
                        return(-1);
                }
                if(open_pd(db_name) < -1)
                        return(-1);
		strcpy(mod_file, mod_name);
	}

	/* open modification file */
	if (!isalpha(*mod_file))
		 return(-1);
	if ((mf = open_file(mod_file,0,xname)) < 0 )
		 return(-1);

	/* create error file for modification file */
	if ((mfe_fd = create_file_force(xname, mod_file,"err", err_file,
		 sizeof(err_file))) < 0){
		return(-1);
	}
	suppress_errors1= TRUE;
	suppress_errors = TRUE;
	dct_mode_save = dct_mode;	/* switch dct modes */
	dct_mode = DCT_NONINTERACTIVE;
	dis_range_check=1;		/* so convert_fld won't complain when
					one of the limits of a variable range
					haven't been read in yet */
	for ( pass = 1; pass < 3; pass++) {
		state = LOOK_PV;
		index = 0;
		line_pos = 0;
		sprintf(buffer, "-----PASS %d-----\n",pass);
		write(mfe_fd, buffer, strlen(buffer));
		printf("\n%s", buffer);
		if (pass > 1) {
			dis_range_check=0;
			if ( seek_file(mf, mod_file, xname, 0L))
				break;
		}
	/* main loop to read in file a line at a time */
	while ( readline(mf,&index,line) > 0){
		if (*line == '$') continue; /* skip lines with interleaf cmds */
		while (split_line(line,word,&line_pos) > 0){
			switch(state){
			case LOOK_PV:  /* look for a new process variable */
			   if (strcmp(word,"PV:")==0 || strcmp(word,"PV")==0){
				split_line(line,word,&line_pos);
				strcpy(pv_name,word);

				if(errors == YES) report_error(15,pv_name);
				report_error(0,pv_name); /*no error,
								 just info*/
				printf("\nProcess Variable: %s",pv_name);
				state = LOOK_TYPE;
			   }
			   break;
			case LOOK_TYPE:  /* look for process variable's type */
			   if (strcmp(word,"Type:")==0||strcmp(word,"Type")==0){
				split_line(line,word,&line_pos);
				strcpy(t_name,word);
		 	   	line_pos = strlen(line);
			   	if((val_new_pv(pv_name,t_name,&type,
							xname))
					>= 0)
					state = LOOK_FIELD;
			   	else 
					 state = LOOK_PV;
			   }else{
				report_error(1,pv_name);
				state = LOOK_PV;
			   }
			   break;
			case LOOK_FIELD:  /* look for fields to be modified */
			  /* new process variable */
			   if (strcmp(word,"PV:")==0 || strcmp(word,"PV")==0){
				split_line(line,word,&line_pos);
				strcpy(pv_name,word);

				if(errors == YES) report_error(15,pv_name);
				report_error(0,pv_name); /*no error,
							 just info*/
				printf("\nProcess Variable: %s",pv_name);
				state = LOOK_TYPE;

				write_record(record_no,xname);
				close_record();
			   /* skip time line */
			   }else if (strcmp(word,"Time:")==0){
				line_pos = strlen(line);
			   /* modify the field */
			   }else{
				fmod_pv(line,type);
				line_pos = strlen(line);
			   }
			   break;
			default:   break;
			}
		}
		line_pos = 0;
	}		/*while ( readline(mf,&index,line) > 0){ */
	/* write last pv if OK */
		if (state == LOOK_FIELD) {
			 write_record(record_no,xname);
			close_record();
		}
		suppress_errors1 = FALSE;
	}		/* for ( pass =1; pass < 3; pass++) */
	dct_mode = dct_mode_save;
	dis_range_check=0;
	if (errors == YES)
 	      printf("\n**** Errors in process variable modification ****\n\n");
	if (tot_error == YES)
		printf("\n\n**** Error descriptions in %s.err ****\n",mod_file);

	close(mf);
	close(mfe_fd);
	if(dct_mode == DCT_INTERACTIVE) {	/* interactive mode */
		printf("\n\nHit any key to continue\n");
		getinpch();
	}
	close_pd(CREATE_DB);
	suppress_errors = FALSE;
	return(0);
}



/*
 *	FADD_PV
 *
 *	adds a process variable to the opendatabase
 *
 */

fadd_pv(ppv,type)
char *ppv;		/* process variable name */
short type;		/* process variable type */
{
	char	*xname = "fadd_pv";


        if ( set_current_record_type(type))
                return(-1);

	/* add new pv if place available */
	if (!get_avl_pv(&record_no))
	{
		/* add pv and save new process variable directory  */
		if (add_pv(ppv,type,record_no,0) < 0){
			report_error(12,ppv);
			return(-1);
		}
		if(save_pvd())
			return(-1);

		/* extend the file to keep the designated number of spares */
		if ((pd_recLoc.no_records - pd_recLoc.no_used) 
			< NUMBER_FREE)
			if(extend_record())
				return(-1);

		init_pv();			/* init field values */
		/* put the name into the record */
   		putfield(type,"NAME",ppv);
		pd_recLoc.no_used++;

		/* save process database header */
        	if(save_pd_header(pvctrl_fd, rec_file, xname, &pd_recLoc)) {
               		 return(-1);
        	}


		return(0);
	}
	else 
	{
		report_error(6,ppv);
		return(-1);
	}
}

/*
 *	FILE_STATUS
 *
 *	returns: -1 if process variable can not be added or modified
 *		  1 if pv can be added
 *		  0 if pv can be modified
 *
 */

file_status(ppv,ptype, pname)
char *ppv;			/* process variable name	*/
short  *ptype;			/* process variable type number */
char	*pname;				/* caller's name */
{
	int   found;		/* signals if pv was found	*/
	short dbtype;	/* type number of pv in database */

	found = pvdGetRec(ppv,&dbtype,&record_no);

	/* if process variable was found but does not have the same type */
	/*  it is a duplicate, thus an error  */
	if (found == 0 && (dbtype != *ptype )) {
		report_error(7,ppv);
		return(-1);
	}

	/* open record of correct type */
	if (open_record(*ptype) < 0)
	{
		report_error(8,ppv);
		return(-1);
	}

	/* if pv exists, read in record */
	if (found == 0)
	{
	 	if(read_record(record_no, pname))
			return(-1);
	}
        if ( set_current_record_type(*ptype))
                return(-1);


	/* return a 0 or 1 (pvdGetRec returns 0 or -1) to modify or add */
	found = -1 * found;
	return(found);

}

/*
 * 	FMOD_PV
 *
 *	updates a field given a line from an update file
 *
 */

fmod_pv(pline,type)
char *pline;			/* line from input file		*/
short type;			/* record type */
{
	char	*xname = "fmod_pv";
	char fld_name[VAL_AR_SZ];
	char value[VAL_AR_SZ];			/* value read from the line	*/
	double outinput[VAL_AR_SZ_DOUBLE];	/* converted stripped value     */	
	int	num;
	int     i;
        struct fldDes  *pfield;
	short index;
	char	fname[FLDNAME_SZ+1];	/* holds null terminated field name*/

	if ( sscanf(pline, " %[A-Z0-9] %n", fld_name, &num) != 1) {
		report_error(9, pline);
		return(-1);
	}
	strcpy(value, pline+num);

	/* pad fld_name array with blanks */
	for (i = strlen(fld_name); i < FLDNAME_SZ; i++)
		*(fld_name+i) = ' ';
	*(fld_name+i) = '\0';


	/* put fields for this type into memory */
        if ( set_current_record_type(type))
                return(-1);

	if ((index = fld_search(fld_name)) < 0){
		report_error(10,fld_name);
		return(-1); 
	}else {
		if(!(pfield = GET_PFLDDES(precorddes, index))) {
			report_error(14, fld_name);
			return(-1);
		}
	}


	/* put new value into field	*/
	if (convert_fld(type,pfield,value,outinput) == 0){
		putfield(type,pfield->fldname,outinput);
		return(0);
	}else{
		report_error(11,fldname_to_string(pfield->fldname, fname, 
			FLDNAME_SZ));
		return(-1);
	}
}


/* 
 *	GET_TYPE_NUM
 *
 *	gets the type number for a type name 
 */

get_type_num(type)
char *type;
{
	int i;
	char	*pstr;

	i = 1;
	while (i < dbRecType->number)
	{
		if (pstr = (char*)XGET_PRECTYPE(i))
			if (!strcmp(type,pstr))
				return(i);
		++i;
	}
	return(-1);
}


static char *pError[] = {
	"Process Variable: ",						/* 0 */
	"---Process variable skipped. Type does not follow the variable ",			/* 1 */
	"---Process variable skipped. Node does not follow type for ",	/* 2 */
	"---Process variable skipped. Non alpha variable, type, or node found.",		/* 3 */
	"---Process variable skipped. Could not find the type ",	/* 4 */
	"---Process variable skipped. Could not find the node ",	/* 5 */
	"---Process variable skipped. No available record in Process Database for ",		/* 6 */
	"---Process variable skipped. PV already exists: ",		/* 7 */
	"---Process variable skipped. could not open ",			/* 8 */
	"***** Error reading field name from the line:",		/* 9 */
	"***** Could not find the field ",				/* 10 */
	"***** Incorrect value for the field ",				/* 11 */
	"---Process variable skipped. Could not add to directory ",	/* 12 */
	"***** Bad input line format",					/* 13 */
	"***** Null pointer for field description"			/* 14 */
	};
 

/*
 * REPORT_ERROR
 *
 * print an error message to error report
 */
void report_error(index,descr)
int index;
char *descr;
{
	char	buffer[80];

	if(pass < 2)		/* don't report errors until pass 2 */
		return;
	if (index != 0){
		errors = YES;
		tot_error = YES;
	}
	if (index == 15){
		printf("\n**** Errors in process variable modification ****\n");
		errors = NO;
		return;
	}

	if (mfe_fd <= 0) return;

	sprintf(buffer,"%s %s\n",pError[index],descr);
	write(mfe_fd,buffer,strlen(buffer));

	if ( (index == 1) || (index == 2)){
		strcpy(buffer,
	"---usage: PV: pv_name   Type: type_name   Node: node_name\n\n");
		write(mfe_fd,buffer,strlen(buffer));
	}

	return;
}


/*
 *	SPLIT_LINE
 *
 *	gets the next word from the line
 */

split_line(pline,pword,line_pos)
char *pline;
char *pword;
short *line_pos;
{

	int	num;	

        /* get word in line */
        if (sscanf(pline+*line_pos, "%s%n", pword, &num) == 1){
		*line_pos += num;
		return(1);
	}
	else
		return(0);	
}

/*
 *	VAL_NEW_PV
 *
 *	checks the validity of the new process variable
 */

val_new_pv(pv_name,t_name,type, pname)
char *pv_name; 			/* process variable name */
char *t_name;			/* field type name */
short *type;			/* record type */
char	*pname;			/* caller's name */
{
	short status;

	/* get the integer representations of the type  */
	if ( (*type = get_type_num(t_name)) < 1) {
		report_error(4, t_name);
		return(-1);
	}
	status = file_status(pv_name, type, pname);

	/* determine if the names are letters */
	if (!legal_name(pv_name) || !isalpha(*t_name) ){
		report_error(3," ");
		return(-1);
	}

	/* check if type exists */
	if (*type == -1){
		report_error(4,t_name);
		return(-2);
	}


	/* check if the pv should be added, modified, or skipped */
	/*  status = -1  pv cannot be added or modified
	    status = 0   pv exists and can be modified
            status = 1   pv does not exist and can be added */

	if (status == -1)
		return(-4);

	if (status == 1){
		if (fadd_pv(pv_name,*type)  < 0)
			return(-5);
	}
	return(0);
}
