static char SccsId[] = "$Id: pdsubs.c,v 1.21 1995/02/28 22:04:46 mrk Exp $ ";

/*  $Id: pdsubs.c,v 1.21 1995/02/28 22:04:46 mrk Exp $
 *	Original Author: Bob Dalesio (LANL)
 *	Current Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	07-22-91	rac	put \n on last line to make sccs happy; put
 *				date at beginning of file
 *	03-16-92	frl 	fixed seg fault introduced by prior fix that
 *				caused dct to seg fault or print corrupted time
 *				in non-interactive mode.
 * .03  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 *				XGET_PRECTYPDES  is old GET_PRECTYPDES
 *				XGET_PRECTYPE is old GET_PRECTYPE
 * .04  08-31-92        rcz     suppressed checksum from dbRead
 */

/*
 * PDSUBS.C
 *
 * The subroutines that access the process database are located in this
 * module
 *
 *
 */
#include        <stdlib.h>
#include        <stddef.h>
#include	<stdlib.h>
#include	<fcntl.h>
#include	<time.h>
#include	<string.h>
#include	<dirent.h>
#include	<ctype.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<curses.h>
#include	<errno.h>
#include	<sdrHeader.h>
#include	<dbRecords.h>
#include	<dbBase.h>
#include	"dbPvd.h"
#include	<dbRecType.h>
#include	<dbRecDes.h>
#include	<link.h>
#include	"dctDef.h"
#include  "dctmacros.h"


/*********************************************/
extern long freeValidDbPtrs();
extern long getValidDbPtrs();
extern struct dbBase *pdbBase;
extern struct choiceSet *choiceCvt;
extern struct arrChoiceSet *choiceGbl;
extern struct choiceRec *choiceRec;
extern struct devChoiceRec *choiceDev;
extern struct arrBrkTable *cvtTable;
extern struct recDes *dbRecDes;
extern struct recType *dbRecType;
extern struct recHeader *dbRecords;
extern struct recDevSup *devSup;
extern struct drvSup *drvSup;
extern struct recSup *recSup;
extern struct pvd *dbPvd;
extern long	dbRead();
/*********************************************/

extern short	write_page();
extern	int     dct_mode;              /* mode flag  0==interactive, 1==batch */
extern int	rep_fd;			/* file descriptor for report */
extern char	curr_dbname[PDNAME_SZ];	/* current process database in memory */
extern char	dctsdr_filename[MAX_FILENAME_SZ];  /* current .dctsdr
						 file name */
extern char     pvd_filename[MAX_FILENAME_SZ];       /* name of current
							 .pvd file */
extern char     rec_file[];           /* name of current pd record
                                                file */
extern short    curr_rectype;           /* current record type in mem */
extern short    curr_rectypedef;        /* current record type
                                                definitions */
extern char	*pd_names[]; /* array of node names	*/
extern	char	*frm_names[];	/* array of form names */
extern short	order[MAX_PV+4]; /* alphabetically ordered  PV names */
extern short	pvctrl_fd;	/* database record control portion file desc */
extern short	suppress_errors;
extern short	long_report;
extern short	precision;
extern char    ctrl_record[MAX_RECORD_SIZE];   /* PV being modified    */

extern struct  recLoc    pd_recLoc;              /* header for a rec db file */
extern struct  sdrHeader pd_sdrHeader;          /* sdr header for a rec dbfile*/
extern struct  recTypDes *precorddes;		/* pointer to current record
						definition structure */

char		pvname_prefix[PVNAME_SZ];

struct	pd_summary {		/* used for summary reports */
	char	*fldname1;	/* fieldname */
	char	*fldname2;	/* alternate field name */
	int	width;		/* width to use for this item */
};

struct	fchoice	forms = {
	8,1,			/* title position */
	2,3,			/* place to start choices */
	0,3,			/* place to start prompt cursor */
	"Select Form To Use",	/* title */
	1,			/* # of choices this gets set later */
	frm_names
};	
struct	fchoice	dbnames_f = {
	8,1,			/* title position */
	2,3,			/* place to start choices */
	0,3,			/* place to start prompt cursor */
	NULL,			/* title */
	0,			/* # of choices this gets set later */
	pd_names
};	
#if     !defined(ERROR) || (ERROR!=(-1))
#define ERROR           (-1)
#endif

/* the following define the menu of operations on pvs */
extern	int	pv_report();
extern	int	pv_create();
extern	int	pv_modify();
extern	int	pv_copy();
extern	int	pv_delete();
extern	int	pv_undelete();
extern	int	pv_rename();
extern	int	loc_xref();
extern	int	display_links();
extern	int	macro_mod();

void rm_pd_name();

int	pv_name_prefix()
{
	char	*pvname_prompt="Enter PV name prefix: ";
	
	clear();
	move(1,1);
	addstr(pvname_prompt);
	getline1(pvname_prefix, PVNAME_SZ-4, 1, strlen(pvname_prompt)+2);
	return(0);
}
int	exit_pv_menu()
{
	return(EXIT);
}
struct	menu_prompt	pv_menu_prompts[] = {
	0, pv_report, NOKEY, "Report a Process Variable",
	1, pv_create, NOKEY, "Create a Process Variable",
	2, pv_modify, NOKEY, "Modify a Process Variable",
	3, pv_copy,   NOKEY, "Copy a Process Variable",
	4, pv_delete, NOKEY, "Delete a Process Variable",
	5, pv_undelete, NOKEY, "Undelete a Process Variable",
	6, pv_rename, NOKEY, "Rename a Process Variable",
	7, loc_xref,  NOKEY, "List all Link References",
	8, display_links, NOKEY, "Display Link Tree for a Process Variable",
	9, macro_mod, NOKEY, "Macro Field Modification",
	10, pv_name_prefix, NOKEY, "Define PV Name Prefix",
	11, exit_pv_menu, EXIT,	"Exit This Menu"
};

char	*pv_title[] = {
	"Process Variable Menu",
	"Database Name:                                  ",
	"PV Name Prefix:                                  ",			
	NULL
};

struct	fmenu	pv_menu_s = {
	8,1,			/* title position */
	2,5,			/* place to start prompt text */
	0,5,			/* place to start prompt cursor */
	pv_title,
	sizeof(pv_menu_prompts)/sizeof(struct menu_prompt),   /* # of prompts*/
	pv_menu_prompts		/* array of prompts */
};



 
/* Subroutines
 *	
 * add_pd_name		add a process database to the PD list
 *	args
 *		pname		pointer to the new Process Database name
 *		node_number	network node number
 *	returns
 *		0		added name to the table
 *		-1		no entries available
 *		-2		node name is already used
 * create_pd		makes a blank database
 * delete_pd		deletes all of the Process Variables in a database
 * macro_mod		modifies specd field of all specd PVs to specd value
 * modify_pd		operations on single PVs in a Process Database
 * node_num_to_name	converts a network number to a node name
 *	args
 *		node	node number
 *		pname	pointer to an ascii string for the name
 *	returns
 *		0	successful
 *		-1	not found
 * open_pd		open a new process database
 *	args
 *		pname		pointer to the new Process Database name
 *	returns
 *		0		new Process Database opened & header read in
 *		-1		could not open "pname"
 * report_pd		reports all of the Process Variables in a database
 * rm_pd_name		remove a process database name from the global table
 *	args
 *		nodename	node name to remove
 * save_pd_header	save the process database header
 * select_db		select a node on which to operate
 *	args
 *		nodename	pointer to the selected node name
 *		pverb		pointer to operation to perform on the node
 *	returns
 *		0		node selected
 *		-1		do not perform operation
 */
  
/*
 *	ADD_PD_NAME
 *
 * add a process database name to the list
 */
add_pd_name(pname)
char	*pname;		/* name string of db ot add */
{
	short	i;

	/* verify name is not used */
	i = 0;
	while ((i < MAXPDS) && (isalpha(*pd_names[i])) ){
		if (strcmp(pd_names[i],pname) == SAME)
			return(-2);
		i++;
	}
	if (i >= MAXPDS) return(-1);

	/* make a new entry into the database name table */
	strcpy(pd_names[i],pname);

	return(0);
}

/*****************************************************************************
 *
 *  close_pd
 * 
 *   closes a process data base and cleans things up
 *
 ******************************************************************************/
int close_pd(flag)
	int	flag;	/* = CREATE_DB then create .database file */
{

	char	*xname = "close_pd";
	static	char	*pstring[] = {
		dctsdr_filename,
		pvd_filename
	};
	int	i;
	
	if(flag == CREATE_DB) {
		create_database_file();
	}
	for(i = 0; i < (sizeof(pstring)/sizeof(xname)); i++) {
		memset(pstring[i], '\0', strlen(pstring[i]));
	}
	curr_rectype = curr_rectypedef = -1;
	memset(curr_dbname, '\0', sizeof(curr_dbname));
}

/*
 *  CREATE_DBFILE
 *
 *   creates an empty database record file of the indicated type 
 *
 *	this proc assumes that the record description structure, dbRecDes
 * 	has already been loaded with a call to sdrLoad()
 *
 *	returns:
 *			file descriptor for created file else
 *		 -1	file create error
 *		 -2 	file already exists 
 *
 */
create_dbfile(pname, type, mtime)
char	*pname;			/* database name */
short	type;			/* record type */
time_t	mtime;			/* time stamp of *.dctsdr */
{

	char	*xname = "create_dbfile";
	int	fd;				/* file descriptor for file
							created */
	struct sdrHeader	sdrHeader;
	struct	recLoc		recLoc;
	struct	recTypDes	*pRecTypDes;	/* set to record description
							struct for this type*/
	char			filename[80];	/* hold complete file name */
	unsigned 		recsize;	/* set to record size */
	char			*buf;		/* used to write null records*/
	char			*rec_type_name; /* string containing record
						type name */
	int			i;
	int			error_flag;

	if(!(pRecTypDes= (struct  recTypDes*)XGET_PRECTYPDES(type))) {
		return(-1);			/* this type not defined in
						record description structures*/ 
	}
 
	if(!(rec_type_name = (char*)XGET_PRECTYPE(type))) {
		return(-1);			/* no record type name defined
							for this type */
	}

	/* set up the header sdrHeader */

	recsize = pRecTypDes->rec_size;
	memset((char *)&sdrHeader, '\0', sizeof(struct sdrHeader));
	sdrHeader.magic = DBMAGIC;
	sdrHeader.nbytes = sizeof(struct recLoc) + NUMBER_FREE*recsize;
	sdrHeader.type = SDR_DB_RECORDS;
	sdrHeader.create_date = mtime;

	/* setup the record type header recLoc */
	((char *)&recLoc, sizeof(struct recLoc));
	recLoc.no_records = NUMBER_FREE;
	recLoc.no_used    = 0;
	recLoc.rec_size   = recsize;
	recLoc.record_type = type;
	recLoc.pFirst     = NULL;

	/* allocate and set up buffer for NULL records */
	if (!(buf= malloc(recsize))) {
		print_error_msg(xname, "malloc() failed");
		return(-2);
	}
	memset(buf, '\0', (int)recsize);

	if ((fd = create_file(xname, pname, rec_type_name,filename,
				sizeof(filename))) < 0 ) {
		if (fd == -2)
			return(-2);
		else
			return(-1);
	}

	error_flag = -1;
	if (! write_file( fd, filename, xname, (char *)&sdrHeader,
		     sizeof(struct sdrHeader))) {
		if(!write_file(fd, filename, xname, (char *)&recLoc, 
			sizeof(struct recLoc)) ) {
			for ( i = 0; i < NUMBER_FREE; i++) {
				if(write_file( fd, filename, xname, buf,
					(int)recsize) )
						break;
			}
			if ( i == NUMBER_FREE) 
				error_flag = 0;
		}
	}	
	free(buf);

	if(close(fd)) {
		print_syserror_msg(xname, "file close error", errno);
		return(-1);
	}

	return(error_flag);
}

/*****************************************************************************
 * check_pd_mtimes
 *
 *	database timestamp checker
 *
 *	
 ****************************************************************************/
int	check_pd_mtimes(pname, mtime)
char	*pname;		/* root name of data base */
time_t	mtime;		/* timestamp of *.dctsdr */
{
	char	*xname = "check_pd_mtimes";
	int	i;
	char	*rec_type;
	char	msg[128];
	int	fd;
	struct	sdrHeader	header;

	if((fd=open_file(pvd_filename, O_RDONLY, xname)) < 0)
		return(-1);
	if(read_sdrHeader(fd, pvd_filename, xname, &header) < 0) {
		close_file(fd, pvd_filename, xname);
		return(-1);
	}
	close_file(fd, pvd_filename, xname);
        if( mtime > header.create_date) {
             string_cat(msg, sizeof(msg), "WARNING-- ", dctsdr_filename,
                                                " has a later mod date than ");
             string_cat(msg+strlen(msg), sizeof(msg)-strlen(msg),pvd_filename,
				"","");
             print_error_msg_p(xname, msg, PAUSE_F);
             return(-1);
	}

	i = 1;
	while(i < dbRecType->number) {
		if(rec_type=(char*)XGET_PRECTYPE(i)) {
			if(!open_record(i)) {
				if( mtime > pd_sdrHeader.create_date) {
				     string_cat(msg, sizeof(msg), "WARNING-- ",
						 dctsdr_filename, 
						" has a later mod date than ");
				     string_cat(msg+strlen(msg), sizeof(msg)-
						strlen(msg), rec_file,"","");
	
				     print_error_msg_p(xname, msg, PAUSE_F);
				     close_record();
				     return(-1);
				}
				close_record();
			}
		}
		i++;
	}
	return(0);
}



/*
 *   create_database_file
 *
 *	creates .database file for current database by concatenating the
 *	.dctsdr, .pvd, and record files.
 *
 */
#define		COPY_BUFFER_SIZE   8192
int	create_database_file()
{
	char	*xname = "create_database_file";
	char	filename[80];
	char	record_filename[80];
	char	*buffer;
	int	dfd;
	int	fd;
	int	status;
	int	i;
	char	*rec_type_name;

	status = 0;
        if(dct_mode == DCT_INTERACTIVE) {       /* interactive mode */
	   print_wait_msg(1);
	   if (!isalpha(*curr_dbname) || (dbRecType == NULL)) {
		print_error_msg_p(xname, "No current open database", PAUSE_F);
		return(-1);
	   }
	}
	if ( (dfd = create_file_force(xname, curr_dbname, "database", filename, 
		sizeof(filename))) < 0) {
		return(-1);
	}
	if ( (fd = open_file(dctsdr_filename, O_RDONLY, xname)) > 0) {
	   if( !(buffer=malloc(COPY_BUFFER_SIZE))) {
		status = -1;
	   }
	   else if (!copy_file(xname, fd, dfd, buffer, COPY_BUFFER_SIZE)){
		close_file(fd, dctsdr_filename, xname);
		if((fd=open_file(pvd_filename, O_RDONLY, xname)) > 0) {
		   if(!copy_file(xname, fd, dfd, buffer, COPY_BUFFER_SIZE)) {
			close_file(fd, pvd_filename, xname);
			for(i=1; i < dbRecType->number; i++) {
			   print_wait_msg(0);
			   if(!(rec_type_name=(char*)XGET_PRECTYPE(i))){
				status = -1;
				break;
			   }
			   if(form_filename(record_filename,
				 sizeof(record_filename), curr_dbname, 
					rec_type_name)) {
					status = -1;
					break;
			   }
			   if((fd=open_file(record_filename, O_RDONLY, xname))
				> 0) {
				if(copy_file(xname, fd, dfd, buffer,
					COPY_BUFFER_SIZE)) {
					status = -1;
					break;
				}
				close_file(fd, record_filename, xname);
			   }
			   else {
				status = -1;
				break;
			   }
			}			/* for loop */
		   } else
			status = -1;
		} else
		   status = -1;
	   } else
		status = -1;
	} else
	   status = -1;

	close_file(dfd, filename, xname);
	close(fd);			/* ignore error if fd is not an active
						descriptor */
	if(status) {
		delete_file(xname, filename);
		print_error_msg_p(xname, "Failed to Create database file",
					PAUSE_F);
	}	
	return(status);
}

/****************************************************************************
 *
 *   create_dctsdr_file()
 *
 *	creates a <database_name>.dctsdr file if it doesn't exist
 *
 *	returns:
 *		-1	bad file name
 *		-2	file already exists
 *		-3	create error
 *		-4	write error 
 *		-5      read error
 *		-6	malloc error
 *
 *
 ****************************************************************************/
int create_dctsdr_file(dbname)
char	*dbname;	/* database name */
{

#define		COPY_BUFFER_SIZE	8192

	char	*xname="create_dctsdr_file";
	char	filename[MAX_FILENAME_SZ];
	char	msg[80];
	int	fd;	/* file descriptor for dbname.dctsdr */
	int	sfd;	/* file descriptor for default.dctsdr */
	int	errstat;
	char	*buffer;

	fd=create_file(xname, dbname, "dctsdr", filename, sizeof(filename));
	if(fd == -2) {
		string_cat(msg, sizeof(msg), "Warning: file ", filename,
			 " exists");
		print_error_msg_p(xname,msg, PAUSE_F);
		return(-2);
	}
	else if(fd < 0)
		return(fd);
	if (( sfd=open("default.dctsdr", O_RDONLY)) < 0 ) {
		print_syserror_msg_p(xname, "Could not open default.dctsdr",
				errno, PAUSE_F);
		close(fd);
		return(-3);
	}
	if(!(buffer=malloc(COPY_BUFFER_SIZE))) {
		print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
		close(fd);
		close(sfd);
		return(-6);
	}
	errstat = copy_file(xname, sfd, fd, buffer, COPY_BUFFER_SIZE);
	if(errstat)
		errstat = errstat -3;
	close(fd);
	close(sfd);
	if(buffer)
		free(buffer);
	return(errstat);
}


/*
 * CREATE_PD
 *
 * create a process database
 */
int	create_pd(db_name)
char	db_name[MAX_FILENAME_SZ];   /* database name for noninteractive mode */
{
	char		*xname = "create_pd";
	short		i;
	short		error;
	char		input[80];
	char		filename[80];
	char		msg[80];
	time_t		mtime;		/* time stamp of *.dctsdr */
	int		errstat;
		
	if(dct_mode == DCT_INTERACTIVE) {	/* interactive mode */
		/* set up the screen */
		clear();

		/* get the database name */
		move(0,0);
		printw("enter the database name: ");
		refresh();
		getline(input,PDNAME_SZ,0,strlen("enter the database name: "));
					/***** ???????? ********/


		if(!strlen(input))	/* 0 length string? */
			return(0);
		else if (!isalpha(*input)) {
			print_error_msg_p(xname,
			    "Sorry, a database name must begin with a letter",
				 PAUSE_F);
			return(-1);
		}
		/* add the process database to the name array */
		if ((error = add_pd_name(input)) < 0){
		    if (error == -1)
			print_error_msg_p( xname,
				 "No entries available in name table", PAUSE_F);
		    else {
			string_cat(msg, sizeof(msg), "Name ", input,
				 " is already used");
			print_error_msg_p(xname, msg, PAUSE_F);
		    }
		    return(-1);
		}

		strcpy(filename,input);
	}else{		/* noninteractive mode */
		if(form_filename(filename, sizeof(filename), db_name, "pvd")) {
			print_err(xname, "Bad filename");
			return(-1);
		}
		if(!check_file_exist(filename)) {
			print_err(xname, "Create error, database exists");
			return(-1);
		}
		strcpy(filename, db_name);

	}
	/* create the files */
	/* load the "name".dctsdr definition file */
	
	print_wait_msg(1);
	errstat=create_dctsdr_file(filename);
	if(errstat == -1 || errstat < -2){
		rm_pd_name(filename);
		return(-1);
	}
	print_wait_msg(0);
	if (load_dctsdr(filename, &mtime)) {
		rm_pd_name(filename);
		getch();
		return(-1);		/* couldn't find "name.dctsdr" or 
						"default.dctsdr" */
	}

	strcpy(curr_dbname,filename);
	print_wait_msg(0);
	/* create the .pvd file */
	if (pvd_create(filename, mtime) < 0) {
		rm_pd_name(filename);
		return(-1);	/* couldn't create/init the .pvd file */
	}


	/* create pv files for each record type */

	for (i=1; i < dbRecType->number; i++){
		print_wait_msg(0);
		create_dbfile( curr_dbname, i, mtime);
	}
	close_pd(CREATE_DB);
	return(0);

}

/*
 *  DELETE_DBFILE
 *
 *   deletes a database record file of the indicated type
 *
 *      this proc assumes that the record description structure, dbRecDes
 *      has already been loaded with a call to sdrLoad()
 *
 *      returns:
 *		  0     success
 *               -1     file delete error
 *
 */
delete_dbfile(pname, type)
char    *pname;                 /* database name */
short   type;                   /* record type */
{

        char    *xname = "delete_dbfile";
        char                    filename[80];   /* hold complete file name */
        char                    *rec_type_name; /* string containing record
                                                type name */
	char			msg[80];

        if(! XGET_PRECTYPDES(type)) {
                return(-1);                     /* this type not defined in
                                                record description structures*/
        }

        if(!(rec_type_name = (char*)XGET_PRECTYPE(type))) {
                return(-1);                     /* no record type name defined
                                                        for this type */
        }

	if( form_filename( filename, sizeof(filename), pname, rec_type_name)) {
		string_cat(msg, sizeof(msg), "Could not form file name ", 
			pname, rec_type_name);
		print_error_msg(xname, msg);
		return(-1);
	}

	if( delete_file(xname, filename)) 
		return(-1);

	return(0);

}

/*
 * DELETE_PD()
 *
 * Delete all of the Process Variables in a Process Database
 */
delete_pd(db_name)
char	*db_name;		/* database name for noninteractive mode */
{
	char	*xname = "delete_pd";
	char	input[80];
	int	i;
	char	database_filename[80];


	if(dct_mode == DCT_INTERACTIVE) {
	/* get the db name and get the process database pvd into memory */
		if (select_db(input,"DELETE") < 0)
			 return(0);
		if(open_pd(input) < -1)
			return(-1);
		clear();

	/* get verification */
		printw("This will delete all PVs in this database. Continue<Y:N>: ");
		refresh();
		getline(input,2,0,
		  strlen("This will delete all PVs in this database. Continue<Y:N>: "));
		if ((input[0] != 'Y') && (input[0] != 'y')) 
			return(0);

	} else {
                if(form_filename(input, sizeof(input), db_name, "pvd")) {
                        print_err(xname, "Bad filename");
                        return(-1);
                }
                if(check_file_exist(input)) {
                       print_err(xname,"Report error, database does not exist");
                        return(-1);
                }
		if(open_pd(db_name) < -1)
			return(-1);
	}
	/* delete the .pvd file */
	delete_file(xname, pvd_filename);

        /* delete the  pv files for each record type */

        for (i=1; i < dbRecType->number; i++){
		delete_dbfile(curr_dbname, i);
        }

	/* delete the database file */

	form_filename(database_filename, sizeof(database_filename), curr_dbname,
		"database");
	delete_file(xname, database_filename);
	/* delete the .dctsdr file */
	form_filename(database_filename, sizeof(database_filename), curr_dbname,
		"dctsdr");
	if(!check_file_exist(database_filename))
		delete_file(xname,database_filename);

	if(dct_mode == DCT_INTERACTIVE) {
		rm_pd_name(curr_dbname);
		memset(curr_dbname, '\0', sizeof(curr_dbname));
	}
	return(0);
}

/****************************************************************************
 *
 *	get_formnames
 *
 *	fetches names of *.frm file in current directory
 *
 *	Returns:
 *		number of file names found
 *		-1 if error
 ****************************************************************************/
int	get_formnames()
{
        extern	char    *string_find();
	char	*xname = "get_formnames";
        DIR     *dirp;
        struct  dirent  *dp;
        int     num_found;

	forms.choice[0]="default";
	forms.nprompts = 1;

        if( (dirp = opendir(".")) == NULL ) {
                print_error_msg_p(xname, "Couldnot open dir\n", PAUSE_F);
                return(-1);
        }


/* lint will complain on this because the lint library still declares
        readdir() as returning (struct direct *) instead of the new
        (struct dirent *) .  sigh!!!!! */

        for(dp=readdir(dirp), num_found=1; (num_found < MAXFORMS) &&
					 (dp != NULL);
                                 dp=readdir(dirp)) {
                if ( string_find( ".frm", dp->d_name) != NULL ) {
                        	strcpy(frm_names[num_found], dp->d_name);
				*strchr(frm_names[num_found],'.') = '\0';
				num_found++;
                }
        }
	forms.nprompts = num_found;
        closedir(dirp);
        return(num_found);
}

/****************************************************************************
 *
 *	get_pdnames
 *
 *	fetches names of *.pvd file in current directory
 *
 *	Returns:
 *		number of file names found
 *		-1 if error
 ****************************************************************************/
int	get_pdnames()
{
        extern	char    *string_find();
	char	*xname = "get_pdnames";
        DIR     *dirp;
        struct  dirent  *dp;
        int     num_found;
	char	name[PDNAME_SZ + 16];

        if( (dirp = opendir(".")) == NULL ) {
                print_error_msg_p(xname, "Couldnot open dir\n", PAUSE_F);
                return(-1);
        }
/* lint will complain on this because the lint library still declares
        readdir() as returning (struct direct *) instead of the new
        (struct dirent *) .  sigh!!!!! */

        for(dp=readdir(dirp), num_found=0; (num_found < MAXPDS) && (dp != NULL);
                                 dp=readdir(dirp)) {
                if ( string_find( ".pvd", dp->d_name) != NULL ) {
			if(get_sdr_filetype(dp->d_name) == SDR_DB_PVD) {
                        	strcpy(name, dp->d_name);
				*strchr(name,'.') = '\0';
				add_pd_name(name);
				num_found++;
			}
                }
        }
        closedir(dirp);
        return(num_found);
}

/****************************************************************************
 *
 *	load_sum_form()
 * 
 *  returns a pointer to a struct containing  a form definition form the 
 *	specified file 
 *	returns NULL if error
 *
 ***************************************************************************/
struct	pd_summary *load_sum_form(form_name)
char	*form_name;		/* name of form to load */
{
#define	LINE_SZ	120		/* max size for a line in a form */
#define	X_SIZE	120		/* max size for a field in a form */
#define	Y_SIZE	32		/* max size for a field name */
	char	*xname="load_sum_form";
	static struct pd_summary    *pform; /* place the form will be stored*/
	static char	*fnames;	    /* place to store field names */
	int	fnamesz;			/* contains size allocated to
						a field name */
	char	*ptr;			/* used to scan through fnames */
	char	msg[120];
	int	i;
	char	name[MAX_FILENAME_SZ];
	int	frm_fd;		/* file descriptor for form file */
	size_t	index;		/* offset within form file */
	char	line[LINE_SZ];	/* buffer to read a line into */
	static	char	field1[X_SIZE];
	static	char	field2[X_SIZE];
	static	char	*pfields[] = {	/* place to parse a line (2 fields to a
						line */
			field1,
			field2
	};
	char	*fielddelim = " ,	";  /* space comma tab  (delimiters) */
	static	char	sfield1[Y_SIZE];	
	static	char	sfield2[Y_SIZE];
	static	char	*psfields[] = {	/* place to parse a field */
			sfield1,
			sfield2
	};

	index=0;
	if (pform)		/* free previous allocated form */
		free((char *)pform);
	if(fnames)
		free(fnames);
	/* allocate memory for the form */
	if (!(pform=
	      		(struct pd_summary *)malloc(sizeof(struct pd_summary)
					*MAXFORMFIELDS))){
		print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
		return(NULL);
	}
	fnamesz = FLDNAME_SZ + 1;
	if (!(fnames = malloc((unsigned)fnamesz*MAXFORMFIELDS*2))) {
		print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
		return(NULL);
	}

	if(form_filename(name, sizeof(name), form_name, FORM_EXT)){
		print_error_msg_p(xname, "Error forming the form filename",
				PAUSE_F);
		return(NULL);
	}

	/* open the form file */
	if( (frm_fd=open_file(name, 0, xname)) < 0)
		return(NULL);
	ptr=fnames;
	i=0;
	while( readline(frm_fd, &index, line) > 0) {
		if(strlen(line) > (unsigned)X_SIZE)
			continue;
		sparse(line, fielddelim, pfields, 2);
		if (*field1 == '#' || *field1 == '$' || *field1 == '%' ||
				 *field1 == '!')
			continue;		/* skip comment lines */
		/* look for a conjunction of 2 fields */
		if(strlen(field1) > (unsigned)Y_SIZE)
			continue;
		sparse(field1,"|", psfields, 2);
		strcpy(ptr, sfield1);
		(pform+i)->fldname1 = ptr;
		if(sscanf(field2, "%d", &((pform+i)->width)) != 1 ) {
			string_cat(msg, sizeof(msg), "Bad field width",line,
					""); 
			print_error_msg_p(xname, msg, PAUSE_F);
			continue;
		}
		ptr +=  fnamesz;
		/* was there a conjunction ?*/
		if(strlen(sfield2)) {
			strcpy(ptr,sfield2);
			(pform+i)->fldname2=ptr;
			ptr += fnamesz;
		} else {
			(pform+i)->fldname2 = NULL;
		}
		i++;
		
	}
	/* end of choices signature */
	(pform+i)->fldname1=NULL;
	(pform+i)->fldname2=NULL;
	(pform+i)->width=0;
	close(frm_fd);
	return(pform);
}

/*
 *	LOAD_DCTSDR
 *
 *	loads record definition files for the database
 *
 *	looks for "name.dctsdr" first, if can't find it, then looks for
 *		"default.dctsdr"
 *	returns  file modification time in mtime
 *
 *	does some rudimentary checking of the load
 *
 *	returns:
 *		0  successful
 *		-1 problem forming file name
 *		-2 could not open either "name".dctsdr or default.dctsdr
 *		-3 load was incomplete
 *
 */
load_dctsdr(pname, pmtime)
char	*pname;		/* root name string of database */
time_t	*pmtime;	/* place to store modification time of *.dctsdr */
{
	char	*xname = "load_dctsdr";
	char	*default_dctsdr = "default";
	char	msg[96];
	struct	stat statbuf;

	if (form_filename(dctsdr_filename, sizeof(dctsdr_filename), pname,
			"dctsdr")) {
		string_cat( msg, sizeof(msg), "Could not form file name:",
				pname, "dctsdr");
		print_error_msg_p(xname, msg, PAUSE_F);
		return(-1);
	}
	/* NULL out some important pointers so can check later */
/*	dbRecType = NULL;
	dbRecDes  = NULL;
*/
	if (stat(dctsdr_filename, &statbuf)){
		string_cat(msg, sizeof(msg), "Couldn't find ", dctsdr_filename,
				" Trying default.dctsdr, hit any key to cont");
		print_error_msg_p(xname, msg, PAUSE_F);
		if (form_filename(dctsdr_filename, sizeof(dctsdr_filename),
				 default_dctsdr, "dctsdr")) {
			string_cat( msg, sizeof(msg),
				 "Could not form file name:", default_dctsdr,
					 ".dctsdr");
			print_error_msg_p(xname, msg, PAUSE_F);
			return(-1);
		}
		if(stat(dctsdr_filename, &statbuf)) {
			print_error_msg_p(xname,
				 "Could not find default.dctsdr", PAUSE_F);
			return(-2);
		}
	}
	freeValidDbPtrs(&pdbBase);
	/* any previous SDR stuff has now been removed from memory */
	/* have to allocate in order to set pvtSumFlag */
	if ((pdbBase = (struct dbBase *)
	    calloc((unsigned) 1, (unsigned) sizeof(struct dbBase))) == NULL) {
	    printf("pvd_load: calloc error\n");
	    return (-1);
	}
	/* this suppresses sum check by dbRead */
	pdbBase->pvtSumFlag = 2;
	if (dbRead(&pdbBase,dctsdr_filename)) {
			getch();
			return(-3);
	}
	getValidDbPtrs(pdbBase);
        if(get_file_mtime(xname, dctsdr_filename, pmtime)) {
		getch();
                return(-1);             /* couldn't get time stamp for *.dctsdr
                                        */
        }

	if ( dbRecType == NULL) {
		print_error_msg_p(xname, "Record Types NOT loaded",PAUSE_F);
		return(-3);
	}
	if ( dbRecDes == NULL) {
		print_error_msg_p(xname, "Record Descriptions NOT loaded",
				PAUSE_F);
		return(-3);
	}
		
	return(0);
}

/*
 * MACRO_MOD
 *
 * modifies a field in all specified process variables
 */
macro_mod(){
	char		*xname = "macro_mod";
	char		pv_specifier[PVNAME_SZ+2];
	char		line[120];
	char		fieldname[10];
	char		value[80];
	int		dct_mode_save;
	short		type,i;
	struct pvd	*ppvd;

        ppvd = dbPvd;           /* this better not be NULL */
                                /* check before calling this */

	fieldname[0] = pv_specifier[0] = 0;

	/* get the Process Variable mask, field name and value */
	clear();
	printw("enter the PV mask: ");
	refresh();
	getline(pv_specifier,PVNAME_SZ,0,strlen("enter the PV mask: "));
	printw("\nenter the field to modify: ");
	refresh();
	getline(fieldname,5,1,strlen("enter the field to modify: "));
	printw("\nenter the new value: ");
	refresh();
	getline(value,80,2,strlen("enter the new value: "));
	if ((fieldname[0] == 0) || (pv_specifier[0] == 0)) 
				return(-1);

	/* order the specified pv_s in this database */
	order_specd_pvs(-1,order,sizeof(order), pv_specifier);

	suppress_errors = TRUE;
	dct_mode_save = dct_mode;
	dct_mode = DCT_NONINTERACTIVE;
	/* modify each record */
	printf("\n\n");
	for (i = 0; order[i] != -1; i++){

		/* get this record into memory */
		type = ppvd->entry[order[i]].record_type;
		if (open_record(type) < 0) 
			continue;
        /* get this record into memory */
        	if(read_record(ppvd->entry[order[i]].record_no, xname) < 0) {
       		        close_record();
			dct_mode = dct_mode_save;
               		return(-1);
        	}


		/* build line to send to fmod */
		strcpy(line,fieldname);
		strcat(line,"  ");
		strcat(line,value);

		/* modify this record and report the results */
		printf("\nProcess Variable to be modified: %s",
		  ppvd->entry[order[i]].name);
		if (fmod_pv(line,type) < 0) printf(" not changed");
		else printf(" changed");

		/* store the modified record */
		if( write_record(ppvd->entry[order[i]].record_no,xname)) {
                	getinpch();
			dct_mode = dct_mode_save;
                	close_record();
                	return(-1);
        	}

		close_record();

	}
	dct_mode = dct_mode_save;
	getinpch();
	suppress_errors = FALSE;
	return(0);
}

/*
 * MODIFY_PD
 *
 * modify a process database
 */
int modify_pd()
{
	char	*xname = "modify_pd";
	char		input[80];
	int	choice;
	short	selected;
		
	/* get the node name and get the process database header into memory */
	if (select_db(input,"OPEN") < 0) return(0);
	if(open_pd(input) < -1)
		return(0);
	
	/* service requests to modify the process database */
	choice=selected = 0;
	while (selected !=EXIT){
		strcpy(pv_menu_s.title[1], "Database:  ");
		strcat(pv_menu_s.title[1], curr_dbname);
		strcpy(pv_menu_s.title[2], "PV Name Prefix:  ");
		strcat(pv_menu_s.title[2], pvname_prefix);

		selected =
		  (*(pv_menu_s.pmt[choice=fmenu(&pv_menu_s, 1,choice)].proc))();
	}
	close_pd(CREATE_DB);
	return(0);
}

/****************************************************************************
 * OPEN_PD
 *
 * open a process database
 *
 *	returns:
 *		0  success
 *		-1 inconsistent timestamps
 *		-2 could not open .dctsdr file
 *		-3 could not read process variable dir 
 *		-4 invalid database name
 *****************************************************************************/
open_pd(pname)
char	*pname;
{
	time_t	mtime;		/* time stamp */


        if(dct_mode == DCT_INTERACTIVE) {       /* interactive mode */
	    if(!isalpha(*pname))
		return(-4);
	}
	if (strcmp(pname,curr_dbname) == SAME)
		return(0);

	/* close the currently opened process database */
	if ( dbRecDes != NULL)
		close_pd(CLEAN_UP);

	/* read in the configuration file */
	if (load_dctsdr(pname, &mtime))
		return(-2);

	/* load the the process variable dir */
	if(pvd_load(pname, pvd_filename, sizeof(pvd_filename))) {
                memset(curr_dbname, '\0',PDNAME_SZ);
		return(-3);
	}

	
	/* mark the current database in memory */
	strcpy(curr_dbname,pname);

	/* check the files' timestamps */
	if (check_pd_mtimes(pname, mtime))
		return(-1);
	

	return(0);
}

/*
 * REPORT_PD
 *
 * Report all of the Process Variables in a Process Database
 */
int report_pd(db_name)
char	*db_name;		/* database name for non-interactive mode */
{
	extern time_t	time();
	char	*xname = "report_pd";
	char	buffer[MAX_FILENAME_SZ];
	short		i,type;
	char	*ptime;
	time_t		timeofday;
	short		fld_type,offset,size;
	struct recLoc	*ppd_header = &pd_recLoc;
	struct pvd		*ppvd;
	char		*pstr;	

	if (dct_mode == DCT_INTERACTIVE) {
		/* get the output device */
		get_report_device();

	/* get the database name  to report*/
		if (select_db(buffer,"REPORT") < 0) 
			return(0);
	} else {
                if(form_filename(buffer, sizeof(buffer), db_name, "pvd")) {
                        print_err(xname, "Bad filename");
                        return(-1);
                }
                if(check_file_exist(buffer)) {
                       print_err(xname,"Report error, database does not exist");
                        return(-1);
                }

		strcpy(buffer, db_name);
	}

	if( open_pd(buffer) < -1) {
		return(-1);
	}
	ppvd = dbPvd;		/* dbPvd set by open_pd() */

	/* set long_report form variable */
	long_report = FALSE;
	timeofday = time((time_t *)0);
	ptime = ctime(&timeofday);

	if(dct_mode == DCT_INTERACTIVE) {
	  	if (rep_fd != STDOUT){
	    		clear();
	    		printw("long form <Y/N>: ");
	    		refresh();
	    		getline(buffer,2,0,strlen("long form <Y/N>: "));
	    		if ((buffer[0] == 'Y') || (buffer[0] == 'y'))
				 long_report = TRUE;
	   	}

			/* print the header page information */
		clear();
		move(0,0);
		printw("Process Database: %s",curr_dbname);
		printw("	%s",ptime);
		i = 1;
		while (i < dbRecType->number){
	    		if (pstr =(char*)XGET_PRECTYPE(i)) {
				if (!open_record(i) ) {
				     printw("%15.15s  defined: %2d free: %d\n",
					  pstr,
					  ppd_header->no_used,
					  ppd_header->no_records
					 - ppd_header->no_used);
				close_record();
				}
	    		}
	    		i++;
		}
		refresh();
		print_help_msg(3);
		if (getinpch() == EXIT) 
			return(0);
	}	/* if(dct_mode ==DCT_INTERACTIVE) */

	if (rep_fd != STDOUT) {	/* interleaf command */
		write(rep_fd,"$$mono\n  Time: ",strlen("$$mono\n  Time: "));
        write(rep_fd,ptime,strlen(ptime));
		if ( dct_mode == DCT_INTERACTIVE)
			print_wait_msg(1);
	}
	/* report every entry in this database by record type */
	for (type = 1; type < dbRecType->number; type++){

	    /* check for type exists */
	    if (!(XGET_PRECTYPE(type))) 
		continue;
	    if ( dct_mode == DCT_INTERACTIVE)
			print_wait_msg(0);
	    /* put records of this type in alphabetical order */
	    if (order_pvs(type, order, MAX_PV+4) == 0)
		continue;

	    /* report each entry */
	    for (i = 0; order[i] != -1; i++){

		/* get this record into memory */
		type = ppvd->entry[order[i]].record_type;
		if (open_record(type) < 0) 
			continue;
        /* get this record into memory */
        	if(read_record(ppvd->entry[order[i]].record_no, xname) < 0) {
                	close_record();
                	return(-1);
        	}

		close_record();

		/* Report this record */
		if (rep_fd == STDOUT){
		    clear();
		    move(0,0);
		}
		/* get the precision */
		if (rep_fd != STDOUT)		/* disable precision for rpt*/
			precision = 6;
		else if (get_recfld(type,"PREC",&fld_type,&offset,&size) >= 0){
			precision = *(short *)(&ctrl_record[offset]);
		}else{
			precision = 2;
		}
		write_page(type,rep_fd);
		if (rep_fd == STDOUT){
		    print_help_msg(3);
		    if(getinpch() == EXIT) 
				return(0);
		}
	    }
	}
	if (rep_fd != STDOUT)	/* interleaf command */
	    write(rep_fd,"\n$$end\n",strlen("\n$$end\n"));
	close_pd(CLEAN_UP);
	return(0);
}

/*
 *	report_unresolv
 *
 *	reports unresolved (ie. pv's defined in another database) process
 *		variables
 *
 */
report_unresolv()
{
	extern	time_t	time();
	char	*xname="report_unresolv";
	char	dbname[MAX_FILENAME_SZ];
	time_t	timeofday;	/* holds time of day in tics since epoch*/
	char	*ptime;		/* pointer to time in ascii */
	char	buffer[2*PVNAME_SZ + 2*(FLDNAME_SZ+2) +32];
	int	i,j;
	short	rectype;
	short	*plinks;	/* pointer to array of offsets of link fields*/
	struct	fldDes	*pfldDes;	/* field description structure */
	struct	link	*plink;		/* a link structure */
	struct	pv_link	*ppv_link;	/* a process variable link structure*/
	short	fld_type, fld_offset, fld_size; /* used by get_recfld call*/
	short	ltype,lrecno;		/* record type and # for pvdGetRec() */
	
	get_report_device();
	if(select_db(dbname, "REPORT UNRESOLVED PVs") < 0)
		return(0);
	if( open_pd(dbname) < -1) {
		return(-1);
	}
	timeofday=time((time_t *) 0);
	ptime = ctime(&timeofday);
	clear();
	move(0,0);
	refresh();
	sprintf(buffer,"Unresolved Process Variables for database %s %s\n",
		dbname, ptime);
	write(rep_fd, buffer, strlen(buffer));

	sprintf(buffer,
 "      process variable                               external pv\n\n");
	write(rep_fd, buffer, strlen(buffer));
	/* do each record type */
	for(rectype = 1; rectype < dbRecType->number; rectype++) {
		/* check if rectype exists */
		if(!XGET_PRECTYPE(rectype))
			continue;
		set_current_record_type(rectype);
		order_pvs(rectype, order, MAX_PV+4);
		if(order[0] == -1)
			continue;
		if(open_record(rectype) < 0)
			continue;
		if(get_recfld(rectype, "NAME", &fld_type, &fld_offset, 
				&fld_size) < 0) {
			print_error_msg_p(xname, 
			       "Couldn't find NAME field in record description",
				 PAUSE_F);
			continue;
		}
		/* do each pv of the current type */
		for(i=0; order[i] != -1; i++) {	
			/* check for links in current pv */
			if(read_record(dbPvd->entry[order[i]].record_no,xname)
					< 0) {
				continue;
			}
			plinks = precorddes->link_ind;
			/* check each link field of the current pv */
			for(j=0; j < precorddes->no_links; j++, plinks++){
				if((pfldDes = GET_PFLDDES(precorddes, *plinks))
					== NULL){
					print_error_msg_p(xname, 
			"NULL pointer to field description encountered",
						PAUSE_F);
					continue;
				}
			     plink=(struct link *)&ctrl_record[pfldDes->offset];
				if(plink->type == PV_LINK){
				    ppv_link=&plink->value.pv_link;
				    /*check if the pv linked to is defined*/
				    if(pvdGetRec(ppv_link->pvname, &ltype,
						&lrecno) ) {
						/* no so print it */
					sprintf(buffer,
				 "%-*.*s.%-*.*s ----> %-*.*s.%-*.*s\n",
						PVNAME_SZ, PVNAME_SZ,
					       (char *)&ctrl_record[fld_offset],
						FLDNAME_SZ, FLDNAME_SZ,
						pfldDes->fldname,
						PVNAME_SZ, PVNAME_SZ,
						ppv_link->pvname,
						FLDNAME_SZ, FLDNAME_SZ,
						 ppv_link->fldname); 
							
					write(rep_fd, buffer, strlen(buffer));
				     }
				}
			
			}
		}
		close_record();
	
	}
	if(rep_fd == STDOUT) {
		printf("\nHit any key to continue\n");
		getinpch();
	}
	close_pd(CLEAN_UP);
	return(0);
}

/*
 *	RM_PD_NAME
 *
 * remove a process database name from the list
 */
void rm_pd_name(pname)
char	*pname;
{
	short	dbinx;

	
	if (dct_mode != DCT_INTERACTIVE) 
		return;
	dbinx = 0;
	for ( dbinx = 0; dbinx < MAXPDS; dbinx++) {
	    if(!isalpha(*pd_names[dbinx]))
		break; 
	    if (strcmp(pd_names[dbinx],pname) == SAME){
		memset(pd_names[dbinx], '\0',PDNAME_SZ);
		while(++dbinx < MAXPDS) {
			if(isalpha(*pd_names[dbinx])) {
				strcpy(pd_names[dbinx-1],pd_names[dbinx]);
				memset(pd_names[dbinx], '\0', PDNAME_SZ);
			} else
				break;
		}
		break;
	    }
	}
}

/*
 * SELECT_DB
 *
 * select the database name
 */
xselect_db(pdbname,pverb)
char	*pdbname;	/* returned database name */
char	*pverb;		/* requested action to take on data base */
{
	short	dbinx;
	static short	selected = 0;

	/* set up the screen */
	clear();
	print_help_msg(1);
	move(0,13);
	printw("Select database to %s",pverb);
	move(0,0);

	/* write out the database names */
	dbinx = 0;
	while (isalpha(pd_names[dbinx][0])){
		move(dbinx,0);
		printw("%12s",pd_names[dbinx]);
		dbinx++;
	}
	refresh();

	if (dbinx == 0) {
		print_error_msg("",
			 "No databases found, press any key to continue");
		getinpch();	
		return(-1);
	}
	if (dbinx == 1){
		strcpy(pdbname,pd_names[0]);
		return(0);
	}

	/* select the database from the menu */
	if (!(selected < dbinx))
		selected = dbinx-1;
	if ((selected = menu_select(12,dbinx-1,selected,FALSE)) < 0)
		return(-1);
	strcpy(pdbname,pd_names[selected]);

	return(0);
}


/*
 * SELECT_DB
 *
 * select the database name
 */
select_db(pdbname,pverb)
char	*pdbname;	/* returned database name */
char	*pverb;		/* requested action to take on data base */
{
	short	dbinx;
	static int	selected = 0;
	char	title[80];
	
	/* setup the title */

	strcpy(title, "Select database to ");
	strcat(title, pverb);
	dbnames_f.title=title;

	/* count the database names  */
	dbinx = 0;
	while (isalpha(pd_names[dbinx][0])){
		dbinx++;
	}
	dbnames_f.nprompts = dbinx;
	if (dbinx == 0) {
		print_error_msg_p("",
			 "No databases found, press any key to continue", 
				PAUSE_F);
		return(-1);
	}
	if (dbinx == 1){
		strcpy(pdbname,pd_names[0]);
		return(0);
	}

	/* select the database from the menu */
	if(selected <0)
		selected=0;
	if (!(selected < dbinx))
		selected = dbinx-1;
	if ((selected = fchoice(&dbnames_f, 1, selected)) < 0)
		return(-1);
	strcpy(pdbname,pd_names[selected]);
	return(0);
}

/*
 *
 *	select_form
 * 
 *	select a form name for a summary report 
 *
 */
int select_form(form_name)
char	*form_name;		/* place to put name of form selected */
{

	int	choice;

	if ( get_formnames() <= 1) {	/* no forms found or error */
		strcpy(form_name, DEFAULT_FORM);
		return(0);
	}
	if((choice = fchoice(&forms,1,0)) < 0)	/* quit? */
		return(-1);
	else {
		strcpy(form_name, forms.choice[choice]);
	}
	return(0);
} 

/*
 * SUMMARIZE_PD
 *
 * Report configurable 1-line summaries  of the Process Variables in a Process
 * Database
 */
int summarize_pd(db_name, form_name)
char	*db_name;		/* database name for non-interactive mode */
char	*form_name;		/* form name to use in non-interactive mode*/
{
	extern	struct	fldDes	*pvdGetFld();
	extern time_t	time();
	struct pd_summary  *load_sum_form();

	char	*xname = "summarize_pd";
	char	buffer[MAX_FILENAME_SZ];
	char	form[MAX_FILENAME_SZ];
	short	i,j,type;
	char	*ptime;
	time_t		timeofday;
	short		fld_type,offset,size;
	struct recLoc	*ppd_header = &pd_recLoc;
	struct pvd		*ppvd;
	char		*pstr;	
	struct	fldDes	*pfield;

static	struct	pd_summary  default_sum[] = {  /* default summary */
		"DESC", NULL, 28,
		"DTYP", NULL, 28,
		"INP", "OUT", 28,
		NULL, NULL, 0		/* indicates end of list */
	};

	struct pd_summary *psum;      /* pointer to summary definition in use*/
	char	*sumstring;		/* place where summary string will be
						built */
	int	stringlen;		/* length of summary string */
	int	numflds;		/* number or fields in summary */
	char	value[80];
	char	tmpstr[80];
	char	tmpstr1[80];


	if (dct_mode == DCT_INTERACTIVE) {
		get_report_device();
		if (select_db(buffer,"SUMMARIZE") < 0) 
			return(0);
		if (select_form(form) < 0)
			return(0);
	} else {
		strcpy(buffer, db_name);
		strcpy(form, form_name);
	}

/* figure out the total line length for the summary string */
	if ( !strcmp(form, DEFAULT_FORM))
		psum=default_sum;
	else if ((psum=load_sum_form(form)) == NULL)
		return(-1);
	i=stringlen=numflds=0;
	while ( (psum+i)->width > 0) {
		numflds++;
		stringlen += (psum+i++)->width +1 ;
	}		 
	stringlen += PVNAME_SZ + MAX_RECNAME_SZ + 3;
	if(numflds == 0)
		return(-1);
	if((sumstring=malloc((unsigned)stringlen+1)) == NULL)
		return(-1);

	if( open_pd(buffer) < -1) {
		return(-1);
	}
	ppvd = dbPvd;		/* dbPvd set by open_pd() */

	if(dct_mode == DCT_INTERACTIVE) {

			/* print the header page information */
		clear();
		move(0,0);
		printw("Process Database: %s",curr_dbname);
		timeofday = time((time_t *)0);
		ptime = ctime(&timeofday);
		printw("	%s",ptime);
		i = 1;
		while (i < dbRecType->number){
	    		if (pstr = (char*)XGET_PRECTYPE(i)) {
				if (!open_record(i) ) {
				     printw("%15.15s  defined: %2d free: %d\n",
					  pstr,
					  ppd_header->no_used,
					  ppd_header->no_records
					 - ppd_header->no_used);
				close_record();
				}
	    		}
	    		i++;
		}
		refresh();
		print_help_msg(3);
		if (getinpch() == EXIT) 
			return(0);
		clear();
		refresh();
	}	/* if(dct_mode ==DCT_INTERACTIVE) */

/*	write header */

	memset(sumstring, '\0', stringlen+1);
	sprintf(sumstring,"\n%-*.*s %-*.*s",
		MAX_RECNAME_SZ, MAX_RECNAME_SZ,"Type", 
					PVNAME_SZ, PVNAME_SZ, "Name");
	for ( i=0; i < numflds; i++) {
		memset(tmpstr, '\0', sizeof(tmpstr));
		strcpy(tmpstr1, (psum+i)->fldname1);
		if ((psum+i)->fldname2) {
			strcat(tmpstr1, "|");
			strcat(tmpstr1, (psum+i)->fldname2);
		}
		sprintf(tmpstr,"%-*.*s ", (psum+i)->width, (psum+i)->width,
				tmpstr1);
		strcat(sumstring, tmpstr);
	}		 
	write(rep_fd, sumstring, strlen(sumstring));

	/* report every entry in this database by record type */

	for (type = 1; type < dbRecType->number; type++){

	    /* check for type exists */
	    if (!(pstr=(char*)XGET_PRECTYPE(type))) 
		continue;

	    /* put records of this type in alphabetical order */
	    if (order_pvs(type, order, MAX_PV+4) == 0)
		continue;

	    /* report each entry */
	    for (i = 0; order[i] != -1; i++){

		/* get this record into memory */
		type = ppvd->entry[order[i]].record_type;
		if (open_record(type) < 0) 
			continue;
        /* get this record into memory */
        	if(read_record(ppvd->entry[order[i]].record_no, xname) < 0) {
                	close_record();
                	return(-1);
        	}

		close_record();

		/* Report this record */
		/* get the precision */
		if (get_recfld(type,"PREC",&fld_type,&offset,&size) >= 0){
			precision = *(short *)(&ctrl_record[offset]);
		}else
			precision = 2;
		/* form summary string */
		memset(tmpstr, '\0', sizeof(tmpstr));
		memset(sumstring, '\0', stringlen+1);
		sprintf(sumstring,"\n%-*.*s %-*.*s",
				MAX_RECNAME_SZ, MAX_RECNAME_SZ,pstr, PVNAME_SZ,
					PVNAME_SZ, ctrl_record);
		for ( j=0; j < numflds; j++) {
			if( ((psum+j)->fldname1 &&
				 (pfield=pvdGetFld(type,(psum+j)->fldname1))) ||
				((psum+j)->fldname2 && 
			      (pfield=pvdGetFld(type, (psum+j)->fldname2)))) { 
				if((prompt_fld(type, pfield) == NOPROMPT) ||
					fld_to_ascii(pfield, value, type)) {
					fill_string(value, (psum+j)->width,
						 ' '); 
				}
			}else {
				      fill_string(value, (psum+j)->width, '*');
			}
			sprintf(tmpstr,"%-*.*s " , (psum+j)->width, 
						(psum+j)->width, value);
			strcat(sumstring, tmpstr);	
							
		}
		write(rep_fd, sumstring, strlen(sumstring));
	    }
	}
	if(rep_fd == STDOUT) {
		printf("\nHit any key to continue \n");
		getinpch();
	}
	close_pd(CLEAN_UP);
	return(0);
}
