static char SccsId[] = "$Id: pvdsubs.c,v 1.15 1995/04/26 13:35:18 mrk Exp $ ";

/*  $Id: pvdsubs.c,v 1.15 1995/04/26 13:35:18 mrk Exp $
 *	Original Author: Bob Dalesio (LANL)
 *	Current Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 *	.01	3/25/92	   frl	Null Terminated PV Name
 * .02  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 * 	...
 */

/*
 *	PVDSUBS.C
 *
 *
 */
#include        <stdlib.h>
#include        <stddef.h>
#include	<stdlib.h>
#include	<fcntl.h>
#include	<ctype.h>	/* SUN support for is* routines */
#include	<curses.h>	/* SUN support for cursor routines */
#include	<errno.h>
#include	<string.h>
#include	<sdrHeader.h>
#include	"dbPvd.h"
#include	<dbBase.h>
#include	"dctDef.h"
#include	<dbRecType.h>
#include	"dctmacros.h"

/* extern procs */

extern	short pvdHash();
extern  short pvdPvSearch();
/*********************************************/
long freeValidDbPtrs();
long getValidDbPtrs();
extern	long  dbRead();
extern struct dbBase *pdbBase;
extern struct choiceSet *choiceCvt;
extern struct arrChoiceSet *choiceGbl;
extern struct choiceRec *choiceRec;
extern struct devChoiceRec *choiceDev;
extern struct arrBrkTable *cvtTable;
extern struct recDes *dbRecDes;
extern struct recType *dbRecType;
extern struct recHeader *dbRecords;
extern struct recDevSup *devSup;
extern struct drvSup *drvSup;
extern struct recSup *recSup;
extern struct pvd *dbPvd;
/*********************************************/
/* global variables - defined in dct.c */
extern char		*pd_names[]; /* database names */
extern char	curr_dbname[];	/* name of current database name */
extern char     pvd_filename[80];       /* name of current .pvd file */
extern struct pvd	*dbPvd;  /* pointer to process var dir struct */
extern  int     dct_mode;              /* mode flag  0==interactive, 1==batch */
char	blanks[] = "                                                                                                                           ";

short		order[MAX_PV+4];	/*******?????????***********/

static findslot();

/*
 * Subroutines
 *
 * addr_to_tae
 *	args
 *		type		record type of this PV
 *		record_no	record_no of this PV
 *		pname		pointer to the PV name
 *	returns
 *		0		successful
 *		-1		not found
 * add_pv	adds a Process Variable to the Directory
 *	arguments
 *		input		name of the Process Variable
 *		type		type of record
 *		record_no	record number
 *	outputs
 *		0		record added successfully
 *		-1		illegal name specified
 *		-2		process variable already defined
 *		-3		process variable directory full
 * del_pv	deletes a Process Variable from the directory
 *	arguments
 *		name		name of the Process Variable
 *	outputs
 *		0		record deleted successfully
 *		-1		record not defined
 * findslot	find the next unused slot in the PVD
 *	returns
 *		>= 0		next available slot
 *		-1		no available slots
 * in_node	return a list of PVs of specified node & type
 *		type		record type to order (0xffff -> all types)
 *		tporder		pointer to the array in which to place the order
 *	returns
 *		no_ordered	the number of PVs in the order array
 * list_pvs	list one page worth of PVs from array pointer spec'd
 *	args
 *		porder		pointer into the order array from which to start
 *		pbegin		pointer to a string to preceed the PV name
 *	returns
 *		i		the number of PVs which were listed
 * order_pvs	order alphabetically the PVs of specified node & type
 *	args
 *		type		record type to order (0xffff -> all types)
 *		tporder		pointer to the array in which to place the order
 *	returns
 *		no_ordered	the number of PVs in the order array
 * order_specd_pvs	order alphabetically the PVs of specd node,type & mask
 *	args
 *		type		record type to order (0xffff -> all types)
 *		tporder		pointer to the array in which to place the order
 *		pv_mask		pointer to a PV mask (* is wildcard)
 *	returns
 *		no_ordered	the number of PVs in the order array
 * pvd_dump	dumps the Process Variable Directory (diagnostic tool)
 * pvd_init	initializes the Process Variable Directory
 * pvd_list	lists the Process Variables in the Directory
 *	args	
 *		type		record type to list - 0 = list all
 *	outputs
 *		-1		record not defined
 *		0		record found
 * save_pvd	write the PVD to disk
 */

/*
 * RECORD_NUM2_NAME (formerly ADDR_TO_TAG)
 *  this routine is not currently used but kept just in case 
 *
 * Given a record type and a record number, this routine searches the pvd for
 * a record of the specified type and record number and returns the pv name
 */
record_num2_name(type,rec_no,pname)
short		type;		/* record type */
short		rec_no;		/* record_no within pvd for this type */
char	*pname;		/* place to return this pv name */
{
	char		*xname = "record_num2_name";
	short		entry;
	struct pvd_entry	*pentry;

	if(pvd_present(xname))
		return(-1);

	/* pass through the pvd once */
	pentry = dbPvd->entry; 
	for (entry = 0; entry < MAX_PV; entry++,pentry++){

		/* check for this PV */
		if (legal_name(pentry->name)){
			if ((pentry->record_type == type)&&
			  (pentry->record_no == rec_no)){
				strcpy(pname,pentry->name);
				return(0);
			}
		}
	}
	return(-1);
}

/*
 *	ADD_PV
 *
 *	add a process variable to the directory
 */
add_pv(input,type,record_no, report_errs)
char	*input;		/* name of pv to add */
short		type;		/* record type */
short		record_no;	/* record # to assign to this pv */
short		report_errs;	/* 0 = suppress error reporting */
{
	char		*xname = "add_pv";
	struct pvd_entry	*pentry;
	short 	i;
	short		hshval;
	unsigned short	curr;
	short		slot;
	char		msg[80];

        if(pvd_present(xname))
                return(-4);

	if ((!legal_name(input)) && report_errs){
		string_cat( msg, sizeof(msg), "Illegal pv name: ", input, "");
		print_error_msg_p(xname, msg, PAUSE_F);
		 return(-1);
	}

	/* get the hash pointer */
	hshval = pvdHash(input);

	/* verify the process variable is not defined */
	for (i = strlen(input); i < PVNAME_SZ+1; i++)	/* NULL fill the name*/
			 *(input + i) = NULL;

	if ((pvdPvSearch(input,hshval) >= 0) && report_errs) {
		string_cat(msg, sizeof(msg), "pv name ", input, 
			" already used");
		print_error_msg_p(xname, msg, PAUSE_F);
		return(-2);
	}


	/* get an available slot in the process variable directory */
	if (((slot = findslot()) < 0) && report_errs) {
		print_error_msg_p(xname, "The pvd is full!", PAUSE_F);
		return(-3);
	}	
	/* chain new PV slot to the collision list in the hash table */
	curr = dbPvd->hash_entry[hshval];
	dbPvd->hash_entry[hshval] = slot;
	
	/* place process variable in directory */
	pentry = &dbPvd->entry[slot];
	pentry->next = curr;
	pentry->record_type = type;
	pentry->record_no = record_no;
	strncpy(pentry->name,input,PVNAME_SZ);

	return(0);
}

/*
 *	DEL_PV
 *
 *	delete a process variable from the directory
 */
del_pv(name)
char	*name;	/* process variable to delete	*/
{
	char		*xname = "del_pv";
	struct pvd_entry	*pentry;
	struct pvd	*ppvd;		/* set to dbPvd */
	short		hshval;		/* hash value for pv	*/
	short	prevptr;	/* pointer to previous pv */
	short	currptr;	/* pointer to current pv */
	short		fwdptr;		/* pointer to next pv	*/
	short		i;

        if(pvd_present(xname))
                return(-4);

	ppvd = dbPvd;

	/* get the hash pointer */
	hshval = pvdHash(name);

	/* find the process variable in the table */
	currptr = ppvd->hash_entry[hshval];
	prevptr = NO_ENTRY;
	for (i = strlen(name); i < PVNAME_SZ; i++)	/* NULL fill the name*/
			 *(name + i) = NULL;
	while ((strncmp(name,ppvd->entry[currptr].name, PVNAME_SZ) != SAME)
	  && (currptr != NO_ENTRY)){
		prevptr = currptr;
		currptr = ppvd->entry[currptr].next;
	}
	if (currptr == NO_ENTRY) {
		print_error_msg(xname, "Could not find that pv name");
		 return(-1);
	}
	fwdptr = ppvd->entry[currptr].next;

	/* unlink pv from hash chain */
	if (prevptr == NO_ENTRY){
		ppvd->hash_entry[hshval] = fwdptr;
	}else{
		ppvd->entry[prevptr].next = fwdptr;
	}
	
	/* delete the pv record */
	pentry = &ppvd->entry[currptr];
	strncpy(pentry->name,blanks,PVNAME_SZ);
	pentry->record_type = 0;
	pentry->record_no = 0;
	pentry->next = NO_ENTRY;
	
	return(0);
}

/*
 *	FINDSLOT
 *
 *	Find next unused slot in process variable directory 
 *
 */
static findslot()
{
	char	*xname = "findslot";
	short	i;
	struct pvd_entry	*pentry;
	
        if(pvd_present(xname))
                return(-1);

	/* get an available elt entry */
	pentry = dbPvd->entry;
	for (i = 0;i < MAX_PV; i++,pentry++)
		if (!legal_name(pentry->name)) return(i);
	
	return(-1);
}

/*
 * FIND_RECORDS_BY_TYPE  (formerly IN_NODE)
 *
 *  searches pvd for all records of a specified type and returns a list of
 *  pvd entries having that type
 *  return val
 *		number of type found
 *		0<  error
 */
int	find_records_by_type(type,list, listsz)
short	type;		/* record type to search for */
unsigned short	*list;	/* array to place pvd indices of this type 
					found */
int	listsz;		/* size of list */

{
	char		*xname = "find_records_by_type";
	int	i;
	short		num_found;
	struct pvd_entry	*pentry;

        if(pvd_present(xname))
                return(-1);

	num_found = 0;
	pentry = dbPvd->entry;

	/* pass through the pvd once */
	i = 0;
	while ( (i < MAX_PV) && (num_found < (listsz-1))) {

		/* get pvs of this  type */
		if  ((pentry->record_type == type) || (type == -1)){

			*(list + num_found) = i;
			num_found++;
		}
		i++;
		pentry++;
	}
	*(list+num_found) = 0xffff;
	return(num_found);
}		



/*
 * LIST_PVS
 *
 * list the process variables
 *
 *	dbPvd must be checked for NULL before calling this routine
 *	returns 
 *		the number listed
 *		0< error
 */
int	list_pvs(list,pbegin)
short	*list; /* list of pvd entry indices to list */
char	*pbegin;	/* string to prepend to each pv name listed */
{
	struct pvd *ppvd;
	int		i;


	ppvd = dbPvd;

	/* list the entries */
	clear();			/* clear the screen */
	i = 0;
	while ((*list != -1) && (i < PAGE_SIZE)){

		/* need to check mask here */

		/* print entry */
		move(i,0);
		printw("%s%.*s",pbegin,PVNAME_SZ,ppvd->entry[*list].name);
		refresh();
		list++;
		i++;
	}
	return(i);
}	

/*
 * ORDER_PVS
 *
 * put in alphabetical order all the pvs of specified type
 *
 *	returns the number of pvs of specified type that have been inserted 
 *      list
 */
order_pvs(type,list, listsz)
short	type;		/* record type */
short	*list;		/* array to hold alphabetized list */
int	listsz;		/* size of list */
{

	short		entry,orderinx,enter;
	short			num_ordered;
	short		*porder;
	struct pvd	*ppvd;
	struct pvd_entry	*pentry;

	ppvd = dbPvd;		/* this better not be NULL */
				/* check before calling this */

	num_ordered = 0;
	pentry = ppvd->entry;

	/* pass through the pvd once */
	for (entry = 0; entry < MAX_PV; entry++,pentry++){

	    /* get pvs of this  type */
	    if  ((pentry->record_type == type) || (type == -1)){

		/* find its place in the order */
		orderinx = 0;
		porder = list;
		while ((orderinx < num_ordered) && 
		  (strcmp(pentry->name,ppvd->entry[*porder].name) > 0)){
		    orderinx++;
		    porder++;
		}

		/* open up the space */
		for (enter = num_ordered; enter > orderinx; enter--)
		    *(list+enter) = *(list+enter-1);

		/* put it there */
		*(list+orderinx) = entry;

		/* increment number ordered */
		if (++num_ordered >= (listsz -1))
			break;
	    }
	}
	*(list+num_ordered) = 0xffff;
	return(num_ordered);
}		


/*
 * ORDER_SPECD_PVS
 *
 * put in alphabetical order all the pvs of specified by wildcards
 */
order_specd_pvs(type,list,listsz,pv_mask)
short		type;		/* record type */
short		*list;		/* list to recieve list of ordered pvs */
int		listsz;		/* size of list */
char		*pv_mask;	/* mask */
{

	short		entry,orderinx,enter;
	short			num_ordered;
	short		*porder;
	struct pvd	*ppvd;
	struct pvd_entry	*pentry;

        ppvd = dbPvd;           /* this better not be NULL */
                                /* check before calling this */

	num_ordered = 0;
	pentry = ppvd->entry;

	/* pass through the pvd once */
	for (entry = 0; entry < MAX_PV; entry++,pentry++){

		/* get pvs of this  type */
		if  ((pentry->record_type != type) && (type != -1))
			continue;

		/* check pv mask */
		if (!specified_by(pv_mask,&pentry->name[0]))
			continue;

		/* find its place in the order */
		orderinx = 0;
		porder = list;
		while ((orderinx < num_ordered) && 
		  (strcmp(pentry->name,&ppvd->entry[*porder].name[0]) > 0)){
			orderinx++;
			porder++;
		}

		/* open up the space */
		for (enter = num_ordered; enter > orderinx; enter--)
			*(list+enter) = *(list+enter-1);

		/* put it there */
		*(list+orderinx) = entry;

		/* increment number ordered */
                if (++num_ordered >= (listsz -1))
                        break;

	}
	*(list+num_ordered) = 0xffff;
	return(num_ordered);
}		

/*
 *
 * PVD_CREATE
 * 
 * create a process variable directory file 
 *
 *	creates a *.pvd file
 *		returns:
 *			0 success
 *			-1 error
 */

int	pvd_create(db_name, mtime)

char	*db_name;	/* process database name */
time_t	mtime;		/* time stamp of *.dctsdr */

{
	char	*xname = "pvd_create";
	int	fd;
        struct sdrHeader        sdrHeader;
        char                    ptr_malloc;  /* indicates if had to allocate
                                                space for pvd */
	int	status;
	int	file_opened;

	file_opened = status = 0;
	
        /* check if we have a buffer area allocated for pvd */

	ptr_malloc = 0;
        if( dbPvd == NULL) {
                if ((dbPvd = (struct pvd *)malloc( sizeof(struct pvd)))
				 == NULL){
                        print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
                        return(-1);
                }
                ptr_malloc = 1;
        }


        /* initialize the sdrHeader header */
        memset((char *)&sdrHeader,'\0',sizeof(struct sdrHeader));
        sdrHeader.magic = DBMAGIC;
        sdrHeader.nbytes = sizeof(struct pvd);      /* size is exclusive
                                                        of header */
        sdrHeader.type = SDR_DB_PVD;
        sdrHeader.create_date = mtime;

	/* create the process variable dir */
	if((fd = create_file(xname, db_name, "pvd", pvd_filename, 
			sizeof(pvd_filename))) < 0) {
		print_error_msg_p(xname, "Could not create process var dir",
					PAUSE_F);
		status++;
		file_opened++;
	}

	/* write the header */
	if (!file_opened) {
	   if( write_file(fd, pvd_filename, xname, (char *)&sdrHeader, 
		sizeof(struct sdrHeader)) ) {
		print_error_msg_p(xname, "Error writing process variable dir",
					PAUSE_F);
		status++;
	    }

	    if(close_file(fd, pvd_filename, xname)) {
		print_error_msg_p(xname, "File close error on process var dir",
					PAUSE_F);
		status++;
	    }
	}


	if(!status) {
	    if (pvd_init()) {
		status++;
	     }
	}
        if(ptr_malloc) {
                free((char *)dbPvd);
		dbPvd = NULL;
	}

	if(status)
		return(-1);
	else
		return(0);
}

/*
 * PVD_DUMP
 *
 * dump the process variable directory
 *
 * This is a diagnostic routine 
 */
pvd_dump(){

	int		i,j;
	int		first_one;
	struct pvd	*ppvd;
	struct pvd_entry	*pentry;
	char		*heading = "Index     PVname                  Rec Type  Rec No.   Next\n\n";

        ppvd = dbPvd;           /* this better not be NULL */
                                /* check before calling this */

#define	NO_HASH_PER_LINE 8

	clear();
	move(0,0);
	refresh();

	/* print the hash table */
	printw("Hash Table\n\n");
	for (i = 0, first_one = 1; i < HASH_NO; i+= NO_HASH_PER_LINE) {
		printw("%4.4d ", i);
		for (j =0 ; j < NO_HASH_PER_LINE; j++ )
			printw("%8.8x ",ppvd->hash_entry[i+j]);
		printw("\n");
		refresh();
		if ((i%(NO_HASH_PER_LINE*(PAGE_SIZE -3))) == 0){
			if (first_one)
				first_one = 0;
			else {
				getinpch();
				clear();
				move(0,0);
				refresh();
			}
		}
	}
	getinpch();
	/* print the entries */
	clear();
	move(0,0);
	refresh();
	printw("PVD Entries\n");
	printw(heading);
	pentry = ppvd->entry;
	for (i = 0, j = 0; i < MAX_PV; i++,pentry++){
		if ( legal_name(pentry->name) || pentry->record_type > 0 ) {
			printw("%5.5d %-*.*s %5.5d    %5.5d   %8.8x\n",
		  		i,
 				 PVNAME_SZ,
				  PVNAME_SZ,
				  pentry->name,
				  pentry->record_type,
				  pentry->record_no,
				  pentry->next);
			j++;
			refresh();
		}
		if ((j%(PAGE_SIZE -3)) == 0) {
			 getinpch();
			clear();
			move(0,0);
			printw(heading);
			refresh();
		}
	}
	getinpch();		
}
	
/*
 * PVD_INIT
 *
 * intialize the process variable directory
 */
pvd_init()
{

	char		*xname = "pvd_init";
	struct pvd_entry	entry;
	struct pvd_entry	*pentry = &entry;
	int		i;



	/* check if we have a buffer area allocated for pvd */

        if (dbPvd == NULL) {
                print_error_msg_p(xname, "No pvd to init", PAUSE_F);
                return(-1);
        }


	/* initialize the hash table */
	for (i = 0; i < HASH_NO; i++)
		dbPvd->hash_entry[i] = NO_ENTRY;


	/* initialize the entries */
	strncpy(pentry->name,blanks,PVNAME_SZ);
	pentry->record_type = 0;
	pentry->record_no = 0;
	pentry->next = NO_ENTRY;
	for (i = 0; i < MAX_PV; i++)
		memcpy((char *)&dbPvd->entry[i], (char *)pentry,
				sizeof(struct pvd_entry));

	return( save_pvd());
}

/*
 * PVD_LIST
 *
 * list the process variable directories
 */
pvd_list()
{
	short	db_inx;
	short	type_inx;
	short	number_listed;
	char		buffer[80];

	/* get pv mask */
	clear();
	printw("Enter PV mask: ");
	refresh();
	getline(buffer,PVNAME_SZ,0,strlen("Enter PV mask: "));
	if (strlen(buffer) == 0) strcpy(buffer,"*");

	/* report the .pvd file for each database */
	db_inx = 0;
	while (db_inx < MAXPDS){
	    while(!isalpha(pd_names[db_inx][0])) {
		db_inx++;
	    	if (db_inx >= MAXPDS) 
			return(-1);
	    }
	    if (open_pd(pd_names[db_inx]) < -1)
			continue;

	    /* report each type */
	    for (type_inx = 1; type_inx < dbRecType->number; type_inx++){

		/* check for type exists */
		while ((type_inx < dbRecType->number)
                && (!(char*)XGET_PRECTYPE(type_inx)))
		    type_inx++;

		if (type_inx >= dbRecType->number)
		    break;

		/* alphabetical order */
		number_listed =
			 order_specd_pvs(type_inx,order,sizeof(order),buffer);

		/* print them */
		number_listed = 0;
		while (order[number_listed] != -1){
		    clear();
		    number_listed +=
		      list_pvs(&order[number_listed],"");

		    /* header - include node name and type name later */
		    move(0,PVNAME_SZ+10);
		    printw("database %s type %s",
		      pd_names[db_inx],(char*)XGET_PRECTYPE(type_inx));	
		    refresh();

		    if (getinpch() == EXIT) 
				return(0);
		}

	    }
	    db_inx++;
/*  ++++++++++++++++++++++++++++++++++++++++++++++++++
	    sdrUnload(pvd_filename);
    +++++++++++++++++++++++++++++++++++++++++++++++++  */
	}
	return(0);
}


/*
 *   PVD_LOAD
 *
 *   load a process variable dir 
 *
 */

int  pvd_load(name, file_name, fnsz)
char	*name;		/* root name,  ".pvd" will be added to this */
char	*file_name;	/* the formed file name will be returned in this str */
int	fnsz;		/* size of file_name */
{
    char           *xname = "pvd_load";
    char            msg[80];
    struct dbBase  *ptmpdbBase = NULL;
    /* load_dctsdr must have been called prior to calling this routine */
    if (!pdbBase) {
	printf("pvd_load: load_dctsdr wasn't called\n");
	return (-1);
    }
    if (form_filename(file_name, fnsz, name, "pvd")) {
	string_cat(msg, sizeof(msg), "Could not form filename ",
		   name, ".pvd");
	print_error_msg_p(xname, msg, PAUSE_F);
	return (-1);
    }
    /* read the pvd (SDR only) into a temp area */
    if ((ptmpdbBase = (struct dbBase *)
	 calloc((unsigned) 1, (unsigned) sizeof(struct dbBase))) == NULL) {
	printf("pvd_load: calloc error\n");
	return (-1);
    }
    /* this suppresses sum check by dbRead */
    ptmpdbBase->pvtSumFlag = 1;
    if (dbRead(&ptmpdbBase, file_name)) {
	string_cat(msg, sizeof(msg), "Could not load ", file_name, "");
	print_error_msg_p(xname, msg, PAUSE_F);
	return (-1);
    }
    /* check pointers for deallocation */
    if (pdbBase->pdbPvd != dbPvd) {
	printf("pvd_load: pdbBase->pdbPvd != dbPvd\n");
	return (-1);
    } else if (pdbBase->pdbPvd) {	/* implies previous  allocation*/
	free(pdbBase->pdbPvd);
    }
    /* ready to xfer pvd pointers */
    pdbBase->pdbPvd = ptmpdbBase->pdbPvd;
    ptmpdbBase->pdbPvd = NULL;
    dbPvd = pdbBase->pdbPvd;
    /* free up the temp area*/
    freeValidDbPtrs(&ptmpdbBase);
    return (0);
}


/*
 *
 *    PVD_PRESENT
 *
 *	checks  if pvd is present
 *   	only method avail is to check if dbPvd is NULL
 */
int	pvd_present(pname)
char	*pname;		/* name of the caller */
{

        if ( dbPvd == NULL ) {
                print_error_msg(pname, "No pvd loaded");
                return(-1);
        }
	return(0);
}


/*
 *	SAVE_PVD
 *
 *	save a copy of the pvd onto the disk
 */
save_pvd()
{

	char	*xname = "save_pvd";
	int     fd;             /* file descriptor of pvd file */



	if (dbPvd == NULL) {
		print_error_msg_p(xname, "No pvd to save", PAUSE_F);
		return(-1);
	}

        if(dct_mode == DCT_INTERACTIVE) {       /* interactive mode */
	
	   if(!isalpha(curr_dbname[0])) {
		print_error_msg_p(xname, "No current database open", PAUSE_F);
		return(-1);
	   }
	   if(!isalpha(pvd_filename[0])) {
		print_error_msg_p(xname, "No pvd filename", PAUSE_F);
		return(-1);
	   }
	}
	if ( (fd = open_file( pvd_filename, O_RDWR, xname)) < 0 ) {
		print_error_msg_p(xname, "Could not open process variable dir",
				PAUSE_F);
		return(-1);
	}
	if (seek_file( fd, pvd_filename, xname, (long)sizeof(struct sdrHeader))
			 ) {
		print_error_msg_p(xname, "Seek error on process variable dir",
				PAUSE_F);
		return(-1);
	}
	if (write_file( fd, pvd_filename, xname, (char *)dbPvd,
				 sizeof(struct pvd))) {
		print_error_msg_p(xname, "Could not write process variable dir",
			PAUSE_F);
		return(-1);
	}
	if (close_file(fd, pvd_filename, xname) ) {
		return(-1);
	}
	return(0);
}

/* unload non NULL pointers from dbBase structure */
long getValidDbPtrs(pdbBase)
struct dbBase *pdbBase;
{
	if ( !pdbBase )
		return(0);
#if 0
	dbPvd=NULL;
	choiceCvt=NULL;
	choiceGbl=NULL;
	choiceRec=NULL;
	choiceDev=NULL;
	cvtTable=NULL;
	dbRecDes=NULL;
	dbRecType=NULL;
	dbRecords=NULL;
	devSup=NULL;
	drvSup=NULL;
	recSup=NULL;
#endif
	if ( pdbBase->pdbPvd != NULL ) {
		dbPvd = pdbBase->pdbPvd;
	}
	if ( pdbBase->pchoiceCvt != NULL ) {
		choiceCvt = pdbBase->pchoiceCvt;
	}
	if ( pdbBase->pchoiceGbl != NULL ) {
		choiceGbl = pdbBase->pchoiceGbl;
	}
	if ( pdbBase->pchoiceRec != NULL ) {
		choiceRec = pdbBase->pchoiceRec;
	}
	if ( pdbBase->pchoiceDev != NULL ) {
		choiceDev = pdbBase->pchoiceDev;
	}
	if ( pdbBase->pcvtTable != NULL ) {
		cvtTable = pdbBase->pcvtTable;
	}
	if ( pdbBase->precDes != NULL ) {
		dbRecDes = pdbBase->precDes;
	}
	if ( pdbBase->precType != NULL ) {
		dbRecType = pdbBase->precType;
	}
	if ( pdbBase->precHeader != NULL ) {
		dbRecords = pdbBase->precHeader;
	}
	if ( pdbBase->precDevSup != NULL ) {
		devSup = pdbBase->precDevSup;
	}
	if ( pdbBase->pdrvSup != NULL ) {
		drvSup = pdbBase->pdrvSup;
	}
	if ( pdbBase->precSup != NULL ) {
		recSup = pdbBase->precSup;
	}
	return(0);
}

long freeValidDbPtrs(ppdbBase)
struct dbBase **ppdbBase;
{
	struct dbBase *pdbBase=(struct dbBase*)*ppdbBase;
	if ( !pdbBase )
		return(0);
	if ( pdbBase->pdbPvd != NULL ) {
		free( pdbBase->pdbPvd );
		dbPvd = NULL;
	}
	if ( pdbBase->pchoiceCvt != NULL ) {
		free( pdbBase->pchoiceCvt );
		choiceCvt = NULL;
	}
	if ( pdbBase->pchoiceGbl != NULL ) {
		free( pdbBase->pchoiceGbl );
		choiceGbl = NULL;
	}
	if ( pdbBase->pchoiceRec != NULL ) {
		free( pdbBase->pchoiceRec );
		choiceRec = NULL;
	}
	if ( pdbBase->pchoiceDev != NULL ) {
		free( pdbBase->pchoiceDev );
		choiceDev = NULL;
	}
	if ( pdbBase->pcvtTable != NULL ) {
		free( pdbBase->pcvtTable );
		cvtTable = NULL;
	}
	if ( pdbBase->precDes != NULL ) {
		free( pdbBase->precDes );
		dbRecDes = NULL;
	}
	if ( pdbBase->precType != NULL ) {
		free( pdbBase->precType );
		dbRecType = NULL;
	}
	if ( pdbBase->precHeader != NULL ) {
		free( pdbBase->precHeader );
		dbRecords = NULL;
	}
	if ( pdbBase->precDevSup != NULL ) {
		free( pdbBase->precDevSup );
		devSup = NULL;
	}
	if ( pdbBase->pdrvSup != NULL ) {
		free( pdbBase->pdrvSup );
		drvSup = NULL;
	}
	if ( pdbBase->precSup != NULL ) {
		free( pdbBase->precSup );
		recSup = NULL;
	}
	if ( pdbBase != NULL ) {
		free( pdbBase );
		*ppdbBase = 0;
	}
	return(0);
}
