static char SccsId[] = "$Id: pvsubs.c,v 1.15 1995/04/26 13:35:21 mrk Exp $ ";

/*  $Id: pvsubs.c,v 1.15 1995/04/26 13:35:21 mrk Exp $
 *	Original Author: Bob Dalesio (LANL)
 *	Current Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * .03	3-16-92		frl     modified pv_report to include Interleaf Header/Trailer
 *				short reports
 * .04  05-18-92        rcz     New database access (changed sdrLoad to dbRead)
 *				XGET_PRECTYPE is old GET_PRECTYPE
 * 	...
 */

/*
 * PVSUBS.C
 *
 * process variable subroutines
 * these routines support the operations on a single process variable
 * in the process database
 *
 */

#include        <stdlib.h>
#include        <stddef.h>
#include        <stdio.h>
#include        <fcntl.h>
#include	<time.h>
#include	<ctype.h>	/* SUN support for is* routines */
#include	<curses.h>	/* SUN support for cursor routines */
#include	<string.h>
#include	"dbPvd.h"
#include	<dbRecType.h>
#include	<dbRecords.h>
#include	<sdrHeader.h>
#include	"dctDef.h"
#include	"dctmacros.h"

extern struct recType *dbRecType;
extern	short	write_page();
extern int	rep_fd;		/* report file descriptor */
extern int	pvctrl_fd;	/* process variable control file descriptor */
extern struct  recLoc    pd_recLoc;       /* header for current rec db file */
extern struct  sdrHeader pd_sdrHeader;    /* sdr header for a rec dbfile*/
extern struct  pvd	*dbPvd;		/* process variable directory */
extern char	ctrl_record[];	/* current database record in memory */
extern char	bu_record[];	/* deleted database record in memory */
extern short	bu_type;	/* deleted record type */
extern char	curr_dbname[]; /* current database open */
extern char     rec_file[MAX_FILENAME_SZ];      /* name of current
                                                         pd file */


extern short	order[];   /* array into which PV names are ordered */
extern short	precision;

extern char	pvname_prefix[];



/* Subroutines
 *
 * close_record close the display and control files currently open
 * dump_record	dumps the display and control portion of a record (Diagnostic)
 *	args
 *		type		record type of the current record
 * extend_record extend the files for the specified record type
 *	args
 * get_avl_pv	get an available record of specified type
 *	args
 *		precord_no	pointer to this record number
 *	returns
 *		0		available record found
 *		-1		no available records of specified type
 * open_loc_record	open the files for the specified record type
 *	args
 *		type		record type of PV
 *	returns
 *		file descriptor
 *		0		open successfully
 *		-1		open failed
 * open_record	open the files for the specified record type
 *	args
 *		type		record type of PV
 *	returns
 *		0		open successfully
 *		-1		open failed
 * pv_copy	create a new Process Variable from an existing one
 * pv_create	create a new Process Variable
 * pv_delete	delete a Process Variable
 * pv_modify	modify the fields of a Process Variable
 * pv_report	report a Process Variable to the printer
 * read_loc_record	read the display and control portion of the PV off disk
 *	args
 *		fd		file descriptor
 *		pbuffer		pointer to the record buffer
 *		record_no	record number of the PV
 *		type		record type of the PV
 *	returns
 *		0		read successful
 *		-1		read failed
 * read_record	read the display and control portion of the PV off disk
 *	args
 *		record_no	record number of the PV
 *		pname		name of caller
 *	returns
 *		0		read successful
 *		-1		read failed
 * select_a_pv	selects a PV from the existing PVs in this node
 *	args
 *		precord_no	pointer to the record number of the selected PV
 *		prectype	record type of the selected PV
 *		poperation	pointer to the operation to be performed on PV
 *		pname		name  of pv
 *	returns
 *		entry_number	entry number of the found Process Variable
 *		-1		operation aborted
 * select_type	select a record type
 *	args
 *		prectype	pointer to the record type
 *	returns
 *		>0		the selected record type
 *		-1		operation aborted
 * write_loc_record write the display and control portions of the PV in memory
 *	args
 *		fd		file_descriptor
 *		pbuffer		pointer to the record buffer
 *		record_no	record number of the PV
 *		type		record type of the PV
 * write_record write the display and control portions of the PV in memory
 *	args
 *		record_no	record number of the PV
 *		pname		name of caller
 */

/*
 * CLOSE_RECORD
 *
 * close the display and control portions of this record
 */
int close_record(){

	char	*xname = "close_record";

	if (close_file(pvctrl_fd, rec_file, xname)) {
		return(-1);
	}
	pvctrl_fd = -1;
	memset((char *)&pd_recLoc, '\0', sizeof(struct recLoc));
	memset(rec_file, '\0', MAX_FILENAME_SZ);
	return(0);
}

/*
 * DUMP_RECORD
 *
 * dump the record to the screen diagnostic
 */
dump_record()
{
	register short i;

	clear();
	move(0,0);
	refresh();
	printf("\n\n");
	for (i = 0; i < pd_recLoc.rec_size; i++){
		printf("%x ",ctrl_record[i]);
		if (((i % 32) == 0) && (i != 0)) printf("\n");
	}
	getinpch();
}

/*
 * EXTEND_RECORD
 *
 * extend the database file to leave some free records
 */
extend_record()
{
	char		*xname = "extend_record";
	char		buffer[MAX_RECORD_SIZE];
	size_t		end_inx;
	struct	recLoc  *pdheader = &pd_recLoc;

	memset(buffer, '\0', MAX_RECORD_SIZE);

	/* add a record to the end of the file */
	end_inx = (pdheader->rec_size * pdheader->no_records)
	  + sizeof(struct recLoc) + sizeof(struct sdrHeader);
	if ( seek_file(pvctrl_fd, rec_file, xname, end_inx))
		return(-1);
	if ( write_file(pvctrl_fd, rec_file, xname, buffer,
				(int)pdheader->rec_size))
		return(-1);

	/* increase the size in sdrHeader */
        if ( seek_file(pvctrl_fd, rec_file, xname, 0L))
                return(-1);
	if ( (read_file(pvctrl_fd, rec_file, xname, (char *)&pd_sdrHeader,
			sizeof(struct sdrHeader))) != sizeof(struct sdrHeader))
		return(-1);

	pd_sdrHeader.nbytes += pdheader->rec_size; 

        if ( seek_file(pvctrl_fd, rec_file, xname, 0L))
                return(-1);
	if ( write_file(pvctrl_fd, rec_file, xname, (char *)&pd_sdrHeader,
			sizeof(struct sdrHeader)) )
		return(-1);

	/* adjust the process database header */
	pdheader->no_records++;
	return(0);
}

/*
 * GET_AVL_PV
 *
 * get an available pv of the specified record type from the process
 * database
 */
get_avl_pv(precord_no)
short	*precord_no;
{
	char	*xname = "get_avl_pv";

	/* find an available record */
	*precord_no = 0;
	if( read_record(*precord_no, xname) < 0) 
		return(-1);

	/* assumes the first field in each record is NAME */
	while (legal_name(&ctrl_record[0])){
		*precord_no += 1;
		if (read_record(*precord_no, xname) < 0)
			return(-1);
	}

	return(0);
}

/*
 * OPEN_LOC_RECORD
 *
 * open the file for this record
 */
int	open_loc_record(type, psdrHeader, precLoc)
short type;
struct	sdrHeader	*psdrHeader;
struct	recLoc		*precLoc;
{
	char	*xname = "open_loc_record";
	char	*rec_type;
	char	msg[80];
	int	fd;

	/* open the file */
	if ( !(rec_type = (char*)XGET_PRECTYPE(type))) {
		string_cat(msg, sizeof(msg), "Record type ", rec_type, 
			" not defined for this db");
		print_error_msg(xname, msg);
		return(-1);
	}
	if (form_filename(rec_file, MAX_FILENAME_SZ, curr_dbname, rec_type)) {
		string_cat(msg, sizeof(msg), "Couldn't form filename from ",
			curr_dbname, rec_type);
		print_error_msg(xname, msg);
		return(-1);
	}
        if(( fd = open_file(rec_file, O_RDWR, xname)) < 0) {
                return(-1);
        }

       /* read the file header */
        if(read_sdrHeader(fd, rec_file, xname, psdrHeader)) {
                return(-1);
        }

        /* read the pd header */
        if(read_pd_header(fd, rec_file, xname, precLoc)) {
                return(-1);
        }

        return(fd);

}

/*
 * OPEN_RECORD
 *
 * open the open the database file for this record type
 */
open_record(type)
short type;
{
	char	*xname = "open_record";
	char	*rec_type;
	char	msg[80];

	/* check files already openned */
	if (pvctrl_fd > 0) {
		if (pd_recLoc.record_type == type)
			return(0);		/* this type currently opened*/
		close_record();		/* some other type is opened, close it*/
	}

	/* open the  file */
        if ( !(rec_type = (char*)XGET_PRECTYPE(type))) {
                string_cat(msg, sizeof(msg), "Record type ", rec_type,
                        " not defined for this db");
                print_error_msg(xname, msg);
                return(-1);
        }
        if (form_filename(rec_file, MAX_FILENAME_SZ, curr_dbname, rec_type)) {
                string_cat(msg, sizeof(msg), "Couldn't form filename from ",
                        curr_dbname, rec_type);
                print_error_msg(xname, msg);
                return(-1);
        }
        if(( pvctrl_fd = open_file(rec_file, O_RDWR, xname)) < 0) {
                getinpch();
                return(-1);
        }


	/* read the file header */
	if(read_sdrHeader(pvctrl_fd, rec_file, xname, &pd_sdrHeader)) {
                getinpch();
                return(-1);
        }

	/* read the pd header */
	if(read_pd_header(pvctrl_fd, rec_file, xname, &pd_recLoc)) {
                getinpch();
                return(-1);
        }
		
	return(0);
}

/*
 * PV_COPY
 *
 * copy an existing process variable to a new process variable
 */
pv_copy(){
	char		*xname = "pv_copy";
	short		rectype;
	short		recordno;
	short		new_recordno;
	char		input[80];
	char		msg[80];

	/* get the process variable name */
	if (select_a_pv(&recordno,&rectype,"COPY") < 0)
		 return(0);

	/* check for available record of this type */
	if( open_record(rectype))
		return(-1);

	if (get_avl_pv(&new_recordno) < 0){
		string_cat( msg, sizeof(msg), "No available ", 
			(char*)XGET_PRECTYPE(rectype), " records");
		print_error_msg(xname, msg);
		getinpch();
		return(-1);
	}

	/* get this record into memory */
	if(read_record(recordno, xname) < 0) {
		close_record();
		return(-1);
	}

	/* get the record name and add it to the PVD */
	clear();
	move(0,0);
	printw("enter the new process variable name: ");
	refresh();
	strcpy(input,pvname_prefix);
	getline1(input,PVNAME_SZ,0,strlen("enter the new process variable name: "));
	if (( add_pv(input,rectype,new_recordno,1)) < 0){
		getinpch();
		return(-1);
	}

	/* extend the file to keep the designated number of spares */
	if ((pd_recLoc.no_records-pd_recLoc.no_used) < NUMBER_FREE)
		if(extend_record())
			return(-1);

	/* put the name into the record */
	putfield(rectype,"NAME",input);
	pd_recLoc.no_used++;

	/* change fields */
	change_flds(rectype);

	/* write out the new record */
	if(write_record(new_recordno, xname)){
                getinpch();
		close_record();
		return(-1);
	}
        if(save_pd_header(pvctrl_fd, rec_file, xname, &pd_recLoc)) {
                getinpch();
                return(-1);
        }
	if(close_record())
		return(-1);

	/* save pvd  */
	if(save_pvd())
		return(-1);
	return(0);
}

/*
 * PV_CREATE
 *
 * create a new process variable
 */
pv_create(){

	char		*xname = "pv_create";
	short		rectype;
	short		record_no;
	char		input[80];
	char		msg[80];

	/* get the record type into memory */
	if (select_type(&rectype) < 0) 
		return(-1);
	if ( set_current_record_type(rectype))
		return(-1);

	/* open the file for this record type */
	if (open_record(rectype) < 0){
		getinpch();
		 return(-1);
	}

	/* check for available record of this type */
	if (get_avl_pv(&record_no) < 0){
		string_cat( msg, sizeof(msg), "No available ", 
			(char*)XGET_PRECTYPE(rectype), " records");
		print_error_msg(xname, msg);
		getinpch();
		return(-1);
	}

	/* get the record name and add it to the PVD */
	clear();
	move(0,0);
	refresh();
	printw("enter the process variable name: ");
	refresh();
	strcpy(input,pvname_prefix);
	getline1(input,PVNAME_SZ,0,strlen("enter the process variable name: "));
	if (strlen(input) < 1)
		return(-1);
	else if (( add_pv(input,rectype,record_no,1)) < 0){
		getinpch();
		return(-1);
	}

	/* extend the file to keep the designated number of spares */
	if ((pd_recLoc.no_records - pd_recLoc.no_used) < NUMBER_FREE)
		if( extend_record())
			return(-1);
	
	/* initialize the pv fields */
	init_pv();

	/* put the name into the record */
	putfield(rectype,"NAME",input);
	pd_recLoc.no_used++;

	/* change fields */
	change_flds(rectype);

	/* write out the new record */
	if(write_record(record_no, xname)) {
		getinpch();
		close_record();
		return(-1);
	}


	/* save the process database header */
        if(save_pd_header(pvctrl_fd, rec_file, xname, &pd_recLoc)) {
                getinpch();
                return(-1);
        }
	if(close_record())
		return(-1);

	/* save pvd  */
	if(save_pvd())
		return(-1);
	return(0);
}

/*
 * PV_UNDELETE
 *
 * undelete a  process variable
 *  this is not currently used .  
 */
pv_undelete(){

	char		*xname = "pv_undelete";
	short		record_no;
	char		msg[80];

	/* check for record deleted */
	if (bu_type == -1){
		print_error_msg_p(xname, "No record to undelete", PAUSE_F);
		return(-1);
	}

	/* get the record type into memory */
        if ( set_current_record_type(bu_type))
                return(-1);

	/* open the file for this record type */
	if (open_record(bu_type) < 0) { 
		getinpch();
		return(-1);
	}

	/* check for available record of this type */
	if (get_avl_pv(&record_no) < 0){
		string_cat( msg, sizeof(msg), "No available ", 
			(char*)XGET_PRECTYPE(bu_type), " records");
		print_error_msg_p(xname, msg, PAUSE_F);
		return(-1);
	}

	/* get the record name and add it to the PVD */
	if (( add_pv(&bu_record[0],bu_type,record_no,1)) < 0){
		getinpch();
		return(-1);
	}

	/* extend the file to keep the designated number of spares */
	if ((pd_recLoc.no_records-pd_recLoc.no_used) < NUMBER_FREE)
		if( extend_record())
			return(-1);

	/* restore the old record */
	memcpy(ctrl_record,bu_record,MAX_RECORD_SIZE);

	/* change fields */
	change_flds(bu_type);

	/* write out the new record */
	if(write_record(record_no, xname)) {
		getinpch();
		close_record();
		return(-1);
	}


	/* save pvd and the process database header */
        if(save_pd_header(pvctrl_fd, rec_file, xname, &pd_recLoc)) {
                getinpch();
                return(-1);
        }

	if(close_record())
		return(-1);

	/* save pvd  */
	if(save_pvd())
		return(-1);

	/* mark this back up copy as restored */
	bu_type = -1;
	return(0);
}

/*
 * PV_DELETE
 *
 * delete a process variable
 */
pv_delete(){

	char			*xname = "pv_delete";
	short			rectype;
	short			record_no;
	char			answer;
	short		no_links;
	short		entry_no;
	struct pvd	*ppvd = dbPvd;


	/* get the process variable name */
	if ((entry_no = select_a_pv(&record_no,&rectype,"DELETE")) >= 0){

	    /* read this record into memory */
	    if (open_record(rectype) < 0) {
		 return(-1);
	    }
	    if( read_record(record_no, xname)<0) {
		close_record();
		return(-1);
		}

	    /* check for any records linked to this record */
	    clear();
	    move(0,0);
	    refresh();
	    no_links = loc_single_ref(ctrl_record);
	    if (no_links > 0){
		printf("\n%d records linked - do you still want to delete this? ",no_links);
		answer = getinpch();
		if ((answer != 'Y') && (answer != 'y'))
			 return(0);
	    }

	    /* save a copy of this record */
	    memcpy(bu_record,ctrl_record,MAX_RECORD_SIZE);
	    bu_type = rectype;

	    /* delete all references to this record */
	    if(no_links > 0) {
		printf("\nDo you want to delete the link references to this record? (y or n) ");
		answer = getinpch();
		if((answer == 'y') || (answer == 'Y')) {
	    		if( del_refs(ctrl_record) < 0) {
				print_error_msg_p(xname,
		 "Error deleting links to this pv, pv not deleted", PAUSE_F);
				return(-1);
			}
	    	}
	    }

	    /* verify this record is defined in the database */
	    if (legal_name(&ctrl_record[0])){


		/* mark this record as free */
		memset(ctrl_record, '\0',(int)pd_recLoc.rec_size);
		if(write_record(record_no, xname)) {
			close_record();
			return(-1);
		}
	    	if (--pd_recLoc.no_used < 0)
			pd_recLoc.no_used = 0;
       	    	if(save_pd_header(pvctrl_fd, rec_file, xname, &pd_recLoc)) {
               		 return(-1);
             	}
		close_record();
	    }


	    /* delete from the PV Directory */
	    del_pv(ppvd->entry[entry_no].name);


	    /* save pvd  */
	    save_pvd();

	}
	return(0);
}

/*
 * PV_MODIFY
 *
 * modify a process variable
 */
pv_modify(){
	char		*xname = "pv_modify";
	short		rectype;
	short		record_no;


	/* get the process variable name */
	while (select_a_pv(&record_no,&rectype,"MODIFY") >= 0){

		/* open the file for this record type */
		if (open_record(rectype) < 0) {
			getinpch();
			return(-1);
		}
		if( read_record(record_no, xname) < 0) {
			close_record();
			return(-1);
		}

		/* change fields */
		if(change_flds(rectype) == 0){ /* write record if not Quit *?

		/* write out the new record */

        		if(write_record(record_no, xname)) {
                		getinpch();
               			close_record();
                		return(-1);
        		}
		} else {
			print_error_msg_p(xname,
				 "Quiting, Process Variable NOT modified",
					 PAUSE_F);
		}
                close_record();

	}
	return(0);
}

/*
 * PV_RENAME
 *
 * rename a process variable
 */
pv_rename(){

	char		*xname = "pv_rename";
	short		rectype;
	short		record_no;
	char		answer;
	char		input[80];
	char		*prompt1 = "Enter the new process variable name: ";
	short		no_links;
	short		entry_no;
	struct pvd	*ppvd = dbPvd;


	/* get the process variable name */
	if ((entry_no = select_a_pv(&record_no,&rectype,"RENAME")) >= 0){

	    /* read this record into memory */
	    if (open_record(rectype) < 0) {
		 return(-1);
	    }
	    if( read_record(record_no, xname)<0) {
		close_record();
		return(-1);
		}

	    clear();
	    move(0,0);
	    refresh();
	    /* verify this record is defined in the database */
	    if (!legal_name(&ctrl_record[0])){
		print_error_msg_p(xname, "Database inconsistency detected -- control record does not contain a pv name!", PAUSE_F);
		close_record();
		return(-1);
	     }
	    /* check for any records linked to this record */
	    no_links = loc_single_ref(ctrl_record);
	    if (no_links > 0){
		printw("\n%d records are linked to %-*.*s\n",no_links, 
			PVNAME_SZ, PVNAME_SZ, ctrl_record);
		printw("\tDo you still want to rename this process variable? (y or n)");
		refresh();
		answer = getinpch();
		if ((answer != 'Y') && (answer != 'y')) {
			 close_record();
			 return(0);
		}
	    }

	    /* save a copy of this record */
	    memcpy(bu_record,ctrl_record,MAX_RECORD_SIZE);
	    bu_type = rectype;

	/* get the new name */
	    clear();
	    move(0,0);
	    printw(prompt1);
	    refresh();
	    strcpy(input, pvname_prefix);
	    getline1(input, PVNAME_SZ, 0, strlen(prompt1));
	    if(strlen(input) < 1) {
		close_record();
		return(-1);
	    } else if ( add_pv(input, rectype, record_no,1) < 0) {
		close_record();
		return(-1);
	    }

	    /* rename all references to this record */
	    if(no_links > 0) {
		printw("\nDo you want to rename the link references to this record? (y or n) ");
		refresh();
		answer = getinpch();
		if((answer == 'y') || (answer == 'Y')) {
	    		if( rename_refs(ctrl_record, input ) < 0) {
				print_error_msg_p(xname,
		 "Error renaming links to this pv, pv not renamed", PAUSE_F);
				if (del_pv(input) < 0) 
					print_error_msg_p("Error backing out new process variable name from pvd", PAUSE_F);
				close_record();
				return(-1);
			}
	    	}
	    }

	    /* change the name in the control record */
	    putfield(rectype, "NAME", input);

	    if(write_record(record_no, xname)) {
			close_record();
			return(-1);
	    }
 	    close_record();


	    /* delete from the PV Directory */
	    del_pv(ppvd->entry[entry_no].name);


	    /* save pvd  */
	    save_pvd();

	}
	return(0);
}

/*
 * PV_REPORT
 *
 * report a process variable
 */
pv_report(){

	char		*xname = "pv_report";
	short		rectype;
	short		record_no;
	short		fld_type,offset,size;
	char		*ptime;
	time_t		timeofday;
	extern	time_t	time();


	/* get a report device */
	get_report_device();
	if ( rep_fd != STDOUT) {
		timeofday = time((time_t *)0);
		ptime = ctime(&timeofday);
		write(rep_fd, "$$mono\n  Time: ", strlen("$$mono\n  Time: "));
		write(rep_fd, ptime, strlen(ptime));
	}

	/* get the process variable being reported */
	while (select_a_pv(&record_no,&rectype,"REPORT") >= 0){

		/* get this record into memory */
		if (open_record(rectype) < 0) 
			continue;
		if(read_record(record_no, xname) < 0) {
			close_record();
			return(-1);
		}
		close_record();

		/* report record */
		if (rep_fd == STDOUT){
			clear();
			move(0,0);
		}
		/* get the precision */
		if (get_recfld(rectype,"PREC",&fld_type,&offset,&size)
				 >= 0){
			precision = *(short *)(&ctrl_record[offset]);
		}else{
			precision = 2;
		}
		write_page(rectype,rep_fd);
		if (rep_fd == STDOUT) 
			getinpch();
	}
	if ( rep_fd != STDOUT) 
		write(rep_fd, "\n$$end\n", strlen ("\n$$end\n"));
	return(0);
}

/*
 * READ_LOC_RECORD
 *
 * read the current database record into memory
 */
read_loc_record(fd,pbuffer,record_no, pname, precLoc)
short	fd;			/* file descriptor of open record */
char	*pbuffer;		/* buffer to place control record */
short		record_no;	/* record number to read */
char	*pname;			/* caller's name */
struct	recLoc	*precLoc;
{

	size_t	ctrl_offset;

	/* get the control portion */
	ctrl_offset = 
	  sizeof(struct sdrHeader) + sizeof(struct recLoc) +
			 (record_no*precLoc->rec_size);

        /*read database record from disk */

        if( seek_file( fd, rec_file, pname, ctrl_offset))
                return(-1);
        if( (read_file(fd, rec_file, pname, pbuffer,
                        (int)precLoc->rec_size)) != precLoc->rec_size)
                return(-1);

	return(0);
}

/****************************************************************************
 *
 *  read_pd_header
 * 
 *  load the recLoc struct for the currently open pd file 
 *
 *	returns:
 *		0  success
 *		-1 error
 *****************************************************************************/
int  read_pd_header(fd, filename, procname, precLoc)
int	fd;		/* file descriptor for the open pd file */
char	*filename;	/* filename for the pd file */
char	*procname;	/* name of the caller */
struct	recLoc *precLoc;	/* place to load the header */
{

	if( seek_file(fd, filename, procname, (size_t)sizeof(struct sdrHeader))) 	{
		getinpch();
		return(-1);
	}
	if ( read_file(fd, filename, procname, (char *)precLoc,
			 sizeof(struct recLoc)) != sizeof(struct recLoc)) {
                getinpch();
                return(-1);
        }
	return(0);
}
	

/*
 * READ_RECORD
 *
 * read the current database record into memory
 */
read_record(record_no, pname)
short	record_no;
char	*pname;			/* caller's name */
{
	size_t	ctrl_offset;

	/* get the control portion */
	ctrl_offset = 
	  sizeof(struct sdrHeader) + sizeof(struct recLoc) + 
		(record_no*pd_recLoc.rec_size);


	/*read database record from disk */

	if( seek_file( pvctrl_fd, rec_file, pname, ctrl_offset))
		return(-1);
	if( (read_file(pvctrl_fd, rec_file, pname, ctrl_record, 
			(int)pd_recLoc.rec_size)) != pd_recLoc.rec_size)
		return(-1);

	return(0);
}

/****************************************************************************
 *
 *  save_pd_header
 * 
 * writes  the recLoc struct for the currently open pd file 
 *
 *	returns:
 *		0  success
 *		-1 error
 *****************************************************************************/
int  save_pd_header(fd, filename, procname, precLoc)
int	fd;		/* file descriptor for the open pd file */
char	*filename;	/* filename for the pd file */
char	*procname;	/* name of the caller */
struct	recLoc *precLoc;	/* the header */
{

	if( seek_file(fd, filename, procname,
			 (size_t)sizeof(struct sdrHeader))) {
		getinpch();
		return(-1);
	}
	if ( write_file(fd, filename, procname, (char *)precLoc,
			 sizeof(struct recLoc)) ) {
                getinpch();
                return(-1);
        }
	return(0);
}
	

/*
 * SELECT_A_PV
 *
 * select a process variable from those in this node
 */
select_a_pv(precord_no,prectype,poperation)
short	*precord_no;		/* returned record # of pv selected */
short	*prectype;		/* returned record type of pv selected */
char	*poperation;
{

	short		page_number;
	short		number_ordered;
	short		new_value;
	short		entry_number;
	short		record_type;
	short		selected = 0;
        struct pvd      *ppvd = dbPvd;


	/* order the pvs in this node */
	if (select_type(&record_type) < 0) return(-1);
	if ((number_ordered = 
	  order_pvs(record_type, order, MAX_PV+4)) == 0)
		return(-1);

	/* page through the defined variables */
	page_number = 0;
	list_pvs( order,"");
	move(0,PVNAME_SZ+10);
	printw("select pv to %s in %s",poperation, curr_dbname);
	print_help_msg(1);	
	refresh();

	/* select a pv from the menu */
	while ((new_value = 
	  menu_select(0,
	    ((((page_number+1)*PAGE_SIZE)>number_ordered)?
	      ((number_ordered%PAGE_SIZE)-1):PAGE_SIZE-1),
	    selected,TRUE))
	  != DONE){
		if (new_value == UP){
			if (page_number > 0){
				page_number--;
				list_pvs(&order[page_number*PAGE_SIZE],"");
				move(0,PVNAME_SZ+10);
				printw("select pv to %s",poperation);
				refresh();
				selected = 0;
			}
		}else if (new_value == DOWN){
			if (((page_number+1) * PAGE_SIZE) < number_ordered){
				page_number++;
				list_pvs(&order[page_number*PAGE_SIZE],"");
				move(0,PVNAME_SZ+10);
				printw("select pv to %s",poperation);
				selected = 0;
				refresh();
			}
		}else{
			selected = new_value;
			entry_number =
				 order[selected + (page_number * PAGE_SIZE)];
			*precord_no = ppvd->entry[entry_number].record_no;
			*prectype = ppvd->entry[entry_number].record_type;
			return(entry_number);
		}
	}
	return(DONE);
}

/*
 * SELECT_TYPE
 *
 * select a record type from those defined
 */
select_type(prectype)
register short	*prectype;
{
	register short	no_types;
	register short	i;
	short		types[MAX_RECORD_TYPES];
	static short	selected = 0;
	char		*pstr;		/* pointer to string containing record
						type name */

	/* display the selection of type names */
	clear();
	no_types = 1;
	i = 0;
	while (no_types < dbRecType->number){
		if (pstr = (char*)XGET_PRECTYPE(no_types)) {
			move(i,0);
			printw("%+15.15s",pstr);
			types[i] = no_types;
			i++;
		}
		no_types++;
	}
	print_help_msg(1);
	refresh();
		
	/* select the record type */
	if ((i = menu_select(16,i-1,selected,FALSE)) < 0){
		return(-1);
	}
	selected = i;
	*prectype = types[i];
	return(*prectype);
}

/*
 * WRITE_LOC_RECORD
 *
 * write the current database record to disk
 */
write_loc_record(fd,pbuffer,record_no, precLoc)
short	fd;
char	*pbuffer;
short		record_no;
struct	recLoc	*precLoc;
{
	char	*xname = "write_loc_record";
	size_t	ctrl_offset;

        /* write the control portion */
        ctrl_offset = (record_no * precLoc->rec_size)
        + sizeof(struct sdrHeader) +sizeof(struct recLoc);

        /* write to disk data base */
        if (seek_file(fd, rec_file, xname, ctrl_offset))
                return(-1);

        if ( write_file(fd, rec_file, xname, pbuffer,
                (int)precLoc->rec_size))
                return(-1);
        return(0);

}

/*
 * WRITE_RECORD
 *
 * write the global record to disk
 */
write_record(record_no,pname)
short	record_no;
char	*pname;		/* name of caller */
{
	size_t	ctrl_offset;

	/* write the control portion */
	ctrl_offset = (record_no * pd_recLoc.rec_size)
  	+ sizeof(struct sdrHeader) +sizeof(struct recLoc);

	/* write to disk data base */
	if (seek_file(pvctrl_fd, rec_file, pname, ctrl_offset))
		return(-1);

	if ( write_file(pvctrl_fd, rec_file, pname, ctrl_record, 
		(int)pd_recLoc.rec_size))
		return(-1);
	return(0);
		
}
