static char SccsId[] = "$Id: rdfsubs.c,v 1.22 1995/04/26 13:35:24 mrk Exp $ ";

/*  $Id: rdfsubs.c,v 1.22 1995/04/26 13:35:24 mrk Exp $
 *	Original Author: Bob Dalesio (LANL)
 *	Current Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  07-22-91        rac     don't put date on report for each record
 * .02	3/16/92		frl	link contstant changed from float to double
 * .03  3/25/92		frl     PV NAME null terminated
 * .04	4/8/92		frl	SPC_HDWLINK support
 * .05  05-18-92        rcz     New database access - XGET_PRECTYPE is old GET_PRECTYPE
 * .06	09-10-92	frl	support for VXI
 * .07	08-26-93	mrk	Changes for link.h; bcopy,bzero=>memcpy,memset
 */

/*
 * RDFSUBS.C
 *
 * Record Definition Subroutines
 *
 */
#include        <stdlib.h>
#include        <stddef.h>
#include        <stdio.h>
#include <time.h>
#include <ctype.h>	/* SUN support for is* routines */
#include <curses.h>	/* SUN support for cursor routines */
#include <string.h>
#include <dbDefs.h>
#include <dbFldTypes.h>
#include <dbRecDes.h>
#include "dbPvd.h"
#include <link.h>
#include <choice.h>
#include <dbRecType.h>
#include <special.h>
#include "dctmacros.h"
#include "dctDef.h"

extern struct dbBase *pdbBase;
extern struct choiceSet *choiceCvt;
extern struct arrChoiceSet *choiceGbl;
extern struct choiceRec *choiceRec;
extern struct devChoiceRec *choiceDev;
extern struct arrBrkTable *cvtTable;
extern struct recDes *dbRecDes;
extern struct recType *dbRecType;
extern struct recHeader *dbRecords;
extern struct recDevSup *devSup;
extern struct drvSup *drvSup;
extern struct recSup *recSup;
extern struct pvd *dbPvd;

extern  char    *get_dct_choice_string();
extern	short	get_dct_choice_val();
extern	struct	fldDes	*pvdGetFld();
extern	long	postfix();
extern	short	write_page();

extern time_t	time();

extern	int		dct_mode;	/* mode flag 0==interactive, 1==batch*/
extern short		long_report;	/* TRUE = long report to disk */
extern char		ctrl_record[];	/* current record in memory */
extern short		curr_rectypedef;  /* record type currently in memory */
extern struct fldDes  **pflddef;	/* current record's field definitions */
extern struct recTypDes *precorddes;	/* pointer to current record def struc*/
extern char		curr_dbname[]; /* current database in memory */
extern int      rep_fd;         /* report file descriptor */



/* input and output link type and precision
 * These flags signify the input and output type and the precision. 
 * The input and output link are used to change the menu so that hardware 
 * dependent prompts only appear when the link is to hardware. The precision
 * changing will also cause the screen to be redisplayed.
 */
short	in_link_type;
short	out_link_type;
short	precision;
short	dtyp_val;		/* used to hold value of DTYP so can determine
					when it's value is changed */
int	inp_out_row;		/* screen row number where INP or OUT field
					is displayed */

/*
 * Subroutines
 *
 *
 * add_sp	adds spaces to field names shorter then 4 characters
 *		 -- also found in brdf.c
 *	args	
 *		pinfld		pointer to input field
 *		poutfld		pointer to output field
 * change_flds		modify the fields of the PV in memory
 *	args
 *		type		record type
 * change_range		visually changes ranges determined by a modified field
 *	args
 *		pfldname	pointer to fieldname
 * check_rewrite	rewrite screen on change between hw and sw links
 *	args
 *		plink		pointer to field value
 *		pfldname	pointer to field name
 *		type		record type
 *		pno_prompts	pointer to the number of fields prompted
 *		row		current row
 * convert_fld 		convert ascii input to PV field
 *	args
 *		type		record type
 *		pfield		pointer to the field definition
 *		pinput		pointer to the ascii input
 *		pvalue		pointer to the converted input
 *	returns
 *		-1		bad value or out of range
 *		0		converted correctly
 * fld_search	search fields for string and return the fields number
 *	args
 *		pstring		name searching for
 *	returns
 *		>=0		field number
 *		-1		string not matched
 * fld_to_ascii		convert a PV field to ascii
 *	args
 *		pfield		pointer to the field definition
 *		buffer		buffer in which to place the value
 *		type		record type
 * get_ab_addr		convert a string to Allen-Bradley address
 *	args
 *		pinput		input string
 *		plink		pointer to AB link number
 *		padapter	pointer to AB adapter
 *		pcard		pointer to card number
 *		psignal		pointer to signal number
 *		pplc_flag	pointer to PLC flag
 *		card_type	card type
 *		pparm		pointer to the parameter
 *
 * get_camac_addr	convert a string to a CAMAC address
 *	args
 *		pinput		input string
 *		pb		pointer to the branch number
 *		pc		pointer to the crate number
 *		pn		pointer to the Station number
 *		pa		pointer to the address
 *		pf		pointer to the function
 *		pparm		pointer to the parameter
 *	returns
 *		0		successful
 *		<0		not successful - issues an error msg on failure
 * get_range_val	return the value from a given field
 *	args
 *		pfield		pointer to field definition
 *		prange		pointer to range
 *	returns
 *		 0		range found A-ok
 *		-1		value of type not checked
 * get_vme_addr		convert a string to a VME address
 *	args
 *		pinput		input string
 *		pslot		pointer to the slot number
 *		pchannel	pointer to the channel number
 *		card_type	card type
 *		pparm		pointer to the parameter
 *	returns
 *		0		successful
 *		<0		not successful - issues an error msg on failure
 * mod_record		get new field specifications for the PV in memory
 *	args
 *		type		type
 *		no_prompts	number of prompted fields
 *
 * prompt_fld		Determines if a field should be prompted
 *	args
 *		prompt		fields' prompt flag
 *	returns
 *		NO		field should not be prompted
 *		YES		prompt field
 *
 * putfield		put a field into a PV
 *	args
 *		type		record type
 *		pfield_name	pointer to the field name
 *		pvalue		pointer to the value buffer
 *
 * rec_name_to_type_num  converts a record name string to a record type index
 *	args
 *		name_string	record name string
 *
 * rotate		change a field with list of selections
 *	args
 *		pfield		pointer to the field definition
 *		direction	next selection or previous selection
 *		pvalue		pointer to the value to change
 *
 * set_current_record_type  sets the current record type
 *	args
 *		record_type    record type index
 *
 * to_fieldname		convert a record type and offset to a field name
 *	args
 *		record_type	record type
 *		offset		field offset
 *		pname		pointer to the field name
 *		field_type	field type
 *	returns
 *		0		found the offset
 *		-1		offset not found
 *
 * write_page		write the fields of the PV in memory to the printer
 *	args
 *		type		record type
 *		fd		file descriptor to which to write the page
 *	returns
 *		no_fields	number of fields which are reported
 * xrectype		convert record type string to number
 *	args
 *		ptypename	pointer to the record type name
 *	returns
 *		>= 0		record type
 *		-1		no match
 */



/*	ADD_SP
 *
 *	adds spaces to field names shorter then 4 char -- also found in brdf.c
 */
add_sp(pinfld,poutfld)
char *pinfld;
char *poutfld;
{
	int i;

	/* put blanks after the name if it is less then four characters */ 
	for (i=0; i<FLDNAME_SZ; i++){
		if ((*(pinfld+i) >= 'A' && *(pinfld+i) <= 'Z') 
		  || (*(pinfld+i) >= '0' && *(pinfld+i) <= '9'))
			*(poutfld+i) = *(pinfld+i);
		else *(poutfld+i) = ' ';
	}
	*(poutfld+i) = '\0';
}

/*
 * CHANGE_FLDS
 *
 * modify the fields of the process variable currently in memory
 */
int change_flds(type)
short	type;
{
	short	no_prompts;	/* num of fields which are prompted */
	short		fld_type,offset,size;

	/* get the precision */
	if (get_recfld(type,"PREC",&fld_type,&offset,&size) >= 0){
		precision = *(short *)(&ctrl_record[offset]);
	}else{
		precision = 2;
	}

	/* display the menu */
	no_prompts = write_page(type,STDOUT);

	/* go through the menu */
	return(mod_record(type,no_prompts));
}

/*
 *	CHANGE_RANGE
 *
 *	visually changes ranges determined by a modified field
 */
change_range(type, pfldname)
short	type;		/* record type */
char *pfldname;		/* name of modified field */
{
	char	*xname = "change_range";
	short	j;	/* field counter	*/
	short	row;	/* row counter		*/
	short	fldnum;	/* field number		*/
	short	change;		/* signals a change needs to be made 	*/
	char	range_str[RANGE_STRING_SZ+1];
	struct fldDes *pfield;

	/* initialize variables */
	row = 1;
	change = FALSE;

	/* find field number of field name */
	fldnum = fld_search(pfldname);

	/* search fields for ranges determined by inputed field */
	for(j=0; j<precorddes->no_fields; j++){

		pfield = pflddef[j];
		/* if range is variable then find its current value */
		if (pfield->lowfl == VAR && fldnum == pfield->range1.fldnum){
			get_range_val(pflddef[fldnum],&pfield->range1.value);
			change = TRUE;
		}

		if (pfield->highfl == VAR && fldnum == pfield->range2.fldnum){
			get_range_val(pflddef[fldnum],&pfield->range2.value);
			change = TRUE;
		}

		/* if there is a range change erase old range and write in new*/
		if ( change == TRUE){

			/* delete old range */
			fill_string(range_str, (short)RANGE_STRING_SZ, ' ');
			move(row,RANGE_X);
			printw("%s",range_str);
			/* create what will be written  */
		build_range_string(xname, pfield->field_type, pfield->cvt_type,
			range_str, &pfield->range1.value, &pfield->range2.value);

			/* write in new range */
			move(row,RANGE_X);
			printw("%s",range_str);

			change = FALSE;
		}

		/* if prompted field, move down a row */
		if (prompt_fld(type, pfield) != NOPROMPT)
			++row;
	}
}

/*
 * CHECK_REWRITE
 *
 * after a field is modified this routine is used to determine 
 * if the input or output links switched between hardware and software
 * inputs - if so the screen is rewritten and the number of fields
 * prompted is returned
 */
check_rewrite(pfield,pfldname,type,pno_prompts,row)
short		*pfield;
char		*pfldname;
short		type;		/* record type index */
short		*pno_prompts;	/* number of prompted fields (returned) */
short		row;
{

	if (strncmp(pfldname,"PREC",4) == SAME){
		/* precision changes the display */
		if (*pfield != precision){
			precision = *pfield;
			*pno_prompts = write_page(type,STDOUT);
		}
	}

	/* move back to original position */
	move(row,VALUE_X);
}

/*
 * CONVERT_FLD
 *
 * convert an input to the correct type for this field
 */
int   convert_fld(type,pfield,pinput,pvalue)
short			type;		/* record type */
struct fldDes	*pfield;	/* pointer to field struct for
						current field */
char		*pinput;	/* character string input by user */
char		*pvalue;	/* pointer to area to receive converted
						value */
{
	char		*xname = "convert_fld";
	short	i;
	int		status;
	int	var_range;	/* != 0 -> current field has a variable range */
	short	dummy1, dummy2;
	union	fld_types  tmp_fld_types;
	static	char	field1[PVNAME_SZ+1];
	static 	char	field2[PVNAME_SZ+1];
	static	char	*pfields[] = {		/* used to receive parsed */
			field1,			/*  pvname */
			field2
			};
	char		*pdelim = " .";		/* field delimiter characters
						for pv names */
	struct link	*pdblink;
        char            *pchoice;
	char		rpbuf[128];		/* used for call to postfix */
	short		error_number;		/* used for call to postfix */
        struct choiceSet        *pchoiceSet;
        struct arrChoiceSet     *parrChoiceSet;
	struct devChoiceSet	*pdevChoiceSet;

	var_range = 0;
	/* if range is variable then find its current value */
	if (pfield->lowfl == VAR){
		get_range_val(pflddef[pfield->range1.fldnum],
				&pfield->range1.value);
		var_range++;
	}
	if (pfield->highfl == VAR){
		get_range_val(pflddef[pfield->range2.fldnum],
				&pfield->range2.value);
		var_range++;
	}

	switch (pfield->field_type) {
	   case (DBF_CHAR):
	   case (DBF_UCHAR):
	   case (DBF_SHORT):
	   case (DBF_USHORT):
	   case (DBF_LONG):
	   case (DBF_ULONG):		
		if( (status = convert_type_to_val( xname, pfield->field_type,
			 pfield->cvt_type, pinput, &tmp_fld_types))
				!= 0 ) 
				if (status == -1) {
				   print_error_msg_p (xname, "illegal value",
						PAUSE_F);
				   return(-1); 
				}
		if( check_field_range(xname, &tmp_fld_types,
			&pfield->range1.value, &pfield->range2.value, 
				 pfield->field_type, var_range) != 0)
				return(-1);
		switch (pfield->field_type) {
	   		case (DBF_CHAR):
				*(char *)pvalue = 
					tmp_fld_types.char_value;
				break;
	   		case (DBF_UCHAR):
				*(unsigned char *)pvalue = 
					tmp_fld_types.uchar_value;
				break;
	   		case (DBF_SHORT):
				*(short *)pvalue = tmp_fld_types.short_value;
				break;
	   		case (DBF_USHORT):
				*(unsigned short *)pvalue = 
					tmp_fld_types.ushort_value;
				break;
	   		case (DBF_LONG):
				*(long *)pvalue = tmp_fld_types.long_value;
				break;
	   		case (DBF_ULONG):		
				*(unsigned long *)pvalue = 
					tmp_fld_types.ulong_value;
				break;
		}
		return(0);
	case (DBF_STRING):
		if(pfield->special == SPC_CALC) {
			memset(rpbuf,'\0', sizeof(rpbuf));
			if(postfix(pinput,rpbuf,&error_number) < 0) {
				error_msg_p(xname, error_number, PAUSE_F);
				return(-1);
			}
		}
		i = 0;
		while (isprintable(*pinput) && (++i < pfield->size))
			*pvalue++ = *pinput++;
		*pvalue = NULL;
		return(0);
	case (DBF_DOUBLE):
		if (sscanf( pinput, "%lf", &tmp_fld_types.double_value) < 1) {
			print_error_msg_p(xname, "illegal value", PAUSE_F);
			return(-1);
		}
		if ( check_field_range(xname, &tmp_fld_types, 
			&pfield->range1.value, &pfield->range2.value,
				pfield->field_type, var_range) != 0)
			return(-1);
		*(double *)pvalue = tmp_fld_types.double_value;
		return(0);
	case (DBF_FLOAT):
		if ( sscanf( pinput, "%f", &tmp_fld_types.float_value) < 1) {
			print_error_msg_p(xname, "illegal value", PAUSE_F);
			return(-1);
		}
		if ( check_field_range(xname, &tmp_fld_types, 
			&pfield->range1.value, &pfield->range2.value,
				pfield->field_type, var_range) != 0)
			return(-1);
		*(float *)pvalue = tmp_fld_types.float_value;
		return(0);
	case (DBF_ENUM):
		if ( sscanf( pinput, "%hu", &tmp_fld_types.enum_value) < 1) {
			print_error_msg_p(xname, "illegal value", PAUSE_F);
			return(-1);
		}
		if ( check_field_range(xname, &tmp_fld_types, 
			&pfield->range1.value, &pfield->range2.value,
				pfield->field_type, var_range) != 0)
			return(-1);
		*(unsigned short *)pvalue = tmp_fld_types.enum_value;
		return(0);
	case (DBF_INLINK):
		pdblink =  (struct link *)pvalue;

	/* check if this is a hardware address spec or softchannel */
		if ( strncmp(pfield->fldname, "OUT ", FLDNAME_SZ) == 0 ) {
			print_error_msg_p(xname,
		 "Field name OUT not allowed to have field type of DBF_INLINK",
					PAUSE_F);
			return(-1);
		} else if (pfield->special == SPC_HDWLINK) {
			return(process_spc_hdwlink(xname, type, pinput,
				pdblink));
		} else if (strncmp(pfield->fldname, "INP ", FLDNAME_SZ) == 0 ) {
			/* yep, this is a hardware add spec, so get the 
					link type */
			return(process_link_field(xname, type, pinput,
					 pdblink, 0 , -1));
		/*  not a hardware address or softchannel  */
		}else {
			return(process_soft_channel(xname, pinput, pdblink));
		}
	case (DBF_OUTLINK):
		pdblink = (struct link *)pvalue;

	/* check if this is a hardware address spec or softchannel */
		if (strncmp(pfield->fldname, "INP ", FLDNAME_SZ) == 0 ) {
			print_error_msg_p(xname,
	 "Field name INP not allowed to have field type of DBF_OUTLINK",
				PAUSE_F);
			return(-1);
		} else if (pfield->special == SPC_HDWLINK) {
			return(process_spc_hdwlink(xname, type, pinput,
				pdblink));
		} else if (strncmp(pfield->fldname, "OUT ", FLDNAME_SZ) == 0 ) {
			/* yep, this is a hardware add spec, so get the link
					 type */
                        return(process_link_field(xname, type, pinput,
                                         pdblink,0, -1));
		/*  not a hardware address or softchannel  */
		}else {
                        return(process_soft_channel(xname, pinput, pdblink));
		}
	case (DBF_FWDLINK):
		pdblink = (struct link *)pvalue;

	/* check if this is a hardware address spec or softchannel */
		if ((strncmp(pfield->fldname, "INP ", FLDNAME_SZ) == 0 ) || 
			(strncmp(pfield->fldname, "OUT ", FLDNAME_SZ) == 0 )) {
			print_error_msg_p( xname,
		 "Field name INP or OUT may not be of type DBF_FWDLINK",
					PAUSE_F);
			return(-1);
		
		/*  not a hardware address or softchannel  */
		} else if (dct_mode == DCT_INTERACTIVE &&
				strlen(pinput) == 0) {
			pop_link_window(xname, pdblink,(long)FWDLINKPROMPTS);
			return(0);
		}else if (legal_name_ref(pinput)){      /* process var name */
                        sparse(pinput, pdelim, pfields, 2);
                        pdblink->type = PV_LINK;
                        strcpy(pdblink->value.pv_link.pvname, field1);
			fixup_fldname(field2, pdblink->value.pv_link.fldname);
			if( dct_mode == DCT_INTERACTIVE && pvdGetRec(
			     pdblink->value.pv_link.pvname, &dummy1, &dummy2)) {
				print_error_msg_p(xname, 
		"Warning, this process variable not defined in this database",
						PAUSE_F);
			}

		
		}else{				/* constant */
			if ( sscanf( pinput, "%lf", &tmp_fld_types.double_value) < 1) {
				print_error_msg_p(xname, "illegal value", PAUSE_F);
				return(-1);
			}

			/* set the new structure */
			memset((char *)pdblink,'\0',sizeof(struct link));
			pdblink->type = CONSTANT;
			pdblink->value.value =
				 tmp_fld_types.double_value;
		}
		return(0);
	case (DBF_GBLCHOICE):
		if( !(pchoiceSet = GET_PCHOICE_SET(choiceGbl, 
				pfield->choice_set))) {
			print_error_msg_p(xname, "NULL pointer for choiceGbl",
				PAUSE_F);
			return(-1);
	 	} 
		if (dct_mode == DCT_INTERACTIVE) {
			pop_choice_window(xname, (short *)pvalue, pchoiceSet);
			return(0);
		}else {
		   i = 0;
		   while ( i < pchoiceSet->number) {
		    if( (pchoice = GET_CHOICE(pchoiceSet, i))) {
			if (strcmp(pinput,pchoice) == SAME){
				*(short *)pvalue = i;
				return(0);
			}
		    }
		    i++;
		   }
		   print_error_msg_p(xname, "String not found", PAUSE_F);
			return(-1);
		}
	case (DBF_CVTCHOICE):
		if(!choiceCvt) {
			print_error_msg_p(xname, "NULL pointer for choiceCvt",
				PAUSE_F);
			return(-1);
		}
		if (dct_mode == DCT_INTERACTIVE) {
			pop_choice_window(xname, (short *)pvalue, choiceCvt);
			return(0);
		}else {
		    i = 0;
		    while ( i < choiceCvt->number) {
		    	if( (pchoice = GET_CHOICE(choiceCvt, i))) {
				if (strcmp(pinput,pchoice) == SAME){
                 	               *(short *)pvalue = i;
					return(0);
				}
		    	}
		    	i++;
		    }
		    print_error_msg_p(xname, "String not found", PAUSE_F);
		    return(-1);
		}
	case (DBF_RECCHOICE):
		if( !(parrChoiceSet = GET_PARR_CHOICE_SET(choiceRec, 
				type))) {
			print_error_msg_p(xname,
				 "NULL pointer for choiceRec", PAUSE_F);
			return(-1);
		} 
		if( !(pchoiceSet = GET_PCHOICE_SET(parrChoiceSet, 
				pfield->choice_set))) {
			print_error_msg_p(xname,
				 "NULL pointer for choiceSet in choiceRec",
				PAUSE_F);
			return(-1);
		} 
		if (dct_mode == DCT_INTERACTIVE) {
			pop_choice_window(xname, (short *)pvalue, pchoiceSet);
			return(0);
		}else {
		i = 0;
		  while ( i < pchoiceSet->number) {
		    if( (pchoice = GET_CHOICE(pchoiceSet, i))) {
			if (strcmp(pinput,pchoice) == SAME){
                                *(short *)pvalue = i;
				return(0);
			}
		    }
		    i++;
		  }
		  print_error_msg_p(xname, "String not found", PAUSE_F);
		  return(-1);
		}
	case (DBF_DEVCHOICE):
		if( !(pdevChoiceSet = GET_PDEV_CHOICE_SET(choiceDev, type))) {
			print_error_msg_p(xname, "NULL pointer for choiceDev",
				PAUSE_F);
			return(-1);
		} 
		if (dct_mode == DCT_INTERACTIVE) {
			pop_devchoice_window(xname,
				 (short *)pvalue,pdevChoiceSet);
			return(0);
		}else {
		  i = 0;
		  while ( i < pdevChoiceSet->number) {
		    if( (pchoice = (GET_DEV_CHOICE(pdevChoiceSet,
						 i)->pchoice))) {
			if (strcmp(pinput,pchoice) == SAME){
                                *(short *)pvalue = i;
				return(0);
			}
		     }
		     i++;
		  }
		  print_error_msg_p(xname, "String not found", PAUSE_F);
		  return(-1);
		}
	}
	return(0);
}


/****************************************************************************
 *
 * process_spc_hdwlink
 *
 *	used by convert_fld to process DBF_INLINK and DBF_OUTLINK with SPC_HDWLINK
 ****************************************************************************/
int	process_spc_hdwlink(pname,type, pinput, pdblink)
char			*pname;		/* name of caller */
short                   type;           /* record type */
char			*pinput;	/* character string input by user */
struct link    *pdblink;		/* place to store link fields */
{
	char	*xname="process_spc_hdwlink";
	struct devChoiceSet	*pdevChoiceSet;
	long		link_choice;
	char		ptrc[MAX_STRING_LENGTH];
	char		*pptr,*pinputp;
	int		cnt;
	static	char	field1[PVNAME_SZ+1];
	static 	char	field2[PVNAME_SZ+1];
	static 	char	field3[PVNAME_SZ+1];
	static	char	*pfields[] = {		/* used to receive parsed */
			field1,			/*  pvname */
			field2,
			field3
			};
	char		*pdelim = " ";		/* field delimiter characters */

	if (dct_mode == DCT_INTERACTIVE) {
		if(pop_linktype_choice_window(xname,
				 &link_choice, pdblink))
			return(-1);
	} else {
		/*  the following is an attempt to determine the link type from the address string */
		if ( *pinput == '#' ) {
			strncpy(ptrc, pinput+1, MAX_STRING_LENGTH);
			sparse(ptrc, pdelim, pfields, 3);
			switch ( *field1) {
				case ('C'):
					link_choice = VME_IO;
					break;
				case ('B'):
					link_choice = CAMAC_IO;
					break;
				case ('L'):
					switch(*field2) {
						case('A'):
							switch(*field3) {
								case('@'):
									link_choice=GPIB_IO;
									break;	
								case('C'):
									link_choice=AB_IO;
									break;	
								default:
							print_error_msg_p(xname,"Cannot decode #Lnn Ann ...",
								PAUSE_F);
									return(-1);
							}
							break;
						case('N'):
							link_choice = BITBUS_IO;
							break;
						case('P'):
							link_choice = AB_IO;
							break;

						case('B'):
							link_choice = BBGPIB_IO;
							break;
						default:
						print_error_msg_p(xname, "Cannot decode #L...", PAUSE_F);
							return(-1);
					}
					break;
				case ('V') :
					link_choice = VXI_IO;
					break;
				case ('@') :
					link_choice = INST_IO;
					break;
				default:
					print_error_msg_p(xname, "Unknown link type", PAUSE_F);
					return(-1);
			}
		} else {
			if(*pinput=='@') 
				link_choice = INST_IO;
			else if (! legal_name_ref(pinput))
				link_choice = CONSTANT;
			else {
				/* the following kludge is an attempt to distiguish between a pvlink and 
					an instk_io link by the # of fields (.'s)  */
				pinputp=pinput;
				cnt=0;
				while ( pptr=strchr(pinputp, '.')) {
					cnt++;
					pinputp = ++pptr;
				}
				if (cnt > 2) 
					link_choice = CONSTANT;
				else
					link_choice = INST_IO;
			}
		}
	}
	return(process_link_field(xname, type, pinput, pdblink, link_choice, SPC_HDWLINK));
}


/****************************************************************************
 *
 * process_link_field
 *
 *	used by convert_fld to process DBF_INLINK and DBF_OUTLINK
 ****************************************************************************/
int	process_link_field(pname,type, pinput, pdblink, link_choice, mode)
char			*pname;		/* name of caller */
short                   type;           /* record type */
char			*pinput;	/* character string input by user */
struct link    *pdblink;		/* place to store link fields */
long			link_choice;	/* type of link */
int			mode;		/* mode */
{
	double		double_temp;
        short           slot,channel;
        short           frame,la;
        short           branch,crate,station,address,function;
        short           iolink,adapter,card,signal,plc_flag;
        short           addr, flag;
	unsigned char	ulink,unode,uport,usignal;
	unsigned char	ubbaddr, ugpibaddr;
	char		param_string[INSTIO_FLD_SZ];
        short           fld_type,fld_offset,fld_size; /* used to get DTYP fld */
        short           card_type;
	short		dummy1, dummy2;
        short           shortval;
        static  char    field1[PVNAME_SZ+1];
        static  char    field2[PVNAME_SZ+1];
        static  char    field3[PVNAME_SZ+1];
        static  char    field4[PVNAME_SZ+1];
        static  char    *pfields[] = {          /* used to receive parsed */
                        field1,                 /*  pvname */
                        field2,
                        field3,                 /*  pvname */
                        field4
                        };
        char            *pdelim = " .";         /* field delimiter characters
                                                for pv names */
        short           numfields;              /* number of fields parsed from
                                                a pvlink field */
        long            link_type;    /* loaded with val from choiceDev struc*/



		  if (mode != SPC_HDWLINK) {
	      		if( get_recfld(type, "DTYP", &fld_type, &fld_offset,
					 &fld_size)) {
				print_error_msg_p(pname,
				       "Required field DTYP missing", PAUSE_F);
				return(-1);
			}
			
			card_type = *(short *)(&ctrl_record[fld_offset]);
	
			/* if no dev choice set for this type assume CONSTANT*/
			if(choiceDev->papDevChoiceSet[type]) 
			   link_type = GET_DEV_CHOICE_LINK_VAL(type, card_type);
			else
			   link_type = CONSTANT;

		} else
			link_type = link_choice;
		if (dct_mode == DCT_INTERACTIVE &&
					 strlen(pinput) == 0) {
                       	pop_link_window(pname, pdblink, link_type);
		}else {

			switch (link_type){
			/* vme hardware channel specification */
			case (VME_IO):

				/* convert value */
				if (get_vme_addr(pinput,&card,&signal,
				   param_string, card_type) < 0) return(-1);

				/* set the new structure */
				memset((char *)pdblink,'\0',sizeof(struct link));
				pdblink->type = VME_IO;
				pdblink->value.vmeio.card = card;
				pdblink->value.vmeio.signal = signal;
				memcpy( pdblink->value.vmeio.parm,param_string, 
					VME_PARAM_SZ);
				break;

			/* Allen Bradley harware channel specification */
			case (AB_IO):

				/* convert value */
				if (get_ab_addr(pinput,&iolink,&adapter,&card,
				  &signal,&plc_flag,param_string,
				  card_type) < 0) return(-1);

				/* set the new structure */
				memset((char *)pdblink,'\0',sizeof(struct link));
				pdblink->type = AB_IO;
				pdblink->value.abio.link = iolink;
				pdblink->value.abio.adapter = adapter;
				pdblink->value.abio.card = card;
				pdblink->value.abio.signal = signal;
				pdblink->value.abio.plc_flag = plc_flag;
				memcpy(pdblink->value.abio.parm,param_string, 
					AB_PARAM_SZ);
				break;

			/* camac harware channel specification */
			case (CAMAC_IO):

				/* convert value */
				if (get_camac_addr(pinput,&branch,&crate,
					&station,&address,&function,
					param_string) < 0) return(-1);

				/* set the new structure */
				memset((char *)pdblink,'\0',sizeof(struct link));
				pdblink->type = CAMAC_IO;
				pdblink->value.camacio.b= branch;
				pdblink->value.camacio.c= crate;
				pdblink->value.camacio.n= station;
				pdblink->value.camacio.a= address;
				pdblink->value.camacio.f= function;
				memcpy(pdblink->value.camacio.parm,param_string,
					CAMAC_PARAM_SZ);
				break;
			case (GPIB_IO):
				if (get_gpib_addr( pinput, &iolink, 
					&addr, param_string, card_type) <0)
						return(-1);
                                memset((char *)pdblink,'\0',sizeof(struct link));
				pdblink->type = GPIB_IO;
				pdblink->value.gpibio.link = iolink;
                                pdblink->value.gpibio.addr = addr;
				memcpy(pdblink->value.gpibio.parm,param_string,
					LINK_PARAM_SZ);
				break;
			case (BITBUS_IO):
                                if (get_bitbus_addr( pinput, &ulink,
                                        &unode, &uport, &usignal,  param_string,
						 card_type) <0)
                                                return(-1);
                                memset((char *)pdblink,'\0',sizeof(struct link));
                                pdblink->type = BITBUS_IO;
                                pdblink->value.bitbusio.link = ulink;
                                pdblink->value.bitbusio.node = unode;
                                pdblink->value.bitbusio.port = uport;
                                pdblink->value.bitbusio.signal = usignal;
				memcpy(pdblink->value.bitbusio.parm,
					param_string,LINK_PARAM_SZ);
				break;
			case (BBGPIB_IO):	/* bitbus to gpib io */
                                if (get_bbgpib_addr( pinput, &ulink,
                                        &ubbaddr, &ugpibaddr, param_string,
						 card_type) <0)
                                                return(-1);
                                memset((char *)pdblink,'\0',sizeof(struct link));
                                pdblink->type = BBGPIB_IO;
                                pdblink->value.bbgpibio.link = ulink;
                                pdblink->value.bbgpibio.bbaddr = ubbaddr;
                                pdblink->value.bbgpibio.gpibaddr = ugpibaddr;
				memcpy(pdblink->value.bbgpibio.parm,
					param_string, LINK_PARAM_SZ);
				break;
			case (INST_IO) :
				memset(pdblink->value.instio.string,'\0',
						 INSTIO_FLD_SZ);
				pdblink->type = INST_IO;
				if(*pinput=='@') pinput++;
				strncpy(pdblink->value.instio.string,pinput,INSTIO_FLD_SZ-1);
				break;
			case (VXI_IO):
				if (get_vxi_addr(pinput, &frame, &slot, &la,
				    &signal,&flag, param_string, card_type) <0)
						return(-1);
                                memset((char *)pdblink,'\0',sizeof(struct link));
				pdblink->type = VXI_IO;
				pdblink->value.vxiio.frame = frame;
                                pdblink->value.vxiio.slot = slot;
                                pdblink->value.vxiio.signal = signal;
                                pdblink->value.vxiio.la = la;
                                pdblink->value.vxiio.flag = flag;
				memcpy(pdblink->value.vxiio.parm,param_string,
					VXI_PARAM_SZ);
				break;
			case (CONSTANT):	/* soft channel */
				if (legal_name_ref(pinput)){  /* process var name */
					numfields= sparse(pinput, pdelim,
						 pfields, 4);
                                        pdblink->type = PV_LINK;
                                        strcpy(pdblink->value.pv_link.pvname,
						field1);
					fixup_fldname(field2, 
					     pdblink->value.pv_link.fldname);
					/* if process_passive and maximze_sevr
						exist, process them */
					if (numfields > 2) {
					    if ((shortval=get_dct_choice_val(
							0, field3)) >=0)
						pdblink->
						  value.pv_link.process_passive=
							shortval;
					     else
						return(-1); 
					    if ((shortval=get_dct_choice_val(
							1, field4)) >=0)
						pdblink->
						  value.pv_link.maximize_sevr=
							shortval;
					     else
						return(-1); 
					}else {    /* set default values */
					 	pdblink->
                                                  value.pv_link.process_passive=
                                                        1;
		                                 pdblink->
                                                  value.pv_link.maximize_sevr=0 ;
					}
	
					if( dct_mode == DCT_INTERACTIVE
						 && pvdGetRec(
			     			pdblink->value.pv_link.pvname,
							 &dummy1, &dummy2)) {
						print_error_msg_p(pname, 
		"Warning, this process variable not defined in this database",
						PAUSE_F);
						}
				} else {		/* constant */
					if ( sscanf( pinput, "%lf",
							 &double_temp) < 1) {
						print_error_msg_p( pname,
							"Illegal value",
							PAUSE_F);
						return(-1);
					}

					/* set the new structure */
					memset((char *)pdblink,'\0',
						sizeof(struct link));
					pdblink->type = CONSTANT;
					pdblink->value.value =
							 double_temp;
				}
				break;
			default:
				return(-1);
			}
		     }
	return(0);
}

/****************************************************************************
 *
 * process_soft_channel
 *
 *	used by convert_fld to process DBF_INLINK and DBF_OUTLINK
 ****************************************************************************/
int	process_soft_channel(pname, pinput, pdblink)
char			*pname;		/* name of caller */
char			*pinput;	/* character string input by user */
struct link    *pdblink;		/* place to store link fields */
{
	double		double_temp;
        short           shortval;
	short		dummy1, dummy2;
        static  char    field1[PVNAME_SZ+1];
        static  char    field2[PVNAME_SZ+1];
        static  char    field3[PVNAME_SZ+1];
        static  char    field4[PVNAME_SZ+1];
        static  char    *pfields[] = {          /* used to receive parsed */
                        field1,                 /*  pvname */
                        field2,
                        field3,                 /*  pvname */
                        field4
                        };
        char            *pdelim = " .";         /* field delimiter characters
                                                for pv names */
        short           numfields;              /* number of fields parsed from
                                                a pvlink field */
		if(dct_mode == DCT_INTERACTIVE && 
					strlen(pinput) == 0) {
                        	pop_link_window(pname, pdblink,(long) CONSTANT);
		}else {
				if (legal_name_ref(pinput)){    /* process var name */
                        		numfields=sparse(pinput, pdelim,
							 pfields, 4);
                        		pdblink->type = PV_LINK;
                        		strcpy(pdblink->value.pv_link.pvname,
						 field1);
					fixup_fldname(field2,
					    pdblink->value.pv_link.fldname);
                                        /* if process_passive and maximze_sevr
                                                exist, process them */
                                        if (numfields > 2) {
                                            if ((shortval=get_dct_choice_val(
                                                        0, field3)) >=0)
                                                pdblink->
                                                  value.pv_link.process_passive=
                                                        shortval;
                                             else
                                                return(-1);
                                            if ((shortval=get_dct_choice_val(
                                                        1, field4)) >=0)
                                                pdblink->
                                                  value.pv_link.maximize_sevr=
                                                        shortval;
                                             else
                                                return(-1);
                                        }else {    /* set default values */
                                                pdblink->
                                                  value.pv_link.process_passive=
                                                        1;
                                                 pdblink->
                                                  value.pv_link.maximize_sevr=0 ;
					}
					if( dct_mode == DCT_INTERACTIVE
						 && pvdGetRec(
			     			pdblink->value.pv_link.pvname,
							 &dummy1, &dummy2)) {
						print_error_msg_p(pname, 
		"Warning, this process variable not defined in this database",
						PAUSE_F);
						}

				}else{				/* constant */
					if ( sscanf( pinput, "%lf", &double_temp) < 1) {
						print_error_msg_p(pname, 
						    "Illegal value", PAUSE_F);
						return(-1);
					}

					/* set the new structure */
					memset((char *)pdblink,'\0',
						sizeof(struct link));
					pdblink->type = CONSTANT;
					pdblink->value.value =
							 double_temp;
				}
			}
	return(0);
}

/*
 * FLD_SEARCH
 *
 *	search fields for string and return the fields number
 */
fld_search(pstring)
char *pstring;
{
	char		add_name[FLDNAME_SZ+1];
	short	index;
	struct fldDes  	*pfields;

	/* make input string like the stored field name */
	add_sp(pstring,add_name);
	for (index=0; index < precorddes->no_fields; index++){

		pfields = pflddef[index];
		/* return index if input is field name in array */
		if (!isalpha(pfields->fldname[0])) return(-1);
		if (bufcmp(pfields->fldname,add_name,4) == SAME)
			return(index); 
	}
	return(-1);
}

/*
 * FLD_TO_ASCII
 *
 * convert a database field to ascii
 */
int	fld_to_ascii(pfield,pbuffer,type)
struct fldDes	*pfield;	/* pointer to field struct for
							current field */
char		*pbuffer;		/* place to put ascii string */
short		type;	/* record type */
{
	char	*xname = "fld_to_ascii";
	short	i;
	char		*ptr;
	struct link	*plink;
        char            *pchoice;
	short		*pshort;
        struct choiceSet        *pchoiceSet;
        struct arrChoiceSet     *parrChoiceSet;
	struct devChoiceSet	*pdevChoiceSet;

	ptr = &ctrl_record[0];
	*pbuffer = '\0';

	switch (pfield->field_type){
	case (DBF_STRING):
		i = 0;
		while (i < pfield->size) {
			*pbuffer++ = *(ptr + pfield->offset + i);
			i++;
		}
		*pbuffer = '\0';
		break;
	case (DBF_CHAR):
	case (DBF_UCHAR):
	case (DBF_SHORT):
	case (DBF_USHORT):
	case (DBF_LONG):
	case (DBF_ULONG):
		if (convert_type_to_ascii( xname, pfield->field_type,
			pfield->cvt_type, "", pbuffer, "", 
			(union fld_types *)(ptr+pfield->offset)) != 0)
			return(-1);
		break;
	case (DBF_DOUBLE):
		if (rep_fd == STDOUT) {
		   if( (*(double *)(ptr+pfield->offset) >= MIN_F_FORMAT_PVAL && 
			*(double *)(ptr+pfield->offset) <= MAX_F_FORMAT_PVAL) ||
		       (*(double *)(ptr+pfield->offset) >= MIN_F_FORMAT_NVAL &&
			 *(double *)(ptr+pfield->offset) <= MAX_F_FORMAT_NVAL))
			sprintf(pbuffer,"%.*f",precision > DOUBLE_E_FORMAT_PREC
					? DOUBLE_E_FORMAT_PREC : precision,
				*(double *)(ptr+pfield->offset));
		   else
			sprintf(pbuffer,"%.*e",precision > DOUBLE_E_FORMAT_PREC
                                        ? DOUBLE_E_FORMAT_PREC : precision,
				*(double *)(ptr+pfield->offset));

		} else {
		   if( (*(double *)(ptr+pfield->offset) >= MIN_F_FORMAT_PVAL && 
			*(double *)(ptr+pfield->offset) <= MAX_F_FORMAT_PVAL) ||
		       (*(double *)(ptr+pfield->offset) >= MIN_F_FORMAT_NVAL &&
			 *(double *)(ptr+pfield->offset) <= MAX_F_FORMAT_NVAL))
			sprintf(pbuffer,"%.*f",DOUBLE_E_FORMAT_PREC,
				*(double *)(ptr+pfield->offset));
		   else
			sprintf(pbuffer,"%.*e",DOUBLE_E_FORMAT_PREC,
				*(double *)(ptr+pfield->offset));
		}
		break;
	case (DBF_FLOAT):
                if (rep_fd == STDOUT) {
		   if( (*(float *)(ptr+pfield->offset) >= MIN_F_FORMAT_PVAL && 
			*(float *)(ptr+pfield->offset) <= MAX_F_FORMAT_PVAL) ||
			(*(float *)(ptr+pfield->offset) >= MIN_F_FORMAT_NVAL &&
			 *(float *)(ptr+pfield->offset) <= MAX_F_FORMAT_NVAL))
			sprintf(pbuffer,"%.*f",precision > FLOAT_E_FORMAT_PREC
					? FLOAT_E_FORMAT_PREC : precision,
				*(float *)(ptr+pfield->offset));
		   else
			sprintf(pbuffer,"%.*e",precision > FLOAT_E_FORMAT_PREC
                                        ? FLOAT_E_FORMAT_PREC : precision,
				*(float *)(ptr+pfield->offset));
		} else {
		   if( (*(float *)(ptr+pfield->offset) >= MIN_F_FORMAT_PVAL && 
			*(float *)(ptr+pfield->offset) <= MAX_F_FORMAT_PVAL) ||
			(*(float *)(ptr+pfield->offset) >= MIN_F_FORMAT_NVAL &&
			 *(float *)(ptr+pfield->offset) <= MAX_F_FORMAT_NVAL))
			sprintf(pbuffer,"%.*f",FLOAT_E_FORMAT_PREC,
				*(float *)(ptr+pfield->offset));
		   else
			sprintf(pbuffer,"%.*e",FLOAT_E_FORMAT_PREC,
				*(float *)(ptr+pfield->offset));
		}
		break;
	case (DBF_ENUM):
		sprintf(pbuffer, "%u", *(unsigned short *)(ptr+pfield->offset));		break;	
	case (DBF_INLINK):
	case (DBF_OUTLINK):
	case (DBF_FWDLINK):
		plink = (struct link *)(ptr+pfield->offset);
		switch (plink->type){
		case (VME_IO):
			sprintf(pbuffer,"#C%d S%d @%s",
				  plink->value.vmeio.card,
				  plink->value.vmeio.signal,
				  plink->value.vmeio.parm);
			break;
		case (CAMAC_IO):
			sprintf(pbuffer,"#B%d C%d N%d A%d F%d @%s",
			  plink->value.camacio.b,
			  plink->value.camacio.c,
			  plink->value.camacio.n,
			  plink->value.camacio.a,
			  plink->value.camacio.f,
			  plink->value.camacio.parm);
			break;
		case (AB_IO):
			if (plink->value.abio.plc_flag == 0)
				sprintf(pbuffer,"#L%d A%d C%d S%d @%s",
				  plink->value.abio.link,
				  plink->value.abio.adapter,
				  plink->value.abio.card,
				  plink->value.abio.signal,
				  plink->value.abio.parm);
			else
				sprintf(pbuffer,"#L%d P%d C%d S%d F%d @%s",
				  plink->value.abio.link,
				  plink->value.abio.adapter,
				  plink->value.abio.card,
				  plink->value.abio.signal,
				  plink->value.abio.plc_flag,
				  plink->value.abio.parm);
			break;
		case (GPIB_IO):
			sprintf(pbuffer,"#L%d A%d @%s",
			plink->value.gpibio.link,
			plink->value.gpibio.addr,
			plink->value.gpibio.parm);
			break;
		case (BITBUS_IO):
			sprintf(pbuffer,"#L%u N%u P%u S%u @%s",
			plink->value.bitbusio.link,
			plink->value.bitbusio.node,
			plink->value.bitbusio.port,
			plink->value.bitbusio.signal,
			plink->value.bitbusio.parm);
			break;
		case (BBGPIB_IO):
			sprintf(pbuffer,"#L%u B%u G%u @%s",
			plink->value.bbgpibio.link,
			plink->value.bbgpibio.bbaddr,
			plink->value.bbgpibio.gpibaddr,
			plink->value.bbgpibio.parm);
			break;
		case (INST_IO):
			sprintf(pbuffer,"@%s", plink->value.instio.string);
			break;
		case (VXI_IO):
			if (plink->value.vxiio.flag == VXIDYNAMIC)
			   sprintf(pbuffer,"#V%d C%d S%d @%s",
			   plink->value.vxiio.frame,
			   plink->value.vxiio.slot,
			   plink->value.vxiio.signal,
			   plink->value.vxiio.parm);
			else
			   sprintf(pbuffer,"#V%d S%d @%s",
			   plink->value.vxiio.la,
			   plink->value.vxiio.signal,
			   plink->value.vxiio.parm);
			break;
		case (CONSTANT):
			if (rep_fd == STDOUT) {
			   if( (plink->value.value >=
						 MIN_F_FORMAT_PVAL && 
				plink->value.value <= 
						MAX_F_FORMAT_PVAL) ||
				(plink->value.value >=
						 MIN_F_FORMAT_NVAL &&
				plink->value.value <=
						 MAX_F_FORMAT_NVAL))
				sprintf(pbuffer,"%.*f",
					precision > CONSTANT_FORMAT_PREC
					? CONSTANT_FORMAT_PREC : precision,
					plink->value.value);
			   else
				sprintf(pbuffer,"%.*e",
					precision > CONSTANT_FORMAT_PREC
                                        ? CONSTANT_FORMAT_PREC : precision,
					plink->value.value);
			} else {
			   if( (plink->value.value >=
						 MIN_F_FORMAT_PVAL && 
				plink->value.value <=
						 MAX_F_FORMAT_PVAL) ||
				(plink->value.value >=
						 MIN_F_FORMAT_NVAL &&
				plink->value.value <=
						 MAX_F_FORMAT_NVAL))
				sprintf(pbuffer,"%.*f",
					CONSTANT_FORMAT_PREC,
					plink->value.value);
			   else
				sprintf(pbuffer,"%.*e",
					CONSTANT_FORMAT_PREC,
					plink->value.value);
			}
			break;
		case (PV_LINK):
                        if(pfield->field_type == DBF_FWDLINK)
                          sprintf(pbuffer, "%.*s.%.*s", 
						PVNAME_SZ,
						plink->value.pv_link.pvname,
						FLDNAME_SZ,
                                                plink->value.pv_link.fldname);
                        else
                          sprintf(pbuffer, "%.*s.%.*s.%s.%s",
				PVNAME_SZ,
                                 plink->value.pv_link.pvname,
				FLDNAME_SZ,
                                 plink->value.pv_link.fldname,
                                 get_dct_choice_string(0,
                                        plink->value.pv_link.process_passive),
                                 get_dct_choice_string(1,
                                        plink->value.pv_link.maximize_sevr));
			break;
		}
		break;
	case (DBF_GBLCHOICE):
		pshort = (short *)(ptr+pfield->offset);
                if( !(pchoiceSet = GET_PCHOICE_SET(choiceGbl,
                                pfield->choice_set))) {
			print_error_msg_p(xname, "NULL pointer for choiceGbl",
				PAUSE_F);
			return(-1);
		}
                if( (pchoice = GET_CHOICE(pchoiceSet, *pshort))) {
			strcpy(pbuffer,pchoice);
		} else {
			strcpy(pbuffer,"");
		}
		break;
        case (DBF_CVTCHOICE):
		if(!choiceCvt) {
			print_error_msg_p(xname, "NULL pointer for choiceCvt",
				PAUSE_F);
			return(-1);
		}
                pshort = (short *)(ptr+pfield->offset);
                if( (pchoice = GET_CHOICE(choiceCvt, *pshort))) {
                        strcpy(pbuffer,pchoice);
                } else {
                        strcpy(pbuffer,"");
                }
		break;
        case (DBF_RECCHOICE):
                pshort = (short *)(ptr+pfield->offset);
                if( !(parrChoiceSet = GET_PARR_CHOICE_SET(choiceRec,
                                type))) {
			print_error_msg_p(xname,
				 "NULL pointer for choiceRec",
					 PAUSE_F);
			return(-1);
                }
                if( !(pchoiceSet = GET_PCHOICE_SET(parrChoiceSet,
                                pfield->choice_set))) {
			print_error_msg_p(xname,
				 "NULL pointer for choiceSet in choiceRec",
				PAUSE_F);
			return(-1);
                }
                if( (pchoice = GET_CHOICE(pchoiceSet, *pshort))) {
                        strcpy(pbuffer,pchoice);
                } else {
                        strcpy(pbuffer,"");
                }
		break;
        case (DBF_DEVCHOICE):
                pshort = (short *)(ptr+pfield->offset);
                if( !(pdevChoiceSet = GET_PDEV_CHOICE_SET(choiceDev,
                                type))) {
			print_error_msg_p(xname, "NULL pointer for choiceDev",
				PAUSE_F);
			return(-1);
                }
                if( pchoice = (GET_DEV_CHOICE(pdevChoiceSet,
				 *pshort)->pchoice)) {
                        strcpy(pbuffer,pchoice);
                        break;
                } else {
                        strcpy(pbuffer,"");
                        break;
                }
		break;


	}
	return(0);
}

/*
 * GET_AB_ADDR
 *
 * extract a Allen-Bradley address from an input
 * format <L## A## C## S## F##>
 */
get_ab_addr(pinput,plink,padapter,pcard,psignal,pplc_flag,pparm,card_type)
char		*pinput;
short		*plink;
short		*padapter;
short		*pcard;
short		*psignal;
short		*pplc_flag;
char		*pparm;	/* parameter string*/
short		card_type;
{
	short i;


	/* init card and signal */
	*plink = 0;
	*padapter = 0;
	*pcard = 0;
	*psignal = 0;
	*pplc_flag = 0;
	memset(pparm,'\0',AB_PARAM_SZ);
	/* check for auto address selection */
/***********************************
	if ((i = cinstr(pinput,'*')) >= 0){
		status = next_vme_addr(io_type,card_type,pcard,psignal,no_channels);
		return(status);
	}
***************************************/

	/* get link */
	if ((i = cinstr(pinput,'L')) >= 0)
		*plink = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'l')) >= 0)
		*plink = to_short(pinput+i+1);

	/* get adapter */
	if ((i = cinstr(pinput,'A')) >= 0)
		*padapter = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'a')) >= 0)
		*padapter = to_short(pinput+i+1);

	/* get PLC */
	if ((i = cinstr(pinput,'P')) >= 0){
		*padapter = to_short(pinput+i+1);
		*pplc_flag = TRUE;
	}else if ((i = cinstr(pinput,'p')) >= 0){
		*padapter = to_short(pinput+i+1);
		*pplc_flag = TRUE;
	}

	/* get card */
	if ((i = cinstr(pinput,'C')) >= 0)
		*pcard = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'c')) >= 0)
		*pcard = to_short(pinput+i+1);

	/* get signal */
	if ((i = cinstr(pinput,'S')) >= 0)
		*psignal = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'s')) >= 0)
		*psignal = to_short(pinput+i+1);

	/* get plc_flag */
	if ((i = cinstr(pinput,'F')) >= 0)
		*pplc_flag = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'f')) >= 0)
		*pplc_flag = to_short(pinput+i+1);
	else
		*pplc_flag = 0;

        /* get parameter string*/
        if ((i = cinstr(pinput,'@')) >= 0)
                memcpy(pparm,pinput+i+1,
                         (strlen(pinput+i+1) >= AB_PARAM_SZ) ?
                                 AB_PARAM_SZ-1 : strlen(pinput+i+1));


	/* verify a good address was entered */
/*
	status = verify_hw_addr(io_type,card_type,pcard,psignal);	
	return(status);
*/
	return(0);
}


/*
 * GET_BITBUS_ADDR
 *
 * extract a bitbus address from an input
 * format <L##  N## P##  S## @parameter_string>
 */
get_bitbus_addr(pinput,plink,pnode,pport,psignal,pparm,card_type)
char	*pinput;	/* input string to be converted */
unsigned char   *plink;	/* pointer to bitbus link number */
unsigned char	*pnode;	/* pointer to bitbus node address */
unsigned char	*pport; /* pointer to port */
unsigned char	*psignal;	/* pointer to signal */
char		*pparm;	/* parameter string*/
short		card_type;	/* card type index */
{
	short i;
	short	temp;

	/* init stuff */
	*plink = 0;
	*pnode = 0;
	*pport = 0;
	*psignal=0;
	memset(pparm,'\0', LINK_PARAM_SZ);

	/* get link */
	if((i = cinstr(pinput,'L')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*plink=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'l')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*plink=(unsigned char)temp;
	}

	/* get bitbus  node */
	if((i = cinstr(pinput,'N')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pnode=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'n')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pnode=(unsigned char)temp;
	}
	/* get bitbus  port */
	if((i = cinstr(pinput,'P')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pport=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'p')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pport=(unsigned char)temp;
	}

	/* get signal number*/
	if((i = cinstr(pinput,'S')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*psignal=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'s')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*psignal=(unsigned char)temp;
	}
	
        /* get parameter string*/
        if ((i = cinstr(pinput,'@')) >= 0)
                memcpy(pparm,pinput+i+1,
                         (strlen(pinput+i+1) >= LINK_PARAM_SZ) ?
                                 LINK_PARAM_SZ-1 : strlen(pinput+i+1));


	return(0);
}


/*
 * GET_BBGPIB_ADDR
 *
 * extract a bitbus address from an input
 * format <L##  B## G## @parameter_string>
 */
get_bbgpib_addr(pinput,plink,pbbaddr,pgpibaddr,pparm,card_type)
char	*pinput;	/* input string to be converted */
unsigned char   *plink;	/* pointer to bitbus link number */
unsigned char	*pbbaddr;	/* pointer to bitbus node address */
unsigned char	*pgpibaddr; /* pointer to gpib address */
char		*pparm;	/* parameter string*/
short		card_type;	/* card type index */
{
	short i;
	short	temp;

	/* init stuff */
	*plink = 0;
	*pbbaddr = 0;
	*pgpibaddr = 0;
	memset(pparm,'\0', LINK_PARAM_SZ);

	/* get link */
	if((i = cinstr(pinput,'L')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*plink=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'l')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*plink=(unsigned char)temp;
	}

	/* get bitbus  node */
	if((i = cinstr(pinput,'B')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pbbaddr=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'b')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pbbaddr=(unsigned char)temp;
	}
	/* get gpib address */
	if((i = cinstr(pinput,'G')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pgpibaddr=(unsigned char)temp;
	}
	else if ((i = cinstr(pinput,'g')) >= 0) {
		sscanf(pinput+i+1, "%hu", &temp);
		*pgpibaddr=(unsigned char)temp;
	}

	
        /* get parameter string*/
        if ((i = cinstr(pinput,'@')) >= 0)
                memcpy(pparm,pinput+i+1,
                         (strlen(pinput+i+1) >= LINK_PARAM_SZ) ?
                                 LINK_PARAM_SZ-1 : strlen(pinput+i+1));


	return(0);
}

/*
 * GET_CAMAC_ADDR
 *
 * extract a CAMAC address from an input
 * format <B## C## S## A##>
 */
get_camac_addr(pinput,pbranch,pcrate,pstation,paddress,pfunction,pparm)
char	*pinput;
short	*pbranch;
short	*pcrate;
short	*pstation;
short	*paddress;
short	*pfunction;
char	*pparm; /* parameter string*/
{
	char	*xname = "get_camac_addr";
	short i;

	memset(pparm,'\0',CAMAC_PARAM_SZ);
	/* get branch */
	if ((i = cinstr(pinput,'B')) < 0){
		print_error_msg_p(xname, "Bad branch specification", PAUSE_F);
		return(-1);
	}
	*pbranch = to_short(pinput+i+1);

	/* get crate */
	if ((i = cinstr(pinput,'C')) < 0){
		print_error_msg_p(xname, "Bad crate_specification", PAUSE_F);
		return(-1);
	}
	*pcrate = to_short(pinput+i+1);
        /* get station number*/
        if ((i = cinstr(pinput,'N')) < 0){
		print_error_msg_p(xname, "bad station # specification", 
				PAUSE_F);
                return(-1);
        }
	*pstation = to_short(pinput+i+1);
	if ((*pstation < 1) || (*pstation > 24)){
		print_error_msg_p(xname, "bad station # specification",
				 PAUSE_F);
		return(-1);
	}
        /* get subaddress */
        if ((i = cinstr(pinput,'A')) < 0){
		/*  no sub address set default value to 0 */
		*paddress = 0;
        } else  {
		*paddress = to_short(pinput+i+1);
	}
	if ((*paddress < 0) || (*paddress > 15)){
		print_error_msg_p(xname, "bad subaddress specification",
					 PAUSE_F);
		return(-1);
	}
        /* get function */
        if ((i = cinstr(pinput,'F')) < 0){
		/*  no function set default value to 0 */
		*pfunction = 0;
        } else {
		*pfunction = to_short(pinput+i+1);
	}
	if ((*pfunction < 0) || (*pfunction > 31)){
		print_error_msg_p(xname, "bad function specification",
					 PAUSE_F);
		return(-1);
	}
	if ((i = cinstr(pinput,'@')) >= 0)
	    memcpy(pparm,pinput+i+1,
		(strlen(pinput+i+1) >= CAMAC_PARAM_SZ) ?
		CAMAC_PARAM_SZ-1 : strlen(pinput+i+1));
	return(0);
}


/*
 * GET_GPIB_ADDR
 *
 * extract a GPIB address from an input
 * format <L##  A##  @parameter_string>
 */
get_gpib_addr(pinput,plink,paddr,psignal,card_type)
char	*pinput;	/* input string to be converted */
short	*plink;		/* pointer to gpib link number */
short	*paddr;		/* pointer to device address */
char	*psignal;	/* pointer to parameter string */
short	card_type;	/* card type index */
{
	short i;

	/* init stuff */
	*plink = 0;
	*paddr = 0;
	memset(psignal,'\0', LINK_PARAM_SZ);

	/* get link */
	if((i = cinstr(pinput,'L')) >= 0)
		*plink = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'l')) > 0)
		*plink = to_short(pinput+i+1);

	/* get device  address */
	if ((i = cinstr(pinput,'A')) >= 0)
		*paddr = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'a')) >= 0)
		*paddr = to_short(pinput+i+1);

	/* get parameter string*/
	if ((i = cinstr(pinput,'@')) >= 0)
		memcpy( psignal,pinput+i+1,
			 (strlen(pinput+i+1) >= LINK_PARAM_SZ) ?
				 LINK_PARAM_SZ-1 : strlen(pinput+i+1));
	
	return(0);
}

/*
 * GET_VXI_ADDR
 *
 * extract a VXI address from an input
 * format <V##  C##  @parameter_string>
 *		or
 * format <V##  @parameter_string>
 */
get_vxi_addr(pinput,pframe,pslot,pla,psignal, pflag, pparm,card_type)
char	*pinput;	/* input string to be converted */
short	*pframe;	/* pointer to frame number */
short	*pslot;		/* pointer to slot number */
short	*pla;		/* pointer to logical address */
short	*psignal;	/* pointer to signal number */
short	*pflag;		/* pointer to flag */
char	*pparm;	/* pointer to parameter string */
short	card_type;	/* card type index */
{
	short i;
	short	tla;
	short	tcard;

	/* init stuff */
	*pframe = 0;
	*pslot = 0;
	*psignal = 0;
	*pla = 0;
	*pflag = 0;
	memset(pparm,'\0', LINK_PARAM_SZ);

	/* get first field */
	if((i = cinstr(pinput,'V')) >= 0)
		 tla = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'v')) > 0)
		tla = to_short(pinput+i+1);

	/* check for card field */
	if ((i = cinstr(pinput,'C')) >= 0) {
		 *pslot = to_short(pinput+i+1);
		 *pframe = tla;
		 *pflag = VXIDYNAMIC;
	} else if ((i = cinstr(pinput,'c')) >= 0) {
		 *pslot = to_short(pinput+i+1);
		 *pframe = tla;
		 *pflag = VXIDYNAMIC;
	} else {				/* ok must be static address */
		*pla = tla;
		*pflag = VXISTATIC;
	}

	if ((i = cinstr(pinput,'S')) >= 0) 
		 *psignal = to_short(pinput+i+1);
	/* get parameter string*/
	if ((i = cinstr(pinput,'@')) >= 0)
		memcpy( pparm,pinput+i+1,
			 (strlen(pinput+i+1) >= VXI_PARAM_SZ) ?
				 VXI_PARAM_SZ-1 : strlen(pinput+i+1));
	
	return(0);
}


/*
 *	GET_RANGE_VAL
 *
 *	return the value from a given field
 */
get_range_val(pfield,prange)
struct fldDes
			*pfield;  	/* pointer to field description 
						structure */
union	fld_types		*prange;	  /* place to return value */
{
	char	*ptr;


	ptr = &ctrl_record[0];

	switch (pfield->field_type){
	case (DBF_DOUBLE):
		prange->double_value = *(double *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_FLOAT):
		prange->float_value  = *(float *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_CHAR):
		prange->char_value = *(char *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_UCHAR):
		prange->uchar_value = *(unsigned char *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_ENUM):
		prange->enum_value = *(unsigned short *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_SHORT):
		prange->short_value = *(short *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_USHORT):
		prange->ushort_value = *(unsigned short *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_ULONG):
		prange->ulong_value = *(unsigned long *)(ptr+pfield->offset);
		return(0);
		break;
	case (DBF_LONG):
		prange->long_value = *(long *)(ptr+pfield->offset);
		return(0);
		break;
	default:
		printf("error getting range\n");
		return(-1);
	}
}

/*
 * GET_VME_ADDR
 *
 * extract a VME address from an input
 * format C##  <S## or B##>  still accepts "B##" notation but puts out only "S##" notation
 */
get_vme_addr(pinput,pcard,psignal,pparm,card_type)
char	*pinput;
short	*pcard;
short	*psignal;
char    *pparm; /* parameter string*/
short	card_type;
{
	short i;

	/* init card and signal */
	*pcard = 0;
	*psignal = 0;
	memset(pparm,'\0',VME_PARAM_SZ);
	/* get card */
	if ((i = cinstr(pinput,'C')) >= 0)
		*pcard = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'c')) >= 0)
		*pcard = to_short(pinput+i+1);

	/* get signal */
	if ((i = cinstr(pinput,'S')) >= 0)
		*psignal = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'s')) >= 0)
		*psignal = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'B')) >= 0)
		*psignal = to_short(pinput+i+1);
	else if ((i = cinstr(pinput,'b')) >= 0)
		*psignal = to_short(pinput+i+1);

	if ((i = cinstr(pinput,'@')) >= 0)
	    memcpy(pparm,pinput+i+1,
		(strlen(pinput+i+1) >= VME_PARAM_SZ) ?
		 VME_PARAM_SZ-1 : strlen(pinput+i+1));
	/* verify the vme address */
/*
	status = verify_hw_addr(io_type,card_type,pcard,psignal);	
	return(status);
*/
	return(0);
}


/****************************************************************************
 *
 *  init_pv
 *
 *   this routine initializes the field values for a process variable 
 *
 *	this routine assumes that set_current_record type has been called
 ***************************************************************************/
 init_pv()
{

	struct	fldDes	*pfield;
	int		i;
	char		*ptr;

	memset(ctrl_record,'\0', precorddes->rec_size);
	for ( i=0; i<precorddes->no_fields; i++ ) {
		pfield = pflddef[i];
		ptr = &ctrl_record[pfield->offset];
		switch(pfield->field_type) {
			case (DBF_CHAR) :
				*(char *)ptr = pfield->initial.char_value;
				break;
			case (DBF_UCHAR) :
				*(unsigned char *)ptr =
					 pfield->initial.uchar_value;
				break;
			case (DBF_SHORT) :
				*(short *)ptr = pfield->initial.short_value;
				break;
			case (DBF_USHORT) :
				*(unsigned short *)ptr =
					 pfield->initial.ushort_value;
				break;
			case (DBF_LONG) :
				*(long *)ptr = pfield->initial.long_value;
				break;
			case (DBF_ULONG) :
				*(unsigned long *)ptr =
					 pfield->initial.ulong_value;
				break;
			case (DBF_FLOAT) :
				*(float *)ptr = pfield->initial.float_value;
				break;
			case (DBF_DOUBLE) :
				*(double *)ptr = pfield->initial.double_value;
				break;
			case (DBF_GBLCHOICE) :
			case (DBF_CVTCHOICE) :
			case (DBF_RECCHOICE) :
			case (DBF_DEVCHOICE) :
			case (DBF_ENUM) :
				*(unsigned short *)ptr =
					 pfield->initial.enum_value;
				break;
			case (DBF_INLINK) :
			case (DBF_OUTLINK) :
			case (DBF_FWDLINK) :
				((struct link *)ptr)->value.value=0.0;
				break;
		}
	}
			
}

/*
 * MOD_RECORD
 *
 * modify the fields of the current record
 */
mod_record(type,ino_prompts)
short	type;		/* record type		*/
short		ino_prompts;	/* initial number of prompted fields */
{
	short		i,row;
	short			no_prompts;
	short		selected;
	char			value[MAX_STRING_LENGTH];
	char			input[MAX_STRING_LENGTH];
	char			*ptr;	
	struct fldDes		*pfield;

	/* go through each field in the file */
	row = 1;
	i = 0;
	selected=0;
	pfield = pflddef[0];
	no_prompts = ino_prompts;

	/* get first prompted field */
	while (prompt_fld(type, pfield) == NOPROMPT)
		pfield= pflddef[++i];

	/* stay in this loop until the user exits or quits */
	while (1){
	
	    /* get user input */
	    move(row,VALUE_X);
	    refresh();
	    input[1] = 0;		 /* Null terminated for getline1 */
	    input[0] = getinpch();

	    /* cursor movement to next entry */
	    switch (input[0]){
 	    case (NEXT):
		clear_error();
		selected=0;
		ptr = &ctrl_record[pfield->offset];
		check_inp_out(type, pfield, ptr);
 		row++;
 		do{
 		    if (++i >= precorddes->no_fields){
 			i = 0;
 			row = 1;
 		    }
		    pfield = pflddef[i];
 		} while (prompt_fld(type, pfield) == NOPROMPT);
		break;

	    /* cursor movement to previous entry */
 	    case (PREVIOUS):
		clear_error();
		selected=0;
		ptr = &ctrl_record[pfield->offset];
		check_inp_out(type, pfield, ptr);
		row--;
		do{
 		    if (--i < 0){
 		       	i=precorddes->no_fields-1;
 			row = no_prompts;
 		    }	
		    pfield = pflddef[i];
 		}while (prompt_fld(type, pfield) == NOPROMPT);
		break;

	    /* cursor movement to the last entry */
 	    case (LAST):
		clear_error();
		selected=0;
		i = precorddes->no_fields-1;
 		row = no_prompts;
 		pfield = pflddef[i];
 		while (prompt_fld(type, pfield) == NOPROMPT){
 		    pfield = pflddef[--i];
 		}
		break;

	    /* cursor movement to the first entry */
 	    case (FIRST):
 		clear_error();
		selected=0;
		i = 0;
 		row = 1;
 		pfield = pflddef[0];
 		while (prompt_fld(type, pfield) == NOPROMPT){
 		    pfield = pflddef[++i];
 		}
		break;

	    /* exit */
 	    case (EXIT):
		ptr = &ctrl_record[pfield->offset];
		check_inp_out(type, pfield, ptr);
		return(0);

	    /* quit without write */
	    case (QUIT1):
	    case (QUIT2):
	    case (QUIT3):
		return(-1);

	    /* change the value using the forward and backward arrow */
	    case (NEXT_ITEM):
	    case (PREV_ITEM):
		clear_error();

		/* determine the new value via a arrow key */
		ptr = &ctrl_record[pfield->offset];
		if(rotate(type,pfield,input[0],(short *)ptr))
				break;

		/* clear the old value */
		move(row,VALUE_X);
		clrtoeol();

		/* print the current value */
		if( fld_to_ascii(pfield,value,type) == 0) {
			printw("%s",value);	
		}
		refresh();
		selected=1;
		break;	

	    case(EXECUTE):
		switch(pfield->field_type) {
			case (DBF_INLINK):
			case (DBF_OUTLINK):
			case (DBF_FWDLINK):
			case (DBF_GBLCHOICE):
			case (DBF_CVTCHOICE):
			case (DBF_RECCHOICE):
			case (DBF_DEVCHOICE):
				clear_error();
				ptr = &ctrl_record[pfield->offset];
				input[0]=0;
				if(!selected) {
				   if(convert_fld(type,pfield,input,ptr) < 0) {
					move(row,VALUE_X);
					clrtoeol();
					break;
				   }
				   move(row, VALUE_X);
				   clrtoeol();
				   if(fld_to_ascii(pfield,value,type) < 0) {
					move(row,VALUE_X);
					clrtoeol();
					break;
				   }
				   printw("%s",value);
				   refresh();
				}
				selected=0;
				check_inp_out(type, pfield, ptr);
				break;
			default:
				break;
		}
		break;
	    /* changing the value of this field */
	    default:
		if (!isprintable(input[0])) continue;
		clrtoeol();
		clear_error();
		refresh();

		/* get the new value from the user */
		getline1(input,sizeof(input),row,INPUT_X);
		/* convert and store it */
		ptr = &ctrl_record[pfield->offset];
		if (convert_fld(type,pfield,input,ptr) < 0) {
			move(row,VALUE_X);
			clrtoeol();
			break;
		}

		/* clear the old value off the screen */
		move(row,VALUE_X);
		clrtoeol();

		/* print the current value */
		move(row,VALUE_X);	/* print the value */
		if( fld_to_ascii(pfield,value,type) == 0 ) {
			printw("%s",value);
		}
		/* check if DTYP changed */
		check_inp_out(type, pfield, ptr);
		/* check if any ranges need changing */
		change_range(type, pfield->fldname);			

		/* check if screen needs to be rewritten */
		check_rewrite((short *)ptr,pfield->fldname,type,&no_prompts,row);

		/* go to the next line */
		row++;
 		do{
 		    if (++i >= precorddes->no_fields){
 			i = 0;
 			row = 1;
 		    }
		    pfield = pflddef[i];
 		} while (prompt_fld(type, pfield) == NOPROMPT);

		refresh();
		break;
	    }

	}
}
/*****************************************************************************
 *
 *	check_inp_out
 *
 *	checks to see it field DTYP has changed, if so prompts for updating
 *		INP or OUT
 *
 *****************************************************************************/
int	check_inp_out(type, pfield, ptr)
short	type;			/* record type */
struct	fldDes	*pfield;	/* current field description structure */
char	*ptr;			/* pointer to current field in ctrl_record */
{
	struct	fldDes	*pfield1;
	struct	link	*pdblink;
	char	input[MAX_STRING_LENGTH];
	char	value[MAX_STRING_LENGTH];
	int	y,x;

	if (strncmp("DTYP",pfield->fldname, FLDNAME_SZ) == SAME) {
		if( dtyp_val != *(short *)ptr) {
			if(( pfield1 = pvdGetFld(type,"INP "))
                 	        || (pfield1=pvdGetFld(type,"OUT "))) {
					dtyp_val = *(short *)ptr;
                        		pdblink = (struct link *)
						&ctrl_record[pfield1->offset]; 
					getyx(stdscr,y,x);
					move(inp_out_row,VALUE_X);
					clrtoeol();
					memset(input,'\0', MAX_STRING_LENGTH);
					if(convert_fld(type,pfield1,input,
							(char *)pdblink)
						 >= 0) {
						if(fld_to_ascii(pfield1,
							value,type)==0){
							move(inp_out_row,
								VALUE_X);
						 	printw("%s",value);
						}
					}
                                        move(y, VALUE_X);
                                        refresh();
			}
       		}
	}
}

/*
 *	PROMPT_FLD
 *
 *	Determines if a field should be prompted
 *
 *     the following test 
 *	     if ((pfield->field_type == DBF_DEVCHOICE) &&
 *                                (!GET_PDEV_CHOICE_SET(choiceDev, type)))
 *	if a work around (ie kludge) because the field "DTYP" appears in 
 *      dbCommon.ascii and therefor in every record type.  Field DTYP is 
 *      is always a PROMPT field, but for many record types, no choices are
 * 	defined in devSup.ascii.  For these record types the pointer to 
 *	the choice set is NULL and hence the above test is done to avoid
 *	presenting the user with the DTYP field.  A side effect of this is
 *	that real errors such as an absence of device choice tables  for 
 *	records such as ai, ao, etc are not detected.   
 *
 *
 */
prompt_fld(type, pfield)
short	type;			/* record type */
struct	fldDes	*pfield;		/* pointer to field description struct */
{
	char	*xname = "prompt_fld";

	if (pfield->promptflag == NOPROMPT)
		return(NOPROMPT);
	else if (pfield->promptflag > NOPROMPT) {
		if ((pfield->field_type == DBF_DEVCHOICE) && 
				(!GET_PDEV_CHOICE_SET(choiceDev, type)))
			return(NOPROMPT);
		else
			return(pfield->promptflag);
	} else{
		print_error_msg_p( xname, "Bad promptflag value", PAUSE_F);
		return(NOPROMPT);
	}
}

/*
 * PUTFIELD
 *
 * put a field into the database
 */
putfield(type,pfield_name,pvalue)
short	type;
char	*pfield_name;
char	*pvalue;
{
	short		fld_type,offset,size;
	char	*ptr;

	/* get the field information */
	get_recfld(type,pfield_name,&fld_type,&offset,&size);

	/* store value */
	ptr = &ctrl_record[0];
	memcpy(ptr+offset,pvalue,size);

}

/*
 *  REC_NAME_TO_TYPE_NUM
 *
 *  converts a record name string to an index within the record description
 * structure  This routine searches struct recType for the "record_name" and
 * returns the index within recType where the string is found else -1
 */
int	rec_name_to_type_num(name_string)
char	*name_string;		/* pointer to string containing name of desired
				record */
{
	short	i;

	for ( i=1; i < dbRecType->number; i++)
		if ( !strcmp( name_string, (char*)XGET_PRECTYPE(i)))
			return(i);
	return(-1);
}

/*
 * ROTATE
 *
 * change a field which has a list of selections
 */
int rotate(type,pfield,direction,pvalue)
short		type;		/* record type */
struct fldDes	*pfield;
char		direction;
short		*pvalue;
{
	char	*xname = "rotate";
	char		*pchar;
	unsigned char	*puchar;
	long		*plong;
	unsigned long	*pulong;
	unsigned short  *pushort;
	short		*pshort;
	short		maxNulls;	/* max # of NULL filled choices */
	short		i;
	char		*pchoice;
	struct choiceSet	*pchoiceSet;
	struct arrChoiceSet	*parrChoiceSet;
	struct devChoiceSet	*pdevChoiceSet;


	switch (pfield->field_type){
	case (DBF_CHAR):
		if (pfield->cvt_type == CT_HEX) {
			pchar = (char *)pvalue;
			if (direction == NEXT_ITEM){
				*pchar >>= 1;
			}else if (direction == PREV_ITEM){
				*pchar <<= 1;
			}
		}
		break;
	case (DBF_UCHAR):
		if (pfield->cvt_type == CT_HEX) {
			puchar = (unsigned char *)pvalue;
			if (direction == NEXT_ITEM){
				*puchar >>= 1;
			}else if (direction == PREV_ITEM){
				*puchar <<= 1;
			}
		}
		break;
	case (DBF_SHORT):
		if (pfield->cvt_type == CT_HEX) {
			pshort = (short *)pvalue;
			if (direction == NEXT_ITEM){
				*pshort >>= 1;
			}else if (direction == PREV_ITEM){
				*pshort <<= 1;
			}
		}
		break;
	case (DBF_USHORT):
		if (pfield->cvt_type == CT_HEX) {
			pushort = (unsigned short *)pvalue;
			if (direction == NEXT_ITEM){
				*pushort >>= 1;
			}else if (direction == PREV_ITEM){
				*pushort <<= 1;
			}
		}
		break;
	case (DBF_LONG):
		if(pfield->cvt_type == CT_HEX) {
			plong = (long *)pvalue;
			if (direction == NEXT_ITEM){
				*plong >>= 1;
			}else if (direction == PREV_ITEM){
				*plong <<= 1;
			}
		}
		break;	
	case (DBF_ULONG):
		if(pfield->cvt_type == CT_HEX) {
			pulong = (unsigned long *)pvalue;
			if (direction == NEXT_ITEM){
				*pulong >>= 1;
			}else if (direction == PREV_ITEM){
				*pulong <<= 1;
			}
		}
		break;	
	case (DBF_GBLCHOICE):
		if( !(pchoiceSet = GET_PCHOICE_SET(choiceGbl, 
				pfield->choice_set))) {
			print_error_msg_p(xname, "NULL pointer for choiceGbl",
				PAUSE_F);
			break;
		} 
		if ( (maxNulls = pchoiceSet->number +1) == 1) {
			print_error_msg_p(xname, "NULL choice table", PAUSE_F);
			return(-1);
		}
		i = 0;
		if (direction == NEXT_ITEM){
		   do {
			*pvalue = *pvalue + 1;
                        if ( *pvalue >= pchoiceSet->number) {
                                *pvalue = 0;
			}
			pchoice = GET_CHOICE(pchoiceSet, *pvalue);
			i++;
		    } while ( pchoice == NULL && i <= maxNulls);
 		}else if (direction == PREV_ITEM){
		   do {
			*pvalue = *pvalue - 1;
                        if ( *pvalue < 0 ) {
                                *pvalue = pchoiceSet->number -1;
			}
			pchoice = GET_CHOICE(pchoiceSet, *pvalue);
		    } while ( pchoice == NULL && i <= maxNulls);
		 }
		if (i >= maxNulls) {
			print_error_msg_p(xname, "NULL filled choice table",
				PAUSE_F);
			return(-1);
		}
		break;
	case (DBF_CVTCHOICE):
		if (!choiceCvt) {
			print_error_msg_p(xname, "NULL pointer for choiceCvt",
				PAUSE_F);
			return(-1);
		}
		if ( (maxNulls = choiceCvt->number +1) == 1) {
			print_error_msg_p(xname, "NULL choice table",PAUSE_F);
			return(-1);
		}
		i = 0;
                if (direction == NEXT_ITEM){
                   do {
                        *pvalue = *pvalue + 1;
                        if ( *pvalue >= choiceCvt->number) {
                                *pvalue = 0;
                        }
                        pchoice = GET_CHOICE(choiceCvt, *pvalue);
                        i++;
                    } while ( pchoice == NULL && i <= maxNulls);
                }else if (direction == PREV_ITEM){
                   do {
                        *pvalue = *pvalue - 1;
                        if ( *pvalue < 0 ) {
                                *pvalue = choiceCvt->number -1;
                        }
                        pchoice = GET_CHOICE(choiceCvt, *pvalue);
		    } while ( pchoice == NULL && i <= maxNulls);
                }

		if (i >= maxNulls) {
			print_error_msg_p(xname, "NULL filled choice table",
				PAUSE_F);
			return(-1);
		}
		break;
	case (DBF_RECCHOICE):
		if( !(parrChoiceSet = GET_PARR_CHOICE_SET(choiceRec, 
				type))) {
			print_error_msg_p(xname,
				 "NULL pointer for choiceRec",
				 PAUSE_F);
			return(-1);
		} 
		if( !(pchoiceSet = GET_PCHOICE_SET(parrChoiceSet, 
				pfield->choice_set))) {
			print_error_msg_p(xname,
				 "NULL pointer for choiceSet in choiceRec",
				PAUSE_F);
			return(-1);
		} 
		if ( (maxNulls = pchoiceSet->number +1) == 1) {
			print_error_msg_p(xname, "NULL choice table", PAUSE_F);
			return(-1);
		}
		i = 0;
                if (direction == NEXT_ITEM){
                   do {
                        *pvalue = *pvalue + 1;
                        if ( *pvalue >= pchoiceSet->number) {
                                *pvalue = 0;
                        }
                        pchoice = GET_CHOICE(pchoiceSet, *pvalue);
                        i++;
                    } while ( pchoice == NULL && i <= maxNulls);
                }else if (direction == PREV_ITEM){
                   do {
                        *pvalue = *pvalue - 1;
                        if ( *pvalue < 0 ) {
                                *pvalue = pchoiceSet->number -1;
                        }
                        pchoice = GET_CHOICE(pchoiceSet, *pvalue);
                    } while ( pchoice == NULL && i <= maxNulls);
                }


		if (i >= maxNulls) {
			print_error_msg_p(xname, "NULL filled choice table",
				PAUSE_F);
			return(-1);
		}
		break;
	case (DBF_DEVCHOICE):
		if( !(pdevChoiceSet = GET_PDEV_CHOICE_SET(choiceDev, type))) {
			print_error_msg_p(xname, "NULL pointer for choiceDev",
				PAUSE_F);	
			return(-1);
		} 
		if ( (maxNulls = pdevChoiceSet->number +1) == 1) {
			print_error_msg_p(xname, "NULL choice table", PAUSE_F);
			return(-1);
		}
		i = 0;
                if (direction == NEXT_ITEM){
                   do {
                        *pvalue = *pvalue + 1;
                        if ( *pvalue >= pdevChoiceSet->number) {
                                *pvalue = 0;
                        }
                        pchoice = GET_DEV_CHOICE(pdevChoiceSet, *pvalue)->pchoice;
                        i++;
                    } while ( pchoice == NULL && i <= maxNulls);
                }else if (direction == PREV_ITEM){
                   do {
                        *pvalue = *pvalue - 1;
                        if ( *pvalue < 0 ) {
                                *pvalue = pdevChoiceSet->number -1;
                        }
                        pchoice = GET_DEV_CHOICE(pdevChoiceSet, *pvalue)->pchoice;
                    } while ( pchoice == NULL && i <= maxNulls);
                }

		if (i >= maxNulls) {
			print_error_msg_p(xname, "NULL filled choice table",
				PAUSE_F);
			return(-1);
		}
		break;
	}

	return(0);
}

/*
 *  SET_CURRENT_RECORD_TYPE
 *
 *  sets up pointers for desired record type
 *
 *  Returns 0 if success else  -1 
 */
int	set_current_record_type(record_type)
short record_type;	/* record type index */
{
	char	*xname = "set_current_type";

	if ((precorddes = XGET_PRECTYPDES(record_type)) == NULL){
		print_error_msg_p(xname, "Uknown record type", PAUSE_F);
		return(-1);
	}
	if ((pflddef = precorddes->papFldDes) == NULL) {
		print_error_msg_p(xname, "NULL pointer to Record Field Descrip",			PAUSE_F);
		return(-1);
	}
	curr_rectypedef = record_type;
	return(0);	
}

/*
 *	TO_FIELDNAME
 *
 *	convert a record type and offset to a field name
 */
to_fieldname(record_type,offset,pname)
short	record_type;	/* record type index */
short	offset;		/* field offset */
char	*pname;
{
	char	*xname = "to_fieldname";
	short		old_rectype;
	short	i;
	struct fldDes	*pfield;

	/* save the current record type */
	old_rectype = curr_rectypedef;

	/* get this record type into memory */
	set_current_record_type(record_type);

	/* find this field */
	for (i=0; i<precorddes->no_fields; i++){
		pfield = pflddef[i];
		if (offset == pfield->offset){
			memcpy(pname,pfield->fldname,FLDNAME_SZ);
			*(pname+FLDNAME_SZ) = NULL;
			set_current_record_type(old_rectype);
			return(0);
		}
	}
	set_current_record_type(old_rectype);
	return(-1);
}

/*
 * WRITE_PAGE
 *
 * display a page of the record definition prompts on the screen
 */
short write_page(type,fd)
short	type;		/* record type */
short	fd;		/* file descriptor to write to */
{
	char		*xname = "write_page";
	short			i;
	struct fldDes	*pfield;
	struct link		*plink;
	char				buffer[VALUE_X+VALUE_STRING_SZ+8];
	char				range_str[RANGE_STRING_SZ+8];
	char				value_str[VALUE_STRING_SZ+8];
	char			*ptime;
	time_t			timeofday;
	short			no_fields_written;
	char			*pstr;	/* pointer to string containing record
						type name */

	/* read the field definitions for this record */
        if ( set_current_record_type(type))
                return(-1);

	/* set pointer to record type name string */

	pstr = (char*)XGET_PRECTYPE(type);

	/* write the header for this process variable */
	if (fd == STDOUT){
		sprintf(buffer,"Process Variable: %-*.*s Type: %-*.*s (%s)",
		  PVNAME_SZ, PVNAME_SZ, &ctrl_record[0], MAX_RECNAME_SZ,
		  MAX_RECNAME_SZ, pstr, curr_dbname);
		clear();
		move(0,0);
		addstr(buffer);
		print_help_msg(2);
	}else{
		timeofday = time((time_t *)0);
		ptime = ctime(&timeofday);
		sprintf(buffer,"\f\n\nPV: %-*.*s Type: %-*.*s (%s) \n",
		  PVNAME_SZ, PVNAME_SZ, &ctrl_record[0], MAX_RECNAME_SZ, 
			MAX_RECNAME_SZ, pstr, curr_dbname);
		write(fd,buffer,strlen(buffer));
	}


	/* go through each field in the file */
	no_fields_written = 0;
	for (i=0; i<precorddes->no_fields; i++){
		pfield = pflddef[i];
		/* print the field if it needs to be entered */
		if (prompt_fld(type,pfield) == NOPROMPT) 
			continue;
		no_fields_written++;

 		/* if range is variable then find its current value */
		if (pfield->lowfl == VAR){
			get_range_val(pflddef[pfield->range1.fldnum],
				&pfield->range1.value);
		}
		if (pfield->highfl == VAR){
			get_range_val(pflddef[pfield->range2.fldnum],
				&pfield->range2.value);
		}

		switch (pfield->field_type){
		case (DBF_STRING):
			sprintf(range_str, "<%d chars>", pfield->size);
			break;
		case (DBF_CHAR):
		case (DBF_UCHAR):
		case (DBF_SHORT):
		case (DBF_USHORT):
		case (DBF_LONG):
		case (DBF_ULONG):
		case (DBF_DOUBLE):
		case (DBF_FLOAT):
		case (DBF_ENUM):
			build_range_string(xname, pfield->field_type,
				pfield->cvt_type, range_str,
				 &pfield->range1.value, &pfield->range2.value);
			break;
		case (DBF_INLINK):
		case (DBF_OUTLINK):
		case (DBF_FWDLINK):
			sprintf(range_str,"<ADDRESS>");
			break;
		case (DBF_GBLCHOICE):
                case (DBF_CVTCHOICE):
                case (DBF_RECCHOICE):
                case (DBF_DEVCHOICE):
                        sprintf(range_str,"<SELECT>");
                        break;
		case (DBF_NOACCESS):
		default:
			sprintf(range_str,"<BAD>");
			break;
		}
                if (strncmp(pfield->fldname,"DTYP", FLDNAME_SZ) == SAME){
                        dtyp_val = *(short *)&ctrl_record[pfield->offset];
                }
		if (strncmp(pfield->fldname,"INP ", FLDNAME_SZ) == SAME){
			plink = (struct link*)&ctrl_record[pfield->offset];
			in_link_type = plink->type;
			inp_out_row = no_fields_written;
		}
		if (strncmp(pfield->fldname,"OUT ", FLDNAME_SZ) == SAME){
			plink = (struct link*)&ctrl_record[pfield->offset];
			out_link_type = plink->type;		
			inp_out_row = no_fields_written;
		}
		
		fld_to_ascii(pfield,value_str,type);
		if (fd == STDOUT){
			sprintf(buffer,"%-*.*s %-*.*s  %-*.*s %-*.*s",
			  PROMPT_SZ, PROMPT_SZ,pfield->prompt,
			  FLDNAME_SZ, FLDNAME_SZ,pfield->fldname,
			  RANGE_STRING_SZ, RANGE_STRING_SZ,range_str,
			  VALUE_STRING_SZ, VALUE_STRING_SZ,value_str);
			move(no_fields_written, 0);
			addstr(buffer);
			refresh(); 
		}else if (long_report){
			sprintf(buffer,"\n%-*.*s %-*.*s  %-*.*s %-*.*s",
			  PROMPT_SZ, PROMPT_SZ,pfield->prompt,
			  FLDNAME_SZ, FLDNAME_SZ,pfield->fldname,
			  RANGE_STRING_SZ, RANGE_STRING_SZ, range_str,
			  VALUE_STRING_SZ, VALUE_STRING_SZ,value_str);
			write(fd,buffer,strlen(buffer));
		}else{
			sprintf(buffer,"\n%.*s %.*s",FLDNAME_SZ,
				pfield->fldname,
					VALUE_STRING_SZ,value_str);
			write(fd,buffer,strlen(buffer));
		}
	}
 

	return(no_fields_written);
}



/*
 * XRECTYPE
 *
 * Translate a record type string into a record number
 */
xrectype(pstring)
char	*pstring;
{
	short	i;

	for (i=1; i < dbRecType->number; i++)
		if (strcmp(pstring,(char*)XGET_PRECTYPE(i)) == SAME)
			return(i);
	return(-1);
}
