static char SccsId[] = "$Id: xref.c,v 1.16 1995/02/28 19:23:11 mrk Exp $ ";

/*  $Id: xref.c,v 1.16 1995/02/28 19:23:11 mrk Exp $
 *	Original Author: Bob Dalesio (LANL)
 *	Current Author: Frank Lenkszus (ANL)
 *	Date:	5/29/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	3/25/92		frl	Null terminated PV Name
 * .02  05-18-92        rcz     New database access - XGET_PRECTYPE is old GET_PRECTYP
 * 	...
 */

/*
 * xref.c
 *
 * cross reference of links reporting utility
 *
 */
#include        <stdlib.h>
#include        <stddef.h>
#include        <stdio.h>
#include	<ctype.h>	/* SUN support for is* routines */
#include	<curses.h>	/* SUN support for cursor routines */
#include	<string.h>
#include	<errno.h>
#include	"dbPvd.h"
#include	<link.h>
#include	<dbRecords.h>
#include	<sdrHeader.h>
#include	<dbRecDes.h>
#include	<dbRecType.h>
#include	<dbFldTypes.h>
#include	"dctDef.h"
#include	"dctmacros.h"

extern struct dbBase *pdbBase;
extern struct choiceSet *choiceCvt;
extern struct arrChoiceSet *choiceGbl;
extern struct choiceRec *choiceRec;
extern struct devChoiceRec *choiceDev;
extern struct arrBrkTable *cvtTable;
extern struct recDes *dbRecDes;
extern struct recType *dbRecType;
extern struct recHeader *dbRecords;
extern struct recDevSup *devSup;
extern struct drvSup *drvSup;
extern struct recSup *recSup;
extern struct pvd *dbPvd;

#define		MAX_LINK_MAP_ENTRIES	128	/* sets max # of links that
						dct can handle when displaying
						the link map for a PV */
extern	struct	fldDes  *pvdGetFld();
extern	char	*fldname_to_string();
extern	char	*get_dct_choice_string();
extern	short	get_pvd_index_by_pvname();

/* global variables - defined in dct.c */
extern  struct  recTypDes       *precorddes;            /* current record type
                                                definition structure */
extern struct pvd	pvd;		/* the process variable directory */
extern int		rep_fd;		/* report file descriptor */
extern char		curr_dbname[];
extern short		order[MAX_PV+4];	/* pv alpha order */
extern char		ctrl_record[];

struct	link_map_entry {
	short	level;			/* level within tree */
	short	rectype;		/* record type */
	short	scantype;		/* scan type for this record */
	char	pscantype[16];		/* scan type string */
	char	pvname[PVNAME_SZ+1];	/* the pv name */
	char	fldname[FLDNAME_SZ+2];	/* the field name */
	short	fldtype;		/* the link type for the field */
	short	llevel;			/* level of PV linked to */
	short	lrectype;		/* record type of pv linked to */
	short	lscantype;		/* scan type for linked record */
	char	plscantype[16];		/* scan type string */
	char	lpvname[PVNAME_SZ+1];	/* name of pv linked to */
	char	lfldname[FLDNAME_SZ+2];  /* field of the pv linked to */
	short	proc_passive;		/* process passive val for this link*/
	short	exist;			/* != 0 -> exists in this database */
	short	lfldtype;		/* type of field linked to */
};
struct	link_map_entry	**pmap;		/* this will point to an array of
					pointers to struct link_map */

struct	temp_hold_pv_info {		/* struct to hold some pv info for for
						left hand side resolution */
	short	rectype;		/* record type */
	short	record_no;		/* record number */
	char	pvname[PVNAME_SZ+1];	/* the pv name */
	short	scantype;		/* scan type for this pv */
	char	pscantype[16];		/* string equivalent of scan type */
};
struct	link_tree_node   {		/* struct to hold a link tree node */
	char	pvname[PVNAME_SZ+2];	/* pv name associated with this node*/
	short	level;			/* level within the tree */
	short	link;			/* link # for this node */
	struct link_tree_node *parent;  /* parent to this node */
	struct link_tree_node *child;	/* first child to this node */
	struct link_tree_node *prev;	/* previous node at this level */
	struct link_tree_node *next;	/* next node at this level */
};

/*
 * Subroutines
 *
 * del_refs		delete all references to a record
 *	args
 *		record_no	record number
 *		test_type	recortd type
 * loc_single_ref	list the references for a given record
 *	args
 *		record_no	record number
 *		test_type	record type
 *	returns
 *		numlinks	number of links to this record
 * loc_xref		list all cross references in the current node
 */

/*
 *	DEL_REFS
 *
 *	delete all the references to the specified record in the current db
 */
del_refs(pvname)
char	*pvname;	/* name of process variable being deleted */
{
	char		*xname = "del_refs";
	short		chkorder[200];
	short		ordinx;
	short		type_inx;
	short		link;		/* link counter */
	short		nolinks;	/* num of links for this record type */
	short		*linkinx;	/* pointer to array of indexes within
					array of pointers to struct fldDes */
					/* for this record type */
	struct link	*plink; /* pointer to a link field structure */
	struct pv_link	*ppv_link;
	struct fldDes   *pfldDes;
	short	fd;
	char	record[MAX_RECORD_SIZE];
	struct	sdrHeader	lsdrHeader;
	struct	recLoc		lrecLoc;

	/* check each type */
	for (type_inx = 1; type_inx < dbRecType->number; type_inx++){

	    /* check for type exists */
	    while ((type_inx < dbRecType->number)
		&& (!(char*)XGET_PRECTYPE(type_inx)))
	        type_inx++;
	    if (type_inx >= dbRecType->number) continue;

	    set_current_record_type(type_inx);
	    nolinks = precorddes->no_links;
		
	    /* open the file for this record type */
	    if ((fd = open_loc_record(type_inx, &lsdrHeader, &lrecLoc)) < 0)
			 return(-1);

	    /* alphabetical order */
	    order_pvs(type_inx, chkorder, MAX_PV+4);

	    /* check each record */
	    for (ordinx = 0; chkorder[ordinx] != -1; ordinx++){

	        /* get this record into memory */
	        if ( read_loc_record(fd,record,
	          dbPvd->entry[chkorder[ordinx]].record_no, xname, &lrecLoc) 
			< 0) {
			close(fd);
			return(-1);
		}
	        /* for each link field */
	        linkinx = precorddes->link_ind;
	        for (link = 0; link < nolinks; link++,linkinx++){

			if ( ( pfldDes = GET_PFLDDES( precorddes, *linkinx)) ==
				NULL) {
				print_error_msg_p(xname,
			       "NULL pointer to field description encountered",
					PAUSE_F);
				continue;
			}
	            /* pointer to the link field */
	            plink = 
		      (struct link *)&record[pfldDes->offset];
	            ppv_link = &plink->value.pv_link;

	            /* is linked */
	            if  (plink->type == PV_LINK) {

			/* to the record being deleted */
			if ( !strcmp(pvname, ppv_link->pvname)) {

	                    /* delete it */
				memset((char *)plink, '\0', sizeof(struct link));
	                    if(write_loc_record(fd, record,
	                      dbPvd->entry[chkorder[ordinx]].record_no,
					&lrecLoc) < 0) {
					close(fd);
					return(-1);
			    }
	                }
	            }
	        }
	    }
	    close(fd);
	}
	return(0);
}

/*
 *	RENAME_REFS
 *
 *	rename all the references to the specified record in the current db
 */
rename_refs(pvname, newname)
char	*pvname;	/* name of process variable being renamed */
char	*newname;	/* new name */
{
	char		*xname = "rename_refs";
	short		chkorder[200];
	short		ordinx;
	short		type_inx;
	short		link;		/* link counter */
	short		nolinks;	/* num of links for this record type */
	short		*linkinx;	/* pointer to array of indexes within
					array of pointers to struct fldDes */
					/* for this record type */
	struct link	*plink; /* pointer to a link field structure */
	struct pv_link	*ppv_link;
	struct fldDes   *pfldDes;
	short	fd;
	char	record[MAX_RECORD_SIZE];
	struct	sdrHeader	lsdrHeader;
	struct	recLoc		lrecLoc;

	/* check each type */
	for (type_inx = 1; type_inx < dbRecType->number; type_inx++){

	    /* check for type exists */
	    while ((type_inx < dbRecType->number)
		&& (!(char*)XGET_PRECTYPE(type_inx)))
	        type_inx++;
	    if (type_inx >= dbRecType->number) continue;

	    set_current_record_type(type_inx);
	    nolinks = precorddes->no_links;
		
	    /* open the file for this record type */
	    if ((fd = open_loc_record(type_inx, &lsdrHeader, &lrecLoc)) < 0)
			 return(-1);

	    /* alphabetical order */
	    order_pvs(type_inx, chkorder, MAX_PV+4);
	    /* check each record */
	    for (ordinx = 0; chkorder[ordinx] != -1; ordinx++){

	        /* get this record into memory */
	        if ( read_loc_record(fd,record,
	          dbPvd->entry[chkorder[ordinx]].record_no, xname, &lrecLoc) 
			< 0) {
			close(fd);
			return(-1);
		}
	        /* for each link field */
	        linkinx = precorddes->link_ind;
	        for (link = 0; link < nolinks; link++,linkinx++){

			if ( ( pfldDes = GET_PFLDDES( precorddes, *linkinx)) ==
				NULL) {
				print_error_msg_p(xname,
			       "NULL pointer to field description encountered",
					PAUSE_F);
				continue;
			}
	            /* pointer to the link field */
	            plink = 
		      (struct link *)&record[pfldDes->offset];
	            ppv_link = &plink->value.pv_link;

	            /* is linked */
	            if  (plink->type == PV_LINK) {

			/* to the record being deleted */
			if ( !strcmp(pvname, ppv_link->pvname)) {

	                    /* rename it */
				strncpy(ppv_link->pvname, newname, PVNAME_SZ+1);
	                    if(write_loc_record(fd, record,
	                      dbPvd->entry[chkorder[ordinx]].record_no,
					&lrecLoc) < 0) {
					close(fd);
					return(-1);
			    }
	                }
	            }
	        }
	    }
	    close(fd);
	}
	return(0);
}

/****************************************************************************
 *
 *  add_hold_pv_entry()
 *
 ****************************************************************************/
int	add_hold_pv_entry(phold, pinx, rectype, recno, pvname, scantype,
		pscantype, pname)
struct temp_hold_pv_info	*phold[];  /* the a hold array */
int	*pinx;			/* the next available location in phold */
/*   the following stuff is saved in the next slot in phold */
short	rectype;		/* record type */
short	recno;			/* record # */
char	*pvname;		/* the pv name to save */
short	scantype;		/* the scan type */
char	*pscantype;		/* the scan type string */
char	*pname;			/* the name of the caller */
{
	int i;

	/* check if we already have this one */
	for ( i = 0; i < *pinx; i++)
		if(!strcmp(phold[i]->pvname, pvname))
			return(0);		/* yep, we already have this*/

	if((phold[*pinx]=(struct temp_hold_pv_info *) calloc(1,
				sizeof(struct temp_hold_pv_info))) == NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	phold[*pinx]->rectype=rectype;
	phold[*pinx]->record_no=recno;
	strcpy(phold[*pinx]->pvname,pvname);
	phold[*pinx]->scantype=scantype;
	strcpy(phold[*pinx]->pscantype, pscantype);
	(*pinx)++;
	return(0);
}

/*****************************************************************************
 *
 *	add_link_map_entry()
 *
 *****************************************************************************/
int	add_link_map_entry(pmap, pinx, level, rectype, recno, pvname, pfldDes,
		 phold, phinx, pname)
struct	link_map_entry	*pmap[];	/* array of pointers to map entries */
int	*pinx;			/* index of next available slot in pmap[] */
int	level;			/* level of this link */
short	rectype;		/* record type of current record */
short	recno;			/* record number */
char	*pvname;		/* current pv name */
struct	fldDes	*pfldDes;	/* pointer to field description of current 
					field */
struct	temp_hold_pv_info **phold;  /* the hold array */
int	*phinx;			/* the next available slot in phold */
char	*pname;			/* name of caller */
{
	struct	link	*plink;	/* the current link field */
	struct	pv_link	*ppv_link;  /* PV_LINK struct within plink */
	struct	fldDes	*pflddes;
	int	status;
 
	if((pmap[*pinx]=(struct link_map_entry *)calloc(1,
			 sizeof(struct link_map_entry))) == NULL) {
		print_syserror_msg_p(pname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	plink=(struct link *)&ctrl_record[pfldDes->offset];
	ppv_link=&plink->value.pv_link;
	pmap[*pinx]->level = level;
	pmap[*pinx]->rectype = rectype;
	strcpy(pmap[*pinx]->pvname, pvname);
	fldname_to_string(pfldDes->fldname, pmap[*pinx]->fldname, FLDNAME_SZ);
	pmap[*pinx]->fldtype = pfldDes->field_type;
	strcpy(pmap[*pinx]->lpvname, ppv_link->pvname);
	fldname_to_string(ppv_link->fldname, pmap[*pinx]->lfldname, FLDNAME_SZ);
	pmap[*pinx]->proc_passive = ppv_link->process_passive;
	pmap[*pinx]->exist = 0;		/* set to "1" later if found */
	if(!(pflddes=pvdGetFld(rectype, "SCAN"))) {
		print_error_msg_p(pname,
		 "Couldn't find SCAN field in record description", PAUSE_F);
		return(-1);
	}
	pmap[*pinx]->scantype=(short)ctrl_record[pflddes->offset];
	if(fld_to_ascii(pflddes, pmap[*pinx]->pscantype , rectype))
		strcpy(pmap[*pinx]->pscantype, "BAD");
	status =add_hold_pv_entry(phold, phinx, rectype, recno, pvname, 
			pmap[*pinx]->scantype, pmap[*pinx]->pscantype,
				pname);
	(*pinx)++; 
	return(status);
}

/*****************************************************************************
 *
 *	print_link_map_entry()
 *
 *****************************************************************************/
int	print_link_map_entry(pmap, inx, pbuffer, pname)
struct	link_map_entry	*pmap;	/* pointer to map entrie */
int	inx;			/* index of the link */
char	*pbuffer;		/* place to build display string */
char	*pname;				/* name of caller */
{
	char	*xname="print_link_map_entry";
	char	*prectype;
	char	*plrectype;
	char	field_type[32];
	char	lfield_type[32];
	char	*ppassive;	/* pointer to process passive string */

	if(pmap != NULL) {
		if((prectype=(char*)XGET_PRECTYPE(pmap->rectype)) == NULL)
			strcpy(prectype,"BAD");
		fieldtype_to_string(pmap->fldtype, field_type, xname);
		if(pmap->fldtype != DBF_FWDLINK)
			ppassive=get_dct_choice_string(0,pmap->proc_passive);
		else
			ppassive="";
		if(pmap->exist != 0) {
			if((plrectype=(char*)XGET_PRECTYPE(pmap->lrectype)) == NULL)
				strcpy(plrectype,"BAD");
			fieldtype_to_string(pmap->lfldtype, lfield_type,
				 xname);
		    sprintf(pbuffer,
			"%3.3d %s.%s (%s,%s,%s) -> %s.%s.%s (%s,%s,%s)\n",
				inx,  pmap->pvname, pmap->fldname, 
				prectype, pmap->pscantype, field_type, 
				pmap->lpvname,pmap->lfldname, ppassive,
				plrectype, pmap->plscantype,lfield_type);
		} else {
			
		     sprintf(pbuffer,
			 "%3.3d %s.%s (%s,%s,%s) -> %s.%s.%s (External)\n",
				inx, pmap->pvname, pmap->fldname, 
				prectype, pmap->pscantype, field_type, 
				pmap->lpvname,pmap->lfldname, ppassive);
		}
		write(rep_fd, pbuffer, strlen(pbuffer));
	}
	return(0);
}

/****************************************************************************
 *
 *	print_link_groups()
 *
 ****************************************************************************/
int	print_link_groups(pmap, num, pvname, pvnamesz, phold, hnum,pname,header)
struct	link_map_entry  *pmap[];	/* array of link groups */
int	num;				/* size of array */
char	*pvname[];			/* array of pvnames */
int	pvnamesz;
struct	temp_hold_pv_info *phold[];	/* array of record info */
int	hnum;				/* number of entries in phold*/
char	*pname;				/* name of caller */
char	*header;		/* page header to print */
{
	char	*xname="print_link_groups";
	int i,j,k;
	char	*pbuffer;		/* place to build display string */
	char	buffer[PVNAME_SZ + 48];
	int	first_one;		/* 1 = first time through for loop */
	int	found;
	char	*prectype;
	
	i = 0;
	if((pbuffer=malloc(2*(PVNAME_SZ+1) +2*(FLDNAME_SZ+1) +2*((3*16)+5) +12))
			 == NULL) {
		print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	if(rep_fd == STDOUT) {
		clear();
		move(0,0);
		refresh();
	}
	write(rep_fd, header, strlen(header));
	for( i=0, j=0, first_one=1; i < num -1 ; i++, first_one=0) {
		if(pmap[i] == NULL || first_one) {
			if(pvname[j] != NULL) {
			     for(k=0,found=0; k < hnum; k++) {
				if(!strcmp(phold[k]->pvname,pvname[j])){
					found=1;
					if((prectype=(char*)XGET_PRECTYPE(
						phold[k]->rectype)) == NULL)
					            prectype="BAD RECORD TYPE";					
					break;
				}
			     }
		 	     if(found)
				sprintf(buffer,"\nLinks for %s (%s,%s)\n",
					pvname[j], prectype,
						 phold[k]->pscantype);
			     else
				sprintf(buffer,"\nLinks for %s (External)\n",
						 pvname[j]);
			}else 
				sprintf(buffer,"\nLinks for BAD PV NAME\n");
			write(rep_fd, buffer, strlen(buffer));
			j++;
		}	
		if(pmap[i] != NULL)
			print_link_map_entry(pmap[i],0, pbuffer, pname);
	}

	if(pbuffer)
		free(pbuffer);
	return(0);
}

/*****************************************************************************
 *
 *	print_link_map()
 *
 *****************************************************************************/
int	print_link_map(pmap,next,pname, header)
struct	link_map_entry	*pmap[];	/* array of pointers to map entries */
int	next;				/* next empty slot */
char	*pname;				/* name of caller */
char	*header;			/* header string */
{
	char	*xname="print_link_map";
	int 	i;
	char	*pbuffer;		/* place to build display string */
	static	char *info[] = {
		"Legend:\n",
		"\tLink# pvname.fldname (record_type, scan_mech, fld_type)",
		" <refers to>\n",
		"\t\tpvname.fldname.pp (record_type, scan_mech, fld_type)\n\n",
		NULL
	}; 

	i = 0;
	if((pbuffer=malloc(2*(PVNAME_SZ+1) + 2*(FLDNAME_SZ+1) +2*((3*16)+5) +12))
			 == NULL) {
		print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
		return(-1);
	}
	if(rep_fd == STDOUT) {
		clear();
		move(0,0);
		refresh();
	}
	write(rep_fd, header, strlen(header));
	print_string_array(info, sizeof(info)/sizeof(char *), rep_fd);
	while (pmap[i] != NULL) {
		print_link_map_entry(pmap[i],i, pbuffer, pname);
		i++;
	}
	if(pbuffer)
		free(pbuffer);
	return(0);
}

/*****************************************************************************
 *
 *   print_hold_struct()
 *
 *****************************************************************************/
int	print_hold_struct(phold, num)
struct	temp_hold_pv_info  *phold[];	/* the hold array */
int	num;				/* number of entries */
{
	int i;
	char	*prectype;

	for (i =0; i < num; i++) {
		if(( prectype = (char*)XGET_PRECTYPE(phold[i]->rectype)) == NULL )
			prectype= "BAD RECORD TYPE";
		printf("%s (%s,%s)\n", phold[i]->pvname, prectype,
				 phold[i]->pscantype);	
	}
	return(0);
}

/*****************************************************************************
 *
 *  check_if_already_processed()
 *
 *****************************************************************************/
static int check_if_already_processed(pvname, parray1, parray2, parray3) 
char	*pvname;		/* pvname to look for */
char	**parray1,**parray2,**parray3;  /* arrays to look in */
{
	if(lookup_string_array(parray1, pvname, MAX_LINK_MAP_ENTRIES) >= 0)
		return(1);
	if(lookup_string_array(parray2, pvname, MAX_LINK_MAP_ENTRIES) >= 0)
		return(2);
	if(lookup_string_array(parray3, pvname, MAX_LINK_MAP_ENTRIES) >= 0)
		return(3);
	return(0);
}

/*****************************************************************************
 *
 *  alloc_link_tree_node()
 *
 *****************************************************************************/
struct link_tree_node	*alloc_link_tree_node(pnode, pinx, size, pvname, level,
		 link)
struct	link_tree_node	*pnode[];	/* place to store node pointers */
int		*pinx;			/* index of next available location*/
int		size;			/* size of pnode[] */
char		*pvname;		/* pv for this node */
short		level;			/* level of this node */
short		link;			/* link for this node */
{
	char	*xname="alloc_link_tree_node";
	struct	link_tree_node	*p;

	if( *pinx >= size-1) {
		print_error_msg_p(xname, "Link tree node table overflow",
				PAUSE_F);
		return(NULL);
	}
	if((p=(struct link_tree_node *)calloc(1,
			sizeof(struct link_tree_node))) == NULL) {
		print_syserror_msg_p(xname, "calloc", errno, PAUSE_F);
		return(NULL);
	}
	pnode[(*pinx)++]=p;
	strcpy(p->pvname, pvname);
	p->level = level;
	p->link	 = link;
	p->parent = NULL;
	p->child = NULL;
	p->next = NULL;
	p->prev = NULL;
	return(p);
}

/*****************************************************************************
 *
 *   find_next_tree()
 *
 ******************************************************************************/
struct link_tree_node  *find_next_tree(pmap, mapsz, pnode, pinx, pnodesz,
			 ppnode, level)
struct	link_map_entry	*pmap[];	/* array of links */
int		mapsz;			/* size of pmap[] */
struct	link_tree_node	*pnode[];	/* array of tree nodes */
int		*pinx;			/* next available slot in pnode[] */
int		pnodesz;		/* size of pnode[] */
struct	link_tree_node	*ppnode;	/* the parent node */
short		level;			/* current level within the tree */
{
	int	i;
	struct	link_tree_node	*p,*p2;

	for ( i=0, p=NULL; i < mapsz; i++) {	/* find parents */
		if(pmap[i]->level >= 1)	/* skip links already used */
			continue;
		if(strcmp(pmap[i]->lpvname, ppnode->pvname))
			continue;
		else {
			pmap[i]->level=level;
		   	if(!(p=alloc_link_tree_node(pnode, pinx, 
					pnodesz, pmap[i]->pvname, level,i)))
					return(NULL);
			if(ppnode->parent)
				   	ppnode->parent->prev=p;
			p->next=ppnode->parent;
			ppnode->parent=p;
		}	/* else */
	}		/* for loop */
	for ( i=0, p=NULL; i < mapsz; i++) {	/* now look for the kids */
		if(pmap[i]->level >= 1)	/* skip links already used */
			continue;
		if(strcmp(pmap[i]->pvname, ppnode->pvname))
			continue;
		else {
			pmap[i]->level=level;
			if(!(p=alloc_link_tree_node(pnode, pinx,
					pnodesz, pmap[i]->lpvname, level,i)))
						return(NULL);
			if(ppnode->child)
		   		ppnode->child->prev=p;
			p->next=ppnode->child;
			ppnode->child=p;
		}	/* else */
	}		/* for loop */
	p2 = ppnode->parent;
	while(p2) {
		find_next_tree(pmap,mapsz,pnode,pinx,pnodesz,p2,level+1);
		p2=p2->next;
	}
	p2=ppnode->child;
	while(p2) {
		find_next_tree(pmap,mapsz,pnode,pinx,pnodesz,p2,level+1);
		p2=p2->next;
	}
	return(p);

}

/****************************************************************************
 *
 *   build_link_tree()
 *
 *****************************************************************************/
struct link_tree_node	*build_link_tree(pmap, mapsz, pnode, pinx, pnodesz,
		 pvname)
struct	link_map_entry  *pmap[];	/* array of links */
int	mapsz;				/* size of pmap[] */
struct  link_tree_node	*pnode[];	/* array of tree nodes */
int	*pinx;				/* next available slot in pnode */
int	pnodesz;			/* size of pnode */
char	*pvname;			/* root pv name */
{
	char	*xname="build_link_tree";
	struct	link_tree_node *p;
	int	i;

	if(!(p=alloc_link_tree_node(pnode,pinx,pnodesz,pvname,0,0)))
		return(NULL);
	find_next_tree(pmap,mapsz,pnode,pinx,pnodesz,p,1);
	for (i=0; i<mapsz; i++) 
		if(pmap[i]->level <0) {
			print_error_msg_p(xname,
		"Warning, some links have been omitted from tree", PAUSE_F);
			break;
		}
	return(p);
}

/*****************************************************************************
 *
 *    print_link_tree()
 *
 *****************************************************************************/
int	print_link_tree(p, buffer, indent, porc, first1)
struct	link_tree_node	*p;
char	*buffer;
char	*indent;
short	porc;
short	first1;
{
	char	ibuffer[120];
	char	sbuffer[PVNAME_SZ + 24];
	char	begin[8];

	if(p) {
		print_link_tree(p->next,buffer,indent,porc,0);
		if(!first1) {
			strcpy(ibuffer,buffer);
			strcat(ibuffer,indent);
		} else
			memset(ibuffer, '\0', sizeof(ibuffer));
		print_link_tree(p->parent,ibuffer, indent, 1,0);
		write(rep_fd, ibuffer, strlen(ibuffer));
		if(first1)
			memset(begin, '\0', sizeof(begin));
		else if(porc)
			strcpy(begin, "v<-");
		else
			strcpy(begin, "^->");
		if(pmap[p->link]) {
		    switch(pmap[p->link]->fldtype) {
			case (DBF_INLINK):
				sprintf(sbuffer, "%s %3.3d (I) %s\n", begin,
					 p->link, p->pvname);
				break;
			case (DBF_OUTLINK):
				sprintf(sbuffer, "%s %3.3d (O) %s\n", begin,
					 p->link, p->pvname);
				break;
			case (DBF_FWDLINK):
				sprintf(sbuffer, "%s %3.3d (F) %s\n", begin,
					 p->link, p->pvname);
				break;
			default :
				sprintf(sbuffer, "%s %3.3d %s\n", begin,
					 p->link, p->pvname);
				break;
		     }
		}
		write(rep_fd, sbuffer, strlen(sbuffer));
		print_link_tree(p->child,ibuffer,indent, 0,0);
	}
	return(0);
}

/*****************************************************************************
 *
 *  print_link_tree_d()
 *
 ****************************************************************************/
int print_link_tree_d(p)
struct link_tree_node	*p;	/* the root node of the link tree */
{
	char	buffer[80];
	static	char	*info[] = {
		"\nLegend:\n",
		"  <Direction Link# (Type) pvname>\n",
		"	Direction: <- refers to, -> is referred by\n",
		"	Link# = reference number for this link\n",
		"	(Type): I = Inlink, O = Outlink, F = Fwdlink\n",
		"	pvname = process variable name\n\n",
		NULL
	};

	if(rep_fd == STDOUT) {
		clear();
		move(0,0);
		refresh();
	}
	if(rep_fd != STDOUT) {
		sprintf(buffer,
		 "\n-------------------------------------------------\n");
		write(rep_fd, buffer, strlen(buffer));
	}
	sprintf(buffer, "\n\nLINK TREE for PV %s\n\n", p->pvname);
	write(rep_fd, buffer, strlen(buffer));
	print_string_array(info, sizeof(info)/sizeof(char *), rep_fd);
	print_link_tree(p,"","     ",1,1);
	if(rep_fd == STDOUT){
		printf("\n\nHit any key to continue\n");
		getinpch();
	}
}

/*****************************************************************************
 *
 *	print_link_map_pvs()
 *
 *****************************************************************************/
int	print_link_map_pvs(phold, pholdsz, ures, unressz)
struct	temp_hold_pv_info	*phold[];	/* holds info on resolved pvs*/
int			pholdsz;		/* size of phold */
char			*ures[];		/* the unresolved pvs */
int			unressz;		/* size of ures */
{
	int	i;
	char	buffer[128];
	char	*prectype;
	static char *info[] = {
		"\n\nThe following process variables are in the tree:\n\n",
		"Legend: pvname [(record_type, scan_mech) | (External)]\n\n",
		NULL
	};

	if(rep_fd== STDOUT) {
		clear();
		move(0,0);
		refresh();
	}
	print_string_array(info, sizeof(info)/sizeof(char *), rep_fd);
	i=0;
	while((phold[i] != NULL) && (i<pholdsz)) {
		if((prectype = (char*)XGET_PRECTYPE(phold[i]->rectype)) == NULL)
			prectype="BAD RECORD_TYPE";
		sprintf(buffer,"%s (%s,%s)\n", 
			 phold[i]->pvname, prectype, phold[i]->pscantype);
		write(rep_fd, buffer, strlen(buffer));
		i++;
	}
	i=0;
	while((ures[i] != NULL) && (i<unressz)) {
		sprintf(buffer,"%s  (External)\n", ures[i++]);
		write(rep_fd, buffer, strlen(buffer));
	}
	if(rep_fd == STDOUT) {
		printf("\n\nHit any key to continue\n");
		getinpch();
	}

	
}


/******************************************************************************
 *
 *	display_links
 *
 *	displays the link tree for a single process variable 
 *
 *****************************************************************************/
int	display_links()
{
	char	*xname="display_links";
	short	rectype;	/* record type of PV to list */
	short	record_no;	/* database file record number */
	int	i,j;
	int	next;		/* next empty slot in pmap[] */
	char	**pvs_to_do;	/* array of pv's yet to process fwd direction */
	char	**pvs_done;	/* array of pv's done */
	char	**unresolved_pvs; /* array of unresolved pvs */
	struct	temp_hold_pv_info  **phold; /* will hold pv info to use to 
					complete left-hand sides of links */
	struct	link_map_entry	**pwork;  /* working array used to group pv
					links */
	struct	link_tree_node	**pnode;  /* array of pointers to link-tree
						nodes */
	struct	link_tree_node	*root_node; /* root node of link tree */
	int	to_do_next;	/* next empty slot in forward to do list */
	int	hold_next;	/* next empty slot in backward to do list */
	int	done_next;	/* next empty slot in done list */
	int	ures_pvs_next;	/* next empty slot in uresolved list */
	int	work_next;	/* next empty slot in pwork */
	int	nodes_next;	/* next empty slot in pnode */
	int	status;		/* used to indicate error */	
	short	level= -1;	/*  current link tree node level */
	short	fld_type;	/* field type returned by get_recfld() */
	short	fld_offset;	/* field offset returned by get_recfld() */
	short	fld_size;	/* field size returned by get_recfld() */
	short	*plinks;	/* pointer to array of offsets of link fields*/
	struct	fldDes	*pfldDes;	/* field description structure */
	struct	link	*plink;		/* link structure */
	struct	pv_link	*ppv_link; 
	char	root_pv_name[PVNAME_SZ+1];  /* the pv who's link map we're 
						finding */
	short	pvd_index;		/* pvd index of current pv */
	short	crec_type;		/* record type of current pv */
	short	crec_no;		/* record number of current pv */
	short	type_inx;
	char	buffer[128];		/* temp buffer for printing */

	get_report_device();
	while(select_a_pv(&record_no,&rectype, "DISPLAY LINKS") >= 0) {
		status = 0;
	   if ( (pmap = (struct link_map_entry **)calloc(MAX_LINK_MAP_ENTRIES,
			sizeof(struct link_map_entry *))) == NULL) {
		print_syserror_msg_p(xname,"malloc", errno, PAUSE_F);
		break;
	   }
	   if (( pvs_to_do=(char **) calloc(MAX_LINK_MAP_ENTRIES,
			 sizeof(char *))) == NULL) {
		print_syserror_msg_p(xname,"malloc", errno, PAUSE_F);
		break;
	   }
		
	   if (( phold=(struct temp_hold_pv_info **)calloc(MAX_LINK_MAP_ENTRIES,
			 sizeof(struct temp_hold_pv_info *))) == NULL) {
		print_syserror_msg_p(xname,"malloc", errno, PAUSE_F);
		break;
	   }
		
	   if (( pvs_done=(char **) calloc(MAX_LINK_MAP_ENTRIES,
			 sizeof(char *))) == NULL) {
		print_syserror_msg_p(xname,"malloc", errno, PAUSE_F);
		break;
	   }
	   if (( unresolved_pvs=(char **) calloc(MAX_LINK_MAP_ENTRIES,
			 sizeof(char *))) == NULL) {
		print_syserror_msg_p(xname,"malloc", errno, PAUSE_F);
		break;
	   }
	   to_do_next=hold_next=done_next=ures_pvs_next=next = 0;


		if(set_current_record_type(rectype)){
			status=-1;
			break;
		}
		if (open_record(rectype) < 0) {
			status = -1;
			break;
		}
		if ( read_record(record_no, xname) < 0) {
			status = -1;
			break;
		}

		if(get_recfld(rectype, "NAME", &fld_type, &fld_offset,
				 &fld_size) < 0) {
			print_error_msg_p(xname,
			     "Couldn't find NAME field in record description",
					 PAUSE_F);
			status = -1;
			break;
		}
		if(add_to_string_array(&ctrl_record[fld_offset], pvs_to_do,
			&to_do_next, MAX_LINK_MAP_ENTRIES, PVNAME_SZ, xname)
				< -1){
			status=-1;
			break;
		}
		strcpy(root_pv_name,&ctrl_record[fld_offset]);

		close_record();
		
		/* the main loop to search for links */

		while(to_do_next != 0) {
		  if((pvd_index=get_pvd_index_by_pvname(pvs_to_do[0],xname,1))
			< 0) {
			if(string_array_move(pvs_to_do, 0, &to_do_next,
				MAX_LINK_MAP_ENTRIES, unresolved_pvs, 
				&ures_pvs_next, MAX_LINK_MAP_ENTRIES,xname)) {
					status=-1;
					break;
			}
			continue;
		  } 
		  crec_type=dbPvd->entry[pvd_index].record_type;
		  crec_no  =dbPvd->entry[pvd_index].record_no;
		  if(set_current_record_type(crec_type)) {
			status = -1;
			continue;
		  }
		  if (open_record(crec_type) < 0) {
			status = -1;
			continue;
		  }
		  if ( read_record(crec_no, xname) < 0) {
			status = -1;
			continue;
		  }
		  
		  if(get_recfld(rectype, "NAME", &fld_type, &fld_offset,
				 &fld_size) < 0) {
			print_error_msg_p(xname,
			     "Couldn't find NAME field in record description",
					 PAUSE_F);
			status = -1;
			continue;
		  }
		/* look in this record for links */
		  plinks = precorddes->link_ind;
		  for(i = 0; i < precorddes->no_links; i++, plinks++) {
			if((pfldDes = GET_PFLDDES(precorddes, *plinks))==NULL){
				print_error_msg_p(xname, 
			        "NULL pointer to field description encountered",
					 PAUSE_F);
				continue;
			}
			plink=(struct link *)&ctrl_record[pfldDes->offset];
			if (plink->type == PV_LINK) {
				ppv_link=&plink->value.pv_link;
				add_link_map_entry(pmap, &next, level,crec_type,
				   crec_no, &ctrl_record[fld_offset], pfldDes,
					phold, &hold_next, xname);	
			      /* check if we've looked at this pv name before*/
				if((lookup_string_array(pvs_done,
					 ppv_link->pvname,
						 MAX_LINK_MAP_ENTRIES) < 0) 
					&& (lookup_string_array(pvs_to_do,
						ppv_link->pvname,
						MAX_LINK_MAP_ENTRIES) < 0 )
					&& (lookup_string_array(unresolved_pvs,
						ppv_link->pvname,
						MAX_LINK_MAP_ENTRIES) < 0)) 
					if(add_to_string_array(ppv_link->pvname,
						pvs_to_do,&to_do_next,
						 MAX_LINK_MAP_ENTRIES,
						 PVNAME_SZ,xname)) {
						status = -1;
						continue;
					}
			}				
		  } /* for(i = 0; i < precorddes->no_links; i++, plinks++) {*/
		  close_record();
		/* now scan all the records for links to the current pv */
		  for(type_inx = 1; type_inx < dbRecType->number; type_inx++){
			/* check that type exists */
			
		    	while ((type_inx < dbRecType->number)
               			 && (!(char*)XGET_PRECTYPE(type_inx)))
	        			type_inx++;
	   	 	if (type_inx >= dbRecType->number)
					 continue;

	    		set_current_record_type(type_inx);

	    		/* open the file for this record type */
	    		if ( open_record(type_inx) < 0)
					 continue;

			/* scan the pvd for records of the current type */
			for(i = 0; i < MAX_PV; i++) {
			    if (dbPvd->entry[i].record_type == type_inx) {
				if( read_record(
					dbPvd->entry[i].record_no, xname) < 0){
						continue;
				}
				if(get_recfld(type_inx, "NAME", &fld_type,
					&fld_offset, &fld_size) < 0) {
				   print_error_msg_p(xname,
			 "Couldn't find NAME field in record description",
					 PAUSE_F);
		 		   continue;
				}
				if(check_if_already_processed(
					&ctrl_record[fld_offset], pvs_to_do,
						pvs_done, unresolved_pvs))
						continue;

				plinks=precorddes->link_ind;
				/* scan each link field for the current pvname*/
				for(j = 0; j < precorddes->no_links; j++,
						 plinks++) {
					if((pfldDes = GET_PFLDDES(precorddes,
						*plinks)) == NULL) {
						print_error_msg_p(xname,
		"NULL pointer to field description encountered", PAUSE_F);
						continue;
					}
					plink=(struct link *)&ctrl_record[
						pfldDes->offset];
					if(plink->type==PV_LINK) {
					  ppv_link=&plink->value.pv_link;
					  if(strcmp(ppv_link->pvname,
							pvs_to_do[0]))
							continue;
					  if(add_to_string_array(
					       (char *)&ctrl_record[fld_offset],
						pvs_to_do, &to_do_next,
						MAX_LINK_MAP_ENTRIES,
						 PVNAME_SZ, xname))
							break;
				        }
				} /* for(j = 0; j , precorddes->no_links; j++,*/
			    } /* if (dbPvd->entry[i].record_type == type_inx) */
			} /* for(i = 0; i < MAX_PV; i++) { */
		  } /*for(type_inx = 1; type_inx < dbRecType->number;.... */ 
		  if(string_array_move(pvs_to_do, 0, &to_do_next,
			MAX_LINK_MAP_ENTRIES, pvs_done, 
			&done_next, MAX_LINK_MAP_ENTRIES,xname)) {
				status=-1;
				break;
		  }
		  close_record();
		}  /*while(to_do_next != 0) { */
	/* now complete left hand side of links */
	   for( i=0; i < hold_next; i++)
	     if(phold[i]->pvname != NULL)
		for(j=0; j < next; j++) 
		   if(pmap[j]->lpvname != NULL)  
			if(!strcmp(phold[i]->pvname,pmap[j]->lpvname)) {	
				pmap[j]->lrectype=phold[i]->rectype;
				pmap[j]->exist=1;
				pmap[j]->lscantype = phold[i]->scantype;
				strcpy(pmap[j]->plscantype,phold[i]->pscantype);
				if(!(pfldDes=pvdGetFld(pmap[j]->lrectype,
						pmap[j]->lfldname)))
					pmap[j]->lfldtype = -1;
				else
					pmap[j]->lfldtype=pfldDes->field_type;
			}
	/* now figure out the grouping of link map entries */
	/* first allocate a place to a build a group list of link entries*/
	   if((pwork=(struct link_map_entry **)calloc(2*MAX_LINK_MAP_ENTRIES,
			sizeof(struct link_map_entry *))) == NULL) {
		print_syserror_msg_p(xname, "malloc", errno, PAUSE_F);
		status=-1;
	   }
	   if(pvs_to_do[0] != NULL || to_do_next != 0) {
		print_error_msg_p(xname,
			"Internal DCT error, pvs_to_do not cleared", PAUSE_F);
		status=-1;
	   }
	   if(!status) {
		if(add_to_string_array(root_pv_name, pvs_to_do, &to_do_next,
			MAX_LINK_MAP_ENTRIES, PVNAME_SZ, xname) < 0) {
			status=-1;
		}
	   }
	   work_next=j =0;
	   while(to_do_next != j && work_next < MAX_LINK_MAP_ENTRIES
			 && !status){
		/* look for the current pv */
		for (i=0; i < next; i++) {
			if(!strcmp(pmap[i]->pvname,pvs_to_do[j])) { 
				pwork[work_next++]=pmap[i];
				if(lookup_string_array(pvs_to_do,
					 pmap[i]->lpvname,
					  MAX_LINK_MAP_ENTRIES) < 0)
					if(add_to_string_array(pmap[i]->lpvname,
						pvs_to_do, &to_do_next,
						MAX_LINK_MAP_ENTRIES,
							 PVNAME_SZ,
						xname) < 0){
							status=-1;
							break;
					}
			} else if(!strcmp(pmap[i]->lpvname, pvs_to_do[j])){
				pwork[work_next++]=pmap[i];
				if(lookup_string_array(pvs_to_do,
					 pmap[i]->pvname,
					  MAX_LINK_MAP_ENTRIES) < 0)
					if(add_to_string_array(pmap[i]->pvname,
						pvs_to_do, &to_do_next,
						MAX_LINK_MAP_ENTRIES,
							 PVNAME_SZ,
						xname) < 0){
							status=-1;
							break;
					}
			}
		} /* for (i=0; i < next; i++) { */
		pwork[work_next++]= NULL;	/* to delimit link groups */
		j++;
	   }   /* while(to_do_next != j ){ */
	/* now construct the link tree */
	/* first initialize the level fields in the link map entrys */
	   for ( i=0; i < next; i++)
		pmap[i]->level=-1;
	/* now allocate an array to hold pointers to the tree nodes */
	   if((pnode=(struct link_tree_node **) calloc(MAX_LINK_MAP_ENTRIES,
			sizeof(struct link_tree_node *))) == NULL) {
			print_syserror_msg_p(xname, "calloc", errno, PAUSE_F);
			status= -1;
			break;
	   }
	   nodes_next=0;
	   root_node=build_link_tree(pmap, next, pnode, &nodes_next,
		 MAX_LINK_MAP_ENTRIES, root_pv_name);
	   if(rep_fd != STDOUT) {
	   	sprintf(buffer,"\n****** LINK MAP INFO for PV: %s ******\n\n",
			root_pv_name);
	   	write(rep_fd, buffer, strlen(buffer));
	   }
	   print_link_map_pvs(phold, MAX_LINK_MAP_ENTRIES, unresolved_pvs,
				MAX_LINK_MAP_ENTRIES);
/*
	   print_hold_struct(phold, hold_next);
	   getinpch();
*/
	   if(rep_fd != STDOUT) {
	   	sprintf(buffer,
			"\n---------------------------------------------\n");
	   	write(rep_fd, buffer, strlen(buffer));
	   }
	   print_link_map(pmap,next,xname,
			 "\nThe following links were found\n\n");
	   if(rep_fd == STDOUT) {
	   	printf("\nHit any key to continue\n");
	   	getinpch();
	   }
/*
	   if(rep_fd != STDOUT) {
	   	write(rep_fd, buffer, strlen(buffer));
	   }
	   print_link_groups(pwork,work_next,pvs_to_do, to_do_next, phold,
			hold_next,  xname,
		 "\nThe following groups the links for each PV in the tree\n");
	   if(rep_fd == STDOUT) {
	   	printf("\nHit any key to continue\n");
	   	getinpch();
	   }
*/
	   print_link_tree_d(root_node);
	   free_array((char **)pnode, MAX_LINK_MAP_ENTRIES);
	   free_array((char **)pwork, MAX_LINK_MAP_ENTRIES);
	   free_array(pvs_done, MAX_LINK_MAP_ENTRIES);
	   free_array((char **)phold, MAX_LINK_MAP_ENTRIES);
	   free_array(pvs_to_do, MAX_LINK_MAP_ENTRIES);
	   free_array(unresolved_pvs, MAX_LINK_MAP_ENTRIES);
	   free_array((char **)pmap, MAX_LINK_MAP_ENTRIES);
	} 	/* main while loop */
	return(0);
}

/*
 *	LOC_SINGLE_REF
 *
 *	list all the references to the specified record in the current database
 */
loc_single_ref(pname)
char	*pname;	/* name of the record being tested for links */
{
	char		*xname = "loc_single_ref";
	short		chkorder[200];
	short	ordinx;
	short	type_inx;
	short		numlinks;
	short	link;		/* link counter */
	short		nolinks;	/* num of links for this record type */
        short           *linkinx;       /* pointer to array of indexes within
                                        array of pointers to struct fldDes */
                                        /* for this record type */
	struct link	*plink; /* pointer to a link field structure */
	struct	pv_link	*ppv_link;
        struct fldDes   *pfldDes;
	short		fd;
	char		lrecord[MAX_RECORD_SIZE];
	char		buffer[120];
	struct	recLoc	lrecLoc;
	struct sdrHeader lsdrHeader;

	/* init */
	numlinks = 0;

	/* check each type */
	for (type_inx = 1; type_inx < dbRecType->number; type_inx++){

	    /* check for type exists */
	    while ((type_inx < dbRecType->number)
                && (!(char*)XGET_PRECTYPE(type_inx)))
	        type_inx++;
	    if (type_inx >= dbRecType->number) continue;

	    set_current_record_type(type_inx);
	    nolinks = precorddes->no_links;

	    /* open the file for this record type */
	    if ( (fd=open_loc_record(type_inx, &lsdrHeader, &lrecLoc)) 
			< 0) continue;

	    /* alphabetical order */
	    order_pvs(type_inx,chkorder, MAX_PV+4);

	    /* check each record */
	    for (ordinx = 0; chkorder[ordinx] != -1; ordinx++){

	        /* get this record into memory */
	        read_loc_record( fd, lrecord, 
			dbPvd->entry[chkorder[ordinx]].record_no,
			 xname, &lrecLoc);

		/* for each link field */
		linkinx = precorddes->link_ind;
		for (link = 0; link < nolinks; link++,linkinx++){
                    if ( ( pfldDes = GET_PFLDDES( precorddes, *linkinx)) ==
                                NULL) {
                                print_error_msg_p(xname,
                               "NULL pointer to field description encountered",
					PAUSE_F);
                                continue;
                    }

		/* pointers to the link field */
		    plink =
			(struct link*)&lrecord[pfldDes->offset];
		    ppv_link = &plink->value.pv_link;

		    /* is linked */
		    if (plink->type == PV_LINK){
			/* to this record */
			if( !strcmp(pname, ppv_link->pvname)) {

			    /* report it */
			      if (numlinks == 0) {
				 sprintf(buffer,"\n\t%s",pname);
				 write(rep_fd, buffer, strlen(buffer));
			      }
			      sprintf(buffer,"\n		%-4.4s",
						ppv_link->fldname);
			      write(rep_fd, buffer, strlen(buffer));
			      numlinks++;
			      sprintf(buffer,
		 "	is referenced by	%-*.*s %-*.*s  (type %s)",
					PVNAME_SZ, PVNAME_SZ,
			      		lrecord, FLDNAME_SZ, FLDNAME_SZ,
					 pfldDes->fldname,
			    			 (char*)XGET_PRECTYPE(type_inx));
			      write(rep_fd, buffer, strlen(buffer));
			}
		    }
		}
	    }
	    close(fd);
	}
	return(numlinks);
}

/*
 * LOC_XREF
 *
 * list the cross reference of all link fields in the current pv database
 */
int loc_xref()
{
	short		type_inx;
	short	ordinx;
	short	num;
	short	got_some;
	char	buffer[120];

	/* get report device */
	get_report_device();

	/* report each type */
	clear();
	refresh();
	if ( rep_fd != STDOUT)
		print_wait_msg(1);
	for (type_inx = 1; type_inx < dbRecType->number; type_inx++){

	    /* check for type exists */
	    while ((type_inx < dbRecType->number)
                && (!(char*)XGET_PRECTYPE(type_inx)))
		type_inx++;
	    if (type_inx >= dbRecType->number) continue;

	    /* alphabetical order */
	    order_pvs(type_inx,order, MAX_PV+4);

	    /* get references */
	    if (order[0] == -1) continue;
	    sprintf(buffer, "\ndatabase %s type %s",
              		curr_dbname,(char*)XGET_PRECTYPE(type_inx));
	    write(rep_fd, buffer, strlen(buffer));
	    for(ordinx = 0, got_some = 0; order[ordinx] != -1; ordinx++){
		if ( rep_fd != STDOUT)
			print_wait_msg(0);
		num = loc_single_ref( dbPvd->entry[order[ordinx]].name);
		if (num > 0){
			got_some = 1;
			sprintf(buffer, "\n\t%d references",num);
			write(rep_fd, buffer, strlen(buffer));
		}
	    }
	    if (got_some && (rep_fd == STDOUT)) {
		printf("\nHit any key to continue -- or KPR11 to Quit\n");
		if (getinpch() == EXIT) return(0);
	    }
	}
	return(0);
}
